module Caml = struct
module XXXX = struct
module Result_compat = struct
# 1 "src/stdune/caml/result_compat.ml"
(** Result type for OCaml < 4.03 *)

type ('a, 'error) result =
  | Ok    of 'a
  | Error of 'error
# 11 "boot.ml"
end
module Result : sig
# 1 "src/stdune/caml/result.mli"
include sig
  [@@@warning "-33"]
  (* This open is unused with OCaml >= 4.03 since the stdlib defines a
     result type *)
  open Result_compat
  open Pervasives

  (** The result type.

      It is equal to the result type defined by the standard library
      for OCaml >= 4.03.  *)
  type ('a, 'error) t = ('a, 'error) result =
    | Ok    of 'a
    | Error of 'error
end

type ('a, 'error) result = ('a, 'error) t
# 32 "boot.ml"
end = struct
# 1 "src/stdune/caml/result.ml"
include struct
  [@@@warning "-33"]
  open Result_compat
  open Pervasives

  type ('a, 'error) t = ('a, 'error) result =
    | Ok    of 'a
    | Error of 'error
end

type ('a, 'error) result = ('a, 'error) t
# 46 "boot.ml"
end
end
open XXXX
# 1 "src/stdune/caml/caml.ml"
module Filename = Filename
module String   = String
module Char     = Char
module Result   = Result

type ('a, 'error) result = ('a, 'error) Result.t =
  | Ok    of 'a
  | Error of 'error
# 59 "boot.ml"
end
module Stdune = struct
module XXXX = struct
module Either : sig
# 1 "src/stdune/either.mli"
(** Left or right *)

type ('a, 'b) t =
  | Left  of 'a
  | Right of 'b
# 70 "boot.ml"
end = struct
# 1 "src/stdune/either.ml"
type ('a, 'b) t =
  | Left  of 'a
  | Right of 'b
# 76 "boot.ml"
end
module Ordering : sig
# 1 "src/stdune/ordering.mli"
(** Element ordering *)

type t =
  | Lt (** Lesser than  *)
  | Eq (** Equal        *)
  | Gt (** Greater than *)

val of_int : int -> t
val to_int : t -> int

val neq : t -> bool
# 91 "boot.ml"
end = struct
# 1 "src/stdune/ordering.ml"
type t =
  | Lt
  | Eq
  | Gt

let of_int n =
  if n < 0 then
    Lt
  else if n = 0 then
    Eq
  else
    Gt

let to_int = function
  | Lt -> -1
  | Eq ->  0
  | Gt ->  1

let neq = function
  | Eq -> false
  | Lt | Gt -> true
# 115 "boot.ml"
end
module List : sig
# 1 "src/stdune/list.mli"
type 'a t = 'a list

include module type of struct include ListLabels end

val is_empty : _ t -> bool

val filter_map : 'a t -> f:('a -> 'b option) -> 'b t

val filteri : 'a t -> f:(int -> 'a -> bool) -> 'a t

val concat_map : 'a t -> f:('a -> 'b t) -> 'b t

val     partition_map : 'a t -> f:('a -> ('b, 'c) Either.t) -> 'b t * 'c t
val rev_partition_map : 'a t -> f:('a -> ('b, 'c) Either.t) -> 'b t * 'c t

type ('a, 'b) skip_or_either =
  | Skip
  | Left  of 'a
  | Right of 'b

val filter_partition_map
  :  'a t
  -> f:('a -> ('b, 'c) skip_or_either)
  -> 'b t * 'c t
val rev_filter_partition_map
  :  'a t
  -> f:('a -> ('b, 'c) skip_or_either)
  -> 'b t * 'c t

val find     : 'a t -> f:('a -> bool     ) -> 'a option
val find_map : 'a t -> f:('a -> 'b option) -> 'b option

val last : 'a t -> 'a option

val        sort : 'a t -> compare:('a -> 'a -> Ordering.t) -> 'a t
val stable_sort : 'a t -> compare:('a -> 'a -> Ordering.t) -> 'a t

val compare : 'a t -> 'a t -> compare:('a -> 'a -> Ordering.t) -> Ordering.t
# 157 "boot.ml"
end = struct
# 1 "src/stdune/list.ml"
type 'a t = 'a list

include ListLabels

let is_empty = function
  | [] -> true
  | _  -> false

let rec filter_map l ~f =
  match l with
  | [] -> []
  | x :: l ->
    match f x with
    | None -> filter_map l ~f
    | Some x -> x :: filter_map l ~f

let filteri l ~f =
  let rec filteri l i =
    match l with
    | [] -> []
    | x :: l ->
      let i' = succ i in
      if f i x
      then x :: filteri l i'
      else filteri l i'
  in
  filteri l 0

let concat_map l ~f = concat (map l ~f)

let rev_partition_map =
  let rec loop l accl accr ~f =
    match l with
    | [] -> (accl, accr)
    | x :: l ->
      match (f x : (_, _) Either.t) with
      | Left  y -> loop l (y :: accl) accr ~f
      | Right y -> loop l accl (y :: accr) ~f
  in
  fun l ~f -> loop l [] [] ~f

let partition_map l ~f =
  let l, r = rev_partition_map l ~f in
  (rev l, rev r)

type ('a, 'b) skip_or_either =
  | Skip
  | Left  of 'a
  | Right of 'b

let rev_filter_partition_map =
  let rec loop l accl accr ~f =
    match l with
    | [] -> (accl, accr)
    | x :: l ->
      match f x with
      | Skip    -> loop l accl accr        ~f
      | Left  y -> loop l (y :: accl) accr ~f
      | Right y -> loop l accl (y :: accr) ~f
  in
  fun l ~f -> loop l [] [] ~f

let filter_partition_map l ~f =
  let l, r = rev_filter_partition_map l ~f in
  (rev l, rev r)

let rec find_map l ~f =
  match l with
  | [] -> None
  | x :: l ->
    match f x with
    | None -> find_map l ~f
    | Some _ as res -> res

let rec find l ~f =
  match l with
  | [] -> None
  | x :: l -> if f x then Some x else find l ~f

let rec last = function
  | [] -> None
  | [x] -> Some x
  | _::xs -> last xs

let sort t ~compare =
  sort t ~cmp:(fun a b -> Ordering.to_int (compare a b))

let stable_sort t ~compare =
  stable_sort t ~cmp:(fun a b -> Ordering.to_int (compare a b))

let rec compare a b ~compare:f : Ordering.t =
  match a, b with
  | [], [] -> Eq
  | [], _ :: _ -> Lt
  | _ :: _, [] -> Gt
  | x :: a, y :: b ->
    match (f x y : Ordering.t) with
    | Eq -> compare a b ~compare:f
    | ne -> ne
# 259 "boot.ml"
end
module Staged : sig
# 1 "src/stdune/staged.mli"
type +'a t
val unstage : 'a t -> 'a
val stage : 'a -> 'a t
# 266 "boot.ml"
end = struct
# 1 "src/stdune/staged.ml"
type 'a t = 'a
let unstage t = t
let stage t = t
# 272 "boot.ml"
end
module String : sig
# 1 "src/stdune/string.mli"
include module type of struct include StringLabels end

val compare : t -> t -> Ordering.t

val break : t -> pos:int -> t * t

val is_empty : t -> bool
val is_prefix : t -> prefix:t -> bool
val is_suffix : t -> suffix:t -> bool

val drop_prefix : t -> prefix:t -> t option

(** These only change ASCII charactes *)
val capitalize   : t -> t
val uncapitalize : t -> t
val uppercase    : t -> t
val lowercase    : t -> t

val index : t -> char -> int option

val extract_words : t -> is_word_char:(char -> bool) -> t list
val extract_comma_space_separated_words : t -> t list
val extract_blank_separated_words : t -> t list

val lsplit2 : t -> on:char -> (t * t) option
val lsplit2_exn : t -> on:char -> t * t
val rsplit2 : t -> on:char -> (t * t) option

val split : t -> on:char -> t list
val split_lines : t -> t list

(** Escace ONLY double quotes. {!escape} also escapes '\n',... and
    transforms all chars above '~' into '\xxx' which is not suitable
    for UTF-8 strings. *)
val escape_double_quote : t -> t

(** Return the length of the longest string in the list *)
val longest : string list -> int
val longest_map : 'a list -> f:('a -> string) -> int

val exists : t -> f:(char -> bool) -> bool
# 317 "boot.ml"
end = struct
# 1 "src/stdune/string.ml"
(* Because other the syntax s.[x] causes trouble *)
module String = Caml.String

include struct
  [@@@warning "-32-3"]
  let capitalize_ascii   = String.capitalize
  let uncapitalize_ascii = String.uncapitalize
  let uppercase_ascii    = String.uppercase
  let lowercase_ascii    = String.lowercase
end

include StringLabels

let compare a b = Ordering.of_int (String.compare a b)

let capitalize   = capitalize_ascii
let uncapitalize = uncapitalize_ascii
let uppercase    = uppercase_ascii
let lowercase    = lowercase_ascii

let break s ~pos =
  (sub s ~pos:0 ~len:pos,
   sub s ~pos ~len:(length s - pos))

let is_empty s = length s = 0

let is_prefix s ~prefix =
  let len = length s in
  let prefix_len = length prefix in
  len >= prefix_len &&
  sub s ~pos:0 ~len:prefix_len = prefix

let is_suffix s ~suffix =
  let len = length s in
  let suffix_len = length suffix in
  len >= suffix_len &&
  sub s ~pos:(len - suffix_len) ~len:suffix_len = suffix

let drop_prefix s ~prefix =
  if is_prefix s ~prefix then
    if length s = length prefix then
      Some ""
    else
      Some (sub s ~pos:(length prefix) ~len:(length s - length prefix))
  else
    None

let extract_words s ~is_word_char =
  let rec skip_blanks i =
    if i = length s then
      []
    else if is_word_char s.[i] then
      parse_word i (i + 1)
    else
      skip_blanks (i + 1)
  and parse_word i j =
    if j = length s then
      [sub s ~pos:i ~len:(j - i)]
    else if is_word_char s.[j] then
      parse_word i (j + 1)
    else
      sub s ~pos:i ~len:(j - i) :: skip_blanks (j + 1)
  in
  skip_blanks 0

let extract_comma_space_separated_words s =
  extract_words s ~is_word_char:(function
    | ',' | ' ' | '\t' | '\n' -> false
    | _ -> true)

let extract_blank_separated_words s =
  extract_words s ~is_word_char:(function
    | ' ' | '\t' -> false
    | _ -> true)

let lsplit2 s ~on =
  match index s on with
  | exception Not_found -> None
  | i ->
    Some
      (sub s ~pos:0 ~len:i,
       sub s ~pos:(i + 1) ~len:(length s - i - 1))

let lsplit2_exn s ~on =
  match lsplit2 s ~on with
  | Some s -> s
  | None -> invalid_arg "lsplit2_exn"

let rsplit2 s ~on =
  match rindex s on with
  | exception Not_found -> None
  | i ->
    Some
      (sub s ~pos:0 ~len:i,
       sub s ~pos:(i + 1) ~len:(length s - i - 1))

let index s ch =
  match index s ch with
  | i -> Some i
  | exception Not_found -> None

let split s ~on =
  let rec loop i j =
    if j = length s then
      [sub s ~pos:i ~len:(j - i)]
    else if s.[j] = on then
      sub s ~pos:i ~len:(j - i) :: loop (j + 1) (j + 1)
    else
      loop i (j + 1)
  in
  loop 0 0

let split_lines s =
  let rec loop ~last_is_cr ~acc i j =
    if j = length s then (
      let acc =
        if j = i || (j = i + 1 && last_is_cr) then
          acc
        else
          sub s ~pos:i ~len:(j - i) :: acc
      in
      List.rev acc
    ) else
      match s.[j] with
      | '\r' -> loop ~last_is_cr:true ~acc i (j + 1)
      | '\n' ->
        let line =
          let len = if last_is_cr then j - i - 1 else j - i in
          sub s ~pos:i ~len
        in
        loop ~acc:(line :: acc) (j + 1) (j + 1) ~last_is_cr:false
      | _ ->
        loop ~acc i (j + 1) ~last_is_cr:false
  in
  loop ~acc:[] 0 0 ~last_is_cr:false

let escape_double_quote s =
  let n = ref 0 in
  let len = length s in
  for i = 0 to len - 1 do
    if unsafe_get s i = '"' then incr n;
  done;
  if !n = 0 then s
  else (
    let b = Bytes.create (len + !n) in
    n := 0;
    for i = 0 to len - 1 do
      if unsafe_get s i = '"' then (
        Bytes.unsafe_set b !n '\\';
        incr n;
      );
      Bytes.unsafe_set b !n (unsafe_get s i);
      incr n
    done;
    Bytes.unsafe_to_string b
  )

let longest_map l ~f =
  List.fold_left l ~init:0 ~f:(fun acc x ->
    max acc (length (f x)))

let longest l = longest_map l ~f:(fun x -> x)

let exists s ~f =
  try
    for i=0 to length s - 1 do
      if (f s.[i]) then raise_notrace Exit
    done;
    false
  with Exit ->
    true
# 491 "boot.ml"
end
module Pp : sig
# 1 "src/stdune/pp.mli"
(** Pretty printers *)

(** A document that is not yet rendered. The argument is the type of
    tags in the document. For instance tags might be used for
    styles. *)
type +'tag t

module type Tag = sig
  type t

  module Handler : sig
    type tag = t
    type t

    (** Initial tag handler *)
    val init : t

    (** Handle a tag: return the string that enables the tag, the
        handler while the tag is active and the string to disable the
        tag. *)
    val handle : t -> tag -> string * t * string
  end with type tag := t
end

module Renderer : sig
  module type S = sig
    module Tag : Tag

    val string
      :  unit
      -> (?margin:int -> ?tag_handler:Tag.Handler.t -> Tag.t t -> string)
           Staged.t
    val channel
      :  out_channel
      -> (?margin:int -> ?tag_handler:Tag.Handler.t -> Tag.t t -> unit)
           Staged.t
  end

  module Make(Tag : Tag) : S with module Tag = Tag
end

(** A simple renderer that doesn't take tags *)
module Render : Renderer.S
  with type Tag.t         = unit
  with type Tag.Handler.t = unit

val nop : 'a t
val seq : 'a t -> 'a t -> 'a t
val concat : 'a t list -> 'a t
val vbox : ?indent:int -> 'a t -> 'a t
val hbox : 'a t -> 'a t

val int    : int    -> _ t
val string : string -> _ t
val char   : char   -> _ t
val list   : ('a -> 'b t) -> 'a list -> 'b t

val space   : _ t
val cut     : _ t
val newline : _ t

val text : string -> _ t

val tag : 'a t -> tag:'a -> 'a t
# 559 "boot.ml"
end = struct
# 1 "src/stdune/pp.ml"
type +'a t =
  | Nop
  | Seq of 'a t * 'a t
  | Concat of 'a t list
  | Vbox of int * 'a t
  | Hbox of 'a t
  | Int of int
  | String of string
  | Char of char
  | List : ('a -> 'b t) * 'a list -> 'b t
  | Space
  | Cut
  | Newline
  | Text of string
  | Tag of 'a * 'a t

module type Tag = sig
  type t
  module Handler : sig
    type tag = t
    type t
    val init : t
    val handle : t -> tag -> string * t * string
  end with type tag := t
end

module Renderer = struct
  module type S = sig
    module Tag : Tag

    val string
      :  unit
      -> (?margin:int -> ?tag_handler:Tag.Handler.t -> Tag.t t -> string)
           Staged.t
    val channel
      :  out_channel
      -> (?margin:int -> ?tag_handler:Tag.Handler.t -> Tag.t t -> unit)
           Staged.t
  end

  module Make(Tag : Tag) = struct
    open Format

    module Tag = Tag

    (* The format interface only support string for tags, so we embed
       then as follow:

       - length of opening string on 16 bits
       - opening string
       - closing string
    *)
    external get16 : string -> int -> int         = "%caml_string_get16"
    external set16 : bytes  -> int -> int -> unit = "%caml_string_set16"

    let embed_tag ~opening ~closing =
      let opening_len = String.length opening  in
      let closing_len = String.length closing in
      assert (opening_len <= 0xffff);
      let buf = Bytes.create (2 + opening_len + closing_len) in
      set16 buf 0 opening_len;
      Bytes.blit_string opening 0 buf  2   opening_len;
      Bytes.blit_string closing 0 buf (2 + opening_len) closing_len;
      Bytes.unsafe_to_string buf

    let extract_opening_tag s =
      let open_len = get16 s 0 in
      String.sub s ~pos:2 ~len:open_len

    let extract_closing_tag s =
      let pos = 2 + get16 s 0 in
      String.sub s ~pos ~len:(String.length s - pos)


    let rec pp th ppf t =
      match t with
      | Nop -> ()
      | Seq (a, b) -> pp th ppf a; pp th ppf b
      | Concat l -> List.iter l ~f:(pp th ppf)
      | Vbox (indent, t) ->
        pp_open_vbox ppf indent;
        pp th ppf t;
        pp_close_box ppf ()
      | Hbox t ->
        pp_open_hbox ppf ();
        pp th ppf t;
        pp_close_box ppf ()
     | Int    x -> pp_print_int ppf x
     | String x -> pp_print_string ppf x
     | Char   x -> pp_print_char ppf x
     | List (f, l) -> pp_print_list (fun ppf x -> pp th ppf (f x)) ppf l
     | Space -> pp_print_space ppf ()
     | Cut -> pp_print_cut ppf ()
     | Newline -> pp_force_newline ppf ()
     | Text s -> pp_print_text ppf s
     | Tag (tag, t) ->
       let opening, th, closing = Tag.Handler.handle th tag in
       pp_open_tag ppf (embed_tag ~opening ~closing);
       pp th ppf t;
       pp_close_tag ppf ()

    let setup ppf =
      let funcs = pp_get_formatter_tag_functions ppf () in
      pp_set_mark_tags ppf true;
      pp_set_formatter_tag_functions ppf
        { funcs with
          mark_open_tag  = extract_opening_tag
        ; mark_close_tag = extract_closing_tag
        }

    let string () =
      let buf = Buffer.create 1024 in
      let ppf = formatter_of_buffer buf in
      setup ppf;
      Staged.stage (fun ?(margin=80) ?(tag_handler=Tag.Handler.init) t ->
        pp_set_margin ppf margin;
        pp tag_handler ppf t;
        pp_print_flush ppf ();
        let s = Buffer.contents buf in
        Buffer.clear buf;
        s)

    let channel oc =
      let ppf = formatter_of_out_channel oc in
      setup ppf;
      Staged.stage (fun ?(margin=80) ?(tag_handler=Tag.Handler.init) t ->
        pp_set_margin ppf margin;
        pp tag_handler ppf t;
        pp_print_flush ppf ())
  end
end

module Render = Renderer.Make(struct
    type t = unit
    module Handler = struct
      type t   = unit
      let init = ()
      let handle () () = "", (), ""
    end
  end)

let nop = Nop
let seq a b = Seq (a, b)
let concat l = Concat l
let vbox ?(indent=0) t = Vbox (indent, t)
let hbox t = Hbox t

let int x = Int x
let string x = String x
let char x = Char x
let list f l = List (f, l)

let space = Space
let cut = Cut
let newline = Newline

let text s = Text s

let tag t ~tag = Tag (tag, t)
# 721 "boot.ml"
end
module Ansi_color : sig
# 1 "src/stdune/ansi_color.mli"
module Color : sig
  type t =
    | Default
    | Black
    | Red
    | Green
    | Yellow
    | Blue
    | Magenta
    | Cyan
    | White
    | Bright_black
    | Bright_red
    | Bright_green
    | Bright_yellow
    | Bright_blue
    | Bright_magenta
    | Bright_cyan
    | Bright_white
end

module Style : sig
  type t =
    | Fg of Color.t
    | Bg of Color.t
    | Bold
    | Dim
    | Underlined

  (** Ansi escape sequence that set the terminal style to exactly
      these styles *)
  val escape_sequence : t list -> string
end

module Render : Pp.Renderer.S
  with type Tag.t = Style.t list

(** Filter out escape sequences in a string *)
val strip : string -> string
# 764 "boot.ml"
end = struct
# 1 "src/stdune/ansi_color.ml"
module Color = struct
  type t =
    | Default
    | Black
    | Red
    | Green
    | Yellow
    | Blue
    | Magenta
    | Cyan
    | White
    | Bright_black
    | Bright_red
    | Bright_green
    | Bright_yellow
    | Bright_blue
    | Bright_magenta
    | Bright_cyan
    | Bright_white

  let fg_code = function
    | Black          -> "30"
    | Red            -> "31"
    | Green          -> "32"
    | Yellow         -> "33"
    | Blue           -> "34"
    | Magenta        -> "35"
    | Cyan           -> "36"
    | White          -> "37"
    | Default        -> "39"
    | Bright_black   -> "90"
    | Bright_red     -> "91"
    | Bright_green   -> "92"
    | Bright_yellow  -> "93"
    | Bright_blue    -> "94"
    | Bright_magenta -> "95"
    | Bright_cyan    -> "96"
    | Bright_white   -> "97"

  let bg_code = function
    | Black          -> "40"
    | Red            -> "41"
    | Green          -> "42"
    | Yellow         -> "43"
    | Blue           -> "44"
    | Magenta        -> "45"
    | Cyan           -> "46"
    | White          -> "47"
    | Default        -> "49"
    | Bright_black   -> "100"
    | Bright_red     -> "101"
    | Bright_green   -> "102"
    | Bright_yellow  -> "103"
    | Bright_blue    -> "104"
    | Bright_magenta -> "105"
    | Bright_cyan    -> "106"
    | Bright_white   -> "107"
end

module Style = struct
  type t =
    | Fg of Color.t
    | Bg of Color.t
    | Bold
    | Dim
    | Underlined

  let code = function
    | Bold       -> "1"
    | Dim        -> "2"
    | Underlined -> "4"
    | Fg c       -> Color.fg_code c
    | Bg c       -> Color.bg_code c

  let escape_sequence l =
    let codes = "0" :: List.map l ~f:code in
    Printf.sprintf "\027[%sm" (String.concat codes ~sep:";")
end

module Styles = struct
  type t =
    { fg         : Color.t
    ; bg         : Color.t
    ; bold       : bool
    ; dim        : bool
    ; underlined : bool
    }

  let default =
    { fg         = Default
    ; bg         = Default
    ; bold       = true
    ; dim        = true
    ; underlined = true
    }

  let apply t (style : Style.t) =
    match style with
    | Fg c       -> { t with fg         = c    }
    | Bg c       -> { t with bg         = c    }
    | Bold       -> { t with bold       = true }
    | Dim        -> { t with dim        = true }
    | Underlined -> { t with underlined = true }

  let escape_sequence t =
    let open Style in
    let l = [] in
    let l =
      match t.fg with
      | Default -> l
      | c       -> Fg c :: l
    in
    let l =
      match t.bg with
      | Default -> l
      | c       -> Bg c :: l
    in
    let l =
      if t.bold then
        Bold :: l
      else
        l
    in
    let l =
      if t.bold then
        Dim :: l
      else
        l
    in
    let l =
      if t.underlined then
        Underlined :: l
      else
        l
    in
    Style.escape_sequence l
end

module Render = Pp.Renderer.Make(struct
    type t = Style.t list

    module Handler = struct
      type t = Styles.t * string

      let init = (Styles.default, "")

      let handle (t, seq) styles =
        let t' = List.fold_left styles ~init:t ~f:Styles.apply in
        if t <> t' then
          let seq' = Styles.escape_sequence t' in
          (seq',
           (t', seq'),
           seq)
        else
          ("", (t, seq), "")
    end
  end)

let strip str =
  let len = String.length str in
  let buf = Buffer.create len in
  let rec loop i =
    if i = len then
      Buffer.contents buf
    else
      match str.[i] with
      | '\027' -> skip (i + 1)
      | c      -> Buffer.add_char buf c; loop (i + 1)
  and skip i =
    if i = len then
      Buffer.contents buf
    else
      match str.[i] with
      | 'm' -> loop (i + 1)
      | _   -> skip (i + 1)
  in
  loop 0
# 944 "boot.ml"
end
module Array = struct
# 1 "src/stdune/array.ml"
include ArrayLabels
# 949 "boot.ml"
end
module Char = struct
# 1 "src/stdune/char.ml"
module Char = Caml.Char

include struct
  [@@@warning "-32-3"]
  let uppercase_ascii    = Char.uppercase
  let lowercase_ascii    = Char.lowercase
end

include Char
# 962 "boot.ml"
end
module Comparable : sig
# 1 "src/stdune/comparable.mli"
module type S = sig
  type t
  val compare : t -> t -> Ordering.t
end
# 970 "boot.ml"
end = struct
# 1 "src/stdune/comparable.ml"
module type S = sig
  type t
  val compare : t -> t -> Ordering.t
end
# 977 "boot.ml"
end
module Exn : sig
# 1 "src/stdune/exn.mli"
(** Exceptions *)

type t = exn

external raise         : exn -> _ = "%raise"
external raise_notrace : exn -> _ = "%raise_notrace"
external reraise       : exn -> _ = "%reraise"

val protect : f:(unit -> 'a) -> finally:(unit -> unit) -> 'a
val protectx : 'a -> f:('a -> 'b) -> finally:('a -> unit) -> 'b

val raise_with_backtrace: exn -> Printexc.raw_backtrace -> _
# 993 "boot.ml"
end = struct
# 1 "src/stdune/exn.ml"
type t = exn

external raise         : exn -> _ = "%raise"
external raise_notrace : exn -> _ = "%raise_notrace"
external reraise       : exn -> _ = "%reraise"

let protectx x ~f ~finally =
  match f x with
  | y           -> finally x; y
  | exception e -> finally x; raise e

let protect ~f ~finally = protectx () ~f ~finally

include
  ((struct
    [@@@warning "-32-3"]
    let raise_with_backtrace exn _ = reraise exn
    include Printexc
    let raise_with_backtrace exn bt = raise_with_backtrace exn bt
  end) : (sig
     val raise_with_backtrace: exn -> Printexc.raw_backtrace -> _
   end))
# 1018 "boot.ml"
end
module Filename : sig
# 1 "src/stdune/filename.mli"
include module type of struct include Caml.Filename end

val split_extension : string -> string * string
val split_extension_after_dot : string -> string * string

val extension : string -> string

type program_name_kind =
  | In_path
  | Relative_to_current_dir
  | Absolute

val analyze_program_name : string -> program_name_kind
# 1035 "boot.ml"
end = struct
# 1 "src/stdune/filename.ml"
include Caml.Filename

(* Return the index of the start of the extension, using the same semantic as
   [Filename.extension] in 4.04 *)
let extension_start =
  (* This is from the win32 implementation, but it is acceptable for
     the usage we make of it in this function and covers all
     platforms. *)
  let is_dir_sep = function
    | '/' | '\\' | ':' -> true
    | _ -> false
  in
  let rec check_at_least_one_non_dot s len candidate i =
    if i < 0 then
      len
    else
      match s.[i] with
      | '.' ->
        check_at_least_one_non_dot s len candidate (i - 1)
      | c ->
        if is_dir_sep c then
          len
        else
          candidate
  in
  let rec search_dot s len i =
    if i <= 0 then
      len
    else
      match s.[i] with
      | '.' -> check_at_least_one_non_dot s len i (i - 1)
      | c   -> if is_dir_sep c then len else search_dot s len (i - 1)
  in
  fun s ->
    let len = String.length s in
    search_dot s len (len - 1)

let split_extension fn =
  let i = extension_start fn in
  (String.sub fn ~pos:0 ~len:i,
   String.sub fn ~pos:i ~len:(String.length fn - i))

let split_extension_after_dot fn =
  let i = extension_start fn + 1 in
  let len = String.length fn in
  if i > len then
    (fn, "")
  else
    (String.sub fn ~pos:0 ~len:i,
     String.sub fn ~pos:i ~len:(String.length fn - i))

let extension fn =
  let i = extension_start fn in
  String.sub fn ~pos:i ~len:(String.length fn - i)

type program_name_kind =
  | In_path
  | Relative_to_current_dir
  | Absolute

let analyze_program_name fn =
  if not (is_relative fn) then
    Absolute
  else if String.contains fn '/' || (Sys.win32 && String.contains fn '\\') then
    Relative_to_current_dir
  else
    In_path
# 1105 "boot.ml"
end
module Hashtbl : sig
# 1 "src/stdune/hashtbl.mli"
include module type of struct include MoreLabels.Hashtbl end

val add : ('a, 'b) t -> 'a -> 'b -> unit

val find : ('a, 'b) t -> 'a -> 'b option
val find_or_add : ('a, 'b) t -> 'a -> f:('a -> 'b) -> 'b

val fold  : ('a, 'b) t -> init:'c -> f:(      'b -> 'c -> 'c) -> 'c
val foldi : ('a, 'b) t -> init:'c -> f:('a -> 'b -> 'c -> 'c) -> 'c
# 1118 "boot.ml"
end = struct
# 1 "src/stdune/hashtbl.ml"
include struct
  [@@@warning "-32"]

  let find_opt t key =
    match MoreLabels.Hashtbl.find t key with
    | x -> Some x
    | exception Not_found -> None
end

include MoreLabels.Hashtbl

let find = find_opt

let find_or_add t key ~f =
  match find t key with
  | Some x -> x
  | None ->
    let x = f key in
    add t ~key ~data:x;
    x

let add t key data = add t ~key ~data

let foldi t ~init ~f = fold  t ~init ~f:(fun ~key ~data acc -> f key data acc)
let fold  t ~init ~f = foldi t ~init ~f:(fun _ x -> f x)
# 1146 "boot.ml"
end
module Int : sig
# 1 "src/stdune/int.mli"
type t = int
val compare : t -> t -> Ordering.t
# 1152 "boot.ml"
end = struct
# 1 "src/stdune/int.ml"
type t = int
let compare (a : int) b : Ordering.t =
  if a < b then
    Lt
  else if a = b then
    Eq
  else
    Gt
# 1163 "boot.ml"
end
module Io : sig
# 1 "src/stdune/io.mli"
(** IO operations *)

val open_in  : ?binary:bool (* default true *) -> string -> in_channel
val open_out : ?binary:bool (* default true *) -> string -> out_channel

val close_in  : in_channel  -> unit
val close_out : out_channel -> unit

val with_file_in  : ?binary:bool (* default true *) -> string -> f:(in_channel -> 'a) -> 'a
val with_file_out : ?binary:bool (* default true *) -> string -> f:(out_channel -> 'a) -> 'a

val with_lexbuf_from_file : string -> f:(Lexing.lexbuf -> 'a) -> 'a

val lines_of_file : string -> string list

val read_file : string -> string
val write_file : string -> string -> unit

val compare_files : string -> string -> Ordering.t

val write_lines : string -> string list -> unit

val copy_channels : in_channel -> out_channel -> unit

val copy_file : src:string -> dst:string -> unit

val read_all : in_channel -> string
# 1194 "boot.ml"
end = struct
# 1 "src/stdune/io.ml"
module P = Pervasives

let open_in ?(binary=true) fn =
  if binary then P.open_in_bin fn else P.open_in fn

let open_out ?(binary=true) fn =
  if binary then P.open_out_bin fn else P.open_out fn

let close_in  = close_in
let close_out = close_out

let with_file_in ?binary fn ~f =
  Exn.protectx (open_in ?binary fn) ~finally:close_in ~f

let with_file_out ?binary fn ~f =
  Exn.protectx (open_out ?binary fn) ~finally:close_out ~f

let with_lexbuf_from_file fn ~f =
  with_file_in fn ~f:(fun ic ->
    let lb = Lexing.from_channel ic in
    lb.lex_curr_p <-
      { pos_fname = fn
      ; pos_lnum  = 1
      ; pos_bol   = 0
      ; pos_cnum  = 0
      };
    f lb)

let input_lines =
  let rec loop ic acc =
    match input_line ic with
    | exception End_of_file -> List.rev acc
    | line ->
       loop ic (line :: acc)
  in
  fun ic -> loop ic []

let read_all ic =
  let len = in_channel_length ic in
  really_input_string ic len

let read_file fn = with_file_in fn ~f:read_all

let lines_of_file fn = with_file_in fn ~f:input_lines ~binary:false

let write_file fn data = with_file_out fn ~f:(fun oc -> output_string oc data)

let write_lines fn lines =
  with_file_out fn ~f:(fun oc ->
    List.iter ~f:(fun line ->
      output_string oc line;
      output_string oc "\n"
    ) lines
  )

let copy_channels =
  let buf_len = 65536 in
  let buf = Bytes.create buf_len in
  let rec loop ic oc =
    match input ic buf 0 buf_len with
    | 0 -> ()
    | n -> output oc buf 0 n; loop ic oc
  in
  loop

let copy_file ~src ~dst =
  with_file_in src ~f:(fun ic ->
    let perm = (Unix.fstat (Unix.descr_of_in_channel ic)).st_perm in
    Exn.protectx (P.open_out_gen
                    [Open_wronly; Open_creat; Open_trunc; Open_binary]
                    perm
                    dst)
      ~finally:close_out
      ~f:(fun oc ->
        copy_channels ic oc))

(* TODO: diml: improve this *)
let compare_files fn1 fn2 = String.compare (read_file fn1) (read_file fn2)
# 1275 "boot.ml"
end
module Result : sig
# 1 "src/stdune/result.mli"
(** Result type *)

type ('a, 'error) t = ('a, 'error) Caml.result =
  | Ok    of 'a
  | Error of 'error

val is_ok    : _ t -> bool
val is_error : _ t -> bool

val ok_exn : ('a, exn) t -> 'a

module O : sig
  val ( >>| ) : ('a, 'error) t -> ('a -> 'b) -> ('b, 'error) t
  val ( >>= ) : ('a, 'error) t -> ('a -> ('b, 'error) t) -> ('b, 'error) t
end

val map  : ('a, 'error) t -> f:('a -> 'b) -> ('b, 'error) t
val bind : ('a, 'error) t -> f:('a -> ('b, 'error) t) -> ('b, 'error) t

val map_error : ('a, 'error1) t -> f:('error1 -> 'error2) -> ('a, 'error2) t

val all : ('a, 'error) t list -> ('a list, 'error) t

val concat_map
  :  'a list
  -> f:('a -> ('b list, 'error) t)
  -> ('b list, 'error) t

(** For compatibility with some other code *)
type ('a, 'error) result = ('a, 'error) t
# 1309 "boot.ml"
end = struct
# 1 "src/stdune/result.ml"
type ('a, 'error) t = ('a, 'error) Caml.result =
  | Ok    of 'a
  | Error of 'error

let is_ok = function
  | Ok    _ -> true
  | Error _ -> false

let is_error = function
  | Ok    _ -> false
  | Error _ -> true

let ok_exn = function
  | Ok    x -> x
  | Error e -> raise e

let bind t ~f =
  match t with
  | Ok x -> f x
  | Error _ as t -> t

let map x ~f =
  match x with
  | Ok x -> Ok (f x)
  | Error _ as x -> x

let map_error x ~f =
  match x with
  | Ok _ as res -> res
  | Error x -> Error (f x)

module O = struct
  let ( >>= ) t f = bind t ~f
  let ( >>| ) t f = map  t ~f
end

open O

let all =
  let rec loop acc = function
    | [] -> Ok (List.rev acc)
    | t :: l ->
      t >>= fun x ->
      loop (x :: acc) l
  in
  fun l -> loop [] l

let concat_map =
  let rec loop f acc = function
    | [] -> Ok (List.rev acc)
    | x :: l ->
      f x >>= fun y ->
      loop f (List.rev_append y acc) l
  in
  fun l ~f -> loop f [] l

type ('a, 'error) result = ('a, 'error) t
# 1369 "boot.ml"
end
module Map_intf = struct
# 1 "src/stdune/map_intf.ml"
module type S = sig
  type key
  and (+'a) t
  val empty     : 'a t
  val is_empty  : 'a t -> bool
  val mem       : 'a t -> key -> bool
  val add       : 'a t -> key -> 'a -> 'a t
  val update    : 'a t -> key -> f:('a option -> 'a option) -> 'a t
  val singleton : key -> 'a -> 'a t
  val remove    : 'a t -> key -> 'a t
  val add_multi : 'a list t -> key -> 'a -> 'a list t

  val merge
    :  'a t
    -> 'b t
    -> f:(key -> 'a option -> 'b option -> 'c option)
    -> 'c t
  val union
    :  'a t
    -> 'a t
    -> f:(key -> 'a -> 'a -> 'a option)
    -> 'a t

  val compare : 'a t -> 'a t -> compare:('a -> 'a -> Ordering.t) -> Ordering.t
  val equal   : 'a t -> 'a t -> equal:('a -> 'a -> bool) -> bool

  val iter  : 'a t -> f:(       'a -> unit) -> unit
  val iteri : 'a t -> f:(key -> 'a -> unit) -> unit
  val fold
    :  'a t
    -> init:'b
    -> f:('a -> 'b -> 'b)
    -> 'b
  val foldi
    :  'a t
    -> init:'b
    -> f:(key -> 'a -> 'b -> 'b)
    -> 'b

  val for_all    : 'a t -> f:(       'a -> bool) -> bool
  val for_alli   : 'a t -> f:(key -> 'a -> bool) -> bool
  val exists     : 'a t -> f:(       'a -> bool) -> bool
  val existsi    : 'a t -> f:(key -> 'a -> bool) -> bool
  val filter     : 'a t -> f:(       'a -> bool) -> 'a t
  val filteri    : 'a t -> f:(key -> 'a -> bool) -> 'a t
  val partition  : 'a t -> f:(       'a -> bool) -> 'a t * 'a t
  val partitioni : 'a t -> f:(key -> 'a -> bool) -> 'a t * 'a t

  val cardinal  : 'a t -> int

  val to_list : 'a t -> (key * 'a) list
  val of_list : (key * 'a) list -> ('a t, key * 'a * 'a) Result.t
  val of_list_exn : (key * 'a) list -> 'a t

  val of_list_multi  : (key * 'a) list -> 'a list t
  val of_list_reduce : (key * 'a) list -> f:('a -> 'a -> 'a) -> 'a t

  val keys   : 'a t -> key list
  val values : 'a t -> 'a list

  val min_binding : 'a t -> (key * 'a) option
  val max_binding : 'a t -> (key * 'a) option
  val choose      : 'a t -> (key * 'a) option

  val split: 'a t -> key -> 'a t * 'a option * 'a t
  val find : 'a t -> key -> 'a option

  val map  : 'a t -> f:(       'a -> 'b) -> 'b t
  val mapi : 'a t -> f:(key -> 'a -> 'b) -> 'b t

  val filter_map  : 'a t -> f:(       'a -> 'b option) -> 'b t
  val filter_mapi : 'a t -> f:(key -> 'a -> 'b option) -> 'b t
end
# 1446 "boot.ml"
end
module Option : sig
# 1 "src/stdune/option.mli"
(** Optional values *)

type 'a t = 'a option =
  | None
  | Some of 'a

module O : sig
  val (>>|) : 'a t -> ('a -> 'b  ) -> 'b t
  val (>>=) : 'a t -> ('a -> 'b t) -> 'b t
end

val map  : 'a t -> f:('a -> 'b  ) -> 'b t
val bind : 'a t -> f:('a -> 'b t) -> 'b t

val iter : 'a t -> f:('a -> unit) -> unit

val value : 'a t -> default:'a -> 'a
val value_exn : 'a t -> 'a

val some : 'a -> 'a t
val some_if : bool -> 'a -> 'a t

val is_some : _ t -> bool
val is_none : _ t -> bool

val both : 'a t -> 'b t -> ('a * 'b) t

val to_list : 'a t -> 'a list
# 1478 "boot.ml"
end = struct
# 1 "src/stdune/option.ml"
type 'a t = 'a option =
  | None
  | Some of 'a

module O = struct
  let (>>|) t f =
    match t with
    | None   -> None
    | Some a -> Some (f a)

  let (>>=) t f =
    match t with
    | None   -> None
    | Some a -> f a
end

let map  t ~f = O.(>>|) t f
let bind t ~f = O.(>>=) t f

let iter t ~f =
  match t with
  | None -> ()
  | Some x -> f x

let value t ~default =
  match t with
  | Some x -> x
  | None -> default

let value_exn = function
  | Some x -> x
  | None -> invalid_arg "Option.value_exn"

let some x = Some x

let some_if cond x =
  if cond then Some x else None

let is_some = function
  | None   -> false
  | Some _ -> true

let is_none = function
  | None   -> true
  | Some _ -> false

let both x y =
  match x, y with
  | Some x, Some y -> Some (x, y)
  | _ -> None

let to_list = function
  | None -> []
  | Some x -> [x]
# 1535 "boot.ml"
end
module Map : sig
# 1 "src/stdune/map.mli"
module type S = Map_intf.S

module Make(Key : Comparable.S) : S with type key = Key.t
# 1542 "boot.ml"
end = struct
# 1 "src/stdune/map.ml"
module type S = Map_intf.S

module Make(Key : Comparable.S) : S with type key = Key.t = struct
  module M = MoreLabels.Map.Make(struct
      type t = Key.t
      let compare a b = Ordering.to_int (Key.compare a b)
    end)

  include struct
    [@@@warning "-32"]

    let find_opt key t =
      match M.find key t with
      | x -> Some x
      | exception Not_found -> None

    let to_opt f t =
      match f t with
      | x -> Some x
      | exception Not_found -> None
    let choose_opt t = to_opt M.choose t
    let min_binding_opt t = to_opt M.min_binding t
    let max_binding_opt t = to_opt M.max_binding t

    let update ~key ~f t =
      match f (find_opt key t) with
      | None -> M.remove key t
      | Some v -> M.add t ~key ~data:v

    let union ~f a b =
      M.merge a b ~f:(fun k a b ->
        match a, b with
        | None  , None   -> None
        | Some v, None
        | None  , Some v -> Some v
        | Some a, Some b -> f k a b)
  end

  include M

  let find key t = find_opt t key

  let mem t k = mem k t
  let add t k v = add ~key:k ~data:v t
  let update t k ~f = update ~key:k ~f t
  let remove t k = remove k t

  let add_multi t key x =
    let l = Option.value (find t key) ~default:[] in
    add t key (x :: l)

  let merge a b ~f = merge a b ~f
  let union a b ~f = union a b ~f

  let compare a b ~compare =
    M.compare a b ~cmp:(fun a b -> Ordering.to_int (compare a b))
    |> Ordering.of_int
  let equal a b ~equal = M.equal a b ~cmp:equal

  let iteri t ~f = iter  t ~f:(fun ~key ~data -> f key data)
  let iter  t ~f = iteri t ~f:(fun _ x -> f x)
  let foldi t ~init ~f = fold  t ~init ~f:(fun ~key ~data acc -> f key data acc)
  let fold  t ~init ~f = foldi t ~init ~f:(fun _ x acc -> f x acc)

  let for_alli   t ~f = for_all    t ~f
  let for_all    t ~f = for_alli   t ~f:(fun _ x -> f x)
  let existsi    t ~f = exists     t ~f
  let exists     t ~f = existsi    t ~f:(fun _ x -> f x)
  let filteri    t ~f = filter     t ~f
  let filter     t ~f = filteri    t ~f:(fun _ x -> f x)
  let partitioni t ~f = partition  t ~f
  let partition  t ~f = partitioni t ~f:(fun _ x -> f x)

  let to_list = bindings

  let of_list =
    let rec loop acc = function
      | [] -> Result.Ok acc
      | (k, v) :: l ->
        match find acc k with
        | None       -> loop (add acc k v) l
        | Some v_old -> Error (k, v_old, v)
    in
    fun l -> loop empty l

  let of_list_exn l =
    match of_list l with
    | Ok    x -> x
    | Error _ -> invalid_arg "Map.of_list_exn"

  let of_list_reduce l ~f =
    List.fold_left l ~init:empty ~f:(fun acc (key, data) ->
      match find acc key with
      | None   -> add acc key data
      | Some x -> add acc key (f x data))

  let of_list_multi l =
    List.fold_left (List.rev l) ~init:empty ~f:(fun acc (key, data) ->
      add_multi acc key data)

  let keys   t = foldi t ~init:[] ~f:(fun k _ l -> k :: l) |> List.rev
  let values t = foldi t ~init:[] ~f:(fun _ v l -> v :: l) |> List.rev

  let min_binding = min_binding_opt
  let max_binding = max_binding_opt
  let choose      = choose_opt

  let split k t = split t k

  let map  t ~f = map  t ~f
  let mapi t ~f = mapi t ~f

  let filter_mapi t ~f =
    merge t empty ~f:(fun key data _always_none ->
      match data with
      | None      -> assert false
      | Some data -> f key data)
  let filter_map t ~f = filter_mapi t ~f:(fun _ x -> f x)
end
# 1664 "boot.ml"
end
module Or_exn : sig
# 1 "src/stdune/or_exn.mli"
(** Either a value or an exception *)

type 'a t = ('a, exn) Result.t
# 1671 "boot.ml"
end = struct
# 1 "src/stdune/or_exn.ml"
type 'a t = ('a, exn) Result.t
# 1675 "boot.ml"
end
module Set_intf = struct
# 1 "src/stdune/set_intf.ml"
module type S = sig
  type elt
  and t

  val empty          : t
  val is_empty       : t -> bool
  val mem            : t -> elt -> bool
  val add            : t -> elt -> t
  val singleton      : elt -> t
  val remove         : t -> elt -> t
  val union          : t -> t -> t
  val inter          : t -> t -> t
  val diff           : t -> t -> t
  val compare        : t -> t -> Ordering.t
  val equal          : t -> t -> bool
  val is_subset      : t -> of_:t -> bool
  val iter           : t -> f:(elt -> unit) -> unit
  val map            : t -> f:(elt -> elt) -> t
  val fold           : t -> init:'a -> f:(elt -> 'a -> 'a) -> 'a
  val for_all        : t -> f:(elt -> bool) -> bool
  val exists         : t -> f:(elt -> bool) -> bool
  val filter         : t -> f:(elt -> bool) -> t
  val partition      : t -> f:(elt -> bool) -> t * t
  val cardinal       : t -> int
  val min_elt        : t -> elt option
  val max_elt        : t -> elt option
  val choose         : t -> elt option
  val split          : t -> elt -> t * bool * t
  val of_list        : elt list -> t
  val to_list        : t -> elt list
end
# 1710 "boot.ml"
end
module Set : sig
# 1 "src/stdune/set.mli"
module type S = Set_intf.S

module Make(Elt : Comparable.S) : S with type elt = Elt.t
# 1717 "boot.ml"
end = struct
# 1 "src/stdune/set.ml"
module type S = Set_intf.S

module Make(Elt : Comparable.S) : S with type elt = Elt.t = struct
  module M = MoreLabels.Set.Make(struct
      type t = Elt.t
      let compare a b = Ordering.to_int (Elt.compare a b)
    end)

  include struct
    [@@@warning "-32"]
    (* [map] is only available since 4.04 *)
    let map ~f t =
      M.elements t
      |> List.map ~f
      |> M.of_list

    (* Since 4.05 *)
    let to_opt f t =
      match f t with
      | x -> Some x
      | exception Not_found -> None
    let choose_opt  = to_opt M.choose
    let min_elt_opt = to_opt M.min_elt
    let max_elt_opt = to_opt M.max_elt
  end

  include M

  let to_list = elements

  let mem t x = mem x t
  let add t x = add x t
  let remove t x = remove x t
  let compare a b = Ordering.of_int (compare a b)
  let is_subset t ~of_ = subset t of_
  let iter t ~f = iter t ~f
  let map t ~f = map t ~f
  let fold t ~init ~f = fold t ~init ~f
  let for_all t ~f = for_all t ~f
  let exists t ~f = exists t ~f
  let filter t ~f = filter t ~f
  let partition t ~f = partition t ~f
  let min_elt = min_elt_opt
  let max_elt = max_elt_opt
  let choose = choose_opt
  let split x t = split t x
end
# 1767 "boot.ml"
end
module Common = struct
# 1 "src/stdune/common.ml"
# 1771 "boot.ml"
end
module Import = struct
# 1 "src/stdune/import.ml"
# 1775 "boot.ml"
end
end
open XXXX
# 1 "src/stdune/stdune.ml"
module Ansi_color = Ansi_color
module Array      = Array
module Comparable = Comparable
module Either     = Either
module Exn        = Exn
module Filename   = Filename
module Hashtbl    = Hashtbl
module Int        = Int
module Io         = Io
module List       = List
module Map        = Map
module Option     = Option
module Or_exn     = Or_exn
module Ordering   = Ordering
module Pp         = Pp
module Result     = Result
module Set        = Set
module Staged     = Staged
module String     = String
module Char       = Char

external reraise : exn -> _ = "%reraise"

let compare a b = Ordering.of_int (compare a b)

(* The following types are re-exported here so that they are always
   available in scope *)

type ('a, 'error) result = ('a, 'error) Result.t =
  | Ok    of 'a
  | Error of 'error

type ('a, 'b) either = ('a, 'b) Either.t =
  | Left  of 'a
  | Right of 'b

type ordering = Ordering.t =
  | Lt
  | Eq
  | Gt
# 1820 "boot.ml"
end
module Fiber = struct
module XXXX = struct
end
open XXXX
include (struct
# 1 "src/fiber/fiber.ml"
open Stdune

module Eq = struct
  type ('a, 'b) t = T : ('a, 'a) t

  let cast (type a) (type b) (T : (a, b) t) (x : a) : b = x
end

module Var0 = struct
  module Key = struct
    type 'a t = ..
  end

  module type T = sig
    type t
    type 'a Key.t += T : t Key.t
    val id : int
  end

  type 'a t = (module T with type t = 'a)

  let next = ref 0

  let create (type a) () =
    let n = !next in
    next := n + 1;
    let module M = struct
      type t = a
      type 'a Key.t += T : t Key.t
      let id = n
    end in
    (module M : T with type t = a)

  let id (type a) (module M : T with type t = a) = M.id

  let eq (type a) (type b)
        (module A : T with type t = a)
        (module B : T with type t = b) : (a, b) Eq.t =
    match A.T with
    | B.T -> Eq.T
    | _ -> assert false
end

module Binding = struct
  type t = T : 'a Var0.t * 'a -> t
end

module Int_map = Map.Make(Int)

module Execution_context : sig
  type t

  val create_initial : unit -> t
  val forward_error : t -> exn -> unit

  val add_refs : t -> int -> unit
  val deref : t -> unit

  (* Create a new context with a new referebce count. [on_release] is called when the
     context is no longer used. *)
  val create_sub
    :  t
    -> on_release:(unit -> unit)
    -> t

  val set_error_handler
    :  t
    -> on_error:(exn -> unit)
    -> t

  val vars : t -> Binding.t Int_map.t
  val set_vars : t -> Binding.t Int_map.t -> t
end = struct
  type t =
    { on_error : exn -> unit (* This callback must never raise *)
    ; fibers   : int ref (* Number of fibers running in this execution
                            context *)
    ; vars     : Binding.t Int_map.t
    ; on_release : unit -> unit
    }

  let vars t = t.vars
  let set_vars t vars = { t with vars }

  let create_initial () =
    { on_error   = reraise
    ; fibers     = ref 1
    ; vars       = Int_map.empty
    ; on_release = ignore
    }

  let add_refs t n = t.fibers := !(t.fibers) + n

  let deref t =
    let n = !(t.fibers) - 1 in
    assert (n >= 0);
    t.fibers := n;
    if n = 0 then t.on_release ()

  let forward_error t exn =
    let bt = Printexc.get_raw_backtrace () in
    try
      t.on_error exn
    with exn2 ->
      (* We can't abort the execution at this point, so we just dump
         the error on stderr *)
      let bt2 = Printexc.get_backtrace () in
      let s =
        (Printf.sprintf "%s\n%s\nOriginal exception was: %s\n%s"
           (Printexc.to_string exn2) bt2
           (Printexc.to_string exn) (Printexc.raw_backtrace_to_string bt))
        |> String.split_lines
        |> List.map ~f:(Printf.sprintf "| %s")
        |> String.concat ~sep:"\n"
      in
      let line = String.make 71 '-' in
      Format.eprintf
        "/%s\n\
         | @{<error>Internal error@}: \
         Fiber.Execution_context.forward_error: error handler raised.\n\
         %s\n\
         \\%s@."
        line s line

  let forward_error t exn =
    forward_error t exn;
    deref t

  let create_sub t ~on_release =
    { t with on_release; fibers = ref 1 }

  let set_error_handler t ~on_error =
    { t with on_error }
end

module EC = Execution_context

type 'a t = Execution_context.t -> ('a -> unit) -> unit

let return x _ k = k x

let never _ _ = ()

let catch f ctx k =
  try
    f () ctx k
  with exn ->
    EC.forward_error ctx exn

module O = struct
  let (>>>) a b ctx k =
    a ctx (fun () -> b ctx k)

  let (>>=) t f ctx k =
    t ctx (fun x -> f x ctx k)

  let (>>|) t f ctx k =
    t ctx (fun x -> k (f x))
end

open O

let both a b =
  a >>= fun x ->
  b >>= fun y ->
  return (x, y)

let all l =
  let rec loop l acc =
    match l with
    | [] -> return (List.rev acc)
    | t :: l -> t >>= fun x -> loop l (x :: acc)
  in
  loop l []

let all_unit l = List.fold_left l ~init:(return ()) ~f:(>>>)

type ('a, 'b) fork_and_join_state =
  | Nothing_yet
  | Got_a of 'a
  | Got_b of 'b

let fork_and_join fa fb ctx k =
  let state = ref Nothing_yet in
  EC.add_refs ctx 1;
  begin
    try
      fa () ctx (fun a ->
        match !state with
        | Nothing_yet -> EC.deref ctx; state := Got_a a
        | Got_a _ -> assert false
        | Got_b b -> k (a, b))
    with exn ->
      EC.forward_error ctx exn
  end;
  fb () ctx (fun b ->
    match !state with
    | Nothing_yet -> EC.deref ctx; state := Got_b b
    | Got_a a -> k (a, b)
    | Got_b _ -> assert false)

let fork_and_join_unit fa fb ctx k =
  let state = ref Nothing_yet in
  EC.add_refs ctx 1;
  begin
    try
      fa () ctx (fun () ->
        match !state with
        | Nothing_yet -> EC.deref ctx; state := Got_a ()
        | Got_a _ -> assert false
        | Got_b b -> k b)
    with exn ->
      EC.forward_error ctx exn
  end;
  fb () ctx (fun b ->
    match !state with
    | Nothing_yet -> EC.deref ctx; state := Got_b b
    | Got_a () -> k b
    | Got_b _ -> assert false)

let list_of_option_array =
  let rec loop arr i acc =
    if i = 0 then
      acc
    else
      let i = i - 1 in
      match arr.(i) with
      | None -> assert false
      | Some x ->
        loop arr i (x :: acc)
  in
  fun a -> loop a (Array.length a) []

let parallel_map l ~f ctx k =
  match l with
  | [] -> k []
  | [x] -> f x ctx (fun x -> k [x])
  | _ ->
    let n = List.length l in
    EC.add_refs ctx (n - 1);
    let left_over = ref n in
    let results = Array.make n None in
    List.iteri l ~f:(fun i x ->
      try
        f x ctx (fun y ->
          results.(i) <- Some y;
          decr left_over;
          if !left_over = 0 then
            k (list_of_option_array results)
          else
            EC.deref ctx)
      with exn ->
        EC.forward_error ctx exn)

let parallel_iter l ~f ctx k =
  match l with
  | [] -> k ()
  | [x] -> f x ctx k
  | _ ->
    let n = List.length l in
    EC.add_refs ctx (n - 1);
    let left_over = ref n in
    let k () =
      decr left_over;
      if !left_over = 0 then k () else EC.deref ctx
    in
    List.iter l ~f:(fun x ->
      try
        f x ctx k
      with exn ->
        EC.forward_error ctx exn)

module Var = struct
  include Var0

  let find ctx var =
    match Int_map.find (EC.vars ctx) (id var) with
    | None -> None
    | Some (Binding.T (var', v)) ->
      let eq = eq var' var in
      Some (Eq.cast eq v)

  let find_exn ctx var =
    match Int_map.find (EC.vars ctx) (id var) with
    | None -> failwith "Fiber.Var.find_exn"
    | Some (Binding.T (var', v)) ->
      let eq = eq var' var in
      Eq.cast eq v

  let get     var ctx k = k (find     ctx var)
  let get_exn var ctx k = k (find_exn ctx var)

  let set (type a) (var : a t) x fiber ctx k =
    let (module M) = var in
    let data = Binding.T (var, x) in
    let ctx = EC.set_vars ctx (Int_map.add (EC.vars ctx) M.id data) in
    fiber ctx k
end

let with_error_handler f ~on_error ctx k =
  let on_error exn =
    try
      on_error exn
    with exn ->
      (* Increase the ref-counter of the parent context since this error doesn't originate
         from a fiber and so doesn't change the number of running fibers. *)
      EC.add_refs ctx 1;
      EC.forward_error ctx exn
  in
  let ctx = EC.set_error_handler ctx ~on_error in
  try
    f () ctx k
  with exn ->
    EC.forward_error ctx exn

let wait_errors t ctx k =
  let result = ref (Result.Error ()) in
  let on_release () =
    try
      k !result
    with exn ->
      EC.forward_error ctx exn
  in
  let sub_ctx = EC.create_sub ctx ~on_release in
  t sub_ctx (fun x ->
    result := Ok x;
    EC.deref sub_ctx)

let fold_errors f ~init ~on_error =
  let acc = ref init in
  let on_error exn =
    acc := on_error exn !acc
  in
  wait_errors (with_error_handler f ~on_error)
  >>| function
  | Ok _ as ok -> ok
  | Error ()   -> Error !acc

let collect_errors f =
  fold_errors f
    ~init:[]
    ~on_error:(fun e l -> e :: l)

let finalize f ~finally =
  wait_errors (catch f) >>= fun res ->
  finally () >>= fun () ->
  match res with
  | Ok x -> return x
  | Error () -> never

module Handler = struct
  type 'a t =
    { run : 'a -> unit
    ; ctx : Execution_context.t
    }

  let run t x =
    try
      t.run x
    with exn ->
      EC.forward_error t.ctx exn
end

module Ivar = struct
  type 'a state =
    | Full  of 'a
    | Empty of 'a Handler.t Queue.t

  type 'a t = { mutable state : 'a state }

  let create () = { state = Empty (Queue.create ()) }

  let fill t x _ctx k =
    match t.state with
    | Full  _ -> failwith "Fiber.Ivar.fill"
    | Empty q ->
      t.state <- Full x;
      Queue.iter
        (fun handler ->
           Handler.run handler x)
        q;
      k ()

  let read t ctx k =
    match t.state with
    | Full  x -> k x
    | Empty q ->
      Queue.push { Handler. run = k; ctx } q

  let peek t =
    match t.state with
    | Full  x -> Some x
    | Empty _ -> None
end

module Future = struct
  type 'a t = 'a Ivar.t

  let wait = Ivar.read
  let peek = Ivar.peek
end

let fork f ctx k =
  let ivar = Ivar.create () in
  EC.add_refs ctx 1;
  begin
    try
      f () ctx (fun x -> Ivar.fill ivar x ctx ignore)
    with exn ->
      EC.forward_error ctx exn
  end;
  k ivar

let nfork_map l ~f ctx k =
  match l with
  | [] -> k []
  | [x] -> fork (fun () -> f x) ctx (fun ivar -> k [ivar])
  | l ->
    let n = List.length l in
    EC.add_refs ctx (n - 1);
    let ivars =
      List.map l ~f:(fun x ->
        let ivar = Ivar.create () in
        begin
          try
            f x ctx (fun x -> Ivar.fill ivar x ctx ignore)
          with exn ->
            EC.forward_error ctx exn
        end;
        ivar)
    in
    k ivars

let nfork l : _ Future.t list t = nfork_map l ~f:(fun f -> f ())

module Mutex = struct
  type t =
    { mutable locked  : bool
    ; mutable waiters : unit Handler.t Queue.t
    }

  let lock t ctx k =
    if t.locked then
      Queue.push { Handler. run = k; ctx } t.waiters
    else begin
      t.locked <- true;
      k ()
    end

  let unlock t _ctx k =
    assert t.locked;
    if Queue.is_empty t.waiters then
      t.locked <- false
    else
      Handler.run (Queue.pop t.waiters) ();
    k ()

  let with_lock t f =
    lock t >>= fun () ->
    finalize f ~finally:(fun () -> unlock t)

  let create () =
    { locked  = false
    ; waiters = Queue.create ()
    }
end

let suspended = ref []

let yield () ctx k = suspended := { Handler. ctx; run = k } :: !suspended

exception Never

let run t =
  let result = ref None in
  let ctx = EC.create_initial () in
  begin
    try
      t ctx (fun x -> result := Some x)
    with exn ->
      EC.forward_error ctx exn
  end;
  let rec loop () =
    match !result with
    | Some x -> x
    | None ->
      match List.rev !suspended with
      | [] -> raise Never
      | to_run ->
        suspended := [];
        List.iter to_run ~f:(fun h -> Handler.run h ());
        loop ()
  in
  loop ()
# 2322 "boot.ml"
end : sig
# 1 "src/fiber/fiber.mli"
(** Concurrency library *)

open Stdune

(** {1 Generals} *)

(** Type of fiber. A fiber represent a suspended computation. Note that using the same
    fiber twice will execute it twice, which is probably not what you want. To share the
    result of a fiber, use an [Ivar.t].  *)
type 'a t

(** Create a fiber that has already terminated. *)
val return : 'a -> 'a t

(** Fiber that never completes. *)
val never : 'a t

module O : sig
  (** [>>>] is a sequencing operator. [a >>> b] is the fiber that
      first executes [a] and then [b]. *)
  val (>>>) : unit t -> 'a t -> 'a t

  (** [>>=] is similar to [>>>] except that the result of the first
      fiber is used to create the second one. *)
  val (>>=) : 'a t -> ('a -> 'b t) -> 'b t

  (** [t >>| f] is the same as [t >>= fun x -> return (f x)] but
      slightly more efficient. *)
  val (>>|) : 'a t -> ('a -> 'b) -> 'b t
end

(** {1 Forking execution} *)

module Future : sig
  type 'a fiber

  (** A future represent a promise that will eventually yield a value. It is used to
      represent the result of a fiber running in the background. *)
  type 'a t

  (** Wait for the given future to yield a value. *)
  val wait : 'a t -> 'a fiber

  (** Return [Some x] if [t] has already returned. *)
  val peek : 'a t -> 'a option
end with type 'a fiber := 'a t

(** [fork f] creates a sub-fiber and return a [Future.t] to wait its result. *)
val fork : (unit -> 'a t) -> 'a Future.t t

(** [nfork l] is similar to [fork] but creates [n] sub-fibers. *)
val nfork : (unit -> 'a t) list -> 'a Future.t list t

(** [nfork_map l ~f] is the same as [nfork (List.map l ~f:(fun x () -> f x))] but more
    efficient. *)
val nfork_map : 'a list -> f:('a -> 'b t) -> 'b Future.t list t

(** {1 Joining} *)

(** The following combinators are helpers to combine the result of several fibers into
    one. Note that they do not introduce parallelism. *)

val both : 'a t -> 'b t -> ('a * 'b) t
val all : 'a t list -> 'a list t
val all_unit : unit t list -> unit t

(** {1 Forking + joining} *)

(** The following functions combine forking 2 or more fibers followed by joining the
    results. For every function, we give an equivalent implementation using the more basic
    functions as documentation. Note however that these functions are implemented as
    primitives and so are more efficient that the suggested implementation. *)

(** For two fibers and wait for their results:

    {[
      let fork_and_join f g =
        fork f >>= fun a ->
        fork g >>= fun b ->
        both (Future.wait a) (Future.wait b)
      ]}
*)
val fork_and_join : (unit -> 'a t) -> (unit -> 'b t) -> ('a * 'b) t

(** Same but assume the first fiber returns [unit]:

    {[
      let fork_and_join_unit f g =
        fork f >>= fun a ->
        fork g >>= fun b ->
        Future.wait a >>> Future.wait b
    ]}
*)
val fork_and_join_unit : (unit -> unit t) -> (unit -> 'a t) -> 'a t

(** Map a list in parallel:

    {[
      let parallel_map l ~f =
        nfork_map l ~f >>= fun futures ->
        all (List.map futures ~f:Future.wait)
    ]}
*)
val parallel_map : 'a list -> f:('a -> 'b t) -> 'b list t

(** Iter over a list in parallel:

    {[
      let parallel_iter l ~f =
        nfork_map l ~f >>= fun futures ->
        all_unit (List.map futures ~f:Future.wait)
    ]}
*)
val parallel_iter : 'a list -> f:('a -> unit t) -> unit t

(** {1 Local storage} *)

(** Variables local to a fiber *)
module Var : sig
  type 'a fiber = 'a t
  type 'a t

  (** Create a new variable *)
  val create : unit -> 'a t

  (** [get var] is a fiber that reads the value of [var] *)
  val get : 'a t -> 'a option fiber

  (** Same as [get] but raises if [var] is unset. *)
  val get_exn : 'a t -> 'a fiber

  (** [set var value fiber] sets [var] to [value] during the execution
      of [fiber].

      For instance, the following fiber always evaluate to [true]:

      {[
        set v x (get_exn v >>| fun y -> x = y)
      ]}
 *)
  val set : 'a t -> 'a -> 'b fiber -> 'b fiber
end with type 'a fiber := 'a t

(** {1 Error handling} *)

(** [with_error_handler f ~on_error] calls [on_error] for every exception raised during
    the execution of [f]. This include exceptions raised when calling [f ()] or during the
    execution of fibers after [f ()] has returned. Exceptions raised by [on_error] are
    passed on to the parent error handler.

    It is guaranteed that after the fiber has returned a value, [on_error] will never be
    called.
*)
val with_error_handler
  :  (unit -> 'a t)
  -> on_error:(exn -> unit)
  -> 'a t

(** If [t] completes without raising, then [wait_errors t] is the same as [t () >>| fun x
    -> Ok x]. However, if the execution of [t] is aborted by an exception, then
    [wait_errors t] will complete and yield [Error ()].

    Note that [wait_errors] only completes after all sub-fibers have completed. For
    instance, in the following code [wait_errors] will only complete after 3s:

    {[
      wait_errors
        (fork_and_join
           (fun () -> sleep 1 >>| fun () -> raise Exit)
           (fun () -> sleep 3))
    ]}

    same for this code:

    {[
      wait_errors
        (fork (fun () -> sleep 3) >>= fun _ ->
         raise Exit)
    }]
*)
val wait_errors : 'a t -> ('a, unit) Result.t t

(** [fold_errors f ~init ~on_error] calls [on_error] for every exception raised during the
    execution of [f]. This include exceptions raised when calling [f ()] or during the
    execution of fibers after [f ()] has returned.

    Exceptions raised by [on_error] are passed on to the parent error handler. *)
val fold_errors
  :  (unit -> 'a t)
  -> init:'b
  -> on_error:(exn -> 'b -> 'b)
  -> ('a, 'b) Result.t t

(** [collect_errors f] is:

    {[
      fold_errors f
        ~init:[]
        ~on_error:(fun e l -> e :: l)
    ]}
*)
val collect_errors
  :  (unit -> 'a t)
  -> ('a, exn list) Result.t t

(** [finalize f ~finally] runs [finally] after [f ()] has terminated,
    whether it fails or succeeds. *)
val finalize
  :  (unit -> 'a t)
  -> finally:(unit -> unit t)
  -> 'a t

(** {1 Synchronization} *)

(** Write once variables *)
module Ivar : sig
  type 'a fiber = 'a t

  (** A ivar is a synchronization variable that can be written only
      once. *)
  type 'a t

  (** Create a new empty ivar. *)
  val create : unit -> 'a t

  (** Read the contents of the ivar. *)
  val read : 'a t -> 'a fiber

  (** Fill the ivar with the following value. This can only be called
      once for a given ivar. *)
  val fill : 'a t -> 'a -> unit fiber

  (** Return [Some x] is [fill t x] has been called previously. *)
  val peek : 'a t -> 'a option
end with type 'a fiber := 'a t

module Mutex : sig
  type 'a fiber = 'a t
  type t
  val create : unit -> t
  val with_lock : t -> (unit -> 'a fiber) -> 'a fiber
end with type 'a fiber := 'a t

(** {1 Running fibers} *)

(** Wait for one iteration of the scheduler *)
val yield : unit -> unit t

(** [run t] runs a fiber until it yield a result. If it becomes clear that the execution
    of the fiber will never terminate, raise [Never]. *)
val run : 'a t -> 'a

exception Never
# 2578 "boot.ml"
end)
end
module Usexp = struct
module XXXX = struct
module Sexp_ast = struct
# 1 "src/usexp/sexp_ast.ml"
module Loc = struct
  type t =
    { start : Lexing.position
    ; stop  : Lexing.position
    }
end

type atom = A of string [@@unboxed]

type t =
  | Atom of Loc.t * atom
  | Quoted_string of Loc.t * string
  | List of Loc.t * t list
# 2598 "boot.ml"
end
module Parser_automaton_internal : sig
# 1 "src/usexp/parser_automaton_internal.mli"
(** Internal bits used by the generated automaton, not part of the public API *)

(** Internal state of the automaton *)
type 'a state

type 'a mode =
  | Single : Sexp_ast.t mode
  | Many   : Sexp_ast.t list mode

type stack
val empty_stack : stack

val new_state
  :  fname:string
  -> 'a mode
  -> 'a state

val mode : 'a state -> 'a mode

(** Number of characters fed to the parser *)
val offset : _ state -> int

(** Position in the text *)
val line   : _ state -> int
val column : _ state -> int

(** Whether there are some unclosed parentheses *)
val has_unclosed_paren : _ state -> bool

val set_error_state : _ state -> unit

val sexp_of_stack : stack -> Sexp_ast.t
val sexps_of_stack : stack -> Sexp_ast.t list

val automaton_state : _ state -> int

module Error : sig
  type t

  val position : t -> Lexing.position
  val message  : t -> string

  module Reason : sig
    type t =
      | Unexpected_char_parsing_hex_escape
      | Unexpected_char_parsing_dec_escape
      | Unterminated_quoted_string
      | Unterminated_block_comment
      | Escape_sequence_out_of_range
      | Unclosed_paren
      | Too_many_sexps
      | Closed_paren_without_opened
      | Comment_token_in_unquoted_atom
      | Sexp_comment_without_sexp
      | Unexpected_character_after_cr
      | No_sexp_found_in_input
      | Automaton_in_error_state
  end

  val raise : _ state -> at_eof:bool -> Reason.t -> _
end

exception Parse_error of Error.t

val set_automaton_state : _ state -> int -> unit

(** Advance the position counters. [advance_eol] is for when we read a newline
    character. *)
val advance     : _ state -> unit
val advance_eol : _ state -> unit

(** Number of opened #| *)
val block_comment_depth : _ state -> int

type 'a action         = 'a state -> char -> stack -> stack
type 'a epsilon_action = 'a state ->         stack -> stack

(** Add a character to the atom buffer. [add_quoted_atom_char] does the same for quoted
    atoms *)
val add_atom_char : _ action
val add_quoted_atom_char : _ action

(** Add a character that just follows a '\\' and the '\\' itself if necessary. *)
val add_escaped : _ action

(** [escaped_value <- escaped_value * 10 + (char - '0')]

    These functions make the assumption that [char] is between '0' and '9'.
    [add_dec_escape_char] also assumes the result doesn't overflow. The automaton
    definition must make sure this is the case.

    [add_last_dec_escape_char] also adds the resulting character to the atom buffer.
*)
val add_dec_escape_char      : _ action
val add_last_dec_escape_char : _ action

(** Same but for quoted strings inside comments. Useful because it can fail. *)
val comment_add_last_dec_escape_char : _ action

(** Same as [add_dec_escape_char] but for hexadicemal escape sequences *)
val add_hex_escape_char      : _ action
val add_last_hex_escape_char : _ action

(** Ignore one more full sexp to come *)
val start_sexp_comment : _ action

(** Add the first char of an unquoted atom. *)
val add_first_char      : _ action
val start_quoted_string : _ action

(** Takes note of a control character in quoted atoms or the uninterpreted characters of
    comments, for which there is no corresponding [add_*] call (a backslash and the x in
    "\xff" or any character in a line comment).  This does not get called for the opening
    ([start_quoted_string]) or closing ([push_quoted_atom]) quotes themselves.
*)
val add_token_char : _ action

val opening      : _ action
val closing      : _ action
val push_quoted_atom : _ action

val start_block_comment : _ action
val end_block_comment   : _ action

val start_line_comment : _ action
val end_line_comment   : _ epsilon_action

val eps_push_atom      : _ epsilon_action
val eps_add_first_char_hash      : _ epsilon_action
val eps_eoi_check      : _ epsilon_action

val eps_add_escaped_cr : _ epsilon_action
# 2734 "boot.ml"
end = struct
# 1 "src/usexp/parser_automaton_internal.ml"
type stack =
  | Empty
  | Open of Lexing.position * stack
  | Sexp of Sexp_ast.t * stack
let empty_stack = Empty

type 'a mode =
  | Single : Sexp_ast.t mode
  | Many   : Sexp_ast.t list mode

type 'a state =
  { mutable automaton_state : int
  ; mutable depth           : int
  ; (* Number of opened #| when parsing a block comment *)
    mutable block_comment_depth  : int
  ; (* Number of full sexp to ignore. This correspond to the number of consecutive #;.
       We don't have to track nested #; because they do not change the result. *)
    mutable ignoring       : int
  ; (* Only meaningful when [ignoring] > 0. Number of opened parentheses of the
       outermost sexp comment. *)
    mutable ignoring_depth : int
  ; (* When parsing an escape sequence of the form "\\NNN" or "\\XX", this accumulates
       the computed number *)
    mutable escaped_value  : int
  ; (* Buffer for accumulating atoms *)
    atom_buffer            : Buffer.t
  ; fname                  : string
  ; mutable full_sexps     : int
  ; mutable offset         : int (* global offset *)
  ; mutable line_number    : int
  ; mutable bol_offset     : int (* offset of beginning of line *)
  ; (* Starting positions of the current token *)
    mutable token_start_pos : Lexing.position
  ; mode                   : 'a mode
  }

(* these magic numbers are checked in gen_parser_automaton.ml:
   let () = assert (initial = 0)
   let () = assert (to_int Error = 1) *)
let initial_state = 0
let error_state   = 1

let new_state ~fname mode =
  { depth          = 0
  ; automaton_state = initial_state
  ; block_comment_depth  = 0
  ; ignoring       = 0
  ; ignoring_depth = 0
  ; escaped_value  = 0
  ; atom_buffer    = Buffer.create 128
  ; mode           = mode
  ; full_sexps     = 0
  ; offset         = 0
  ; line_number    = 1
  ; bol_offset     = 0
  ; fname          = fname
  ; token_start_pos = { pos_fname = fname; pos_cnum = 0; pos_lnum = 1; pos_bol = 0 }
  }

let mode t = t.mode

let offset t = t.offset
let line   t = t.line_number
let column t = t.offset - t.bol_offset

let position t =
  { Lexing.
    pos_fname = t.fname
  ; pos_cnum  = t.offset
  ; pos_lnum  = t.line_number
  ; pos_bol   = t.bol_offset
  }

let has_unclosed_paren state = state.depth > 0

let set_error_state state = state.automaton_state <- error_state

let sexp_of_stack : stack -> Sexp_ast.t = function
  | Sexp (sexp, Empty) -> sexp
  | _ -> failwith "Parser_automaton.sexp_of_stack"

let sexps_of_stack =
  let rec loop acc : stack -> Sexp_ast.t list = function
    | Empty -> acc
    | Open _ -> failwith "Parser_automaton.sexps_of_stack"
    | Sexp (sexp, stack) -> loop (sexp :: acc) stack
  in
  fun stack -> loop [] stack

let automaton_state state = state.automaton_state

module Error = struct
  type t =
    { position       : Lexing.position
    ; message        : string
    }

  exception Parse_error of t

  module Reason = struct
    (* To be kept in sync with the Error module in gen/gen_parser_automaton.ml *)
    type t =
      | Unexpected_char_parsing_hex_escape
      | Unexpected_char_parsing_dec_escape
      | Unterminated_quoted_string
      | Unterminated_block_comment
      | Escape_sequence_out_of_range
      | Unclosed_paren
      | Too_many_sexps
      | Closed_paren_without_opened
      | Comment_token_in_unquoted_atom
      | Sexp_comment_without_sexp
      | Unexpected_character_after_cr
      | No_sexp_found_in_input
      | Automaton_in_error_state
  end

  let raise state ~at_eof (reason : Reason.t) =
    set_error_state state;
    let message =
      (* These messages where choosen such that we can build the various Sexplib parsing
         functions on top of Parsexp and keep the same exceptions.

         At the time of writing this, a simple layer on top of parsexp to implement the
         sexplib API is passing all the sexplib tests.

         Note that parsexp matches the semantic of Sexp.parse which is slightly
         different from the ocamllex/ocamlyacc based parser of Sexplib. The latter one
         is less tested and assumed to be less used. *)
      match reason with
      | Unexpected_char_parsing_hex_escape ->
        "unterminated hexadecimal escape sequence"
      | Unexpected_char_parsing_dec_escape ->
        "unterminated decimal escape sequence"
      | Unterminated_quoted_string ->
        "unterminated quoted string"
      | Unterminated_block_comment ->
        "unterminated block comment"
      | Escape_sequence_out_of_range ->
        "escape sequence in quoted string out of range"
      | Unclosed_paren ->
        "unclosed parentheses at end of input"
      | Too_many_sexps ->
        "s-expression followed by data"
      | Closed_paren_without_opened ->
        "unexpected character: ')'"
      | Comment_token_in_unquoted_atom ->
        if Buffer.contents state.atom_buffer = "|" then
          "illegal end of comment"
        else
          "comment tokens in unquoted atom"
      | Sexp_comment_without_sexp ->
        "unterminated sexp comment"
      | Unexpected_character_after_cr ->
        if at_eof then
          "unexpected end of input after carriage return"
        else
          "unexpected character after carriage return"
      | No_sexp_found_in_input ->
        "no s-expression found in input"
      | Automaton_in_error_state ->
        failwith "Parsexp.Parser_automaton: parser is dead"
    in
    let position = position state in
    raise (Parse_error { position; message })

  let position       t = t.position
  let message        t = t.message
end

exception Parse_error = Error.Parse_error

type 'a action         = 'a state -> char -> stack -> stack
type 'a epsilon_action = 'a state ->         stack -> stack

let current_pos ?(delta=0) state : Lexing.position =
  let offset = state.offset + delta in
  { pos_fname = state.fname
  ; pos_lnum  = state.line_number
  ; pos_cnum  = offset
  ; pos_bol   = state.bol_offset
  }

let set_automaton_state state x = state.automaton_state <- x

let advance state = state.offset <- state.offset + 1

let advance_eol state =
  let newline_offset = state.offset in
  state.offset <- newline_offset + 1;
  state.bol_offset <- state.offset;
  state.line_number <- state.line_number + 1

let block_comment_depth state = state.block_comment_depth

let add_token_char _state _char stack = stack

let add_atom_char state c stack =
  Buffer.add_char state.atom_buffer c;
  stack

let add_quoted_atom_char state c stack =
  Buffer.add_char state.atom_buffer c;
  add_token_char state c stack

let check_new_sexp_allowed : type a. a state -> unit = fun state ->
  let is_single = match state.mode with Single -> true | _ -> false in
  if is_single && state.full_sexps > 0 && state.ignoring = 0 then
    Error.raise state ~at_eof:false Too_many_sexps

let add_first_char state char stack =
  check_new_sexp_allowed state;
  Buffer.add_char state.atom_buffer char;
  state.token_start_pos <- current_pos state;
  stack

let eps_add_first_char_hash state stack =
  check_new_sexp_allowed state;
  Buffer.add_char state.atom_buffer '#';
  state.token_start_pos <- current_pos state ~delta:(-1);
  stack

let start_quoted_string state _char stack =
  check_new_sexp_allowed state;
  state.token_start_pos <- current_pos state;
  stack

let add_escaped state c stack =
  let c' =
    match c with
    | 'n' -> '\n'
    | 'r' -> '\r'
    | 'b' -> '\b'
    | 't' -> '\t'
    | '\\' | '\'' | '"' -> c
    | _ -> Buffer.add_char state.atom_buffer '\\'; c
  in
  Buffer.add_char state.atom_buffer c';
  add_token_char state c stack

let eps_add_escaped_cr state stack =
  Buffer.add_char state.atom_buffer '\r';
  stack

let dec_val c = Char.code c - Char.code '0'

let hex_val c =
  match c with
  | '0'..'9' -> Char.code c - Char.code '0'
  | 'a'..'f' -> Char.code c - Char.code 'a' + 10
  | _        -> Char.code c - Char.code 'A' + 10

let add_dec_escape_char state c stack =
  state.escaped_value <- state.escaped_value * 10 + dec_val c;
  add_token_char state c stack

let add_last_dec_escape_char state c stack =
  let value = state.escaped_value * 10 + dec_val c in
  state.escaped_value <- 0;
  if value > 255 then Error.raise state ~at_eof:false Escape_sequence_out_of_range;
  Buffer.add_char state.atom_buffer (Char.unsafe_chr value);
  add_token_char state c stack

let comment_add_last_dec_escape_char state c stack =
  let value = state.escaped_value * 10 + dec_val c in
  state.escaped_value <- 0;
  if value > 255 then Error.raise state ~at_eof:false Escape_sequence_out_of_range;
  add_token_char state c stack

let add_hex_escape_char state c stack =
  state.escaped_value <- (state.escaped_value lsl 4) lor hex_val c;
  add_token_char state c stack

let add_last_hex_escape_char state c stack =
  let value = (state.escaped_value lsl 4) lor hex_val c in
  state.escaped_value <- 0;
  Buffer.add_char state.atom_buffer (Char.unsafe_chr value);
  add_token_char state c stack

let opening state _char stack =
  check_new_sexp_allowed state;
  state.depth <- state.depth + 1;
  if state.ignoring = 0 then
    Open (current_pos state, stack)
  else
    stack

let sexp_or_comment_added ~is_comment state stack ~delta:_ =
  if state.ignoring lor state.depth <> 0 then
    stack
  else begin
    if not is_comment then state.full_sexps <- state.full_sexps + 1;
    stack
  end

let sexp_added state stack ~delta =
  let stack = sexp_or_comment_added ~is_comment:false state stack ~delta in
  if state.ignoring <> 0 && state.ignoring_depth = state.depth then begin
    state.ignoring <- state.ignoring - 1;
    stack
  end else stack

let rec make_list stop acc : stack -> stack = function
  | Empty               -> assert false
  | Open (start, stack) -> Sexp (List ({ start; stop }, acc), stack)
  | Sexp (sexp, stack)  -> make_list stop (sexp :: acc) stack

let closing state _char stack =
  if state.depth > 0 then begin
    let stack =
      if state.ignoring = 0 then
        make_list (current_pos state ~delta:1) [] stack
      else
        stack
    in
    if state.ignoring <> 0 && state.ignoring_depth = state.depth then
      Error.raise state ~at_eof:false Sexp_comment_without_sexp;
    state.depth <- state.depth - 1;
    sexp_added state stack ~delta:1
  end else
    Error.raise state ~at_eof:false Closed_paren_without_opened

let make_loc ?(delta=0) state : Sexp_ast.Loc.t =
  { start = state.token_start_pos
  ; stop  = current_pos state ~delta
  }

let eps_push_atom state stack =
  let str = Buffer.contents state.atom_buffer in
  Buffer.clear state.atom_buffer;
  let stack =
    if state.ignoring = 0 then
      Sexp (Atom (make_loc state, A str), stack)
    else
      stack
  in
  sexp_added state stack ~delta:0

let push_quoted_atom state _char stack =
  let str = Buffer.contents state.atom_buffer in
  Buffer.clear state.atom_buffer;
  let stack =
    if state.ignoring = 0 then
      Sexp (Quoted_string (make_loc state ~delta:1, str), stack)
    else
      stack
  in
  sexp_added state stack ~delta:1

let start_sexp_comment state _char stack =
  if state.ignoring = 0 || state.ignoring_depth = state.depth then begin
    state.ignoring <- state.ignoring + 1;
    state.ignoring_depth <- state.depth
  end;
  stack

let start_block_comment state _char stack =
  state.block_comment_depth <- state.block_comment_depth + 1;
  stack

let end_block_comment state _char stack =
  state.block_comment_depth <- state.block_comment_depth - 1;
  stack

let start_line_comment _state _char stack = stack
let   end_line_comment _state       stack = stack

let eps_eoi_check : type a. a state -> stack -> stack = fun state stack ->
  if state.depth      > 0 then Error.raise state ~at_eof:true Unclosed_paren;
  if state.ignoring   > 0 then Error.raise state ~at_eof:true Sexp_comment_without_sexp;
  if state.full_sexps = 0 then (
    match state.mode with
    | Many -> ()
    | Single ->
      Error.raise state ~at_eof:true No_sexp_found_in_input
  );
  stack
# 3114 "boot.ml"
end
module Table : sig
# 1 "src/usexp/table.mli"
val transitions     : Obj.t Parser_automaton_internal.action         array
val transitions_eoi : Obj.t Parser_automaton_internal.epsilon_action array
# 3120 "boot.ml"
end = struct
# 1 "src/usexp/table.ml"
(* generated by ./gen/gen_parser_automaton.exe *)

open Parser_automaton_internal

let raise = Parser_automaton_internal.Error.raise

let tr_00 state char stack =
  let stack = add_first_char state char stack in
  set_automaton_state state 3;
  advance state;
  stack
let tr_01 state _char stack =
  set_automaton_state state 0;
  advance state;
  stack
let tr_02 state _char stack =
  set_automaton_state state 0;
  advance_eol state;
  stack
let tr_03 state _char stack =
  set_automaton_state state 2;
  advance state;
  stack
let tr_04 state char stack =
  let stack = start_quoted_string state char stack in
  set_automaton_state state 8;
  advance state;
  stack
let tr_05 state _char stack =
  set_automaton_state state 7;
  advance state;
  stack
let tr_06 state char stack =
  let stack = opening state char stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_07 state char stack =
  let stack = closing state char stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_08 state char stack =
  let stack = start_line_comment state char stack in
  set_automaton_state state 6;
  advance state;
  stack
let tr_09 state char stack =
  let stack = add_first_char state char stack in
  set_automaton_state state 5;
  advance state;
  stack
let tr_10 _state _char _stack =
  raise _state ~at_eof:false Automaton_in_error_state
let tr_11 _state _char _stack =
  raise _state ~at_eof:false Unexpected_character_after_cr
let tr_12 state char stack =
  let stack = add_atom_char state char stack in
  set_automaton_state state 3;
  advance state;
  stack
let tr_13 state _char stack =
  let stack = eps_push_atom state stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_14 state _char stack =
  let stack = eps_push_atom state stack in
  set_automaton_state state 0;
  advance_eol state;
  stack
let tr_15 state _char stack =
  let stack = eps_push_atom state stack in
  set_automaton_state state 2;
  advance state;
  stack
let tr_16 state char stack =
  let stack = eps_push_atom state stack in
  let stack = start_quoted_string state char stack in
  set_automaton_state state 8;
  advance state;
  stack
let tr_17 state char stack =
  let stack = add_atom_char state char stack in
  set_automaton_state state 4;
  advance state;
  stack
let tr_18 state char stack =
  let stack = eps_push_atom state stack in
  let stack = opening state char stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_19 state char stack =
  let stack = eps_push_atom state stack in
  let stack = closing state char stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_20 state char stack =
  let stack = eps_push_atom state stack in
  let stack = start_line_comment state char stack in
  set_automaton_state state 6;
  advance state;
  stack
let tr_21 state char stack =
  let stack = add_atom_char state char stack in
  set_automaton_state state 5;
  advance state;
  stack
let tr_22 _state _char _stack =
  raise _state ~at_eof:false Comment_token_in_unquoted_atom
let tr_23 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 6;
  advance state;
  stack
let tr_24 state _char stack =
  let stack = end_line_comment state stack in
  set_automaton_state state 0;
  advance_eol state;
  stack
let tr_25 state _char stack =
  let stack = end_line_comment state stack in
  set_automaton_state state 2;
  advance state;
  stack
let tr_26 state char stack =
  let stack = eps_add_first_char_hash state stack in
  let stack = add_atom_char state char stack in
  set_automaton_state state 3;
  advance state;
  stack
let tr_27 state _char stack =
  let stack = eps_add_first_char_hash state stack in
  let stack = eps_push_atom state stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_28 state _char stack =
  let stack = eps_add_first_char_hash state stack in
  let stack = eps_push_atom state stack in
  set_automaton_state state 0;
  advance_eol state;
  stack
let tr_29 state _char stack =
  let stack = eps_add_first_char_hash state stack in
  let stack = eps_push_atom state stack in
  set_automaton_state state 2;
  advance state;
  stack
let tr_30 state char stack =
  let stack = eps_add_first_char_hash state stack in
  let stack = eps_push_atom state stack in
  let stack = start_quoted_string state char stack in
  set_automaton_state state 8;
  advance state;
  stack
let tr_31 state char stack =
  let stack = eps_add_first_char_hash state stack in
  let stack = add_atom_char state char stack in
  set_automaton_state state 4;
  advance state;
  stack
let tr_32 state char stack =
  let stack = eps_add_first_char_hash state stack in
  let stack = eps_push_atom state stack in
  let stack = opening state char stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_33 state char stack =
  let stack = eps_add_first_char_hash state stack in
  let stack = eps_push_atom state stack in
  let stack = closing state char stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_34 state char stack =
  let stack = start_sexp_comment state char stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_35 state char stack =
  let stack = start_block_comment state char stack in
  set_automaton_state state 16;
  advance state;
  stack
let tr_36 state char stack =
  let stack = add_quoted_atom_char state char stack in
  set_automaton_state state 8;
  advance state;
  stack
let tr_37 state char stack =
  let stack = add_quoted_atom_char state char stack in
  set_automaton_state state 8;
  advance_eol state;
  stack
let tr_38 state char stack =
  let stack = push_quoted_atom state char stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_39 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 9;
  advance state;
  stack
let tr_40 state char stack =
  let stack = add_escaped state char stack in
  set_automaton_state state 8;
  advance state;
  stack
let tr_41 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 15;
  advance_eol state;
  stack
let tr_42 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 10;
  advance state;
  stack
let tr_43 state char stack =
  let stack = add_dec_escape_char state char stack in
  set_automaton_state state 11;
  advance state;
  stack
let tr_44 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 13;
  advance state;
  stack
let tr_45 state char stack =
  let stack = eps_add_escaped_cr state stack in
  let stack = add_quoted_atom_char state char stack in
  set_automaton_state state 8;
  advance state;
  stack
let tr_46 state char stack =
  let stack = eps_add_escaped_cr state stack in
  let stack = push_quoted_atom state char stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_47 state char stack =
  let stack = eps_add_escaped_cr state stack in
  let stack = add_token_char state char stack in
  set_automaton_state state 9;
  advance state;
  stack
let tr_48 _state _char _stack =
  raise _state ~at_eof:false Unexpected_char_parsing_dec_escape
let tr_49 state char stack =
  let stack = add_dec_escape_char state char stack in
  set_automaton_state state 12;
  advance state;
  stack
let tr_50 state char stack =
  let stack = add_last_dec_escape_char state char stack in
  set_automaton_state state 8;
  advance state;
  stack
let tr_51 _state _char _stack =
  raise _state ~at_eof:false Unexpected_char_parsing_hex_escape
let tr_52 state char stack =
  let stack = add_hex_escape_char state char stack in
  set_automaton_state state 14;
  advance state;
  stack
let tr_53 state char stack =
  let stack = add_last_hex_escape_char state char stack in
  set_automaton_state state 8;
  advance state;
  stack
let tr_54 state char stack =
  let stack = add_quoted_atom_char state char stack in
  set_automaton_state state 8;
  advance state;
  stack
let tr_55 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 15;
  advance state;
  stack
let tr_56 state char stack =
  let stack = add_quoted_atom_char state char stack in
  set_automaton_state state 8;
  advance_eol state;
  stack
let tr_57 state char stack =
  let stack = push_quoted_atom state char stack in
  set_automaton_state state 0;
  advance state;
  stack
let tr_58 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 9;
  advance state;
  stack
let tr_59 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 16;
  advance state;
  stack
let tr_60 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 16;
  advance_eol state;
  stack
let tr_61 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 19;
  advance state;
  stack
let tr_62 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 18;
  advance state;
  stack
let tr_63 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 17;
  advance state;
  stack
let tr_64 state char stack =
  let stack = end_block_comment state char stack in
  set_automaton_state state (if block_comment_depth state <> 0 then 16 else 0);
  advance state;
  stack
let tr_65 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 19;
  advance_eol state;
  stack
let tr_66 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 20;
  advance state;
  stack
let tr_67 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 26;
  advance_eol state;
  stack
let tr_68 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 21;
  advance state;
  stack
let tr_69 state char stack =
  let stack = add_dec_escape_char state char stack in
  set_automaton_state state 22;
  advance state;
  stack
let tr_70 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 24;
  advance state;
  stack
let tr_71 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 19;
  advance state;
  stack
let tr_72 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 16;
  advance state;
  stack
let tr_73 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 20;
  advance state;
  stack
let tr_74 state char stack =
  let stack = add_dec_escape_char state char stack in
  set_automaton_state state 23;
  advance state;
  stack
let tr_75 state char stack =
  let stack = comment_add_last_dec_escape_char state char stack in
  set_automaton_state state 19;
  advance state;
  stack
let tr_76 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 25;
  advance state;
  stack
let tr_77 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 26;
  advance state;
  stack
let tr_78 state char stack =
  let stack = add_token_char state char stack in
  set_automaton_state state 19;
  advance_eol state;
  stack

let tr_eoi_00 state stack =
  eps_eoi_check state stack
let tr_eoi_01 state _stack =
  raise state ~at_eof:true Automaton_in_error_state
let tr_eoi_02 state _stack =
  raise state ~at_eof:true Unexpected_character_after_cr
let tr_eoi_03 state stack =
  let stack = eps_push_atom state stack in
  eps_eoi_check state stack
let tr_eoi_04 state stack =
  let stack = end_line_comment state stack in
  eps_eoi_check state stack
let tr_eoi_05 state stack =
  let stack = eps_add_first_char_hash state stack in
  let stack = eps_push_atom state stack in
  eps_eoi_check state stack
let tr_eoi_06 state _stack =
  raise state ~at_eof:true Unterminated_quoted_string
let tr_eoi_07 state _stack =
  raise state ~at_eof:true Unterminated_block_comment

let transitions =
  [| tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_01; tr_02; tr_00; tr_01; tr_03; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_01; tr_00; tr_04; tr_05; tr_00; tr_00; tr_00; tr_00
   ; tr_06; tr_07; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_08; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_09; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00; tr_00
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10; tr_10
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_02; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11; tr_11
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_13; tr_14; tr_12; tr_13; tr_15; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_13; tr_12; tr_16; tr_17; tr_12; tr_12; tr_12; tr_12
   ; tr_18; tr_19; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_20; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_21; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_13; tr_14; tr_12; tr_13; tr_15; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_13; tr_12; tr_16; tr_17; tr_12; tr_12; tr_12; tr_12
   ; tr_18; tr_19; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_20; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_22; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_13; tr_14; tr_12; tr_13; tr_15; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_13; tr_12; tr_16; tr_22; tr_12; tr_12; tr_12; tr_12
   ; tr_18; tr_19; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_20; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_21; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12; tr_12
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_24; tr_23; tr_23; tr_25; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23; tr_23
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_27; tr_28; tr_26; tr_27; tr_29; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_27; tr_26; tr_30; tr_31; tr_26; tr_26; tr_26; tr_26
   ; tr_32; tr_33; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_34; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_35; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26; tr_26
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_37; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_38; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_39; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36; tr_36
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_41; tr_40; tr_40; tr_42; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_43; tr_43; tr_43; tr_43; tr_43; tr_43; tr_43; tr_43
   ; tr_43; tr_43; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_44; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40; tr_40
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_41; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_46; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_47; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45; tr_45
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_49; tr_49; tr_49; tr_49; tr_49; tr_49; tr_49; tr_49
   ; tr_49; tr_49; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_50; tr_50; tr_50; tr_50; tr_50; tr_50; tr_50; tr_50
   ; tr_50; tr_50; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_52; tr_52; tr_52; tr_52; tr_52; tr_52; tr_52; tr_52
   ; tr_52; tr_52; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_52; tr_52; tr_52; tr_52; tr_52; tr_52; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_52; tr_52; tr_52; tr_52; tr_52; tr_52; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_53; tr_53; tr_53; tr_53; tr_53; tr_53; tr_53; tr_53
   ; tr_53; tr_53; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_53; tr_53; tr_53; tr_53; tr_53; tr_53; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_53; tr_53; tr_53; tr_53; tr_53; tr_53; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_55; tr_56; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_55; tr_54; tr_57; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_58; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54; tr_54
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_60; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_61; tr_62; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_63; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_60; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_61; tr_64; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_63; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_60; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_61; tr_62; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_35; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59; tr_59
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_65; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_59; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_66; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_67; tr_61; tr_61; tr_68; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_69; tr_69; tr_69; tr_69; tr_69; tr_69; tr_69; tr_69
   ; tr_69; tr_69; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_70; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_67; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_72; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_73; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_74; tr_74; tr_74; tr_74; tr_74; tr_74; tr_74; tr_74
   ; tr_74; tr_74; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_75; tr_75; tr_75; tr_75; tr_75; tr_75; tr_75; tr_75
   ; tr_75; tr_75; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48; tr_48
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_76; tr_76; tr_76; tr_76; tr_76; tr_76; tr_76; tr_76
   ; tr_76; tr_76; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_76; tr_76; tr_76; tr_76; tr_76; tr_76; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_76; tr_76; tr_76; tr_76; tr_76; tr_76; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61
   ; tr_61; tr_61; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_61; tr_61; tr_61; tr_61; tr_61; tr_61; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51; tr_51
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_77; tr_78; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_77; tr_71; tr_72; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_73; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
   ; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71; tr_71
  |]

let transitions_eoi =
  [| tr_eoi_00
   ; tr_eoi_01
   ; tr_eoi_02
   ; tr_eoi_03
   ; tr_eoi_03
   ; tr_eoi_03
   ; tr_eoi_04
   ; tr_eoi_05
   ; tr_eoi_06
   ; tr_eoi_06
   ; tr_eoi_06
   ; tr_eoi_06
   ; tr_eoi_06
   ; tr_eoi_06
   ; tr_eoi_06
   ; tr_eoi_06
   ; tr_eoi_07
   ; tr_eoi_07
   ; tr_eoi_07
   ; tr_eoi_07
   ; tr_eoi_07
   ; tr_eoi_07
   ; tr_eoi_07
   ; tr_eoi_07
   ; tr_eoi_07
   ; tr_eoi_07
   ; tr_eoi_07
  |]

# 4442 "boot.ml"
end
end
open XXXX
include (struct
# 1 "src/usexp/usexp.ml"
module UnlabeledBytes = Bytes
open StdLabels

module Bytes = struct
  include StdLabels.Bytes

  (* [blit_string] was forgotten from the labeled version in OCaml
     4.02—4.04. *)
  let blit_string ~src ~src_pos ~dst ~dst_pos ~len =
    UnlabeledBytes.blit_string src src_pos dst dst_pos len
end

module A = Parser_automaton_internal

module Atom = struct
 type t = Sexp_ast.atom = A of string [@@unboxed]

 let is_valid str =
   let len = String.length str in
   len > 0 &&
   let rec loop ix =
     match str.[ix] with
     | '"' | '(' | ')' | ';' -> true
     | '|' -> ix > 0 && let next = ix - 1 in str.[next] = '#' || loop next
     | '#' -> ix > 0 && let next = ix - 1 in str.[next] = '|' || loop next
     | ' ' | '\t' | '\n' | '\012' | '\r' -> true
     | _ -> ix > 0 && loop (ix - 1)
   in
   not (loop (len - 1))

 (* XXX eventually we want to report a nice error message to the user
     at the point the conversion is made. *)
  let of_string s =
    if is_valid s then A s
    else invalid_arg(Printf.sprintf "Usexp.Atom.of_string: %S" s)

  let of_int i = A (string_of_int i)
  let of_float x = A (string_of_float x)
  let of_bool x = A (string_of_bool x)
  let of_int64 i = A (Int64.to_string i)
  let of_digest d = A (Digest.to_hex d)

  let to_string (A s) = s
end

type t =
  | Atom of Atom.t
  | Quoted_string of string
  | List of t list

type sexp = t

let atom s =
  if Atom.is_valid s then Atom (A s)
  else invalid_arg "Usexp.atom"

let unsafe_atom_of_string s = Atom(A s)

let should_be_atom str =
  let len = String.length str in
  len > 0 &&
  let rec loop ix =
    match str.[ix] with
    | '"' | '(' | ')' | ';' | '\\' -> true
    | '|' -> ix > 0 && let next = ix - 1 in str.[next] = '#' || loop next
    | '#' -> ix > 0 && let next = ix - 1 in str.[next] = '|' || loop next
    | '\000' .. '\032' | '\127' .. '\255' -> true
    | _ -> ix > 0 && loop (ix - 1)
  in
  not (loop (len - 1))

let atom_or_quoted_string s =
  if should_be_atom s then Atom (A s)
  else Quoted_string s

let quote_length s =
  let n = ref 0 in
  for i = 0 to String.length s - 1 do
    n := !n + (match String.unsafe_get s i with
               | '\"' | '\\' | '\n' | '\t' | '\r' | '\b' -> 2
               | ' ' .. '~' -> 1
               | _ -> 4)
  done;
  !n

let escape_to s ~dst:s' ~ofs =
  let n = ref ofs in
  for i = 0 to String.length s - 1 do
    begin match String.unsafe_get s i with
    | ('\"' | '\\') as c ->
       Bytes.unsafe_set s' !n '\\'; incr n; Bytes.unsafe_set s' !n c
    | '\n' ->
       Bytes.unsafe_set s' !n '\\'; incr n; Bytes.unsafe_set s' !n 'n'
    | '\t' ->
       Bytes.unsafe_set s' !n '\\'; incr n; Bytes.unsafe_set s' !n 't'
    | '\r' ->
       Bytes.unsafe_set s' !n '\\'; incr n; Bytes.unsafe_set s' !n 'r'
    | '\b' ->
       Bytes.unsafe_set s' !n '\\'; incr n; Bytes.unsafe_set s' !n 'b'
    | (' ' .. '~') as c -> Bytes.unsafe_set s' !n c
    | c ->
       let a = Char.code c in
       Bytes.unsafe_set s' !n '\\';
       incr n;
       Bytes.unsafe_set s' !n (Char.unsafe_chr (48 + a / 100));
       incr n;
       Bytes.unsafe_set s' !n (Char.unsafe_chr (48 + (a / 10) mod 10));
       incr n;
       Bytes.unsafe_set s' !n (Char.unsafe_chr (48 + a mod 10));
    end;
    incr n
  done

(* Escape [s] if needed. *)
let escaped s =
  let n = quote_length s in
  if n = 0 || n > String.length s then
    let s' = Bytes.create n in
    escape_to s ~dst:s' ~ofs:0;
    Bytes.unsafe_to_string s'
  else s

(* Surround [s] with quotes, escaping it if necessary. *)
let quoted s =
  let len = String.length s in
  let n = quote_length s in
  let s' = Bytes.create (n + 2) in
  Bytes.unsafe_set s' 0 '"';
  if len = 0 || n > len then
    escape_to s ~dst:s' ~ofs:1
  else
    Bytes.blit_string ~src:s ~src_pos:0 ~dst:s' ~dst_pos:1 ~len;
  Bytes.unsafe_set s' (n + 1) '"';
  Bytes.unsafe_to_string s'

let rec to_string = function
  | Atom (A s) -> s
  | Quoted_string s -> quoted s
  | List l -> Printf.sprintf "(%s)" (List.map l ~f:to_string |> String.concat ~sep:" ")

let rec pp ppf = function
  | Atom (A s) ->
    Format.pp_print_string ppf s
  | Quoted_string s ->
    Format.pp_print_string ppf (quoted s)
  | List [] ->
    Format.pp_print_string ppf "()"
  | List (first :: rest) ->
    Format.pp_open_box ppf 1;
    Format.pp_print_string ppf "(";
    Format.pp_open_hvbox ppf 0;
    pp ppf first;
    List.iter rest ~f:(fun sexp ->
      Format.pp_print_space ppf ();
      pp ppf sexp);
    Format.pp_close_box ppf ();
    Format.pp_print_string ppf ")";
    Format.pp_close_box ppf ()

let split_string s ~on =
  let rec loop i j =
    if j = String.length s then
      [String.sub s ~pos:i ~len:(j - i)]
    else if s.[j] = on then
      String.sub s ~pos:i ~len:(j - i) :: loop (j + 1) (j + 1)
    else
      loop i (j + 1)
  in
  loop 0 0

let pp_print_quoted_string ppf s =
  if String.contains s '\n' then begin
    match split_string s ~on:'\n' with
    | [] -> Format.pp_print_string ppf (quoted s)
    | first :: rest ->
       Format.fprintf ppf "@[<hv 1>\"@{<atom>%s" (escaped first);
       List.iter rest ~f:(fun s ->
           Format.fprintf ppf "@,\\n%s" (escaped s));
       Format.fprintf ppf "@}\"@]"
  end else
    Format.pp_print_string ppf (quoted s)

let rec pp_split_strings ppf = function
  | Atom (A s) -> Format.pp_print_string ppf s
  | Quoted_string s -> pp_print_quoted_string ppf s
  | List [] ->
    Format.pp_print_string ppf "()"
  | List (first :: rest) ->
    Format.pp_open_box ppf 1;
    Format.pp_print_string ppf "(";
    Format.pp_open_hvbox ppf 0;
    pp_split_strings ppf first;
    List.iter rest ~f:(fun sexp ->
      Format.pp_print_space ppf ();
      pp_split_strings ppf sexp);
    Format.pp_close_box ppf ();
    Format.pp_print_string ppf ")";
    Format.pp_close_box ppf ()

type formatter_state =
  | In_atom
  | In_makefile_action
  | In_makefile_stuff

let prepare_formatter ppf =
  let state = ref [] in
  Format.pp_set_mark_tags ppf true;
  let ofuncs = Format.pp_get_formatter_out_functions ppf () in
  let tfuncs = Format.pp_get_formatter_tag_functions ppf () in
  Format.pp_set_formatter_tag_functions ppf
    { tfuncs with
      mark_open_tag  = (function
        | "atom" -> state := In_atom :: !state; ""
        | "makefile-action" -> state := In_makefile_action :: !state; ""
        | "makefile-stuff" -> state := In_makefile_stuff :: !state; ""
        | s -> tfuncs.mark_open_tag s)
    ; mark_close_tag = (function
        | "atom" | "makefile-action" | "makefile-stuff" -> state := List.tl !state; ""
        | s -> tfuncs.mark_close_tag s)
    };
  Format.pp_set_formatter_out_functions ppf
    { ofuncs with
      out_newline = (fun () ->
        match !state with
        | [In_atom; In_makefile_action] ->
          ofuncs.out_string "\\\n\t" 0 3
        | [In_atom] ->
          ofuncs.out_string "\\\n" 0 2
        | [In_makefile_action] ->
          ofuncs.out_string " \\\n\t" 0 4
        | [In_makefile_stuff] ->
          ofuncs.out_string " \\\n" 0 3
        | [] ->
          ofuncs.out_string "\n" 0 1
        | _ -> assert false)
    ; out_spaces = (fun n ->
        ofuncs.out_spaces
          (match !state with
           | In_atom :: _ -> max 0 (n - 2)
           | _ -> n))
    }

module Loc = Sexp_ast.Loc

module Ast = struct
  type t = Sexp_ast.t =
    | Atom of Loc.t * Atom.t
    | Quoted_string of Loc.t * string
    | List of Loc.t * t list

  let atom_or_quoted_string loc s =
    if should_be_atom s then Atom (loc, A s)
    else Quoted_string (loc, s)

let loc (Atom (loc, _) | Quoted_string (loc, _) | List (loc, _)) = loc

  let rec remove_locs : t -> sexp = function
    | Atom (_, s) -> Atom s
    | Quoted_string (_, s) -> Quoted_string s
    | List (_, l) -> List (List.map l ~f:remove_locs)

  module Token = struct
    type t =
      | Atom   of Loc.t * Atom.t
      | String of Loc.t * string
      | Lparen of Loc.t
      | Rparen of Loc.t
  end

  let tokenize =
    let rec loop acc t =
      match t with
      | Atom (loc, s) -> Token.Atom (loc, s) :: acc
      | Quoted_string (loc, s) -> Token.String (loc, s) :: acc
      | List (loc, l) ->
        let shift (pos : Lexing.position) delta =
          { pos with pos_cnum = pos.pos_cnum + delta }
        in
        let l_loc = { loc with stop  = shift loc.start  1  } in
        let r_loc = { loc with start = shift loc.stop (-1) } in
        let acc = Token.Lparen l_loc :: acc in
        let acc = List.fold_left l ~init:acc ~f:loop in
        let acc = Token.Rparen r_loc :: acc in
        acc
    in
    fun t -> loop [] t |> List.rev
end

let rec add_loc t ~loc : Ast.t =
  match t with
  | Atom s -> Atom (loc, s)
  | Quoted_string s -> Quoted_string (loc, s)
  | List l -> List (loc, List.map l ~f:(add_loc ~loc))

module Parser = struct
  module Error = A.Error
  exception Error = A.Parse_error

  module Mode = struct
    type 'a t = 'a A.mode =
      | Single : Ast.t t
      | Many   : Ast.t list t
  end

  module Stack = struct
    type t = A.stack
    let empty = A.empty_stack
  end

  type 'a t = 'a A.state
  let create ~fname ~mode = A.new_state ~fname mode

  let feed : type a. a A.action = fun state char stack ->
    let idx = (A.automaton_state state lsl 8) lor (Char.code char) in
    (* We need an Obj.magic as the type of the array can't be generalized.
       This problem will go away when we get immutable arrays. *)
    (Obj.magic (Table.transitions.(idx) : Obj.t A.action) : a A.action) state char stack
  [@@inline always]

  let feed_eoi : type a. a t -> Stack.t -> a = fun state stack ->
    let stack =
      (Obj.magic (Table.transitions_eoi.(A.automaton_state state)
                  : Obj.t A.epsilon_action)
       : a A.epsilon_action) state stack
    in
    A.set_error_state state;
    match A.mode state with
    | Mode.Single -> A.sexp_of_stack stack
    | Mode.Many   -> A.sexps_of_stack stack

  let rec feed_substring_unsafe str state stack i stop =
    if i < stop then
      let c = String.unsafe_get str i in
      let stack = feed state c stack in
      feed_substring_unsafe str state stack (i + 1) stop
    else
      stack

  let rec feed_subbytes_unsafe str state stack i stop =
    if i < stop then
      let c = Bytes.unsafe_get str i in
      let stack = feed state c stack in
      feed_subbytes_unsafe str state stack (i + 1) stop
    else
      stack

  let feed_substring state str ~pos ~len stack =
    let str_len = String.length str in
    if pos < 0 || len < 0 || pos > str_len - len then
      invalid_arg "Jbuilder_sexp.feed_substring";
    feed_substring_unsafe str state stack pos (pos + len)

  let feed_subbytes state str ~pos ~len stack =
    let str_len = Bytes.length str in
    if pos < 0 || len < 0 || pos > str_len - len then
      invalid_arg "Jbuilder_sexp.feed_subbytes";
    feed_subbytes_unsafe str state stack pos (pos + len)

  let feed_string state str stack =
    feed_substring_unsafe str state stack 0 (String.length str)

  let feed_bytes state str stack =
    feed_subbytes_unsafe str state stack 0 (Bytes.length str)
end

let parse_string ~fname ~mode str =
  let p = Parser.create ~fname ~mode in
  let stack = Parser.feed_string p str Parser.Stack.empty in
  Parser.feed_eoi p stack
# 4817 "boot.ml"
end : sig
# 1 "src/usexp/usexp.mli"
(** Parsing of s-expressions *)

module Atom : sig
  type t = private A of string [@@unboxed]
  (** Acceptable atoms are composed of chars in the range ['!' .. '~'] excluding
      [' ' '"' '(' ')' ';' '\\'], and must be nonempty. *)

  val is_valid : string -> bool
  (** [is_valid s] checks that [s] respects the constraints to be an atom. *)

 val of_string : string -> t
  (** Convert a string to an atom.  If the string contains invalid
     characters, raise [Invalid_argument]. *)

  val of_int : int -> t
  val of_float : float -> t
  val of_bool : bool -> t
  val of_int64 : Int64.t -> t
  val of_digest : Digest.t -> t

  val to_string : t -> string
end

module Loc : sig
  type t =
    { start : Lexing.position
    ; stop  : Lexing.position
    }
end

(** The S-expression type *)
type t =
  | Atom of Atom.t
  | Quoted_string of string
  | List of t list

val atom : string -> t
(** [atom s] convert the string [s] to an Atom.
    @raise Invalid_argument if [s] does not satisfy [Atom.is_valid s]. *)

val atom_or_quoted_string : string -> t

val unsafe_atom_of_string : string -> t

(** Serialize a S-expression *)
val to_string : t -> string

(** Serialize a S-expression using indentation to improve readability *)
val pp : Format.formatter -> t -> unit

(** Same as [pp], but split long strings. The formatter must have been
    prepared with [prepare_formatter]. *)
val pp_split_strings : Format.formatter -> t -> unit

(** Prepare a formatter for [pp_split_strings]. Additionaly the
    formatter escape newlines when the tags "makefile-action" or
    "makefile-stuff" are active. *)
val prepare_formatter : Format.formatter -> unit

(** Abstract syntax tree *)
module Ast : sig
  type sexp = t
  type t =
    | Atom of Loc.t * Atom.t
    | Quoted_string of Loc.t * string
    | List of Loc.t * t list

  val atom_or_quoted_string : Loc.t -> string -> t

  val loc : t -> Loc.t

  val remove_locs : t -> sexp

  module Token : sig
    type t =
      | Atom   of Loc.t * Atom.t
      | String of Loc.t * string
      | Lparen of Loc.t
      | Rparen of Loc.t
  end

  val tokenize : t -> Token.t list
end with type sexp := t

val add_loc : t -> loc:Loc.t -> Ast.t

module Parser : sig
  module Error : sig
    type t

    val position : t -> Lexing.position
    val message  : t -> string
  end

  (** Exception raised in case of a parsing error *)
  exception Error of Error.t

  module Mode : sig
    type sexp = t
    type 'a t =
      | Single : Ast.t t
      | Many   : Ast.t list t
  end with type sexp := t

  module Stack : sig
    (** Parser stack. The stack is not in [state] for optimization purposes. *)
    type t

    val empty : t
  end

  type 'a t

  (** Create a new parser state. [fname] is the filename the input is from. *)
  val create : fname:string -> mode:'a Mode.t -> 'a t

  (** Feed one character to the parser. In case of error, it raises [Parse_error] *)
  val feed : _ t -> char -> Stack.t -> Stack.t

  (** Instruct the parser that the end of input was reached. In case of error, it raises
      [Parse_error] *)
  val feed_eoi : 'a t -> Stack.t -> 'a

  (** {3 Convenience functions} *)

  val feed_string    : _ t -> string                       -> Stack.t -> Stack.t
  val feed_substring : _ t -> string -> pos:int -> len:int -> Stack.t -> Stack.t
  val feed_bytes     : _ t -> bytes                        -> Stack.t -> Stack.t
  val feed_subbytes  : _ t -> bytes -> pos:int -> len:int  -> Stack.t -> Stack.t
end

val parse_string : fname:string -> mode:'a Parser.Mode.t -> string -> 'a
# 4952 "boot.ml"
end)
end
module Ocaml_config = struct
module XXXX = struct
end
open XXXX
include (struct
# 1 "src/ocaml-config/ocaml_config.ml"
open Stdune
open Result.O

module Prog_and_args = struct
  type t =
    { prog : string
    ; args : string list
    }
end
open Prog_and_args

module Value = struct
  type t =
    | Bool          of bool
    | Int           of int
    | String        of string
    | Words         of string list
    | Prog_and_args of Prog_and_args.t

  let sexp_of_t : t -> Usexp.t = function
    | Bool   x -> Atom (Usexp.Atom.of_bool x)
    | Int    x -> Atom (Usexp.Atom.of_int x)
    | String x -> Usexp.atom_or_quoted_string x
    | Words  x -> List (List.map x ~f:Usexp.atom_or_quoted_string)
    | Prog_and_args x ->
      List (Usexp.atom_or_quoted_string x.prog
            :: List.map x.args ~f:Usexp.atom_or_quoted_string)

  let to_string = function
    | Bool   x -> string_of_bool x
    | Int    x -> string_of_int x
    | String x -> x
    | Words  x -> String.concat x ~sep:" "
    | Prog_and_args x -> String.concat ~sep:" " (x.prog :: x.args)
end

type t =
  { version                  : int * int * int
  ; version_string           : string
  ; standard_library_default : string
  ; standard_library         : string
  ; standard_runtime         : string
  ; ccomp_type               : string
  ; c_compiler               : string
  ; ocamlc_cflags            : string list
  ; ocamlopt_cflags          : string list
  ; bytecomp_c_compiler      : Prog_and_args.t
  ; bytecomp_c_libraries     : string list
  ; native_c_compiler        : Prog_and_args.t
  ; native_c_libraries       : string list
  ; cc_profile               : string list
  ; architecture             : string
  ; model                    : string
  ; int_size                 : int
  ; word_size                : int
  ; system                   : string
  ; asm                      : Prog_and_args.t
  ; asm_cfi_supported        : bool
  ; with_frame_pointers      : bool
  ; ext_exe                  : string
  ; ext_obj                  : string
  ; ext_asm                  : string
  ; ext_lib                  : string
  ; ext_dll                  : string
  ; os_type                  : string
  ; default_executable_name  : string
  ; systhread_supported      : bool
  ; host                     : string
  ; target                   : string
  ; profiling                : bool
  ; flambda                  : bool
  ; spacetime                : bool
  ; safe_string              : bool
  ; exec_magic_number        : string
  ; cmi_magic_number         : string
  ; cmo_magic_number         : string
  ; cma_magic_number         : string
  ; cmx_magic_number         : string
  ; cmxa_magic_number        : string
  ; ast_impl_magic_number    : string
  ; ast_intf_magic_number    : string
  ; cmxs_magic_number        : string
  ; cmt_magic_number         : string
  ; natdynlink_supported     : bool
  }

let version                  t = t.version
let version_string           t = t.version_string
let standard_library_default t = t.standard_library_default
let standard_library         t = t.standard_library
let standard_runtime         t = t.standard_runtime
let ccomp_type               t = t.ccomp_type
let c_compiler               t = t.c_compiler
let ocamlc_cflags            t = t.ocamlc_cflags
let ocamlopt_cflags          t = t.ocamlopt_cflags
let bytecomp_c_compiler      t = t.bytecomp_c_compiler
let bytecomp_c_libraries     t = t.bytecomp_c_libraries
let native_c_compiler        t = t.native_c_compiler
let native_c_libraries       t = t.native_c_libraries
let cc_profile               t = t.cc_profile
let architecture             t = t.architecture
let model                    t = t.model
let int_size                 t = t.int_size
let word_size                t = t.word_size
let system                   t = t.system
let asm                      t = t.asm
let asm_cfi_supported        t = t.asm_cfi_supported
let with_frame_pointers      t = t.with_frame_pointers
let ext_exe                  t = t.ext_exe
let ext_obj                  t = t.ext_obj
let ext_asm                  t = t.ext_asm
let ext_lib                  t = t.ext_lib
let ext_dll                  t = t.ext_dll
let os_type                  t = t.os_type
let default_executable_name  t = t.default_executable_name
let systhread_supported      t = t.systhread_supported
let host                     t = t.host
let target                   t = t.target
let profiling                t = t.profiling
let flambda                  t = t.flambda
let spacetime                t = t.spacetime
let safe_string              t = t.safe_string
let exec_magic_number        t = t.exec_magic_number
let cmi_magic_number         t = t.cmi_magic_number
let cmo_magic_number         t = t.cmo_magic_number
let cma_magic_number         t = t.cma_magic_number
let cmx_magic_number         t = t.cmx_magic_number
let cmxa_magic_number        t = t.cmxa_magic_number
let ast_impl_magic_number    t = t.ast_impl_magic_number
let ast_intf_magic_number    t = t.ast_intf_magic_number
let cmxs_magic_number        t = t.cmxs_magic_number
let cmt_magic_number         t = t.cmt_magic_number
let natdynlink_supported     t = t.natdynlink_supported

let to_list t : (string * Value.t) list =
  [ "version"                  , String        t.version_string
  ; "standard_library_default" , String        t.standard_library_default
  ; "standard_library"         , String        t.standard_library
  ; "standard_runtime"         , String        t.standard_runtime
  ; "ccomp_type"               , String        t.ccomp_type
  ; "c_compiler"               , String        t.c_compiler
  ; "ocamlc_cflags"            , Words         t.ocamlc_cflags
  ; "ocamlopt_cflags"          , Words         t.ocamlopt_cflags
  ; "bytecomp_c_compiler"      , Prog_and_args t.bytecomp_c_compiler
  ; "bytecomp_c_libraries"     , Words         t.bytecomp_c_libraries
  ; "native_c_compiler"        , Prog_and_args t.native_c_compiler
  ; "native_c_libraries"       , Words         t.native_c_libraries
  ; "cc_profile"               , Words         t.cc_profile
  ; "architecture"             , String        t.architecture
  ; "model"                    , String        t.model
  ; "int_size"                 , Int           t.int_size
  ; "word_size"                , Int           t.word_size
  ; "system"                   , String        t.system
  ; "asm"                      , Prog_and_args t.asm
  ; "asm_cfi_supported"        , Bool          t.asm_cfi_supported
  ; "with_frame_pointers"      , Bool          t.with_frame_pointers
  ; "ext_exe"                  , String        t.ext_exe
  ; "ext_obj"                  , String        t.ext_obj
  ; "ext_asm"                  , String        t.ext_asm
  ; "ext_lib"                  , String        t.ext_lib
  ; "ext_dll"                  , String        t.ext_dll
  ; "os_type"                  , String        t.os_type
  ; "default_executable_name"  , String        t.default_executable_name
  ; "systhread_supported"      , Bool          t.systhread_supported
  ; "host"                     , String        t.host
  ; "target"                   , String        t.target
  ; "profiling"                , Bool          t.profiling
  ; "flambda"                  , Bool          t.flambda
  ; "spacetime"                , Bool          t.spacetime
  ; "safe_string"              , Bool          t.safe_string
  ; "exec_magic_number"        , String        t.exec_magic_number
  ; "cmi_magic_number"         , String        t.cmi_magic_number
  ; "cmo_magic_number"         , String        t.cmo_magic_number
  ; "cma_magic_number"         , String        t.cma_magic_number
  ; "cmx_magic_number"         , String        t.cmx_magic_number
  ; "cmxa_magic_number"        , String        t.cmxa_magic_number
  ; "ast_impl_magic_number"    , String        t.ast_impl_magic_number
  ; "ast_intf_magic_number"    , String        t.ast_intf_magic_number
  ; "cmxs_magic_number"        , String        t.cmxs_magic_number
  ; "cmt_magic_number"         , String        t.cmt_magic_number
  ; "natdynlink_supported"     , Bool          t.natdynlink_supported
  ]

let sexp_of_t t =
  Usexp.List
    (to_list t
     |> List.map ~f:(fun (k, v) ->
       Usexp.List [ Usexp.atom_or_quoted_string k
                  ; Value.sexp_of_t v
                  ]))

exception E of string
let fail fmt =
  Printf.ksprintf (fun msg -> raise (E msg)) fmt

let split_prog s =
  match String.extract_blank_separated_words s with
  | []           -> None
  | prog :: args -> Some { prog; args }

module Vars = struct
  module M = Map.Make(String)

  type t = string M.t

  let of_lines lines =
    let rec loop acc = function
      | [] -> Ok acc
      | line :: lines ->
        match String.index line ':' with
        | Some i ->
          let x =
            (String.sub line ~pos:0 ~len:i,
             String.sub line ~pos:(i + 2) ~len:(String.length line - i - 2))
          in
          loop (x :: acc) lines
        | None ->
          Error (Printf.sprintf "Unrecognized line: %S" line)
    in
    loop [] lines >>= fun vars ->
    Result.map_error (M.of_list vars) ~f:(fun (var, _, _) ->
      Printf.sprintf "Variable %S present twice." var)

  let get_opt t var = M.find t var

  let get t var =
    match get_opt t var with
    | Some s -> s
    | None   -> fail "Variable %S not found." var

  let get_bool t var =
    match get_opt t var with
    | None -> false
    | Some s ->
      match s with
      | "true"  -> true
      | "false" -> false
      | s ->
        fail "Value of %S is neither 'true' neither 'false': %s." var s

  let get_int_opt t var =
    match get_opt t var with
    | None -> None
    | Some s ->
      match int_of_string s with
      | x -> Some x
      | exception _ ->
        fail "Value of %S is not an integer: %s." var s

  let get_words t var =
    match get_opt t var with
    | None   -> []
    | Some s -> String.extract_blank_separated_words s

  let get_prog_or_dummy t var =
    Option.map (get_opt t var) ~f:(fun v ->
      match split_prog v with
      | None ->
        { prog = Printf.sprintf "%s-not-found-in-ocaml-config" var
        ; args = []
        }
      | Some s -> s
    )

  let get_prog_or_dummy_exn t var =
    match get_prog_or_dummy t var with
    | None -> fail "Variable %S not found." var
    | Some s -> s
end

let get_arch_sixtyfour stdlib_dir =
  let files = ["caml/config.h"; "caml/m.h"] in
  let get_arch_sixtyfour_from file =
    let file = Filename.concat stdlib_dir file in
    if Sys.file_exists file then begin
      let rec loop ic =
        match input_line ic with
        | exception End_of_file -> false
        | line ->
          match String.extract_blank_separated_words line with
          | ["#define"; "ARCH_SIXTYFOUR"] -> true
          | _ -> loop ic
      in
      Exn.protectx (open_in file) ~finally:close_in ~f:loop
    end else
      false
  in
  List.exists ~f:get_arch_sixtyfour_from files

let make vars =
  match
    let open Vars in
    let bytecomp_c_compiler =
      get_prog_or_dummy_exn vars "bytecomp_c_compiler" in
    let native_c_compiler =
      get_prog_or_dummy_exn vars "native_c_compiler" in
    let c_compiler, ocamlc_cflags, ocamlopt_cflags =
      match get_prog_or_dummy vars "c_compiler" with
      | Some { prog; args } -> (* >= 4.06 *)
        let get_flags var = args @ get_words vars var in
        (prog,
         get_flags "ocamlc_cflags",
         get_flags "ocamlopt_cflags")
      | None ->
        (bytecomp_c_compiler.prog,
         bytecomp_c_compiler.args,
         native_c_compiler.args)
    in
    let version_string = get vars "version" in
    let version =
      Scanf.sscanf version_string "%u.%u.%u" (fun a b c -> a, b, c)
    in
    let os_type                  = get vars "os_type" in
    let standard_library_default = get vars "standard_library_default" in
    let standard_library         = get vars "standard_library" in
    let standard_runtime         = get vars "standard_runtime" in
    let ccomp_type               = get vars "ccomp_type" in
    let bytecomp_c_libraries     = get_words vars "bytecomp_c_libraries" in
    let native_c_libraries       = get_words vars "native_c_libraries" in
    let cc_profile               = get_words vars "cc_profile" in
    let architecture             = get vars "architecture" in
    let model                    = get vars "model" in
    let system                   = get vars "system" in
    let asm_cfi_supported        = get_bool vars "asm_cfi_supported" in
    let with_frame_pointers      = get_bool vars "with_frame_pointers" in
    let asm                      = get_prog_or_dummy_exn vars "asm" in
    let word_size =
      match get_int_opt vars "word_size" with
      | Some n -> n
      | None ->
        if get_arch_sixtyfour standard_library then
          64
        else
          32
    in
    let int_size =
      match get_int_opt vars "int_size" with
      | Some n -> n
      | None   -> word_size - 1
    in

    let ext_obj = get vars "ext_obj" in
    let ext_asm = get vars "ext_asm" in
    let ext_lib = get vars "ext_lib" in
    let ext_dll = get vars "ext_dll" in
    let ext_exe =
      match get_opt vars "exe_ext" with
      | Some s -> s
      | None -> if os_type = "Win32" then ".exe" else ""
    in

    let default_executable_name = get      vars "default_executable_name" in
    let systhread_supported     = get_bool vars "systhread_supported"     in
    let host                    = get      vars "host"                    in
    let target                  = get      vars "target"                  in
    let profiling               = get_bool vars "profiling"               in
    let flambda                 = get_bool vars "flambda"                 in
    let spacetime               = get_bool vars "spacetime"               in
    let safe_string             = get_bool vars "safe_string"             in
    let exec_magic_number       = get      vars "exec_magic_number"       in
    let cmi_magic_number        = get      vars "cmi_magic_number"        in
    let cmo_magic_number        = get      vars "cmo_magic_number"        in
    let cma_magic_number        = get      vars "cma_magic_number"        in
    let cmx_magic_number        = get      vars "cmx_magic_number"        in
    let cmxa_magic_number       = get      vars "cmxa_magic_number"       in
    let ast_impl_magic_number   = get      vars "ast_impl_magic_number"   in
    let ast_intf_magic_number   = get      vars "ast_intf_magic_number"   in
    let cmxs_magic_number       = get      vars "cmxs_magic_number"       in
    let cmt_magic_number        = get      vars "cmt_magic_number"        in

    let natdynlink_supported =
      Sys.file_exists (Filename.concat standard_library "dynlink.cmxa")
    in
    { version
    ; version_string
    ; standard_library_default
    ; standard_library
    ; standard_runtime
    ; ccomp_type
    ; c_compiler
    ; ocamlc_cflags
    ; ocamlopt_cflags
    ; bytecomp_c_compiler
    ; bytecomp_c_libraries
    ; native_c_compiler
    ; native_c_libraries
    ; cc_profile
    ; architecture
    ; model
    ; int_size
    ; word_size
    ; system
    ; asm
    ; asm_cfi_supported
    ; with_frame_pointers
    ; ext_exe
    ; ext_obj
    ; ext_asm
    ; ext_lib
    ; ext_dll
    ; os_type
    ; default_executable_name
    ; systhread_supported
    ; host
    ; target
    ; profiling
    ; flambda
    ; spacetime
    ; safe_string
    ; exec_magic_number
    ; cmi_magic_number
    ; cmo_magic_number
    ; cma_magic_number
    ; cmx_magic_number
    ; cmxa_magic_number
    ; ast_impl_magic_number
    ; ast_intf_magic_number
    ; cmxs_magic_number
    ; cmt_magic_number
    ; natdynlink_supported
    }
  with
  | t                 -> Ok t
  | exception (E msg) -> Error msg
# 5385 "boot.ml"
end : sig
# 1 "src/ocaml-config/ocaml_config.mli"
(** Represent the output of [ocamlc -config] *)

open Stdune

(** Represent a parsed and interpreted output of [ocamlc -config] *)
type t

val sexp_of_t : t -> Usexp.t

module Prog_and_args : sig
  type t =
    { prog : string
    ; args : string list
    }
end

(** {1 Raw bindings} *)

(** Represent the parsed but uninterpreted output of [ocamlc -config] *)
module Vars : sig
  type t = string Map.Make(String).t

  (** Parse the output of [ocamlc -config] given as a list of lines. *)
  val of_lines : string list -> (t, string) Result.t
end

(** {1 Creation} *)

(** Interpret raw bindings *)
val make : Vars.t -> (t, string) Result.t

(** {1 Query} *)

(** The following parameters match the variables in the output of
    [ocamlc -config] but are stable accross versions of OCaml. *)

val version                  : t -> int * int * int
val version_string           : t -> string
val standard_library_default : t -> string
val standard_library         : t -> string
val standard_runtime         : t -> string
val ccomp_type               : t -> string
val c_compiler               : t -> string
val ocamlc_cflags            : t -> string list
val ocamlopt_cflags          : t -> string list
val bytecomp_c_compiler      : t -> Prog_and_args.t
val bytecomp_c_libraries     : t -> string list
val native_c_compiler        : t -> Prog_and_args.t
val native_c_libraries       : t -> string list
val cc_profile               : t -> string list
val architecture             : t -> string
val model                    : t -> string
val int_size                 : t -> int
val word_size                : t -> int
val system                   : t -> string
val asm                      : t -> Prog_and_args.t
val asm_cfi_supported        : t -> bool
val with_frame_pointers      : t -> bool
val ext_exe                  : t -> string
val ext_obj                  : t -> string
val ext_asm                  : t -> string
val ext_lib                  : t -> string
val ext_dll                  : t -> string
val os_type                  : t -> string
val default_executable_name  : t -> string
val systhread_supported      : t -> bool
val host                     : t -> string
val target                   : t -> string
val profiling                : t -> bool
val flambda                  : t -> bool
val spacetime                : t -> bool
val safe_string              : t -> bool
val exec_magic_number        : t -> string
val cmi_magic_number         : t -> string
val cmo_magic_number         : t -> string
val cma_magic_number         : t -> string
val cmx_magic_number         : t -> string
val cmxa_magic_number        : t -> string
val ast_impl_magic_number    : t -> string
val ast_intf_magic_number    : t -> string
val cmxs_magic_number        : t -> string
val cmt_magic_number         : t -> string
val natdynlink_supported     : t -> bool

(** {1 Values} *)

module Value : sig
  type t =
    | Bool          of bool
    | Int           of int
    | String        of string
    | Words         of string list
    | Prog_and_args of Prog_and_args.t

  val to_string : t -> string
  val sexp_of_t : t -> Usexp.t
end

val to_list : t -> (string * Value.t) list
# 5487 "boot.ml"
end)
end
module Xdg = struct
module XXXX = struct
end
open XXXX
include (struct
# 1 "src/xdg/xdg.ml"
let home =
  try
    Sys.getenv "HOME"
  with Not_found ->
    try
      (Unix.getpwuid (Unix.getuid ())).Unix.pw_dir
    with Unix.Unix_error _ | Not_found ->
      if Sys.win32 then
        try
          Sys.getenv "AppData"
        with Not_found ->
          ""
      else
        ""

let ( / ) = Filename.concat

let get env_var unix_default win32_default =
  try
    Sys.getenv env_var
  with Not_found ->
    if Sys.win32 then win32_default else unix_default

let cache_dir =
  get "XDG_CACHE_HOME"
    (home / ".cache")
    (home / "Local Settings" / "Cache")

let config_dir =
  get "XDG_CONFIG_HOME"
    (home / ".config")
    (home / "Local Settings")

let data_dir =
  get "XDG_DATA_HOME"
    (home / ".local" / "share")
    (try Sys.getenv "AppData" with Not_found -> "")
# 5533 "boot.ml"
end : sig
# 1 "src/xdg/xdg.mli"
(** Implement the XDG specification

    http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html
*)

(** The directory where the application should read/write config files. *)
val config_dir : string

(** The directory where the application should read/write data files. *)
val data_dir : string

(** The directory where the application should read/write cached files. *)
val cache_dir : string
# 5549 "boot.ml"
end)
end
module Clflags : sig
# 1 "src/clflags.mli"
(** Command line flags *)

(** Compilation flags for OCaml files *)
(*val ocaml_comp_flags : string list ref*)

(** [-g] *)
val g : bool ref

(** Print dependency path in case of error *)
val debug_dep_path : bool ref

(** Debug the findlib implementation *)
val debug_findlib : bool ref

(** Compiler warnings *)
val warnings : string ref

(** Whether we are compiling with extra warnings *)
val dev_mode : bool ref

(** The path to the workspace root *)
val workspace_root : string ref

(** The command line for "Hint: try: jbuilder external-lib-deps ..." *)
val external_lib_deps_hint : string list ref

(** Capture the output of sub-commands *)
val capture_outputs : bool ref

(** Always print backtraces, to help debugging jbuilder itself *)
val debug_backtraces : bool ref

(** Command to use to diff things *)
val diff_command : string option ref

(** Automatically promote files *)
val auto_promote : bool ref

(** Force re-running actions associated to aliases *)
val force : bool ref
# 5594 "boot.ml"
end = struct
# 1 "src/clflags.ml"
(*let ocaml_comp_flags = ref ["-g"]*)
let g = ref true
let debug_findlib = ref false
let warnings = ref "-40"
let debug_dep_path = ref false
let dev_mode = ref false
let workspace_root = ref "."
let external_lib_deps_hint = ref []
let capture_outputs = ref true
let debug_backtraces = ref false
let diff_command = ref None
let auto_promote = ref false
let force = ref false
# 5610 "boot.ml"
end
module Errors : sig
# 1 "src/errors.mli"
(** Dealing with errors *)

(* CR-soon diml: stop including this in [Import] *)
(** This module is included in [Import] *)

(* CR-soon diml:
   - Rename to [User_error]
   - change the [string] argument to [Loc.t option * string] and get rid of
   [Loc.Error]. The two are a bit confusing
   - change [string] to [Colors.Style.t Pp.t]
*)
(** A fatal error, that should be reported to the user in a nice way *)
exception Fatal_error of string

(* CR-soon diml: replace the [string] argument by [Usexp.t] *)
(** An programming error in the code of jbuilder, that should be reported upstream. The
    error message shouldn't try to be developper friendly rather than user friendly.  *)
exception Code_error of string

(* CR-soon diml: we won't need this once we can generate rules dynamically *)
(** Raised for errors that have already been reported to the user and shouldn't be
    reported again. This might happen when trying to build a dependency that has already
    failed. *)
exception Already_reported

(* CR-soon diml: Rename to [user_errorf]. *)
(** Raise a [Fatal_error] exception *)
val die : ('a, Format.formatter, unit, 'b) format4 -> 'a

(** Raise a [Code_error] exception *)
val code_errorf : ('a, Format.formatter, unit, 'b) format4 -> 'a

(**/**)
(* Referenced in Ansi_color and Report_error *)
val err_buf : Buffer.t
val err_ppf : Format.formatter
val kerrf
  :  ('a, Format.formatter, unit, 'b) format4
  -> f:(string -> 'b)
  -> 'a
# 5654 "boot.ml"
end = struct
# 1 "src/errors.ml"
exception Fatal_error of string
exception Code_error of string
exception Already_reported

let err_buf = Buffer.create 128
let err_ppf = Format.formatter_of_buffer err_buf
let kerrf fmt ~f =
  Format.kfprintf
    (fun ppf ->
       Format.pp_print_flush ppf ();
       let s = Buffer.contents err_buf in
       Buffer.clear err_buf;
       f s)
    err_ppf fmt

let code_errorf fmt =
  kerrf fmt ~f:(fun s -> raise (Code_error s))

let die fmt =
  kerrf fmt ~f:(fun s -> raise (Fatal_error s))
# 5677 "boot.ml"
end
module Jbuilder_re = struct
# 1 "vendor/boot/jbuilder_re.ml"
module Re = struct
  type t = unit
  type re = unit
  let compile () = ()
  let execp _ _ = false
end
# 5687 "boot.ml"
end
module Import = struct
# 1 "src/import.ml"
include Stdune
include Jbuilder_re
include Errors

(* To make bug reports usable *)
let () = Printexc.record_backtrace true

let sprintf = Printf.sprintf
let ksprintf = Printf.ksprintf

let initial_cwd = Sys.getcwd ()

module String_set = Set.Make(String)
module String_map = struct
  include Map.Make(String)

  let pp f fmt t =
    Format.pp_print_list (fun fmt (k, v) ->
      Format.fprintf fmt "@[<hov 2>(%s@ =@ %a)@]" k f v
    ) fmt (to_list t)
end

module Int_set = Set.Make(Int)
module Int_map = Map.Make(Int)

module Sys = struct
  include Sys

  let force_remove =
    if win32 then
      fun fn ->
        try
          remove fn
        with Sys_error _ ->
          (* Try to remove the "read-only" attribute, then retry. *)
          (try Unix.chmod fn 0o666 with Unix.Unix_error _ -> ());
          remove fn
    else
      remove
end

type ('a, 'b) eq = Eq : ('a, 'a) eq

type nothing = (int, string) eq

let protect  = Exn.protect
let protectx = Exn.protectx

let warn fmt =
  ksprintf (fun msg ->
    prerr_endline ("Warning: jbuild: " ^ msg))
    fmt

type fail = { fail : 'a. unit -> 'a }

let need_quoting s =
  let len = String.length s in
  len = 0 ||
  let rec loop i =
    if i = len then
      false
    else
      match s.[i] with
      | ' ' | '\"' -> true
      | _ -> loop (i + 1)
  in
  loop 0

let quote_for_shell s =
  if need_quoting s then
    Filename.quote s
  else
    s

let suggest_function : (string -> string list -> string list) ref = ref (fun _ _ -> [])

let hint name candidates =
  match !suggest_function name candidates with
  | [] -> ""
  | l ->
    let rec mk_hint = function
      | [a; b] -> sprintf "%s or %s" a b
      | [a] -> a
      | a :: l -> sprintf "%s, %s" a (mk_hint l)
      | [] -> ""
    in
    sprintf "\nHint: did you mean %s?" (mk_hint l)


(* [maybe_quoted s] is [s] if [s] doesn't need escaping according to OCaml lexing
   conventions and [sprintf "%S" s] otherwise. *)
let maybe_quoted s =
  let escaped = String.escaped s in
  if s == escaped || s = escaped then
    s
  else
    sprintf {|"%s"|} escaped

(* Disable file operations to force to use the IO module *)
let open_in      = `Use_Io
let open_in_bin  = `Use_Io
let open_in_gen  = `Use_Io
let open_out     = `Use_Io
let open_out_bin = `Use_Io
let open_out_gen = `Use_Io

(* We open this module at the top of module generating rules, to make sure they don't do
   Io manually *)
module No_io = struct
  module Io = struct end
end

module Fmt = struct
  (* CR-someday diml: we should define a GADT for this:

     {[
       type 'a t =
         | Int : int t
         | Box : ...
         | Colored : ...
     ]}

     This way we could separate the creation of messages from the
     actual rendering.
  *)
  type 'a t = Format.formatter -> 'a -> unit

  let kstrf f fmt =
    let buf = Buffer.create 17 in
    let f fmt = Format.pp_print_flush fmt () ; f (Buffer.contents buf) in
    Format.kfprintf f (Format.formatter_of_buffer buf) fmt

  let failwith fmt = kstrf failwith fmt

  let list = Format.pp_print_list
  let string s ppf = Format.pp_print_string ppf s

  let nl = Format.pp_print_newline

  let prefix f g ppf x = f ppf; g ppf x

  let ocaml_list pp fmt = function
    | [] -> Format.pp_print_string fmt "[]"
    | l ->
      Format.fprintf fmt "@[<hv>[ %a@ ]@]"
        (list ~pp_sep:(fun fmt () -> Format.fprintf fmt "@,; ")
           pp) l

  let quoted fmt = Format.fprintf fmt "%S"

  let const
    : 'a t -> 'a -> unit t
    = fun pp a' fmt () -> pp fmt a'

  let record fmt = function
    | [] -> Format.pp_print_string fmt "{}"
    | xs ->
      let pp fmt (field, pp) =
        Format.fprintf fmt "@[<hov 1>%s@ =@ %a@]"
          field pp () in
      let pp_sep fmt () = Format.fprintf fmt "@,; " in
      Format.fprintf fmt "@[<hv>{ %a@ }@]"
        (Format.pp_print_list ~pp_sep pp) xs
end

(* This is ugly *)
let printer = ref (Printf.eprintf "%s%!")
let print_to_console s = !printer s
# 5859 "boot.ml"
end
module Loc : sig
# 1 "src/loc.mli"
type t = Usexp.Loc.t =
  { start : Lexing.position
  ; stop  : Lexing.position
  }

val of_lexbuf : Lexing.lexbuf -> t

exception Error of t * string

val exnf     : t             -> ('a, Format.formatter, unit, exn) format4 -> 'a
val fail     : t             -> ('a, Format.formatter, unit, 'b ) format4 -> 'a
val fail_lex : Lexing.lexbuf -> ('a, Format.formatter, unit, 'b ) format4 -> 'a
val fail_opt : t option      -> ('a, Format.formatter, unit, 'b ) format4 -> 'a

val in_file : string -> t

(** To be used with [__POS__] *)
val of_pos : (string * int * int * int) -> t

val none : t

val to_file_colon_line : t -> string
val pp_file_colon_line : Format.formatter -> t -> unit

(** Prints "File ..., line ..., characters ...:\n" *)
val print : Format.formatter -> t -> unit

(** Prints a warning *)
val warn : t -> ('a, Format.formatter, unit) format -> 'a
# 5892 "boot.ml"
end = struct
# 1 "src/loc.ml"
open Import

type t = Usexp.Loc.t =
  { start : Lexing.position
  ; stop  : Lexing.position
  }

let of_lexbuf lb =
  { start = Lexing.lexeme_start_p lb
  ; stop  = Lexing.lexeme_end_p   lb
  }

exception Error of t * string

let exnf t fmt =
  Format.pp_print_as err_ppf 7 ""; (* "Error: " *)
  kerrf fmt ~f:(fun s -> Error (t, s))

let fail t fmt =
  Format.pp_print_as err_ppf 7 ""; (* "Error: " *)
  kerrf fmt ~f:(fun s ->
    raise (Error (t, s)))

let fail_lex lb fmt =
  fail (of_lexbuf lb) fmt

let fail_opt t fmt =
  match t with
  | None -> die fmt
  | Some t -> fail t fmt

let in_file fn =
  let pos : Lexing.position =
    { pos_fname = fn
    ; pos_lnum  = 1
    ; pos_cnum  = 0
    ; pos_bol   = 0
    }
  in
  { start = pos
  ; stop = pos
  }

let of_pos (fname, lnum, cnum, enum) =
  let pos : Lexing.position =
    { pos_fname = fname
    ; pos_lnum  = lnum
    ; pos_cnum  = cnum
    ; pos_bol   = 0
    }
  in
  { start = pos
  ; stop  = { pos with pos_cnum = enum }
  }

let none = in_file "<none>"

let print ppf { start; stop } =
  let start_c = start.pos_cnum - start.pos_bol in
  let stop_c  = stop.pos_cnum  - start.pos_bol in
  Format.fprintf ppf
    "@{<loc>File \"%s\", line %d, characters %d-%d:@}@\n"
    start.pos_fname start.pos_lnum start_c stop_c

let warn t fmt =
  Errors.kerrf ~f:print_to_console
    ("%a@{<warning>Warning@}: " ^^ fmt ^^ "@.") print t

let to_file_colon_line t =
  sprintf "%s:%d" t.start.pos_fname t.start.pos_lnum

let pp_file_colon_line ppf t =
  Format.pp_print_string ppf (to_file_colon_line t)
# 5968 "boot.ml"
end
module Sexp : sig
# 1 "src/sexp.mli"
open Import

include module type of struct include Usexp end with module Loc := Usexp.Loc

val code_error : string -> (string * t) list -> _

val load : fname:string -> mode:'a Parser.Mode.t -> 'a
val load_many_as_one : fname:string -> Ast.t

type sexps_or_ocaml_script =
  | Sexps of Ast.t list
  | Ocaml_script

val load_many_or_ocaml_script : string -> sexps_or_ocaml_script

module type Combinators = sig
  type 'a t
  val unit       : unit                      t

  val string     : string                    t
  (** Convert an [Atom] or a [Quoted_string] from/to a string. *)

  val int        : int                       t
  val float      : float                     t
  val bool       : bool                      t
  val pair       : 'a t -> 'b t -> ('a * 'b) t
  val triple     : 'a t -> 'b t -> 'c t -> ('a * 'b * 'c) t
  val list       : 'a t -> 'a list           t
  val array      : 'a t -> 'a array          t
  val option     : 'a t -> 'a option         t

  val string_set : String_set.t            t
  (** [atom_set] is a conversion to/from a set of strings representing atoms. *)

  val string_map : 'a t -> 'a String_map.t   t
  (** [atom_map conv]: given a conversion [conv] to/from ['a], returns
     a conversion to/from a map where the keys are atoms and the
     values are of type ['a]. *)

  val string_hashtbl : 'a t -> (string, 'a) Hashtbl.t t
  (** [atom_hashtbl conv] is similar to [atom_map] for hash tables. *)
end

module To_sexp : sig
  type sexp = t
  include Combinators with type 'a t = 'a -> t

  val record : (string * sexp) list -> sexp

  type field

  val field
    :  string
    -> 'a t
    -> ?equal:('a -> 'a -> bool)
    -> ?default:'a
    -> 'a
    -> field
  val field_o : string -> 'a t-> 'a option -> field

  val record_fields : field list t
end with type sexp := t

module Of_sexp : sig
  type ast = Ast.t =
    | Atom of Loc.t * Atom.t
    | Quoted_string of Loc.t * string
    | List of Loc.t * ast list

  include Combinators with type 'a t = Ast.t -> 'a

  val of_sexp_error  : Ast.t -> string -> _
  val of_sexp_errorf : Ast.t -> ('a, unit, string, 'b) format4 -> 'a

  val located : 'a t -> (Loc.t * 'a) t

  val raw : ast t

  (* Record parsing monad *)
  type 'a record_parser
  val return : 'a -> 'a record_parser
  val ( >>= ) : 'a record_parser -> ('a -> 'b record_parser) -> 'b record_parser

  (** Return the location of the record being parsed *)
  val record_loc : Loc.t record_parser

  module Short_syntax : sig
    type 'a t =
      | Not_allowed
      | This    of 'a
      | Located of (Loc.t -> 'a)
  end

  val field
    :  string
    -> ?short:'a Short_syntax.t
    -> ?default:'a
    -> 'a t
    -> 'a record_parser
  val field_o
    :  string
    -> ?short:'a Short_syntax.t
    -> 'a t
    -> 'a option record_parser
  val field_b : string -> bool record_parser

  val map_validate : 'a record_parser -> f:('a -> ('b, string) result) -> 'b record_parser

  val ignore_fields : string list -> unit record_parser

  val record : 'a record_parser -> 'a t

  module Constructor_spec : sig
    type 'a t
  end

  module Constructor_args_spec : sig
    type ('a, 'b) t
  end

  val nil : ('a, 'a) Constructor_args_spec.t
  val ( @> )
    :  'a t
    -> ('b, 'c) Constructor_args_spec.t
    -> ('a -> 'b, 'c) Constructor_args_spec.t

  val cstr : string -> ('a, 'b) Constructor_args_spec.t -> 'a -> 'b Constructor_spec.t
  val cstr_rest
    :  string
    -> ('a, 'b list -> 'c) Constructor_args_spec.t
    -> 'b t
    -> 'a
    -> 'c Constructor_spec.t

  val cstr_record : string -> 'a record_parser -> 'a Constructor_spec.t

  val cstr_loc
    :  string
    -> ('a, 'b) Constructor_args_spec.t
    -> (Loc.t -> 'a)
    -> 'b Constructor_spec.t

  val cstr_rest_loc
    :  string
    -> ('a, 'b list -> 'c) Constructor_args_spec.t
    -> 'b t
    -> (Loc.t -> 'a)
    -> 'c Constructor_spec.t

  val sum
    :  'a Constructor_spec.t list
    -> 'a t

  val enum : (string * 'a) list -> 'a t
end
# 6127 "boot.ml"
end = struct
# 1 "src/sexp.ml"
open Import

include (Usexp : module type of struct include Usexp end
         with module Loc := Usexp.Loc)

let code_error message vars =
  code_errorf "%a" pp
    (List (Usexp.atom_or_quoted_string message
           :: List.map vars ~f:(fun (name, value) ->
             List [Usexp.atom_or_quoted_string name; value])))

let buf_len = 65_536

let load ~fname ~mode =
  Io.with_file_in fname ~f:(fun ic ->
    let state = Parser.create ~fname ~mode in
    let buf = Bytes.create buf_len in
    let rec loop stack =
      match input ic buf 0 buf_len with
      | 0 -> Parser.feed_eoi state stack
      | n -> loop (Parser.feed_subbytes state buf ~pos:0 ~len:n stack)
    in
    loop Parser.Stack.empty)

let load_many_as_one ~fname =
  match load ~fname ~mode:Many with
  | [] -> Ast.List (Loc.in_file fname, [])
  | x :: l ->
    let last = Option.value (List.last l) ~default:x in
    let loc = { (Ast.loc x) with stop = (Ast.loc last).stop } in
    Ast.List (loc, x :: l)

let ocaml_script_prefix = "(* -*- tuareg -*- *)"
let ocaml_script_prefix_len = String.length ocaml_script_prefix

type sexps_or_ocaml_script =
  | Sexps of Ast.t list
  | Ocaml_script

let load_many_or_ocaml_script fname =
  Io.with_file_in fname ~f:(fun ic ->
    let state = Parser.create ~fname ~mode:Many in
    let buf = Bytes.create buf_len in
    let rec loop stack =
      match input ic buf 0 buf_len with
      | 0 -> Parser.feed_eoi state stack
      | n -> loop (Parser.feed_subbytes state buf ~pos:0 ~len:n stack)
    in
    let rec loop0 stack i =
      match input ic buf i (buf_len - i) with
      | 0 ->
        let stack = Parser.feed_subbytes state buf ~pos:0 ~len:i stack in
        Sexps (Parser.feed_eoi state stack)
      | n ->
        let i = i + n in
        if i < ocaml_script_prefix_len then
          loop0 stack i
        else if Bytes.sub_string buf 0 ocaml_script_prefix_len
                  [@warning "-6"]
                = ocaml_script_prefix then
          Ocaml_script
        else
          let stack = Parser.feed_subbytes state buf ~pos:0 ~len:i stack in
          Sexps (loop stack)
    in
    loop0 Parser.Stack.empty 0)

module type Combinators = sig
  type 'a t
  val unit       : unit                      t
  val string     : string                    t
  val int        : int                       t
  val float      : float                     t
  val bool       : bool                      t
  val pair       : 'a t -> 'b t -> ('a * 'b) t
  val triple     : 'a t -> 'b t -> 'c t -> ('a * 'b * 'c) t
  val list       : 'a t -> 'a list           t
  val array      : 'a t -> 'a array          t
  val option     : 'a t -> 'a option         t
  val string_set : String_set.t            t
  val string_map : 'a t -> 'a String_map.t   t
  val string_hashtbl : 'a t -> (string, 'a) Hashtbl.t t
end

module To_sexp = struct
  type nonrec 'a t = 'a -> t
  let unit () = List []
  let string = Usexp.atom_or_quoted_string
  let int n = Atom (Atom.of_int n)
  let float f = Atom (Atom.of_float f)
  let bool b = Atom (Atom.of_bool b)
  let pair fa fb (a, b) = List [fa a; fb b]
  let triple fa fb fc (a, b, c) = List [fa a; fb b; fc c]
  let list f l = List (List.map l ~f)
  let array f a = list f (Array.to_list a)
  let option f = function
    | None -> List []
    | Some x -> List [f x]
  let string_set set = list atom (String_set.to_list set)
  let string_map f map = list (pair atom f) (String_map.to_list map)
  let record l =
    List (List.map l ~f:(fun (n, v) -> List [Atom(Atom.of_string n); v]))
  let string_hashtbl f h =
    string_map f
      (Hashtbl.foldi h ~init:String_map.empty ~f:(fun key data acc ->
         String_map.add acc key data))

  type field = string * Usexp.t option

  let field name f ?(equal=(=)) ?default v =
    match default with
    | None -> (name, Some (f v))
    | Some d ->
      if equal d v then
        (name, None)
      else
        (name, Some (f v))
  let field_o name f v = (name, Option.map ~f v)

  let record_fields (l : field list) =
    List (List.filter_map l ~f:(fun (k, v) ->
      Option.map v ~f:(fun v -> List[Atom (Atom.of_string k); v])))
end

module Of_sexp = struct
  type ast = Ast.t =
    | Atom of Loc.t * Atom.t
    | Quoted_string of Loc.t * string
    | List of Loc.t * ast list

  type 'a t = ast -> 'a

  let located f sexp =
    (Ast.loc sexp, f sexp)

  let of_sexp_error sexp str = raise (Loc.Error (Ast.loc sexp, str))
  let of_sexp_errorf sexp fmt = ksprintf (of_sexp_error sexp) fmt

  let raw x = x

  let unit = function
    | List (_, []) -> ()
    | sexp -> of_sexp_error sexp "() expected"

  let string = function
    | Atom (_, A s) -> s
    | Quoted_string (_, s) -> s
    | List _ as sexp -> of_sexp_error sexp "Atom or quoted string expected"

  let int sexp = match sexp with
    | Atom (_, A s) -> (try int_of_string s
                        with _ -> of_sexp_error sexp "Integer expected")
    | _ -> of_sexp_error sexp "Integer expected"

  let float sexp = match sexp with
    | Atom (_, A s) -> (try float_of_string s
                        with _ -> of_sexp_error sexp "Float expected")
    | _ -> of_sexp_error sexp "Float expected"

  let bool = function
    | Atom (_, A "true") -> true
    | Atom (_, A "false") -> false
    | sexp -> of_sexp_error sexp "'true' or 'false' expected"

  let pair fa fb = function
    | List (_, [a; b]) -> (fa a, fb b)
    | sexp -> of_sexp_error sexp "S-expression of the form (_ _) expected"

  let triple fa fb fc = function
    | List (_, [a; b; c]) -> (fa a, fb b, fc c)
    | sexp -> of_sexp_error sexp "S-expression of the form (_ _ _) expected"

  let list f = function
    | (Atom _ | Quoted_string _) as sexp -> of_sexp_error sexp "List expected"
    | List (_, l) -> List.map l ~f

  let array f sexp = Array.of_list (list f sexp)

  let option f = function
    | List (_, []) -> None
    | List (_, [x]) -> Some (f x)
    | sexp -> of_sexp_error sexp "S-expression of the form () or (_) expected"

  let string_set sexp = String_set.of_list (list string sexp)
  let string_map f sexp =
    match String_map.of_list (list (pair string f) sexp) with
    | Ok x -> x
    | Error (key, _v1, _v2) ->
      of_sexp_error sexp (sprintf "key %S present multiple times" key)

  let string_hashtbl f sexp =
    let map = string_map f sexp in
    let tbl = Hashtbl.create (String_map.cardinal map + 32) in
    String_map.iteri map ~f:(Hashtbl.add tbl);
    tbl

  type unparsed_field =
    { value : Ast.t option
    ; entry : Ast.t
    }

  module Name = struct
    type t = string
    let compare a b =
      let alen = String.length a and blen = String.length b in
      match Int.compare alen blen with
      | Eq -> String.compare a b
      | ne -> ne
  end

  module Name_map = Map.Make(Name)

  type record_parser_state =
    { loc      : Loc.t
    ; unparsed : unparsed_field Name_map.t
    ; known    : string list
    }

  type 'a record_parser = record_parser_state -> 'a * record_parser_state

  let return x state = (x, state)
  let (>>=) m f state =
    let x, state = m state in
    f x state

  let record_loc state =
    (state.loc, state)

  let consume name state =
    { state with
      unparsed = Name_map.remove state.unparsed name
    ; known    = name :: state.known
    }

  let add_known name state =
    { state with known = name :: state.known }

  let ignore_fields names state =
    let unparsed =
      List.fold_left names ~init:state.unparsed ~f:(fun acc name ->
        Name_map.remove acc name)
    in
    ((),
     { state with
       unparsed
     ; known = List.rev_append names state.known
     })

  let map_validate parse ~f state =
    let x, state' = parse state in
    match f x with
    | Ok x -> x, state'
    | Error msg ->
      let parsed =
        Name_map.merge state.unparsed state'.unparsed ~f:(fun _key before after ->
          match before, after with
          | Some _, None -> before
          | _ -> None)
      in
      let loc =
        match
          Name_map.values parsed
          |> List.map ~f:(fun f -> Ast.loc f.entry)
          |> List.sort ~compare:(fun a b -> compare a.Loc.start.pos_cnum b.start.pos_cnum)
        with
        | [] -> state.loc
        | first :: l ->
          let last = List.fold_left l ~init:first ~f:(fun _ x -> x) in
          { first with stop = last.stop }
      in
      Loc.fail loc "%s" msg

  module Short_syntax = struct
    type 'a t =
      | Not_allowed
      | This    of 'a
      | Located of (Loc.t -> 'a)

    let parse t entry name =
      match t with
      | Not_allowed ->
        Loc.fail (Ast.loc entry) "field %s needs a value" name
      | This    x -> x
      | Located f -> f (Ast.loc entry)
  end

  let field name ?(short=Short_syntax.Not_allowed)
        ?default value_of_sexp state =
    match Name_map.find state.unparsed name with
    | Some { value = Some value; _ } ->
      (value_of_sexp value, consume name state)
    | Some { value = None; entry } ->
      (Short_syntax.parse short entry name,
       consume name state)
    | None ->
      match default with
      | Some v -> (v, add_known name state)
      | None ->
        Loc.fail state.loc "field %s missing" name

  let field_o name ?(short=Short_syntax.Not_allowed) value_of_sexp state =
    match Name_map.find state.unparsed name with
    | Some { value = Some value; _ } ->
      (Some (value_of_sexp value), consume name state)
    | Some { value = None; entry } ->
      (Some (Short_syntax.parse short entry name),
       consume name state)
    | None -> (None, add_known name state)

  let field_b name = field name bool ~default:false ~short:(This true)

  let make_record_parser_state sexp =
    match sexp with
    | Atom _ | Quoted_string _ -> of_sexp_error sexp "List expected"
    | List (loc, sexps) ->
      let unparsed =
        List.fold_left sexps ~init:Name_map.empty ~f:(fun acc sexp ->
          match sexp with
          | List (_, [Atom (_, A name)]) ->
            Name_map.add acc name { value = None; entry = sexp }
          | List (_, [name_sexp; value]) -> begin
              match name_sexp with
              | Atom (_, A name) ->
                Name_map.add acc name { value = Some value; entry = sexp }
              | List _ | Quoted_string _ ->
                of_sexp_error name_sexp "Atom expected"
            end
          | _ ->
            of_sexp_error sexp "S-expression of the form (_ _) expected")
      in
      { loc    = loc
      ; known  = []
      ; unparsed
      }

  let record parse sexp =
    let state = make_record_parser_state sexp in
    let v, state = parse state in
    match Name_map.choose state.unparsed with
    | None -> v
    | Some (name, { entry; _ }) ->
      let name_sexp =
        match entry with
        | List (_, s :: _) -> s
        | _ -> assert false
      in
      of_sexp_errorf name_sexp
        "Unknown field %s%s" name (hint name state.known)

  type ('a, 'b) rest =
    | No_rest : ('a, 'a) rest
    | Many    : 'a t -> ('a list -> 'b, 'b) rest

  module Constructor_args_spec = struct
    type 'a conv = 'a t
    type ('a, 'b) t =
      | Nil  : ('a, 'a) t
      | Cons : 'a conv * ('b, 'c) t -> ('a -> 'b, 'c) t

    let rec convert : type a b c. (a, b) t -> (b, c) rest -> Ast.t -> Ast.t list -> a -> c
      = fun t rest sexp sexps f ->
        match t, rest, sexps with
        | Nil, No_rest, [] -> f
        | Nil, Many _ , [] -> f []
        | Cons _, _, [] -> of_sexp_error sexp "not enough arguments"
        | Nil, No_rest, _ :: _ -> of_sexp_error sexp "too many arguments"
        | Nil, Many conv, l -> f (List.map l ~f:conv)
        | Cons (conv, t), _, s :: sexps ->
          convert t rest sexp sexps (f (conv s))
  end

  let nil = Constructor_args_spec.Nil
  let ( @> ) a b = Constructor_args_spec.Cons (a, b)

  module Constructor_spec = struct
    type ('a, 'b, 'c) tuple =
      { name : string
      ; args : ('a, 'b) Constructor_args_spec.t
      ; rest : ('b, 'c) rest
      ; make : Loc.t -> 'a
      }

    type 'a record =
      { name  : string
      ; parse : 'a record_parser
      }

    type 'a t =
      | Tuple  : (_, _, 'a) tuple -> 'a t
      | Record : 'a record        -> 'a t

    let name = function
      | Tuple  x -> x.name
      | Record x -> x.name
  end
  module C = Constructor_spec

  let cstr_loc name args make =
    C.Tuple { name; args; make; rest = No_rest }
  let cstr_rest_loc name args rest make =
    C.Tuple { name; args; make; rest = Many rest }

  let cstr_record name parse =
    C.Record { name; parse }

  let cstr name args make =
    cstr_loc name args (fun _ -> make)

  let cstr_rest name args rest make =
    cstr_rest_loc name args rest (fun _ -> make)

  let equal_cstr_name a b = Name.compare a b = Eq

  let find_cstr cstrs sexp name =
    match
      List.find cstrs ~f:(fun cstr ->
        equal_cstr_name (C.name cstr) name)
    with
    | Some cstr -> cstr
    | None ->
      of_sexp_errorf sexp
        "Unknown constructor %s%s" name
        (hint
           (String.uncapitalize name)
           (List.map cstrs ~f:(fun c ->
              String.uncapitalize (C.name c))))

  let sum cstrs sexp =
    match sexp with
    | Atom (loc, A s) -> begin
        match find_cstr cstrs sexp s with
        | C.Tuple  t -> Constructor_args_spec.convert t.args t.rest sexp [] (t.make loc)
        | C.Record _ -> of_sexp_error sexp "'%s' expect arguments"
      end
    | Quoted_string _ -> of_sexp_error sexp "Atom expected"
    | List (_, []) -> of_sexp_error sexp "non-empty list expected"
    | List (loc, name_sexp :: args) ->
      match name_sexp with
      | Quoted_string _ | List _ -> of_sexp_error name_sexp "Atom expected"
      | Atom (_, A s) ->
        match find_cstr cstrs sexp s with
        | C.Tuple  t -> Constructor_args_spec.convert t.args t.rest sexp args (t.make loc)
        | C.Record r -> record r.parse (List (loc, args))

  let enum cstrs sexp =
    match sexp with
    | Quoted_string _ | List _ -> of_sexp_error sexp "Atom expected"
    | Atom (_, A s) ->
      match
        List.find cstrs ~f:(fun (name, _) ->
          equal_cstr_name name s)
      with
      | Some (_, value) -> value
      | None ->
        of_sexp_errorf sexp
          "Unknown value %s%s" s
          (hint
             (String.uncapitalize s)
             (List.map cstrs ~f:(fun (name, _) ->
                String.uncapitalize name)))
end
# 6591 "boot.ml"
end
module Path : sig
# 1 "src/path.mli"
open Import

(** In the current workspace (anything under the current project root) *)
module Local : sig
  type t

  val compare : t -> t -> Ordering.t

  module Set : Set.S with type elt = t

  val root : t
  val is_root : t -> bool
  val to_string : t -> string
  val mkdir_p : t -> unit
  val ensure_parent_directory_exists : t -> unit
  val append : t -> t -> t
  val descendant : t -> of_:t -> t option
  val parent : t -> t
end

(** In the outside world *)
module External : sig
  type t

  val to_string : t -> string
end

module Kind : sig
  type t =
    | External of External.t
    | Local    of Local.t
end

type t

val t : t Sexp.Of_sexp.t
val sexp_of_t : t Sexp.To_sexp.t

val compare : t -> t -> Ordering.t
(** a directory is smaller than its descendants *)

module Set : sig
  include Set.S with type elt = t
  val sexp_of_t : t Sexp.To_sexp.t
  val of_string_set : String_set.t -> f:(string -> elt) -> t
end

module Map : Map.S with type key = t


val kind : t -> Kind.t

val of_string : ?error_loc:Loc.t -> string -> t
val to_string : t -> string

(** [to_string_maybe_quoted t] is [maybe_quoted (to_string t)] *)
val to_string_maybe_quoted : t -> string

val root : t
val is_root : t -> bool

val is_local : t -> bool

val relative : ?error_loc:Loc.t -> t -> string -> t

(** Create an external path. If the argument is relative, assume it is
    relative to the initial directory jbuilder was launched in. *)
val absolute : string -> t

(** Convert a path to an absolute filename. Must be called after the
    workspace root has been set. *)
val to_absolute_filename : t -> string

val reach : t -> from:t -> string
val reach_for_running : t -> from:t -> string

val descendant : t -> of_:t -> t option
val is_descendant : t -> of_:t -> bool

val append : t -> t -> t

val basename : t -> string
val parent : t -> t

val extend_basename : t -> suffix:string -> t

(** Extract the build context from a path. For instance, representing paths as strings:

    {[
      extract_build_context "_build/blah/foo/bar" = Some ("blah", "foo/bar")
    ]}
*)
val extract_build_context : t -> (string * t) option

(** Same as [extract_build_context] but return the build context as a path:

    {[
      extract_build_context "_build/blah/foo/bar" = Some ("_build/blah", "foo/bar")
    ]}
*)
val extract_build_context_dir : t -> (t * t) option

(** Drop the "_build/blah" prefix *)
val drop_build_context : t -> t option
val drop_build_context_exn : t -> t

(** Drop the "_build/blah" prefix if present, return [t] otherwise *)
val drop_optional_build_context : t -> t

val explode : t -> string list option
val explode_exn : t -> string list

(** The build directory *)
val build_dir : t

(** [is_in_build_dir t = is_descendant t ~of:build_dir] *)
val is_in_build_dir : t -> bool

(** [is_in_build_dir t = is_local t && not (is_in_build_dir t)] *)
val is_in_source_tree : t -> bool

val is_alias_stamp_file : t -> bool

(**  Split after the first component if [t] is local *)
val split_first_component : t -> (string * t) option

val insert_after_build_dir_exn : t -> string -> t

val exists : t -> bool
val readdir : t -> string list
val is_directory : t -> bool
val rmdir : t -> unit
val unlink : t -> unit
val unlink_no_err : t -> unit
val rm_rf : t -> unit

(** Changes the extension of the filename (or adds an extension if there was none) *)
val change_extension : ext:string -> t -> t

val extension : t -> string

(** maintains the invariant:
    {[
      let suffix = Option.value_exn (Path.drop_prefix t ~prefix) in
      Path.relative prefix suffix = t
    ]}
*)
val drop_prefix : t -> prefix:t -> string option

val pp : t Fmt.t
# 6745 "boot.ml"
end = struct
# 1 "src/path.ml"
open Import

let explode_path =
  let rec loop path acc =
    let dir  = Filename.dirname  path in
    let base = Filename.basename path in
    let acc = base :: acc in
    if dir = Filename.current_dir_name then
      acc
    else
      loop dir acc
  in
  fun path -> loop path []

module External = struct
  type t = string

  let to_string t = t
(*
  let rec cd_dot_dot t =
    match Unix.readlink t with
    | exception _ -> Filename.dirname t
    | t -> cd_dot_dot t

  let relative initial_t path =
    let rec loop t components =
      match components with
      | [] | ["." | ".."] ->
        die "invalid filename concatenation: %s / %s" initial_t path
      | [fn] -> Filename.concat t fn
      | "."  :: rest -> loop t rest
      | ".." :: rest -> loop (cd_dot_dot t) rest
      | comp :: rest -> loop (Filename.concat t comp) rest
    in
    loop initial_t (explode_path path)
*)

  let relative = Filename.concat
end

let is_root = function
  | "" -> true
  | _  -> false

module Local = struct
  (* either "" for root, either a '/' separated list of components other that ".", ".."
     and not containing '/'. *)
  type t = string

  let root = ""

  let is_root = function
    | "" -> true
    | _  -> false

  let to_string = function
    | "" -> "."
    | t  -> t

  let compare = String.compare

  module Set = String_set

  let to_list =
    let rec loop t acc i j =
      if i = 0 then
        String.sub t ~pos:0 ~len:j :: acc
      else
        match t.[i - 1] with
        | '/' -> loop t (String.sub t ~pos:i ~len:(j - i) :: acc) (i - 1) (i - 1)
        | _   -> loop t acc (i - 1) j
    in
    function
    | "" -> []
    | t  ->
      let len = String.length t in
      loop t [] len len

  let parent = function
    | "" ->
      code_errorf "Path.Local.parent called on the root"
    | t ->
      match String.rindex_from t (String.length t - 1) '/' with
      | exception Not_found -> ""
      | i -> String.sub t ~pos:0 ~len:i

  let basename = function
    | "" ->
      code_errorf "Path.Local.basename called on the root"
    | t ->
      let len = String.length t in
      match String.rindex_from t (len - 1) '/' with
      | exception Not_found -> t
      | i -> String.sub t ~pos:(i + 1) ~len:(len - i - 1)

  let relative ?error_loc t path =
    let rec loop t components =
      match components with
      | [] -> Ok t
      | "." :: rest -> loop t rest
      | ".." :: rest ->
        begin match t with
        | "" -> Error ()
        | t -> loop (parent t) rest
        end
      | fn :: rest ->
        match t with
        | "" -> loop fn rest
        | _ -> loop (t ^ "/" ^ fn) rest
    in
    match loop t (explode_path path) with
    | Ok t -> t
    | Error () ->
       Loc.fail_opt error_loc "path outside the workspace: %s from %s" path
         (to_string t)

  let is_canonicalized =
    let rec before_slash s i =
      if i < 0 then
        false
      else
        match s.[i] with
        | '/' -> false
        | '.' -> before_dot_slash s (i - 1)
        | _   -> in_component     s (i - 1)
    and before_dot_slash s i =
      if i < 0 then
        false
      else
        match s.[i] with
        | '/' -> false
        | '.' -> before_dot_dot_slash s (i - 1)
        | _   -> in_component         s (i - 1)
    and before_dot_dot_slash s i =
      if i < 0 then
        false
      else
        match s.[i] with
        | '/' -> false
        | _   -> in_component s (i - 1)
    and in_component s i =
      if i < 0 then
        true
      else
        match s.[i] with
        | '/' -> before_slash s (i - 1)
        | _   -> in_component s (i - 1)
    in
    fun s ->
      let len = String.length s in
      if len = 0 then
        true
      else
        before_slash s (len - 1)

  let of_string ?error_loc s =
    if is_canonicalized s then
      s
    else
      relative "" s ?error_loc

  let rec mkdir_p = function
    | "" -> ()
    | t ->
      try
        Unix.mkdir t 0o777
      with
      | Unix.Unix_error (EEXIST, _, _) -> ()
      | Unix.Unix_error (ENOENT, _, _) as e ->
        match parent t with
        | "" -> raise e
        | p ->
          mkdir_p p;
          Unix.mkdir t 0o777

  let ensure_parent_directory_exists = function
    | "" -> ()
    | t -> mkdir_p (parent t)

  let append a b =
    match a, b with
    | "", x | x, "" -> x
    | _ -> a ^ "/" ^ b

  let descendant t ~of_ =
    match of_ with
    | "" -> Some t
    | _ ->
      let of_len = String.length of_ in
      let t_len = String.length t in
      if t_len = of_len then
        Option.some_if (t = of_) t
      else if (t_len >= of_len && t.[of_len] = '/' && String.is_prefix t ~prefix:of_) then
        Some (String.sub t ~pos:(of_len + 1) ~len:(t_len - of_len - 1))
      else
        None

  let is_descendant t ~of_ =
    match of_ with
    | "" -> true
    | _ ->
      let of_len = String.length of_ in
      let t_len = String.length t in
      (t_len = of_len && t = of_) ||
      (t_len >= of_len && t.[of_len] = '/' && String.is_prefix t ~prefix:of_)

  let reach t ~from =
    let rec loop t from =
      match t, from with
      | a :: t, b :: from when a = b ->
        loop t from
      | _ ->
        match List.fold_left from ~init:t ~f:(fun acc _ -> ".." :: acc) with
        | [] -> "."
        | l -> String.concat l ~sep:"/"
    in
    loop (to_list t) (to_list from)
end

type t = string
let compare = String.compare

module Set = struct
  include String_set
  let sexp_of_t t = Sexp.To_sexp.(list string) (String_set.to_list t)
  let of_string_set = map
end

module Map = String_map

module Kind = struct
  type t =
    | External of External.t
    | Local    of Local.t
end

let is_local t = is_root t || Filename.is_relative t

let kind t : Kind.t =
  if is_local t then
    Local t
  else
    External t

let to_string = function
  | "" -> "."
  | t  -> t

let to_string_maybe_quoted t =
  maybe_quoted (to_string t)

let root = ""

let relative ?error_loc t fn =
  if fn = "" then
    t
  else
    match is_local t, is_local fn with
    | true, true  -> Local.relative t fn ?error_loc
    | _   , false -> fn
    | false, true -> External.relative t fn

let of_string ?error_loc s =
  match s with
  | "" -> ""
  | s  ->
    if Filename.is_relative s then
      Local.of_string s ?error_loc
    else
      s

let t sexp = of_string (Sexp.Of_sexp.string sexp) ~error_loc:(Sexp.Ast.loc sexp)
let sexp_of_t t = Sexp.atom_or_quoted_string (to_string t)

let absolute fn =
  if is_local fn then
    Filename.concat initial_cwd fn
  else
    fn

let to_absolute_filename t =
  if is_local t then begin
    let root = !Clflags.workspace_root in
    assert (not (Filename.is_relative root));
    Filename.concat root (to_string t)
  end else
    t

let reach t ~from =
  match is_local t, is_local from with
  | false, _ -> t
  | true, false ->
    Sexp.code_error "Path.reach called with invalid combination"
      [ "t"   , sexp_of_t t
      ; "from", sexp_of_t from
      ]
  | true, true -> Local.reach t ~from

let reach_for_running t ~from =
  match is_local t, is_local from with
  | false, _ -> t
  | true, false ->
    Sexp.code_error "Path.reach_for_running called with invalid combination"
      [ "t"   , sexp_of_t t
      ; "from", sexp_of_t from
      ]
  | true, true ->
    let s = Local.reach t ~from in
    if String.is_prefix s ~prefix:"../" then
      s
    else
      "./" ^ s

let descendant t ~of_ =
  if is_local t && is_local of_ then
    Local.descendant t ~of_
  else
    None

let is_descendant t ~of_ =
  if is_local t && is_local of_ then
    Local.is_descendant t ~of_
  else
    false

let append a b =
  if not (is_local b) then
    Sexp.code_error "Path.append called with non-local second path"
      [ "a", sexp_of_t a
      ; "b", sexp_of_t b
      ];
  if is_local a then
    Local.append a b
  else
    Filename.concat a b

let basename t =
  if is_local t then
    Local.basename t
  else
    Filename.basename t

let parent t =
  if is_local t then
    Local.parent t
  else
    Filename.dirname t

let build_prefix = "_build/"

let build_dir = "_build"

let is_in_build_dir t =
  String.is_prefix t ~prefix:build_prefix

let is_in_source_tree t = is_local t && not (is_in_build_dir t)

let is_alias_stamp_file t =
  String.is_prefix t ~prefix:"_build/.aliases/"

let extract_build_context t =
  if String.is_prefix t ~prefix:build_prefix then
    let i = String.length build_prefix in
    match String.index_from t i '/' with
    | exception _ ->
      Some
        (String.sub t ~pos:i ~len:(String.length t - i),
         "")
    | j ->
      Some
        (String.sub t ~pos:i ~len:(j - i),
         String.sub t ~pos:(j + 1) ~len:(String.length t - j - 1))
  else
    None

let extract_build_context_dir t =
  if String.is_prefix t ~prefix:build_prefix then
    let i = String.length build_prefix in
    match String.index_from t i '/' with
    | exception _ ->
      Some (t, "")
    | j ->
      Some
        (String.sub t ~pos:0 ~len:j,
         String.sub t ~pos:(j + 1) ~len:(String.length t - j - 1))
  else
    None

let drop_build_context t =
  Option.map (extract_build_context t) ~f:snd

let drop_build_context_exn t =
  match extract_build_context t with
  | None -> Sexp.code_error "Path.drop_build_context_exn" [ "t", sexp_of_t t ]
  | Some (_, t) -> t

let drop_optional_build_context t =
  match extract_build_context t with
  | None -> t
  | Some (_, t) -> t

let split_first_component t =
  if is_local t && not (is_root t)then
    match String.index t '/' with
    | None -> Some (t, root)
    | Some i ->
      Some
        (String.sub t ~pos:0 ~len:i,
         String.sub t ~pos:(i + 1) ~len:(String.length t - i - 1))
  else
    None

let explode t =
  if is_root t then
    Some []
  else if is_local t then
    Some (String.split t ~on:'/')
  else
    None

let explode_exn t =
  if is_root t then
    []
  else if is_local t then
    String.split t ~on:'/'
  else
    Sexp.code_error "Path.explode_exn"
      ["path", Sexp.atom_or_quoted_string t]

let exists t = Sys.file_exists (to_string t)
let readdir t = Sys.readdir (to_string t) |> Array.to_list
let is_directory t =
  try Sys.is_directory (to_string t)
  with Sys_error _ -> false
let rmdir t = Unix.rmdir (to_string t)
let win32_unlink fn =
  try
    Unix.unlink fn
  with Unix.Unix_error (Unix.EACCES, _, _) as e ->
    (* Try removing the read-only attribute *)
    try
      Unix.chmod fn 0o666;
      Unix.unlink fn
    with _ ->
      raise e
let unlink_operation =
  if Sys.win32 then
    win32_unlink
  else
    Unix.unlink
let unlink t =
  unlink_operation (to_string t)
let unlink_no_err t = try unlink t with _ -> ()

let extend_basename t ~suffix = t ^ suffix

let insert_after_build_dir_exn =
  let error a b =
    Sexp.code_error
      "Path.insert_after_build_dir_exn"
      [ "path"  , Sexp.unsafe_atom_of_string a
      ; "insert", Sexp.unsafe_atom_of_string b
      ]
  in
  fun a b ->
    if not (is_local a) || String.contains b '/' then error a b;
    match String.lsplit2 a ~on:'/' with
    | Some ("_build", rest) ->
      sprintf "_build/%s/%s" b rest
    | _ ->
      error a b

let rm_rf =
  let rec loop dir =
    Array.iter (Sys.readdir dir) ~f:(fun fn ->
      let fn = Filename.concat dir fn in
      match Unix.lstat fn with
      | { st_kind = S_DIR; _ } -> loop fn
      | _                      -> unlink_operation fn);
    Unix.rmdir dir
  in
  fun t ->
    let fn = to_string t in
    match Unix.lstat fn with
    | exception Unix.Unix_error(ENOENT, _, _) -> ()
    | _ -> loop fn

let change_extension ~ext t =
  let t = try Filename.chop_extension t with Not_found -> t in
  t ^ ext

let extension = Filename.extension

let pp ppf t = Format.pp_print_string ppf (to_string t)

let drop_prefix t ~prefix =
  let t = to_string t in
  let prefix =
    to_string (
      if String.is_suffix prefix ~suffix:"/" then
        prefix
      else
        prefix ^ "/") in
  String.drop_prefix t ~prefix
# 7252 "boot.ml"
end
module Action_intf = struct
# 1 "src/action_intf.ml"
module Outputs = struct
  type t =
    | Stdout
    | Stderr
    | Outputs (** Both Stdout and Stderr *)
end

module type Ast = sig
  type program
  type path
  type string

  module Diff : sig
    type file = { src : path; dst : path }

    type t =
      { optional : bool
      ; file1    : path
      ; file2    : path
      }
  end

  type t =
    | Run            of program * string list
    | Chdir          of path * t
    | Setenv         of string * string * t
    | Redirect       of Outputs.t * path * t
    | Ignore         of Outputs.t * t
    | Progn          of t list
    | Echo           of string
    | Cat            of path
    | Copy           of path * path
    | Symlink        of path * path
    | Copy_and_add_line_directive of path * path
    | System         of string
    | Bash           of string
    | Write_file     of path * string
    | Rename         of path * path
    | Remove_tree    of path
    | Mkdir          of path
    | Digest_files   of path list
    | Diff           of Diff.t
end

module type Helpers = sig
  type program
  type path
  type string
  type t

  val run : program -> string list -> t
  val chdir : path -> t -> t
  val setenv : string -> string -> t -> t
  val with_stdout_to : path -> t -> t
  val with_stderr_to : path -> t -> t
  val with_outputs_to : path -> t -> t
  val ignore_stdout : t -> t
  val ignore_stderr : t -> t
  val ignore_outputs : t -> t
  val progn : t list -> t
  val echo : string -> t
  val cat : path -> t
  val copy : path -> path -> t
  val symlink : path -> path -> t
  val copy_and_add_line_directive : path -> path -> t
  val system : string -> t
  val bash : string -> t
  val write_file : path -> string -> t
  val rename : path -> path -> t
  val remove_tree : path -> t
  val mkdir : path -> t
  val digest_files : path list -> t
  val diff : ?optional:bool -> Path.t -> Path.t -> t
end
# 7330 "boot.ml"
end
module Env : sig
# 1 "src/env.mli"
open Import

module Var : sig
  type t = string
  val compare : t -> t -> Ordering.t
end

type t

module Map : Map.S with type key = Var.t

(** The environment when the process started *)
val initial : t

val to_unix : t -> string array

val get : t -> Var.t -> string option

val extend : t -> vars:string Map.t -> t

val extend_env : t -> t -> t

val add : t -> var:Var.t -> value:string -> t

val diff : t -> t -> t

val update : t -> var:string -> f:(string option -> string option) -> t

val sexp_of_t : t -> Sexp.t

# 7364 "boot.ml"
end = struct
# 1 "src/env.ml"
open Import

module Var = struct
  type t = string
  let compare =
    if Sys.win32 then (
      fun a b -> String.compare (String.lowercase a) (String.lowercase b)
    ) else (
      String.compare
    )
end

module Map = Map.Make(Var)

type t =
  { vars : string Map.t
  ; mutable unix : string array option
  }

let make vars =
  { vars
  ; unix = None
  }

let get t k = Map.find t.vars k

let to_unix t =
  match t.unix with
  | Some v -> v
  | None ->
    let res =
      Map.foldi ~init:[] ~f:(fun k v acc ->
        (sprintf "%s=%s" k v)::acc
      ) t.vars
      |> Array.of_list in
    t.unix <- Some res;
    res

let of_unix arr =
  Array.to_list arr
  |> List.map ~f:(fun s ->
    match String.lsplit2 s ~on:'=' with
    | None ->
      Sexp.code_error "Env.of_unix: entry without '=' found in the environ"
        ["var", Sexp.To_sexp.string s]
    | Some (k, v) -> (k, v))
  |> Map.of_list_multi
  |> Map.map ~f:(function
    | [] -> assert false
    | x::_ -> x)

let initial = make (of_unix (Unix.environment ()))

let add t ~var ~value =
  make (Map.add t.vars var value)

let extend t ~vars =
  make (Map.union t.vars vars ~f:(fun _ _ v -> Some v))

let extend_env x y =
  extend x ~vars:y.vars

let sexp_of_t t =
  let open Sexp.To_sexp in
  (list (pair string string)) (Map.to_list t.vars)

let diff x y =
  Map.merge x.vars y.vars ~f:(fun _k vx vy ->
    match vy with
    | Some _ -> None
    | None -> vx)
  |> make

let update t ~var ~f =
  make (Map.update t.vars var ~f)
# 7442 "boot.ml"
end
module Config : sig
# 1 "src/config.mli"
(** Configuration parameters *)

open! Import

(** Local installation directory *)
val local_install_dir : context:string -> Path.t

val local_install_bin_dir : context:string -> Path.t
val local_install_man_dir : context:string -> Path.t
val local_install_lib_dir : context:string -> package:string -> Path.t

val dev_null : Path.t

(** When this file is present in a directory jbuilder will delete
    nothing in it if it knows to generate this file. *)
val jbuilder_keep_fname : string

(** Are we running inside an emacs shell? *)
val inside_emacs : bool

(** Are we running insinde Dune? *)
val inside_dune : bool

(** Jbuilder configuration *)

module Display : sig
  type t =
    | Progress (** Single interactive status line *)
    | Short    (** One line per command           *)
    | Verbose  (** Display all commands fully     *)
    | Quiet    (** Only display errors            *)

  val t : t Sexp.Of_sexp.t
  val all : (string * t) list
end

module type S = sig
  type 'a field

  type t =
    { display     : Display.t field
    ; concurrency : int       field
    }
end

include S with type 'a field = 'a

module Partial : S with type 'a field := 'a option

val t : t Sexp.Of_sexp.t

val merge : t -> Partial.t -> t

val default : t
val user_config_file : string
val load_user_config_file : unit -> t
val load_config_file : fname:string -> t

(** Set display mode to [Quiet] if it is [Progress], the output is not
    a tty and we are not running inside emacs. *)
val adapt_display : t -> output_is_a_tty:bool -> t
# 7507 "boot.ml"
end = struct
# 1 "src/config.ml"
open! Import

let local_install_dir =
  let dir = Path.(relative root) "_build/install" in
  fun ~context -> Path.relative dir context

let local_install_bin_dir ~context =
  Path.relative (local_install_dir ~context) "bin"

let local_install_man_dir ~context =
  Path.relative (local_install_dir ~context) "bin"

let local_install_lib_dir ~context ~package =
  Path.relative
    (Path.relative (local_install_dir ~context) "lib")
    package

let dev_null = Path.of_string (if Sys.win32 then "nul" else "/dev/null")

let jbuilder_keep_fname = ".jbuilder-keep"

let inside_emacs = Option.is_some (Env.get Env.initial "INSIDE_EMACS")
let inside_dune  = Option.is_some (Env.get Env.initial "INSIDE_DUNE")

open Sexp.Of_sexp

module Display = struct
  type t =
    | Progress
    | Short
    | Verbose
    | Quiet

  let all =
      [ "progress" , Progress
      ; "verbose"  , Verbose
      ; "short"    , Short
      ; "quiet"    , Quiet
      ]

  let t = enum all
end

module type S = sig
  type 'a field

  type t =
    { display     : Display.t field
    ; concurrency : int       field
    }
end

module rec M : S with type 'a field = 'a = M
include M

module rec Partial : S with type 'a field := 'a option = Partial

let merge t (partial : Partial.t) =
  let field from_t from_partial =
    Option.value from_partial ~default:from_t
  in
  { display     = field t.display     partial.display
  ; concurrency = field t.concurrency partial.concurrency
  }

let default =
  { display     = if inside_dune then Quiet else Progress
  ; concurrency = if inside_dune then 1     else 4
  }

let t =
  record
    (field "display" Display.t ~default:default.display
     >>= fun display ->
     field "jobs" int ~default:default.concurrency
     >>= fun concurrency ->
     return { display
            ; concurrency
            })

let user_config_file = Filename.concat Xdg.config_dir "dune/config"

let load_config_file ~fname =
  t (Sexp.load_many_as_one ~fname)

let load_user_config_file () =
  if Sys.file_exists user_config_file then
    load_config_file ~fname:user_config_file
  else
    default

let adapt_display config ~output_is_a_tty =
  if config.display = Progress &&
     not output_is_a_tty &&
     not inside_emacs
  then
    { config with display = Quiet }
  else
    config
# 7609 "boot.ml"
end
module Bin : sig
# 1 "src/bin.mli"
(** OCaml binaries *)

(** Character used to separate entries in [PATH] and similar
    environment variables *)
val path_sep : char

(** Contents of [PATH] *)
val path : Path.t list

(** Parse a [PATH] like variable *)
val parse_path : ?sep:char -> string -> Path.t list

(** The opam tool *)
val opam : Path.t option

(** Extension to append to executable filenames *)
val exe : string

(** Look for a program in the PATH *)
val which : ?path:Path.t list -> string -> Path.t option

(** Return the .opt version of a tool if available. If the tool is not available at all in
    the given directory, returns [None]. *)
val best_prog : Path.t -> string -> Path.t option

(** "make" program *)
val make : Path.t option

# 7641 "boot.ml"
end = struct
# 1 "src/bin.ml"
open Import

let path_sep =
  if Sys.win32 then
    ';'
  else
    ':'

let parse_path ?(sep=path_sep) s =
  List.map (String.split s ~on:sep) ~f:Path.absolute

let path =
  match Env.get Env.initial "PATH" with
  | None   -> []
  | Some s -> parse_path s

let exe = if Sys.win32 then ".exe" else ""

let best_prog dir prog =
  let fn = Path.relative dir (prog ^ ".opt" ^ exe) in
  if Path.exists fn then
    Some fn
  else
    let fn = Path.relative dir (prog ^ exe) in
    if Path.exists fn then
      Some fn
    else
      None

let which ?(path=path) prog =
  let rec search = function
    | [] -> None
    | dir :: rest ->
      match best_prog dir prog with
      | None -> search rest
      | Some fn -> Some fn
  in
  search path

let opam = which "opam"

let make =
  match which "gmake" with
  | None -> which "make"
  | some -> some
# 7689 "boot.ml"
end
module Colors : sig
# 1 "src/colors.mli"
open Stdune

val colorize : key:string -> string -> string

val stderr_supports_colors : bool Lazy.t

(** [Env.initial] extended with variables to force a few tools to
    print colors *)
val setup_env_for_colors : Env.t -> Env.t

(** Strip colors in [not (Lazy.force stderr_supports_colors)] *)
val strip_colors_for_stderr : string -> string

(** Enable the interpretation of color tags for [Format.err_formatter] *)
val setup_err_formatter_colors : unit -> unit

type styles

val output_filename : styles

val apply_string : styles -> string -> string

module Style : sig
  type t =
    | Loc
    | Error
    | Warning
    | Kwd
    | Id
    | Prompt
    | Details
    | Ok
    | Debug
end

module Render : Pp.Renderer.S
  with type Tag.t = Style.t
# 7730 "boot.ml"
end = struct
# 1 "src/colors.ml"
open Import

type styles = Ansi_color.Style.t list

let apply_string styles str =
  sprintf "%s%s%s"
    (Ansi_color.Style.escape_sequence styles)
    str
    (Ansi_color.Style.escape_sequence [])

let colorize =
  let color_combos =
    let open Ansi_color.Color in
    [| Blue,          Bright_green
     ; Red,           Bright_yellow
     ; Yellow,        Blue
     ; Magenta,       Bright_cyan
     ; Bright_green,  Blue
     ; Bright_yellow, Red
     ; Blue,          Yellow
     ; Bright_cyan,   Magenta
    |]
  in
  fun ~key str ->
    let hash = Hashtbl.hash key in
    let fore, back = color_combos.(hash mod (Array.length color_combos)) in
    apply_string [Fg fore; Bg back] str

let stderr_supports_colors = lazy(
  not Sys.win32        &&
  Unix.(isatty stderr) &&
  match Env.get Env.initial "TERM" with
  | None        -> false
  | Some "dumb" -> false
  | Some _      -> true
)

let strip_colors_for_stderr s =
  if Lazy.force stderr_supports_colors then
    s
  else
    Ansi_color.strip s

(* We redirect the output of all commands, so by default the various
   tools will disable colors. Since we support colors in the output of
   commands, we force it via specific environment variables if stderr
   supports colors. *)
let setup_env_for_colors env =
  let set env var value =
    Env.update env ~var ~f:(function
      | None   -> Some value
      | Some s -> Some s)
  in
  let env = set env "OPAMCOLOR"   "always" in
  let env = set env "OCAML_COLOR" "always" in
  env

module Style = struct
  open Ansi_color.Style

  type t =
    | Loc
    | Error
    | Warning
    | Kwd
    | Id
    | Prompt
    | Details
    | Ok
    | Debug

  let to_styles = function
    | Loc     -> [Bold]
    | Error   -> [Bold; Fg Red]
    | Warning -> [Bold; Fg Magenta]
    | Kwd     -> [Bold; Fg Blue]
    | Id      -> [Bold; Fg Yellow]
    | Prompt  -> [Bold; Fg Green]
    | Details -> [Dim; Fg White]
    | Ok      -> [Dim; Fg Green]
    | Debug   -> [Underlined; Fg Bright_cyan]

  let of_string = function
    | "loc"     -> Some Loc
    | "error"   -> Some Error
    | "warning" -> Some Warning
    | "kwd"     -> Some Kwd
    | "id"      -> Some Id
    | "prompt"  -> Some Prompt
    | "details" -> Some Details
    | "ok"      -> Some Ok
    | "debug"   -> Some Debug
    | _         -> None
end

let styles_of_tag s =
  match Style.of_string s with
  | None -> []
  | Some style -> Style.to_styles style

let setup_err_formatter_colors () =
  let open Format in
  if Lazy.force stderr_supports_colors then begin
    List.iter [err_formatter; err_ppf] ~f:(fun ppf ->
      let funcs = pp_get_formatter_tag_functions ppf () in
      pp_set_mark_tags ppf true;
      pp_set_formatter_tag_functions ppf
        { funcs with
          mark_close_tag = (fun _   -> Ansi_color.Style.escape_sequence [])
        ; mark_open_tag  = (fun tag -> Ansi_color.Style.escape_sequence
                                         (styles_of_tag tag))
        })
  end

let output_filename : styles = [Bold; Fg Green]

module Render = Pp.Renderer.Make(struct
    type t = Style.t

    module Handler = struct
      include Ansi_color.Render.Tag.Handler

      let handle t style = handle t (Style.to_styles style)
    end
  end)
# 7858 "boot.ml"
end
module Meta_lexer : sig
# 1 "src/meta_lexer.mli"
type token =
  | Name   of string
  | String of string
  | Minus
  | Lparen
  | Rparen
  | Comma
  | Equal
  | Plus_equal
  | Eof

val token : Lexing.lexbuf -> token
# 7874 "boot.ml"
end = struct
# 1 "src/meta_lexer.ml"
# 1 "src/meta_lexer.mll"
 
type token =
  | Name   of string
  | String of string
  | Minus
  | Lparen
  | Rparen
  | Comma
  | Equal
  | Plus_equal
  | Eof

let escaped_buf = Buffer.create 256

# 17 "src/meta_lexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\243\255\244\255\001\000\246\255\247\255\248\255\249\255\
    \250\255\251\255\077\000\253\255\001\000\005\000\245\255\002\000\
    \252\255\253\255\254\255\005\000\255\255";
  Lexing.lex_backtrk =
   "\000\000\255\255\255\255\012\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\003\000\255\255\001\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\002\000\255\255";
  Lexing.lex_default =
   "\001\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\000\000\012\000\255\255\000\000\017\000\
    \000\000\000\000\000\000\017\000\000\000";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\013\000\011\000\255\255\018\000\013\000\013\000\018\000\
    \000\000\000\000\013\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \013\000\000\000\009\000\012\000\020\000\013\000\000\000\000\000\
    \007\000\006\000\000\000\003\000\005\000\008\000\010\000\000\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\000\000\000\000\000\000\004\000\014\000\000\000\
    \000\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\000\000\000\000\000\000\019\000\010\000\
    \000\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\000\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \000\000\000\000\000\000\000\000\010\000\000\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \002\000\255\255\016\000\000\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\012\000\015\000\000\000\013\000\019\000\
    \255\255\255\255\013\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\000\000\000\000\015\000\013\000\255\255\255\255\
    \000\000\000\000\255\255\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\000\000\003\000\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\015\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\010\000\255\255\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \255\255\255\255\255\255\255\255\010\000\255\255\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\012\000\015\000\255\255\255\255\019\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_base_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\004\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\004\000\000\000";
  Lexing.lex_default_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\007\000\000\000";
  Lexing.lex_trans_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\001\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_check_code =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\019\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\015\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \019\000";
  Lexing.lex_code =
   "\255\002\255\255\000\001\255\001\002\255";
}

let rec token lexbuf =
   __ocaml_lex_token_rec lexbuf 0
and __ocaml_lex_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 17 "src/meta_lexer.mll"
                     ( token lexbuf )
# 208 "src/meta_lexer.ml"

  | 1 ->
# 18 "src/meta_lexer.mll"
                  ( token lexbuf )
# 213 "src/meta_lexer.ml"

  | 2 ->
# 19 "src/meta_lexer.mll"
         ( Lexing.new_line lexbuf; token lexbuf )
# 218 "src/meta_lexer.ml"

  | 3 ->
let
# 21 "src/meta_lexer.mll"
                                          s
# 224 "src/meta_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 21 "src/meta_lexer.mll"
                                            ( Name s )
# 228 "src/meta_lexer.ml"

  | 4 ->
# 23 "src/meta_lexer.mll"
      ( Buffer.clear escaped_buf;
        string escaped_buf lexbuf )
# 234 "src/meta_lexer.ml"

  | 5 ->
# 25 "src/meta_lexer.mll"
        ( Minus )
# 239 "src/meta_lexer.ml"

  | 6 ->
# 26 "src/meta_lexer.mll"
        ( Lparen )
# 244 "src/meta_lexer.ml"

  | 7 ->
# 27 "src/meta_lexer.mll"
        ( Rparen )
# 249 "src/meta_lexer.ml"

  | 8 ->
# 28 "src/meta_lexer.mll"
        ( Comma )
# 254 "src/meta_lexer.ml"

  | 9 ->
# 29 "src/meta_lexer.mll"
        ( Equal )
# 259 "src/meta_lexer.ml"

  | 10 ->
# 30 "src/meta_lexer.mll"
         ( Plus_equal )
# 264 "src/meta_lexer.ml"

  | 11 ->
# 31 "src/meta_lexer.mll"
        ( Eof )
# 269 "src/meta_lexer.ml"

  | 12 ->
# 32 "src/meta_lexer.mll"
      ( Loc.fail_lex lexbuf "invalid character" )
# 274 "src/meta_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_token_rec lexbuf __ocaml_lex_state

and string buf lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 3 (-1);(* L=1 [1] <- p ;  *)
  lexbuf.Lexing.lex_mem.(1) <- lexbuf.Lexing.lex_curr_pos ;
 __ocaml_lex_string_rec buf lexbuf 15
and __ocaml_lex_string_rec buf lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 36 "src/meta_lexer.mll"
      ( String (Buffer.contents buf) )
# 288 "src/meta_lexer.ml"

  | 1 ->
# 39 "src/meta_lexer.mll"
      ( Lexing.new_line lexbuf;
        Buffer.add_char buf '\n';
        string buf lexbuf )
# 295 "src/meta_lexer.ml"

  | 2 ->
let
# 42 "src/meta_lexer.mll"
               c
# 301 "src/meta_lexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_mem.(0) in
# 44 "src/meta_lexer.mll"
      ( Buffer.add_char buf c;
        string buf lexbuf )
# 306 "src/meta_lexer.ml"

  | 3 ->
# 47 "src/meta_lexer.mll"
      ( Loc.fail_lex lexbuf "unterminated string" )
# 311 "src/meta_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_string_rec buf lexbuf __ocaml_lex_state

;;

# 8193 "boot.ml"
end
module Meta : sig
# 1 "src/meta.mli"
(** META file parsing/printing *)

open! Import

type t =
  { name    : string
  ; entries : entry list
  }

and entry =
  | Comment of string
  | Rule    of rule
  | Package of t

and rule =
  { var        : string
  ; predicates : predicate list
  ; action     : action
  ; value      : string
  }

and action = Set | Add

and predicate =
  | Pos of string
  | Neg of string

module Simplified : sig
  module Rules : sig
    type t =
      { set_rules : rule list
      ; add_rules : rule list
      }
  end

  type t =
    { name : string
    ; vars : Rules.t String_map.t
    ; subs : t list
    }

  val pp : Format.formatter -> t -> unit
end

val load : fn:string -> name:string -> Simplified.t

(** Builtin META files for libraries distributed with the compiler. For when ocamlfind is
    not installed. *)
val builtins : stdlib_dir:Path.t -> Simplified.t String_map.t

val pp : Format.formatter -> entry list -> unit
# 8248 "boot.ml"
end = struct
# 1 "src/meta.ml"
open Import

type t =
  { name    : string
  ; entries : entry list
  }

and entry =
  | Comment of string
  | Rule    of rule
  | Package of t

and rule =
  { var        : string
  ; predicates : predicate list
  ; action     : action
  ; value      : string
  }

and action = Set | Add

and predicate =
  | Pos of string
  | Neg of string

module Parse = struct
  let error = Loc.fail_lex

  let next = Meta_lexer.token

  let package_name lb =
    match next lb with
    | String s ->
      if String.contains s '.' then
        error lb "'.' not allowed in sub-package names";
      s
    | _ -> error lb "package name expected"

  let string lb =
    match next lb with
    | String s -> s
    | _ -> error lb "string expected"

  let lparen lb =
    match next lb with
    | Lparen -> ()
    | _ -> error lb "'(' expected"

  let action lb =
    match next lb with
    | Equal      -> Set
    | Plus_equal -> Add
    | _          -> error lb "'=' or '+=' expected"

  let rec predicates_and_action lb acc =
    match next lb with
    | Rparen -> (List.rev acc, action lb)
    | Name n -> after_predicate lb (Pos n :: acc)
    | Minus  ->
      let n =
        match next lb with
        | Name p -> p
        | _      -> error lb "name expected"
      in
      after_predicate lb (Neg n :: acc)
    | _          -> error lb "name, '-' or ')' expected"

  and after_predicate lb acc =
    match next lb with
    | Rparen -> (List.rev acc, action lb)
    | Comma  -> predicates_and_action lb acc
    | _      -> error lb "')' or ',' expected"

  let rec entries lb depth acc =
    match next lb with
    | Rparen ->
      if depth > 0 then
        List.rev acc
      else
        error lb "closing parenthesis without matching opening one"
    | Eof ->
      if depth = 0 then
        List.rev acc
      else
        error lb "%d closing parentheses missing" depth
    | Name "package" ->
      let name = package_name lb in
      lparen lb;
      let sub_entries = entries lb (depth + 1) [] in
      entries lb depth (Package { name; entries = sub_entries } :: acc)
    | Name var ->
      let predicates, action =
        match next lb with
        | Equal      -> ([], Set)
        | Plus_equal -> ([], Add)
        | Lparen     -> predicates_and_action lb []
        | _          -> error lb "'=', '+=' or '(' expected"
      in
      let value = string lb in
      entries lb depth (Rule { var; predicates; action; value } :: acc)
    | _ ->
      error lb "'package' or variable name expected"
end

let pp_action fmt = function
  | Set -> Format.pp_print_string fmt "Set"
  | Add -> Format.pp_print_string fmt "Add"

let pp_predicate fmt = function
  | Pos s -> Format.fprintf fmt "%S" ("+" ^ s)
  | Neg s -> Format.fprintf fmt "%S" ("-" ^ s)

let pp_rule fmt (t : rule) =
  Fmt.record fmt
    [ "var", (Fmt.const Fmt.quoted t.var)
    ; "predicates", (Fmt.const (Fmt.ocaml_list pp_predicate) t.predicates)
    ; "action", (Fmt.const pp_action t.action)
    ; "value", (Fmt.const Fmt.quoted t.value)
    ]

module Simplified = struct
  module Rules = struct
    type t =
      { set_rules : rule list
      ; add_rules : rule list
      }

    let pp fmt t =
      Fmt.record fmt
        [ "set_rules", Fmt.const (Fmt.ocaml_list pp_rule) t.set_rules
        ; "add_rules", Fmt.const (Fmt.ocaml_list pp_rule) t.add_rules
        ]
  end

  type t =
    { name : string
    ; vars : Rules.t String_map.t
    ; subs : t list
    }

  let rec pp fmt t =
    Fmt.record fmt
      [ "name", Fmt.const Fmt.quoted t.name
      ; "vars", Fmt.const (String_map.pp Rules.pp) t.vars
      ; "subs", Fmt.const (Fmt.ocaml_list pp) t.subs
      ]
end

let rec simplify t =
  List.fold_right t.entries
    ~init:
      { name = t.name
      ; vars = String_map.empty
      ; subs = []
      }
    ~f:(fun entry (pkg : Simplified.t) ->
      match entry with
      | Comment _ -> pkg
      | Package sub ->
        { pkg with subs = simplify sub :: pkg.subs }
      | Rule rule ->
        let rules =
          Option.value (String_map.find pkg.vars rule.var)
            ~default:{ set_rules = []; add_rules = [] }
        in
        let rules =
          match rule.action with
          | Set -> { rules with set_rules = rule :: rules.set_rules }
          | Add -> { rules with add_rules = rule :: rules.add_rules }
        in
        { pkg with vars = String_map.add pkg.vars rule.var rules })

let load ~fn ~name =
  { name
  ; entries =
      Io.with_lexbuf_from_file fn ~f:(fun lb ->
        Parse.entries lb 0 [])
  }
  |> simplify

let rule var predicates action value =
  Rule { var; predicates; action; value }
let requires ?(preds=[]) pkgs =
  rule "requires" preds Set (String.concat ~sep:" " pkgs)
let version     s = rule "version"    []      Set s
let directory   s = rule "directory"  []      Set s
let archive p s   = rule "archive"    [Pos p] Set s
let plugin  p s   = rule "plugin"     [Pos p] Set s
let archives name =
  [ archive "byte"   (name ^ ".cma" )
  ; archive "native" (name ^ ".cmxa")
  ; plugin  "byte"   (name ^ ".cma" )
  ; plugin  "native" (name ^ ".cmxs")
  ]

let builtins ~stdlib_dir =
  let version = version "[distributed with Ocaml]" in
  let simple name ?dir ?(archive_name=name) deps =
    let archives = archives archive_name in
    { name
    ; entries =
        (requires deps ::
         version       ::
         match dir with
         | None -> archives
         | Some d -> directory d :: archives)
    }
  in
  let compiler_libs =
    let sub name deps =
      Package (simple name deps ~archive_name:("ocaml" ^ name))
    in
    { name = "compiler-libs"
    ; entries =
        [ requires []
        ; version
        ; directory "+compiler-libs"
        ; sub "common" []
        ; sub "bytecomp" ["compiler-libs.common"  ]
        ; sub "optcomp"  ["compiler-libs.common"  ]
        ; sub "toplevel" ["compiler-libs.bytecomp"]
        ]
    }
  in
  let str = simple "str" [] ~dir:"+" in
  let unix = simple "unix" [] ~dir:"+" in
  let bigarray = simple "bigarray" ["unix"] ~dir:"+" in
  let threads =
    { name = "threads"
    ; entries =
        [ version
        ; requires ~preds:[Pos "mt"; Pos "mt_vm"   ] ["threads.vm"]
        ; requires ~preds:[Pos "mt"; Pos "mt_posix"] ["threads.posix"]
        ; directory "+"
        ; rule "type_of_threads" [] Set "posix"
        ; rule "error" [Neg "mt"] Set "Missing -thread or -vmthread switch"
        ; rule "error" [Neg "mt_vm"; Neg "mt_posix"] Set "Missing -thread or -vmthread switch"
        ; Package (simple "vm" ["unix"] ~dir:"+vmthreads" ~archive_name:"threads")
        ; Package (simple "posix" ["unix"] ~dir:"+threads" ~archive_name:"threads")
        ]
    }
  in
  let num =
    { name = "num"
    ; entries =
        [ requires ["num.core"]
        ; version
        ; Package (simple "core" [] ~dir:"+" ~archive_name:"nums")
        ]
    }
  in
  let libs =
    (* We do not rely on an "exists_if" ocamlfind variable,
       because it would produce an error message mentioning
       a "hidden" package (which could be confusing). *)
    if Path.exists (Path.relative stdlib_dir "nums.cma") then
      [ compiler_libs; str; unix; bigarray; threads; num ]
    else
      [ compiler_libs; str; unix; bigarray; threads ]
  in
  List.map libs ~f:(fun t -> t.name, simplify t)
  |> String_map.of_list_exn

let string_of_action = function
  | Set -> "="
  | Add -> "+="

let string_of_predicate = function
  | Pos p -> p
  | Neg p -> "-" ^ p

let pp_list f ppf l =
  match l with
  | [] -> ()
  | x :: l ->
    f ppf x;
    List.iter l ~f:(fun x ->
      Format.pp_print_cut ppf ();
      f ppf x)

let pp_print_text ppf s =
  Format.fprintf ppf "\"@[<hv>";
  Format.pp_print_text ppf (String.escape_double_quote s);
  Format.fprintf ppf "@]\""

let pp_print_string ppf s =
  Format.fprintf ppf "\"@[<hv>";
  Format.pp_print_string ppf (String.escape_double_quote s);
  Format.fprintf ppf "@]\""

let pp_quoted_value var =
  match var with
  | "archive" | "plugin" | "requires"
  | "ppx_runtime_deps" | "linkopts" | "jsoo_runtime" ->
     pp_print_text
  | _ ->
     pp_print_string

let rec pp ppf entries =
  Format.fprintf ppf "@[<v>%a@]" (pp_list pp_entry) entries

and pp_entry ppf entry =
  let open Format in
  match entry with
  | Comment s ->
    fprintf ppf "# %s" s
  | Rule { var; predicates = []; action; value } ->
    fprintf ppf "@[%s %s %a@]"
      var (string_of_action action) (pp_quoted_value var) value
  | Rule { var; predicates; action; value } ->
    fprintf ppf "@[%s(%s) %s %a@]"
      var (String.concat ~sep:"," (List.map predicates ~f:string_of_predicate))
      (string_of_action action) (pp_quoted_value var) value
  | Package { name; entries } ->
    fprintf ppf "@[<v 2>package %S (@,%a@]@,)"
      name pp entries
# 8567 "boot.ml"
end
module Arg_spec : sig
# 1 "src/arg_spec.mli"
(** Command line arguments specification *)

(** This module implements a small DSL to specify the command line
    argument of a program as well as the dependencies and targets of
    the program at the same time.

    For instance to represent the argument of [ocamlc -o src/foo.exe
    src/foo.ml], one might write:

    {[
      [ A "-o"
      ; Target (Path.relatie  dir "foo.exe")
      ; Dep    (Path.relative dir "foo.ml")
      ]
    ]}

    This DSL was inspired from the ocamlbuild API.  *)

open! Import

(** [A] stands for "atom", it is for command line arguments that are
    neither dependencies nor targets.

    [Path] is similar to [A] in the sense that it defines a command
    line argument that is neither a dependency or target. However, the
    difference between the two is that [A s] produces exactly the
    argument [s], while [Path p] produces a string that depends on
    where the command is executed. For instance [Path (Path.of_string
    "src/foo.ml")] will translate to "../src/foo.ml" if the command is
    started from the "test" directory.  *)
type 'a t =
  | A        of string
  | As       of string list
  | S        of 'a t list
  | Concat   of string * 'a t list (** Concatenation with a custom separator *)
  | Dep      of Path.t (** A path that is a dependency *)
  | Deps     of Path.t list
  | Target   of Path.t
  | Path     of Path.t
  | Paths    of Path.t list
  | Hidden_deps of Path.t list
  (** Register dependencies but produce no argument *)
  | Dyn      of ('a -> nothing t)

val add_deps    : _ t list -> Path.Set.t -> Path.Set.t
val add_targets : _ t list -> Path.t list -> Path.t list
val expand      : dir:Path.t -> 'a t list -> 'a -> string list * Path.Set.t

(** [quote_args quote args] is [As \[quote; arg1; quote; arg2; ...\]] *)
val quote_args : string -> string list -> _ t

val of_result : 'a t Or_exn.t -> 'a t
val of_result_map : 'a Or_exn.t -> f:('a -> 'b t) -> 'b t
# 8624 "boot.ml"
end = struct
# 1 "src/arg_spec.ml"
open Import

module Pset = Path.Set

type 'a t =
  | A        of string
  | As       of string list
  | S        of 'a t list
  | Concat   of string * 'a t list
  | Dep      of Path.t
  | Deps     of Path.t list
  | Target   of Path.t
  | Path     of Path.t
  | Paths    of Path.t list
  | Hidden_deps of Path.t list
  | Dyn      of ('a -> nothing t)

let rec add_deps ts set =
  List.fold_left ts ~init:set ~f:(fun set t ->
    match t with
    | Dep fn -> Pset.add set fn
    | Deps        fns
    | Hidden_deps fns -> Pset.union set (Pset.of_list fns)
    | S ts
    | Concat (_, ts) -> add_deps ts set
    | _ -> set)

let rec add_targets ts acc =
  List.fold_left ts ~init:acc ~f:(fun acc t ->
    match t with
    | Target fn  -> fn :: acc
    | S ts
    | Concat (_, ts) -> add_targets ts acc
    | _ -> acc)

let expand ~dir ts x =
  let dyn_deps = ref Path.Set.empty in
  let add_dep path = dyn_deps := Path.Set.add !dyn_deps path in
  let rec loop_dyn : nothing t -> string list = function
    | A s  -> [s]
    | As l -> l
    | Dep fn ->
      add_dep fn;
      [Path.reach fn ~from:dir]
    | Path fn -> [Path.reach fn ~from:dir]
    | Deps fns ->
      List.map fns ~f:(fun fn ->
        add_dep fn;
        Path.reach ~from:dir fn)
    | Paths fns ->
      List.map fns ~f:(Path.reach ~from:dir)
    | S ts -> List.concat_map ts ~f:loop_dyn
    | Concat (sep, ts) -> [String.concat ~sep (loop_dyn (S ts))]
    | Target _ -> die "Target not allowed under Dyn"
    | Dyn _ -> assert false
    | Hidden_deps l ->
      dyn_deps := Pset.union !dyn_deps (Pset.of_list l);
      []
  in
  let rec loop = function
    | A s  -> [s]
    | As l -> l
    | (Dep fn | Path fn) -> [Path.reach fn ~from:dir]
    | (Deps fns | Paths fns) -> List.map fns ~f:(Path.reach ~from:dir)
    | S ts -> List.concat_map ts ~f:loop
    | Concat (sep, ts) -> [String.concat ~sep (loop (S ts))]
    | Target fn -> [Path.reach fn ~from:dir]
    | Dyn f -> loop_dyn (f x)
    | Hidden_deps _ -> []
  in
  let l = List.concat_map ts ~f:loop in
  (l, !dyn_deps)

let quote_args =
  let rec loop quote = function
    | [] -> []
    | arg :: args -> quote :: arg :: loop quote args
  in
  fun quote args -> As (loop quote args)

let of_result = function
  | Ok x -> x
  | Error e -> Dyn (fun _ -> raise e)

let of_result_map res ~f =
  match res with
  | Ok    x -> f x
  | Error e -> Dyn (fun _ -> raise e)
# 8715 "boot.ml"
end
module Ml_kind : sig
# 1 "src/ml_kind.mli"
type t = Impl | Intf

val all : t list

(** "" or "i" *)
val suffix : t -> string

val to_string : t -> string

val flag : t -> _ Arg_spec.t
val ppx_driver_flag : t -> _ Arg_spec.t

module Dict : sig
  type kind = t

  type 'a t =
    { impl : 'a
    ; intf : 'a
    }

  val get : 'a t -> kind -> 'a

  val of_func : (ml_kind:kind -> 'a) -> 'a t

  val make_both : 'a -> 'a t

  val map : 'a t -> f:('a -> 'b) -> 'b t
end with type kind := t
# 8747 "boot.ml"
end = struct
# 1 "src/ml_kind.ml"
type t = Impl | Intf

let all = [Impl; Intf]

let choose impl intf = function
  | Impl -> impl
  | Intf -> intf

let suffix = choose "" "i"

let to_string = choose "impl" "intf"

let flag t = choose (Arg_spec.A "-impl") (A "-intf") t

let ppx_driver_flag t = choose (Arg_spec.A "--impl") (A "--intf") t

module Dict = struct
  type 'a t =
    { impl : 'a
    ; intf : 'a
    }

  let get t = function
    | Impl -> t.impl
    | Intf -> t.intf

  let of_func f =
    { impl = f ~ml_kind:Impl
    ; intf = f ~ml_kind:Intf
    }

  let make_both x = { impl = x; intf = x }

  let map t ~f = { impl = f t.impl; intf = f t.intf }
end
# 8785 "boot.ml"
end
module Cm_kind : sig
# 1 "src/cm_kind.mli"
type t = Cmi | Cmo | Cmx

val all : t list

val ext : t -> string
val source : t -> Ml_kind.t

module Dict : sig
  type cm_kind = t

  type 'a t =
    { cmi : 'a
    ; cmo : 'a
    ; cmx : 'a
    }

  val get : 'a t -> cm_kind -> 'a

  val of_func : (cm_kind:cm_kind -> 'a) -> 'a t

  val make_all : 'a -> 'a t
end with type cm_kind := t
# 8811 "boot.ml"
end = struct
# 1 "src/cm_kind.ml"
type t = Cmi | Cmo | Cmx

let all = [Cmi; Cmo; Cmx]

let choose cmi cmo cmx = function
  | Cmi -> cmi
  | Cmo -> cmo
  | Cmx -> cmx

let ext = choose ".cmi" ".cmo" ".cmx"

let source = choose Ml_kind.Intf Impl Impl

module Dict = struct
  type 'a t =
    { cmi : 'a
    ; cmo : 'a
    ; cmx : 'a
    }

  let get t = function
    | Cmi -> t.cmi
    | Cmo -> t.cmo
    | Cmx -> t.cmx

  let of_func f =
    { cmi = f ~cm_kind:Cmi
    ; cmo = f ~cm_kind:Cmo
    ; cmx = f ~cm_kind:Cmx
    }

  let make_all x =
    { cmi = x
    ; cmo = x
    ; cmx = x
    }
end
# 8851 "boot.ml"
end
module Interned : sig
# 1 "src/interned.mli"
(** Interned strings *)

open! Import

module type S = sig
  type t
  val compare : t -> t -> Ordering.t
  val to_string : t -> string

  val make : string -> t

  (** Like [make] except it returns [None] if the string hasn't been
      registered with [make] previously. *)
  val get : string -> t option

  module Set : sig
    include Set.S with type elt = t

    val make : string list -> t
  end

  module Map : Map.S with type key = t

  (** Same as a hash table, but optimized for the case where we are
      using one entry for every possible [t] *)
  module Table : sig
    type key = t
    type 'a t

    val create : default_value:'a -> 'a t

    val get : 'a t -> key -> 'a
    val set : 'a t -> key:key -> data:'a -> unit
  end with type key := t
end

module Make() : S
# 8892 "boot.ml"
end = struct
# 1 "src/interned.ml"
open Import

module type S = sig
  type t
  val compare : t -> t -> Ordering.t
  val to_string : t -> string
  val make : string -> t
  val get : string -> t option
  module Set : sig
    include Set.S with type elt = t
    val make : string list -> t
  end
  module Map : Map.S with type key = t
  module Table : sig
    type key = t
    type 'a t
    val create : default_value:'a -> 'a t
    val get : 'a t -> key -> 'a
    val set : 'a t -> key:key -> data:'a -> unit
  end with type key := t
end

module Make() = struct
  include Int

  let ids = Hashtbl.create 1024
  let next = ref 0

  module Table = struct
    type 'a t =
      { default_value : 'a
      ; mutable data  : 'a array
      }

    let create ~default_value =
      { default_value
      ; data = [||]
      }

    let resize t =
      let increment_size = 512 in
      let n = (!next land (lnot (increment_size - 1))) + (increment_size * 2) in
      let old_data = t.data                       in
      let new_data = Array.make n t.default_value in
      t.data <- new_data;
      Array.blit
        ~src:old_data ~src_pos:0
        ~dst:new_data ~dst_pos:0
        ~len:(Array.length old_data)

    let get t key =
      if key >= Array.length t.data then
        t.default_value
      else
        t.data.(key)

    let set t ~key ~data =
      if key >= Array.length t.data then resize t;
      t.data.(key) <- data
  end

  let names = Table.create ~default_value:""

  let make s =
    Hashtbl.find_or_add ids s ~f:(fun s ->
      let n = !next in
      next := n + 1;
      Table.set names ~key:n ~data:s;
      n)

  let get s = Hashtbl.find ids s

  let to_string t = Table.get names t

  module Set = struct
    include Int_set

    let make l =
      List.fold_left l ~init:empty ~f:(fun acc s -> add acc (make s))
  end

  module Map = Int_map
end
# 8978 "boot.ml"
end
module Variant : sig
# 1 "src/variant.mli"
(** Library variants *)

(** Library variants allow to select the implementation of a library
    at link time.

    They are directly mapped to findlib predicates.
*)

include Interned.S

(** Well-known variants *)
val ppx_driver : t
val mt         : t
val mt_posix   : t
val byte       : t
val native     : t
val plugin     : t
# 8999 "boot.ml"
end = struct
# 1 "src/variant.ml"
include Interned.Make()

let ppx_driver = make "ppx_driver"
let mt         = make "mt"
let mt_posix   = make "mt_posix"
let byte       = make "byte"
let native     = make "native"
let plugin     = make "plugin"
# 9010 "boot.ml"
end
module Mode : sig
# 1 "src/mode.mli"
open! Import

type t = Byte | Native

val t : t Sexp.Of_sexp.t

val all : t list

val compiled_unit_ext : t -> string
val compiled_lib_ext : t -> string
val exe_ext : t -> string
val plugin_ext : t -> string

val cm_kind : t -> Cm_kind.t
val of_cm_kind : Cm_kind.t -> t

val variant : t -> Variant.t

module Dict : sig
  type mode = t

  type 'a t =
    { byte   : 'a
    ; native : 'a
    }

  val get : 'a t -> mode -> 'a

  val of_func : (mode:mode -> 'a) -> 'a t

  val map2 : 'a t -> 'b t -> f:('a -> 'b -> 'c) -> 'c t

  val make_both : 'a -> 'a t

  module Set : sig

    type nonrec t = bool t
    val t : t Sexp.Of_sexp.t
    val all : t
    val is_empty : t -> bool
    val to_list : t -> mode list
    val of_list : mode list -> t
    val iter : t -> f:(mode -> unit) -> unit
  end
end with type mode := t
# 9059 "boot.ml"
end = struct
# 1 "src/mode.ml"
open! Import

type t = Byte | Native

let all = [Byte; Native]

let t =
  let open Sexp.Of_sexp in
  enum
    [ "byte"   , Byte
    ; "native" , Native
    ]

let choose byte native = function
  | Byte   -> byte
  | Native -> native

let compiled_unit_ext = choose ".cmo" ".cmx"
let compiled_lib_ext = choose ".cma" ".cmxa"
let plugin_ext = choose ".cma" ".cmxs"

let variant = choose Variant.byte Variant.native

let cm_kind = choose Cm_kind.Cmo Cmx

let exe_ext = choose ".bc" ".exe"

let of_cm_kind : Cm_kind.t -> t = function
  | Cmi | Cmo -> Byte
  | Cmx -> Native

module Dict = struct
  type 'a t =
    { byte   : 'a
    ; native : 'a
    }

  let get t = function
    | Byte   -> t.byte
    | Native -> t.native

  let of_func f =
    { byte   = f ~mode:Byte
    ; native = f ~mode:Native
    }

  let map2 a b ~f =
    { byte   = f a.byte   b.byte
    ; native = f a.native b.native
    }

  let make_both x =
    { byte   = x
    ; native = x
    }

  module Set = struct
    type nonrec t = bool t

    let all =
      { byte   = true
      ; native = true
      }

    let to_list t =
      let l = [] in
      let l = if t.native then Native :: l else l in
      let l = if t.byte   then Byte   :: l else l in
      l

    let of_list l =
      { byte   = List.mem Byte   ~set:l
      ; native = List.mem Native ~set:l
      }

    let t sexp = of_list (Sexp.Of_sexp.list t sexp)

    let is_empty t = not (t.byte || t.native)

    let iter t ~f =
      if t.byte   then f Byte;
      if t.native then f Native
  end
end
# 9146 "boot.ml"
end
module Findlib : sig
# 1 "src/findlib.mli"
(** Findlib database *)

open Stdune

(** Findlib database *)
type t

val create
  :  stdlib_dir:Path.t
  -> path:Path.t list
  -> t

(** The search path for this DB *)
val path : t -> Path.t list

(** [root_package_name "foo.*"] is "foo" *)
val root_package_name : string -> string

module Package : sig
  (** Representation of a findlib package *)
  type t

  val meta_file        : t -> Path.t
  val name             : t -> string
  val dir              : t -> Path.t
  val version          : t -> string option
  val description      : t -> string option
  val archives         : t -> Path.t list Mode.Dict.t
  val plugins          : t -> Path.t list Mode.Dict.t
  val jsoo_runtime     : t -> Path.t list
  val requires         : t -> string list
  val ppx_runtime_deps : t -> string list
  val dune_file        : t -> Path.t option
end

module Unavailable_reason : sig
  type t =
    | Not_found
    (** The package is hidden because it contains an unsatisfied
        'exist_if' clause *)
    | Hidden of Package.t

  val to_string : t -> string
  val pp : Format.formatter -> t -> unit
end

(** Lookup a package in the given database *)
val find : t -> string -> (Package.t, Unavailable_reason.t) result

val available : t -> string -> bool

(** List all the packages available in this Database *)
val all_packages  : t -> Package.t list

(** List all the packages that are not available in this database *)
val all_unavailable_packages : t -> (string * Unavailable_reason.t) list

(** A dummy package. This is used to implement [external-lib-deps] *)
val dummy_package : t -> name:string -> Package.t

module Config : sig
  type t
  val load : Path.t -> toolchain:string -> context:string -> t
  val get : t -> string -> string option
end
# 9215 "boot.ml"
end = struct
# 1 "src/findlib.ml"
open Import

module P  = Variant
module Ps = Variant.Set

(* An assignment or addition *)
module Rule = struct
  type t =
    { preds_required  : Ps.t
    ; preds_forbidden : Ps.t
    ; value           : string
    }

  let formal_predicates_count t =
    Ps.cardinal t.preds_required + Ps.cardinal t.preds_forbidden

  let matches t ~preds =
    Ps.is_subset t.preds_required ~of_:preds &&
    Ps.is_empty (Ps.inter preds t.preds_forbidden)

  let make (rule : Meta.rule) =
    let preds_required, preds_forbidden =
      List.partition_map rule.predicates ~f:(function
        | Pos x -> Left  x
        | Neg x -> Right x)
    in
    { preds_required  = Ps.make preds_required
    ; preds_forbidden = Ps.make preds_forbidden
    ; value           = rule.value
    }
end

(* Set of rules for a given variable of a package. Implements the
   algorithm described here:

   http://projects.camlcity.org/projects/dl/findlib-1.6.3/doc/ref-html/r729.html
*)
module Rules = struct
  (* To implement the algorithm, [set_rules] is sorted by decreasing
     number of formal predicates, then according to the order of the
     META file. [add_rules] are in the same order as in the META
     file. *)
  type t =
    { set_rules : Rule.t list
    ; add_rules : Rule.t list
    }

  let interpret t ~preds =
    let rec find_set_rule = function
      | [] -> ""
      | rule :: rules ->
        if Rule.matches rule ~preds then
          rule.value
        else
          find_set_rule rules
    in
    let v = find_set_rule t.set_rules in
    List.fold_left t.add_rules ~init:v ~f:(fun v rule ->
      if Rule.matches rule ~preds then
        v ^ " " ^ rule.value
      else
        v)

  let of_meta_rules (rules : Meta.Simplified.Rules.t) =
    let add_rules = List.map rules.add_rules ~f:Rule.make in
    let set_rules =
      List.map rules.set_rules ~f:Rule.make
      |> List.stable_sort ~compare:(fun a b ->
        compare
          (Rule.formal_predicates_count b)
          (Rule.formal_predicates_count a))
    in
    { add_rules; set_rules }
end

module Vars = struct
  type t = Rules.t String_map.t

  let get (t : t) var preds =
    match String_map.find t var with
    | None -> None
    | Some rules -> Some (Rules.interpret rules ~preds)

  let get_words t var preds =
    match get t var preds with
    | None -> []
    | Some s -> String.extract_comma_space_separated_words s
end

module Config = struct
  type t =
    { vars  : Vars.t
    ; preds : Ps.t
    }

  let load path ~toolchain ~context =
    let path = Path.extend_basename path ~suffix:".d" in
    let conf_file = Path.relative path (toolchain ^ ".conf") in
    if not (Path.exists conf_file) then
      die "@{<error>Error@}: ocamlfind toolchain %s isn't defined in %a \
           (context: %s)" toolchain Path.pp path context;
    let vars = (Meta.load ~name:"" ~fn:(Path.to_string conf_file)).vars in
    { vars = String_map.map vars ~f:Rules.of_meta_rules
    ; preds = Ps.make [toolchain]
    }

  let get { vars; preds } var =
    Vars.get vars var preds
end

module Package = struct
  type t =
    { meta_file : Path.t
    ; name      : string
    ; dir       : Path.t
    ; vars      : Vars.t
    }

  let meta_file t = t.meta_file
  let name      t = t.name
  let dir       t = t.dir

  let preds = Ps.of_list [P.ppx_driver; P.mt; P.mt_posix]

  let get_paths t var preds =
    List.map (Vars.get_words t.vars var preds) ~f:(Path.relative t.dir)

  let make_archives t var preds =
    Mode.Dict.of_func (fun ~mode ->
      get_paths t var (Ps.add preds (Mode.variant mode)))

  let version          t = Vars.get       t.vars "version"          Ps.empty
  let description      t = Vars.get       t.vars "description"      Ps.empty
  let jsoo_runtime     t = get_paths      t      "jsoo_runtime"     Ps.empty
  let requires         t = Vars.get_words t.vars "requires"         preds
  let ppx_runtime_deps t = Vars.get_words t.vars "ppx_runtime_deps" preds

  let archives t = make_archives t "archive" preds
  let plugins t =
    Mode.Dict.map2 ~f:(@)
      (make_archives t "archive" (Ps.add preds Variant.plugin))
      (make_archives t "plugin" preds)

  let dune_file t =
    let fn = Path.relative t.dir (sprintf "%s.dune" t.name) in
    Option.some_if (Path.exists fn) fn
end

module Unavailable_reason = struct
  type t =
    | Not_found
    | Hidden of Package.t

  let to_string = function
    | Not_found  -> "not found"
    | Hidden pkg ->
      sprintf "in %s is hidden (unsatisfied 'exist_if')"
        (Path.to_string_maybe_quoted (Package.dir pkg))

  let pp ppf t = Format.pp_print_string ppf (to_string t)
end

type t =
  { stdlib_dir : Path.t
  ; path       : Path.t list
  ; builtins   : Meta.Simplified.t String_map.t
  ; packages   : (string, (Package.t, Unavailable_reason.t) result) Hashtbl.t
  }

let path t = t.path

let root_package_name s =
  match String.index s '.' with
  | None -> s
  | Some i -> String.sub s ~pos:0 ~len:i

let dummy_package t ~name =
  let dir =
    match t.path with
    | [] -> t.stdlib_dir
    | dir :: _ -> Path.relative dir (root_package_name name)
  in
  { Package.
    meta_file = Path.relative dir "META"
  ; name      = name
  ; dir       = dir
  ; vars      = String_map.empty
  }

(* Parse a single package from a META file *)
let parse_package t ~meta_file ~name ~parent_dir ~vars =
  let pkg_dir = Vars.get vars "directory" Ps.empty in
  let dir =
    match pkg_dir with
    | None | Some "" -> parent_dir
    | Some pkg_dir ->
      if pkg_dir.[0] = '+' || pkg_dir.[0] = '^' then
        Path.relative t.stdlib_dir
          (String.sub pkg_dir ~pos:1 ~len:(String.length pkg_dir - 1))
      else if Filename.is_relative pkg_dir then
        Path.relative parent_dir pkg_dir
      else
        Path.absolute pkg_dir
  in
  let pkg =
    { Package.
      meta_file
    ; name
    ; dir
    ; vars
    }
  in
  let exists_if = Vars.get_words vars "exists_if" Ps.empty in
  let exists =
    match exists_if with
    | _ :: _ ->
      List.for_all exists_if ~f:(fun fn ->
        Path.exists (Path.relative dir fn))
    | [] ->
      if not (String_map.mem t.builtins (root_package_name name)) then
        true
      else
        (* The META files for installed packages are sometimes broken,
           i.e. META files for libraries that were not installed by
           the compiler are still present:

           https://github.com/ocaml/dune/issues/563

           To workaround this problem, for builtin packages we check
           that at least one of the archive is present. *)
        match Package.archives pkg with
        | { byte = []; native = [] } -> true
        | { byte; native } -> List.exists (byte @ native) ~f:Path.exists
  in
  let res =
    if exists then
      Ok pkg
    else
      Error (Unavailable_reason.Hidden pkg)
  in
  (dir, res)

(* Parse all the packages defined in a META file and add them to
   [t.packages] *)
let parse_and_acknowledge_meta t ~dir ~meta_file (meta : Meta.Simplified.t) =
  let rec loop ~dir ~full_name (meta : Meta.Simplified.t) =
    let vars = String_map.map meta.vars ~f:Rules.of_meta_rules in
    let dir, res =
      parse_package t ~meta_file ~name:full_name ~parent_dir:dir ~vars
    in
    Hashtbl.add t.packages full_name res;
    List.iter meta.subs ~f:(fun (meta : Meta.Simplified.t) ->
      loop ~dir ~full_name:(sprintf "%s.%s" full_name meta.name) meta)
  in
  loop ~dir ~full_name:meta.name meta

(* Search for a <package>/META file in the findlib search path, parse
   it and add its contents to [t.packages] *)
let find_and_acknowledge_meta t ~fq_name =
  let root_name = root_package_name fq_name in
  let rec loop dirs : (Path.t * Path.t * Meta.Simplified.t) option =
    match dirs with
    | dir :: dirs ->
      let sub_dir = Path.relative dir root_name in
      let fn = Path.relative sub_dir "META" in
      if Path.exists fn then
        Some (sub_dir,
              fn,
              Meta.load ~name:root_name ~fn:(Path.to_string fn))
      else
        (* Alternative layout *)
        let fn = Path.relative dir ("META." ^ root_name) in
        if Path.exists fn then
          Some (dir,
                fn,
                Meta.load ~fn:(Path.to_string fn) ~name:root_name)
        else
          loop dirs
    | [] ->
      match String_map.find t.builtins root_name with
      | Some meta -> Some (t.stdlib_dir, Path.of_string "<internal>", meta)
      | None -> None
  in
  match loop t.path with
  | None ->
    Hashtbl.add t.packages root_name (Error Not_found)
  | Some (dir, meta_file, meta) ->
    parse_and_acknowledge_meta t meta ~meta_file ~dir

let find t name =
  match Hashtbl.find t.packages name with
  | Some x -> x
  | None ->
    find_and_acknowledge_meta t ~fq_name:name;
    match Hashtbl.find t.packages name with
    | Some x -> x
    | None ->
      let res = Error Unavailable_reason.Not_found in
      Hashtbl.add t.packages name res;
      res

let available t name =
  match find t name with
  | Ok    _ -> true
  | Error _ -> false

let root_packages t =
  let pkgs =
    List.concat_map t.path ~f:(fun dir ->
      Sys.readdir (Path.to_string dir)
      |> Array.to_list
      |> List.filter ~f:(fun name ->
        Path.exists (Path.relative dir (name ^ "/META"))))
    |> String_set.of_list
  in
  let pkgs =
    String_set.union pkgs
      (String_set.of_list (String_map.keys t.builtins))
  in
  String_set.to_list pkgs

let load_all_packages t =
  List.iter (root_packages t) ~f:(fun pkg ->
    find_and_acknowledge_meta t ~fq_name:pkg)

let all_packages t =
  load_all_packages t;
  Hashtbl.fold t.packages ~init:[] ~f:(fun x acc ->
    match x with
    | Ok    p -> p :: acc
    | Error _ -> acc)
  |> List.sort ~compare:(fun (a : Package.t) b -> String.compare a.name b.name)

let create ~stdlib_dir ~path =
  { stdlib_dir
  ; path
  ; builtins = Meta.builtins ~stdlib_dir
  ; packages = Hashtbl.create 1024
  }

let all_unavailable_packages t =
  load_all_packages t;
  Hashtbl.foldi t.packages ~init:[] ~f:(fun name x acc ->
    match x with
    | Ok    _ -> acc
    | Error e -> ((name, e) :: acc))
  |> List.sort ~compare:(fun (a, _) (b, _) -> String.compare a b)
# 9565 "boot.ml"
end
module Package : sig
# 1 "src/package.mli"
(** Information about a package defined in the workspace *)

module Name : sig
  type t

  val of_string : string -> t

  val opam_fn : t -> string

  val pp : Format.formatter -> t -> unit

  include Interned.S with type t := t
end

type t =
  { name                   : Name.t
  ; path                   : Path.t
  ; version_from_opam_file : string option
  }

val opam_file : t -> Path.t
# 9590 "boot.ml"
end = struct
# 1 "src/package.ml"

module Name = struct
  include Interned.Make()

  let of_string = make

  let opam_fn (t : t) = to_string t ^ ".opam"

  let pp fmt t = Format.pp_print_string fmt (to_string t)
end


type t =
  { name                   : Name.t
  ; path                   : Path.t
  ; version_from_opam_file : string option
  }

let opam_file t = Path.relative t.path (Name.opam_fn t.name)
# 9612 "boot.ml"
end
module Utils : sig
# 1 "src/utils.mli"
(** Utilities that can't go in [Import] *)

open! Import

(** Return the absolute path to the shell and the argument to pass it
    (-c or /c). Raise in case in cannot be found. *)
val system_shell_exn : needed_to:string -> Path.t * string

(** Same as [system_shell_exn] but for bash *)
val bash_exn : needed_to:string -> Path.t

(** Convert a signal number to a name: INT, TERM, ... *)
val signal_name : int -> string

(** [jbuild_file_in ~dir = Path.relative dir "jbuild"]. *)
val jbuild_file_in : dir:Path.t -> Path.t

(** Nice description of a target *)
val describe_target : Path.t -> string

(** Return the directory where the object files for the given
    library should be stored. *)
val library_object_directory
  :  dir:Path.t
  -> string
  -> Path.t

(** Return the directory where the object files for the given
    executable should be stored. *)
val executable_object_directory
  :  dir:Path.t
  -> string
  -> Path.t

type target_kind =
  | Regular of string (* build context *) * Path.t
  | Alias   of string (* build context *) * Path.t
  | Other of Path.t

(** Return the name of an alias from its stamp file *)
val analyse_target : Path.t -> target_kind

(** Raise an error about a program not found in the PATH or in the tree *)
val program_not_found
  :  ?context:string
  -> ?hint:string
  -> string
  -> _

(** Raise an error about a library not found *)
val library_not_found : ?context:string -> ?hint:string -> string -> _

(** [\["-g"\]] if [!Clflags.g] and [\[\]] otherwise *)
val g : unit -> string list

val install_file
  :  package:Package.Name.t
  -> findlib_toolchain:string option
  -> string

(** Digest files with caching *)
module Cached_digest : sig
  (** Digest the contents of the following file *)
  val file : Path.t -> Digest.t

  (** Clear the following digest from the cache *)
  val remove : Path.t -> unit

  (** Dump/load the cache to/from the disk *)
  val dump : unit -> unit
  val load : unit -> unit
end
# 9688 "boot.ml"
end = struct
# 1 "src/utils.ml"
open Import

let system_shell_exn =
  let cmd, arg, os =
    if Sys.win32 then
      ("cmd", "/c", "on Windows")
    else
      ("sh", "-c", "")
  in
  let bin = lazy (Bin.which cmd) in
  fun ~needed_to ->
    match Lazy.force bin with
    | Some path -> (path, arg)
    | None ->
      die "I need %s to %s but I couldn't find it :(\n\
           Who doesn't have %s%s?!"
        cmd needed_to cmd os

let bash_exn =
  let bin = lazy (Bin.which "bash") in
  fun ~needed_to ->
    match Lazy.force bin with
    | Some path -> path
    | None ->
      die "I need bash to %s but I couldn't find it :("
        needed_to

let signal_name =
  let table =
    let open Sys in
    [ sigabrt   , "ABRT"
    ; sigalrm   , "ALRM"
    ; sigfpe    , "FPE"
    ; sighup    , "HUP"
    ; sigill    , "ILL"
    ; sigint    , "INT"
    ; sigkill   , "KILL"
    ; sigpipe   , "PIPE"
    ; sigquit   , "QUIT"
    ; sigsegv   , "SEGV"
    ; sigterm   , "TERM"
    ; sigusr1   , "USR1"
    ; sigusr2   , "USR2"
    ; sigchld   , "CHLD"
    ; sigcont   , "CONT"
    ; sigstop   , "STOP"
    ; sigtstp   , "TSTP"
    ; sigttin   , "TTIN"
    ; sigttou   , "TTOU"
    ; sigvtalrm , "VTALRM"
    ; sigprof   , "PROF"
    (* These ones are only available in OCaml >= 4.03 *)
    ; -22       , "BUS"
    ; -23       , "POLL"
    ; -24       , "SYS"
    ; -25       , "TRAP"
    ; -26       , "URG"
    ; -27       , "XCPU"
    ; -28       , "XFSZ"
    ]
  in
  fun n ->
    match List.assoc n table with
    | exception Not_found -> sprintf "%d\n" n
    | s -> s

let jbuild_file_in ~dir = Path.relative dir "jbuild"

type target_kind =
  | Regular of string * Path.t
  | Alias   of string * Path.t
  | Other of Path.t

let analyse_target fn =
  match Path.extract_build_context fn with
  | Some (".aliases", sub) -> begin
      match Path.split_first_component sub with
      | None -> Other fn
      | Some (ctx, fn) ->
        if Path.is_root fn then
          Other fn
        else
          let basename =
            match String.rsplit2 (Path.basename fn) ~on:'-' with
            | None -> assert false
            | Some (name, digest) ->
              assert (String.length digest = 32);
              name
          in
          Alias (ctx, Path.relative (Path.parent fn) basename)
    end
  | Some ("install", _) -> Other fn
  | Some (ctx, sub) -> Regular (ctx, sub)
  | None ->
    Other fn

let describe_target fn =
  let ctx_suffix = function
    | "default" -> ""
    | ctx -> sprintf " (context %s)" ctx
  in
  match analyse_target fn with
  | Alias (ctx, p) ->
    sprintf "alias %s%s" (Path.to_string_maybe_quoted p) (ctx_suffix ctx)
  | Regular (ctx, fn) ->
    sprintf "%s%s" (Path.to_string_maybe_quoted fn) (ctx_suffix ctx)
  | Other fn ->
    Path.to_string_maybe_quoted fn

let library_object_directory ~dir name =
  Path.relative dir ("." ^ name ^ ".objs")

(* Use "eobjs" rather than "objs" to avoid a potential conflict with a
   library of the same name *)
let executable_object_directory ~dir name =
  Path.relative dir ("." ^ name ^ ".eobjs")

let program_not_found ?context ?hint prog =
  die "@{<error>Error@}: Program %s not found in the tree or in PATH%s%a"
    (maybe_quoted prog)
    (match context with
     | None -> ""
     | Some name -> sprintf " (context: %s)" name)
    (fun fmt -> function
       | None -> ()
       | Some h -> Format.fprintf fmt "@ Hint: %s" h)
    hint

let library_not_found ?context ?hint lib =
  die "@{<error>Error@}: Library %s not found%s%a" (maybe_quoted lib)
    (match context with
     | None -> ""
     | Some name -> sprintf " (context: %s)" name)
    (fun fmt -> function
       | None -> ()
       | Some h -> Format.fprintf fmt "@ Hint: %s" h)
    hint

let g () =
  if !Clflags.g then
    ["-g"]
  else
    []

let install_file ~(package : Package.Name.t) ~findlib_toolchain =
  let package = Package.Name.to_string package in
  match findlib_toolchain with
  | None -> package ^ ".install"
  | Some x -> sprintf "%s-%s.install" package x

module Cached_digest = struct
  type file =
    { mutable digest            : Digest.t
    ; mutable timestamp         : float
    ; mutable timestamp_checked : bool
    }

  let cache = Hashtbl.create 1024

  let file fn =
    match Hashtbl.find cache fn with
    | Some x ->
      if x.timestamp_checked then
        x.digest
      else begin
        let mtime = (Unix.stat (Path.to_string fn)).st_mtime in
        if mtime <> x.timestamp then begin
          let digest = Digest.file (Path.to_string fn) in
          x.digest    <- digest;
          x.timestamp <- mtime;
        end;
        x.timestamp_checked <- true;
        x.digest
      end
    | None ->
      let digest = Digest.file (Path.to_string fn) in
      Hashtbl.add cache fn
        { digest
        ; timestamp = (Unix.stat (Path.to_string fn)).st_mtime
        ; timestamp_checked = true
        };
      digest

  let remove fn = Hashtbl.remove cache fn

  let db_file = "_build/.digest-db"

  let dump () =
    let module Pmap = Path.Map in
    let sexp =
      Sexp.List (
        Hashtbl.foldi cache ~init:Pmap.empty ~f:(fun key data acc ->
          Pmap.add acc key data)
        |> Path.Map.to_list
        |> List.map ~f:(fun (path, file) ->
          Sexp.List [ Quoted_string (Path.to_string path)
                    ; Atom (Sexp.Atom.of_digest file.digest)
                    ; Atom (Sexp.Atom.of_int64
                              (Int64.bits_of_float file.timestamp))
                    ]))
    in
    if Sys.file_exists "_build" then
      Io.write_file db_file (Sexp.to_string sexp)

  let load () =
    if Sys.file_exists db_file then begin
      let sexp = Sexp.load ~fname:db_file ~mode:Single in
      let bindings =
        let open Sexp.Of_sexp in
        list
          (triple
             Path.t
             (fun s -> Digest.from_hex (string s))
             (fun s -> Int64.float_of_bits (Int64.of_string (string s)))
          ) sexp
      in
      List.iter bindings ~f:(fun (path, digest, timestamp) ->
        Hashtbl.add cache path
          { digest
          ; timestamp
          ; timestamp_checked = false
          });
    end
end
# 9915 "boot.ml"
end
module Log : sig
# 1 "src/log.mli"
(** Log file *)

type t

val no_log : t

val create : ?display:Config.Display.t -> unit -> t

(** Print an information message in the log *)
val info  : t -> string -> unit
val infof : t -> ('a, Format.formatter, unit, unit) format4 -> 'a

(** Print an executed command in the log *)
val command
  :  t
  -> command_line:string
  -> output:string
  -> exit_status:Unix.process_status
  -> unit



# 9941 "boot.ml"
end = struct
# 1 "src/log.ml"
open Import

type real =
  { oc      : out_channel
  ; buf     : Buffer.t
  ; ppf     : Format.formatter
  ; display : Config.Display.t
  }

type t = real option

let no_log = None

let create ?(display=Config.default.display) () =
  if not (Sys.file_exists "_build") then
    Unix.mkdir "_build" 0o777;
  let oc = Io.open_out "_build/log" in
  Printf.fprintf oc "# %s\n# OCAMLPARAM: %s\n%!"
    (String.concat (List.map (Array.to_list Sys.argv) ~f:quote_for_shell) ~sep:" ")
    (match Env.get Env.initial "OCAMLPARAM" with
     | Some s -> Printf.sprintf "%S" s
     | None   -> "unset");
  let buf = Buffer.create 1024 in
  let ppf = Format.formatter_of_buffer buf in
  Some { oc; buf; ppf; display }

let info_internal { ppf; display; _ } str =
  let write ppf =
    List.iter (String.split_lines str) ~f:(function
      | "" -> Format.pp_print_string ppf "#\n"
      | s  -> Format.fprintf ppf "# %s\n" s);
    Format.pp_print_flush ppf ()
  in
  write ppf;
  if display = Verbose then print_to_console (Format.asprintf "%t" write)

let info t str =
  match t with
  | None -> ()
  | Some t -> info_internal t str

let infof t fmt =
  match t with
  | None -> Format.ikfprintf ignore Format.str_formatter fmt
  | Some t ->
    Format.kfprintf
      (fun ppf ->
         Format.pp_print_flush ppf ();
         let s = Buffer.contents t.buf in
         Buffer.clear t.buf;
         info_internal t s)
      t.ppf
      fmt

let command t ~command_line ~output ~exit_status =
  match t with
  | None -> ()
  | Some { oc; _ } ->
    Printf.fprintf oc "$ %s\n" (Ansi_color.strip command_line);
    List.iter (String.split_lines output) ~f:(fun s ->
      match Ansi_color.strip s with
      | "" -> output_string oc ">\n"
      | s  -> Printf.fprintf oc "> %s\n" s);
    (match (exit_status : Unix.process_status) with
     | WEXITED   0 -> ()
     | WEXITED   n -> Printf.fprintf oc "[%d]\n" n
     | WSIGNALED n -> Printf.fprintf oc "[got signal %s]\n" (Utils.signal_name n)
     | WSTOPPED  _ -> assert false);
    flush oc
# 10013 "boot.ml"
end
module Dep_path : sig
# 1 "src/dep_path.mli"
(** Dependency path *)

module Entry : sig
  type t =
    | Path of Path.t
    | Alias of Path.t
    | Library of Path.t * string
    | Preprocess of string list
    | Loc of Loc.t

  (** [jbuild_file_in ~dir = Path (Path.relative dir "jbuild")] *)
  val jbuild_file_in : dir:Path.t -> t

  val to_string : t -> string
  val pp : Format.formatter -> t -> unit
end

module Entries : sig
  type t = Entry.t list

  val pp : Format.formatter -> t -> unit
end

(** Re-raise an exception and augment it's dependency path with the
    given entry. The raised exception will be wrapped. *)
val reraise : exn -> Entry.t -> _

(** Extend the required_by stack of an exception *)
val prepend_exn : exn -> Entry.t -> exn

(** Extract a wrapped exception *)
val unwrap_exn : exn -> exn * Entry.t list option
# 10049 "boot.ml"
end = struct
# 1 "src/dep_path.ml"
open Import

module Entry = struct
  type t =
    | Path of Path.t
    | Alias of Path.t
    | Library of Path.t * string
    | Preprocess of string list
    | Loc of Loc.t

  let jbuild_file_in ~dir = Path (Utils.jbuild_file_in ~dir)

  let to_string = function
    | Path p -> Utils.describe_target p
    | Alias p -> "alias " ^ Utils.describe_target p
    | Library (path, lib_name) ->
      sprintf "library %S in %s" lib_name (Path.to_string_maybe_quoted path)
    | Preprocess l ->
      Sexp.to_string (List [Sexp.unsafe_atom_of_string "pps";
                            Sexp.To_sexp.(list string) l])
    | Loc loc ->
      Loc.to_file_colon_line loc

  let pp ppf x =
    Format.pp_print_string ppf (to_string x)
end

module Entries = struct
  type t = Entry.t list

  let pp ppf t =
    Format.fprintf ppf "@[<v>%a@]"
      (Format.pp_print_list
         (fun ppf x ->
            Format.fprintf ppf "-> required by %a" Entry.pp x))
      t
end

exception E of exn * Entry.t list

let prepend_exn exn entry =
  match exn with
  | E (exn, entries) -> E (exn, entry :: entries)
  | exn -> E (exn, [entry])

let reraise exn entry = reraise (prepend_exn exn entry)

let unwrap_exn = function
  | E (exn, entries) -> (exn, Some entries)
  | exn -> (exn, None)

let () =
  Printexc.register_printer (function
    | E (exn, _) -> Some (Printexc.to_string exn)
    | _ -> None)
# 10107 "boot.ml"
end
module Report_error : sig
# 1 "src/report_error.mli"
(** Error reporting *)

(** Captures the backtrace and report an error.

    Because jbuilder doesn't stop at the first error, it might end up
    reporting the same error twice about missing libraries for
    instance. To avoid this, we keep a cache of reported errors and
    ignore errors that have already been reported.

    We cache what is actually printed to the screen.  *)
val report : exn -> unit

type printer =
  { loc       : Loc.t option
  ; pp        : Format.formatter -> unit
  ; hint      : string option
  ; backtrace : bool
  }

(** Register an error reporter. *)
val register : (exn -> printer option) -> unit

(**/**)
val map_fname : (string -> string) ref
# 10135 "boot.ml"
end = struct
# 1 "src/report_error.ml"
open Import

let map_fname = ref (fun x -> x)

type printer =
  { loc       : Loc.t option
  ; pp        : Format.formatter -> unit
  ; hint      : string option
  ; backtrace : bool
  }

let p =
  { loc       = None
  ; pp        = ignore
  ; hint      = None
  ; backtrace = false
  }

let reporters = ref []
let register f = reporters := f :: !reporters

(* Firt return value is [true] if the backtrace was printed *)
let report_with_backtrace exn =
  match List.find_map !reporters ~f:(fun f -> f exn) with
  | Some p -> p
  | None ->
    match exn with
    | Loc.Error (loc, msg) ->
      let loc =
        { loc with
          start = { loc.start with pos_fname = !map_fname loc.start.pos_fname }
        }
      in
      let pp ppf = Format.fprintf ppf "@{<error>Error@}: %s\n" msg in
      { p with loc = Some loc; pp }
    | Usexp.Parser.Error e ->
      let pos = Usexp.Parser.Error.position e in
      let msg = Usexp.Parser.Error.message e in
      let pos = { pos with pos_fname = !map_fname pos.pos_fname } in
      let loc = { Loc. start = pos; stop = pos } in
      { p with
        loc = Some loc
      ; pp  = fun ppf -> Format.fprintf ppf "@{<error>Error@}: %s\n" msg
      }
    | Fatal_error msg ->
      { p with pp = fun ppf ->
          if msg.[String.length msg - 1] = '\n' then
            Format.fprintf ppf "%s" msg
          else
            Format.fprintf ppf "%s\n" (String.capitalize msg)
      }
    | Code_error msg ->
      { p with
        backtrace = true
      ; pp = fun ppf ->
          Format.fprintf ppf "@{<error>Internal error, please report upstream \
                              including the contents of _build/log.@}\n\
                              Description: %s\n"
            msg
      }
    | Unix.Unix_error (err, func, fname) ->
      { p with pp = fun ppf ->
          Format.fprintf ppf "@{<error>Error@}: %s: %s: %s\n"
            func fname (Unix.error_message err)
      }
    | _ ->
      { p with
        backtrace = true
      ; pp = fun ppf ->
          let s = Printexc.to_string exn in
          if String.is_prefix s ~prefix:"File \"" then
            Format.fprintf ppf "%s\n" s
          else
            Format.fprintf ppf "@{<error>Error@}: exception %s\n" s
      }

let reported = ref String_set.empty

let report exn =
  let exn, dependency_path = Dep_path.unwrap_exn exn in
  match exn with
  | Already_reported -> ()
  | _ ->
    let backtrace = Printexc.get_raw_backtrace () in
    let ppf = err_ppf in
    let p = report_with_backtrace exn in
    let loc = if p.loc = Some Loc.none then None else p.loc in
    Option.iter loc ~f:(fun loc -> Loc.print ppf loc);
    p.pp ppf;
    Format.pp_print_flush ppf ();
    let s = Buffer.contents err_buf in
    (* Hash to avoid keeping huge errors in memory *)
    let hash = Digest.string s in
    if String_set.mem !reported hash then
      Buffer.clear err_buf
    else begin
      reported := String_set.add !reported hash;
      if p.backtrace || !Clflags.debug_backtraces then
        Format.fprintf ppf "Backtrace:\n%s"
          (Printexc.raw_backtrace_to_string backtrace);
      let dependency_path =
        let dependency_path = Option.value dependency_path ~default:[] in
        if !Clflags.debug_dep_path then
          dependency_path
        else
          (* Only keep the part that doesn't come from the build system *)
          let rec drop : Dep_path.Entries.t -> _ = function
            | (Path _ | Alias _) :: l -> drop l
            | l -> l
          in
          match loc with
          | None -> drop dependency_path
          | Some loc ->
            if Filename.is_relative loc.start.pos_fname then
              (* If the error points to a local file, no need to print the
                 dependency stack *)
              []
            else
              drop dependency_path
      in
      if dependency_path <> [] then
        Format.fprintf ppf "%a@\n" Dep_path.Entries.pp
          (List.rev dependency_path);
      Option.iter p.hint ~f:(fun s -> Format.fprintf ppf "Hint: try: %s\n" s);
      Format.pp_print_flush ppf ();
      let s = Buffer.contents err_buf in
      Buffer.clear err_buf;
      print_to_console s
    end
# 10267 "boot.ml"
end
module Scheduler : sig
# 1 "src/scheduler.mli"
(** Scheduling *)

(** [go ?log ?config ?gen_status_line fiber] runs the following fiber until it
    terminates. [gen_status_line] is used to print a status line when [config.display =
    Progress]. *)
val go
  :  ?log:Log.t
  -> ?config:Config.t
  -> ?gen_status_line:(unit -> string option)
  -> 'a Fiber.t
  -> 'a

(** Wait for the following process to terminate *)
val wait_for_process : int -> Unix.process_status Fiber.t

(** Set the status line generator for the current scheduler *)
val set_status_line_generator : (unit -> string option) -> unit Fiber.t

(** Scheduler informations *)
type t

(** Wait until less tham [!Clflags.concurrency] external processes are running and return
    the scheduler informations. *)
val wait_for_available_job : unit -> t Fiber.t

(** Logger *)
val log : t -> Log.t

(** Execute the given callback with current directory temporarily changed *)
val with_chdir : t -> dir:string -> f:(unit -> 'a) -> 'a

(** Display mode for this scheduler *)
val display : t -> Config.Display.t

(** Print something to the terminal *)
val print : t -> string -> unit
# 10307 "boot.ml"
end = struct
# 1 "src/scheduler.ml"
open Import
open Fiber.O

type running_job =
  { pid  : int
  ; ivar : Unix.process_status Fiber.Ivar.t
  }

module Running_jobs : sig
  val add : running_job -> unit
  val wait : unit -> running_job * Unix.process_status
  val count : unit -> int
end = struct
  let all = Hashtbl.create 128

  let add job = Hashtbl.add all job.pid job

  let count () = Hashtbl.length all

  let resolve_and_remove_job pid =
    let job =
      match Hashtbl.find all pid with
      | Some job -> job
      | None -> assert false
    in
    Hashtbl.remove all pid;
    job

  exception Finished of running_job * Unix.process_status

  let wait_nonblocking_win32 () =
    match
      Hashtbl.iter all ~f:(fun ~key:pid ~data:job ->
        let pid, status = Unix.waitpid [WNOHANG] pid in
        if pid <> 0 then
          raise_notrace (Finished (job, status)))
    with
    | () -> None
    | exception (Finished (job, status)) ->
      Hashtbl.remove all job.pid;
      Some (job, status)

  let rec wait_win32 () =
    match wait_nonblocking_win32 () with
    | None ->
      ignore (Unix.select [] [] [] 0.001);
      wait_win32 ()
    | Some x -> x

  let wait_unix () =
    let pid, status = Unix.wait () in
    (resolve_and_remove_job pid, status)

  let wait =
    if Sys.win32 then
      wait_win32
    else
      wait_unix
end

type t =
  { log                       : Log.t
  ; original_cwd              : string
  ; display                   : Config.Display.t
  ; concurrency               : int
  ; waiting_for_available_job : t Fiber.Ivar.t Queue.t
  ; mutable status_line       : string
  ; mutable gen_status_line   : unit -> string option
  }

let log t = t.log
let display t = t.display

let with_chdir t ~dir ~f =
  Sys.chdir dir;
  protectx () ~finally:(fun () -> Sys.chdir t.original_cwd) ~f

let hide_status_line s =
  let len = String.length s in
  if len > 0 then Printf.eprintf "\r%*s\r" len ""

let show_status_line s =
  prerr_string s

let print t msg =
  let s = t.status_line in
  hide_status_line s;
  prerr_string msg;
  show_status_line s;
  flush stderr

let t_var : t Fiber.Var.t = Fiber.Var.create ()

let set_status_line_generator f =
  Fiber.Var.get_exn t_var >>| fun t ->
  t.gen_status_line <- f

let wait_for_available_job () =
  Fiber.Var.get_exn t_var >>= fun t ->
  if Running_jobs.count () < t.concurrency then
    Fiber.return t
  else begin
    let ivar = Fiber.Ivar.create () in
    Queue.push ivar t.waiting_for_available_job;
    Fiber.Ivar.read ivar
  end

let wait_for_process pid =
  let ivar = Fiber.Ivar.create () in
  Running_jobs.add { pid; ivar };
  Fiber.Ivar.read ivar

let rec restart_waiting_for_available_job t =
  if Queue.is_empty t.waiting_for_available_job ||
     Running_jobs.count () >= t.concurrency then
    Fiber.return ()
  else begin
    Fiber.Ivar.fill (Queue.pop t.waiting_for_available_job) t
    >>= fun () ->
    restart_waiting_for_available_job t
  end

let rec go_rec t =
  Fiber.yield ()
  >>= fun () ->
  let count = Running_jobs.count () in
  if count = 0 then begin
    hide_status_line t.status_line;
    flush stderr;
    Fiber.return ()
  end else begin
    if t.display = Progress then begin
      match t.gen_status_line () with
      | None ->
        if t.status_line <> "" then begin
          hide_status_line t.status_line;
          flush stderr
        end
      | Some status_line ->
        let status_line = sprintf "%s (jobs: %u)" status_line count in
        hide_status_line t.status_line;
        show_status_line   status_line;
        flush stderr;
        t.status_line <- status_line;
    end;
    let job, status = Running_jobs.wait () in
    Fiber.Ivar.fill job.ivar status
    >>= fun () ->
    restart_waiting_for_available_job t
    >>= fun () ->
    go_rec t
  end

let go ?(log=Log.no_log) ?(config=Config.default)
      ?(gen_status_line=fun () -> None) fiber =
  Log.info log ("Workspace root: " ^ !Clflags.workspace_root);
  let cwd = Sys.getcwd () in
  if cwd <> initial_cwd then
    Printf.eprintf "Entering directory '%s'\n%!" cwd;
  let t =
    { log
    ; gen_status_line
    ; original_cwd = cwd
    ; display      = config.display
    ; concurrency  = config.concurrency
    ; status_line  = ""
    ; waiting_for_available_job = Queue.create ()
    }
  in
  printer := print t;
  let fiber =
    Fiber.Var.set t_var t
      (Fiber.with_error_handler (fun () -> fiber) ~on_error:Report_error.report)
  in
  Fiber.run
    (Fiber.fork_and_join_unit
       (fun () -> go_rec t)
       (fun () -> fiber))
# 10488 "boot.ml"
end
module Process : sig
# 1 "src/process.mli"
(** Running external programs *)

open Import

type accepted_codes =
  | These of int list
  | All

(** How to handle sub-process failures *)
type ('a, 'b) failure_mode =
  | Strict : ('a, 'a) failure_mode
  (** Fail if the process exits with anything else than [0] *)
  | Accept : accepted_codes -> ('a, ('a, int) result) failure_mode
  (** Accept the following non-zero exit codes, and return [Error
      code] if the process exists with one of these codes. *)

(** Where to redirect standard output *)
type std_output_to =
  | Terminal
  | File        of string
  | Opened_file of opened_file

and opened_file =
  { filename : string
  ; desc     : opened_file_desc
  ; tail     : bool
  (** If [true], the descriptor is closed after starting the command *)
  }

and opened_file_desc =
  | Fd      of Unix.file_descr
  | Channel of out_channel

(** Why a Fiber.t was run *)
type purpose =
  | Internal_job
  | Build_job of Path.t list

(** [run ?dir ?stdout_to prog args] spawns a sub-process and wait for its termination *)
val run
  :  ?dir:string
  -> ?stdout_to:std_output_to
  -> ?stderr_to:std_output_to
  -> env:Env.t
  -> ?purpose:purpose
  -> (unit, 'a) failure_mode
  -> string
  -> string list
  -> 'a Fiber.t

(** Run a command and capture its output *)
val run_capture
  :  ?dir:string
  -> env:Env.t
  -> ?purpose:purpose
  -> (string, 'a) failure_mode
  -> string
  -> string list
  -> 'a Fiber.t
val run_capture_line
  :  ?dir:string
  -> env:Env.t
  -> ?purpose:purpose
  -> (string, 'a) failure_mode
  -> string
  -> string list
  -> 'a Fiber.t
val run_capture_lines
  :  ?dir:string
  -> env:Env.t
  -> ?purpose:purpose
  -> (string list, 'a) failure_mode
  -> string
  -> string list
  -> 'a Fiber.t

# 10568 "boot.ml"
end = struct
# 1 "src/process.ml"
open Import
open Fiber.O

type accepted_codes =
  | These of int list
  | All

let code_is_ok accepted_codes n =
  match accepted_codes with
  | These set -> List.mem n ~set
  | All -> true

type ('a, 'b) failure_mode =
  | Strict : ('a, 'a) failure_mode
  | Accept : accepted_codes -> ('a, ('a, int) result) failure_mode

let accepted_codes : type a b. (a, b) failure_mode -> accepted_codes = function
  | Strict -> These [0]
  | Accept (These codes) -> These (0 :: codes)
  | Accept All -> All

let map_result
  : type a b. (a, b) failure_mode -> int Fiber.t -> f:(unit -> a) -> b Fiber.t
  = fun mode t ~f ->
    match mode with
    | Strict   -> t >>| fun _ -> f ()
    | Accept _ ->
      t >>| function
      | 0 -> Ok (f ())
      | n -> Error n

type std_output_to =
  | Terminal
  | File        of string
  | Opened_file of opened_file

and opened_file =
  { filename : string
  ; desc     : opened_file_desc
  ; tail     : bool
  }

and opened_file_desc =
  | Fd      of Unix.file_descr
  | Channel of out_channel

type purpose =
  | Internal_job
  | Build_job of Path.t list

module Temp = struct
  let tmp_files = ref String_set.empty
  let () =
    at_exit (fun () ->
      let fns = !tmp_files in
      tmp_files := String_set.empty;
      String_set.iter fns ~f:(fun fn ->
        try Sys.force_remove fn with _ -> ()))

  let create prefix suffix =
    let fn = Filename.temp_file prefix suffix in
    tmp_files := String_set.add !tmp_files fn;
    fn

  let destroy fn =
    (try Sys.force_remove fn with Sys_error _ -> ());
    tmp_files := String_set.remove !tmp_files fn
end

module Fancy = struct
  let split_prog s =
    let len = String.length s in
    if len = 0 then
      "", "", ""
    else begin
      let rec find_prog_start i =
        if i < 0 then
          0
        else
          match s.[i] with
          | '\\' | '/' -> (i + 1)
          | _ -> find_prog_start (i - 1)
      in
      let prog_end =
        match s.[len - 1] with
        | '"' -> len - 1
        | _   -> len
      in
      let prog_start = find_prog_start (prog_end - 1) in
      let prog_end =
        match String.index_from s prog_start '.' with
        | exception _ -> prog_end
        | i -> i
      in
      let before = String.sub s ~pos:0 ~len:prog_start in
      let after = String.sub s ~pos:prog_end ~len:(len - prog_end) in
      let prog = String.sub s ~pos:prog_start ~len:(prog_end - prog_start) in
      before, prog, after
    end

  let colorize_prog s =
    let len = String.length s in
    if len = 0 then
      s
    else
      let before, prog, after = split_prog s in
      before ^ Colors.colorize ~key:prog prog ^ after

  let rec colorize_args = function
    | [] -> []
    | "-o" :: fn :: rest ->
      "-o" :: Colors.(apply_string output_filename) fn :: colorize_args rest
    | x :: rest -> x :: colorize_args rest

  let command_line ~prog ~args ~dir ~stdout_to ~stderr_to =
    let quote = quote_for_shell in
    let prog = colorize_prog (quote prog) in
    let s =
      String.concat (prog :: colorize_args (List.map args ~f:quote)) ~sep:" "
    in
    let s =
      match dir with
      | None -> s
      | Some dir -> sprintf "(cd %s && %s)" dir s
    in
    match stdout_to, stderr_to with
    | (File fn1 | Opened_file { filename = fn1; _ }),
      (File fn2 | Opened_file { filename = fn2; _ }) when fn1 = fn2 ->
      sprintf "%s &> %s" s fn1
    | _ ->
      let s =
        match stdout_to with
        | Terminal -> s
        | File fn | Opened_file { filename = fn; _ } -> sprintf "%s > %s" s fn
      in
      match stderr_to with
      | Terminal -> s
      | File fn | Opened_file { filename = fn; _ } -> sprintf "%s 2> %s" s fn

  let pp_purpose ppf = function
    | Internal_job ->
      Format.fprintf ppf "(internal)"
    | Build_job targets ->
      let rec split_paths targets_acc ctxs_acc = function
        | [] -> List.rev targets_acc, String_set.(to_list (of_list ctxs_acc))
        | path :: rest ->
          let add_ctx ctx acc = if ctx = "default" then acc else ctx :: acc in
          match Utils.analyse_target path with
          | Other path ->
            split_paths (Path.to_string path :: targets_acc) ctxs_acc rest
          | Regular (ctx, filename) ->
            split_paths (Path.to_string filename :: targets_acc)
              (add_ctx ctx ctxs_acc) rest
          | Alias (ctx, name) ->
            split_paths (("alias " ^ Path.to_string name) :: targets_acc)
              (add_ctx ctx ctxs_acc) rest
      in
      let target_names, contexts = split_paths [] [] targets in
      let target_names_grouped_by_prefix =
        List.map target_names ~f:Filename.split_extension_after_dot
        |> String_map.of_list_multi
        |> String_map.to_list
      in
      let pp_comma ppf () = Format.fprintf ppf "," in
      let pp_group ppf (prefix, suffixes) =
        match suffixes with
        | [] -> assert false
        | [suffix] -> Format.fprintf ppf "%s%s" prefix suffix
        | _ ->
          Format.fprintf ppf "%s{%a}"
            prefix
            (Format.pp_print_list ~pp_sep:pp_comma Format.pp_print_string)
            suffixes
      in
      let pp_contexts ppf = function
        | [] -> ()
        | ctxs ->
          Format.fprintf ppf " @{<details>[%a]@}"
            (Format.pp_print_list ~pp_sep:pp_comma
               (fun ppf s -> Format.fprintf ppf "%s" s))
            ctxs
      in
      Format.fprintf ppf "%a%a"
        (Format.pp_print_list ~pp_sep:pp_comma pp_group)
        target_names_grouped_by_prefix
        pp_contexts
        contexts;
end

let get_std_output ~default = function
  | Terminal -> (default, None)
  | File fn ->
    let fd = Unix.openfile fn [O_WRONLY; O_CREAT; O_TRUNC; O_SHARE_DELETE] 0o666 in
    (fd, Some (Fd fd))
  | Opened_file { desc; tail; _ } ->
    let fd =
      match desc with
      | Fd      fd -> fd
      | Channel oc -> flush oc; Unix.descr_of_out_channel oc
    in
    (fd, Option.some_if tail desc)

let close_std_output = function
  | None -> ()
  | Some (Fd      fd) -> Unix.close fd
  | Some (Channel oc) -> close_out  oc

let gen_id =
  let next = ref (-1) in
  fun () -> incr next; !next

let run_internal ?dir ?(stdout_to=Terminal) ?(stderr_to=Terminal) ~env ~purpose
      fail_mode prog args =
  Scheduler.wait_for_available_job ()
  >>= fun scheduler ->
  let display = Scheduler.display scheduler in
  let dir =
    match dir with
    | Some "." -> None
    | _ -> dir
  in
  let id = gen_id () in
  let ok_codes = accepted_codes fail_mode in
  let command_line = Fancy.command_line ~prog ~args ~dir ~stdout_to ~stderr_to in
  if display = Verbose then
    Format.eprintf "@{<kwd>Running@}[@{<id>%d@}]: %s@." id
      (Colors.strip_colors_for_stderr command_line);
  let argv = Array.of_list (prog :: args) in
  let output_filename, stdout_fd, stderr_fd, to_close =
    match stdout_to, stderr_to with
    | (Terminal, _ | _, Terminal) when !Clflags.capture_outputs ->
      let fn = Temp.create "jbuilder" ".output" in
      let fd = Unix.openfile fn [O_WRONLY; O_SHARE_DELETE] 0 in
      (Some fn, fd, fd, Some fd)
    | _ ->
      (None, Unix.stdout, Unix.stderr, None)
  in
  let stdout, close_stdout = get_std_output stdout_to ~default:stdout_fd in
  let stderr, close_stderr = get_std_output stderr_to ~default:stderr_fd in
  let run () =
    Unix.create_process_env prog argv (Env.to_unix env)
      Unix.stdin stdout stderr
  in
  let pid =
    match dir with
    | None -> run ()
    | Some dir -> Scheduler.with_chdir scheduler ~dir ~f:run
  in
  Option.iter to_close ~f:Unix.close;
  close_std_output close_stdout;
  close_std_output close_stderr;
  Scheduler.wait_for_process pid
  >>| fun status ->
  let output =
    match output_filename with
    | None -> ""
    | Some fn ->
      let s = Io.read_file fn in
      Temp.destroy fn;
      let len = String.length s in
      if len > 0 && s.[len - 1] <> '\n' then
        s ^ "\n"
      else
        s
  in
  Log.command (Scheduler.log scheduler)
    ~command_line:command_line
    ~output:output
    ~exit_status:status;
  let _, progname, _ = Fancy.split_prog prog in
  let print fmt = Errors.kerrf ~f:(Scheduler.print scheduler) fmt in
  match status with
  | WEXITED n when code_is_ok ok_codes n ->
    if display = Verbose then begin
      if output <> "" then
        print "@{<kwd>Output@}[@{<id>%d@}]:\n%s" id output;
      if n <> 0 then
        print
          "@{<warning>Warning@}: Command [@{<id>%d@}] exited with code %d, \
           but I'm ignoring it, hope that's OK.\n" id n
    end else if output <> "" ||
                (display = Short && purpose <> Internal_job) then begin
      let pad = String.make (max 0 (12 - String.length progname)) ' ' in
      print "%s@{<ok>%s@} %a\n%s" pad progname Fancy.pp_purpose purpose output
    end;
    n
  | WEXITED n ->
    if display = Verbose then
      die "\n@{<kwd>Command@} [@{<id>%d@}] exited with code %d:\n\
           @{<prompt>$@} %s\n%s"
        id n
        (Colors.strip_colors_for_stderr command_line)
        (Colors.strip_colors_for_stderr output)
    else
      die "@{<error>%12s@} %a @{<error>(exit %d)@}\n\
           @{<details>%s@}\n\
           %s"
        progname Fancy.pp_purpose purpose n
        (Ansi_color.strip command_line)
        output
  | WSIGNALED n ->
    if display = Verbose then
      die "\n@{<kwd>Command@} [@{<id>%d@}] got signal %s:\n\
           @{<prompt>$@} %s\n%s"
        id (Utils.signal_name n)
        (Colors.strip_colors_for_stderr command_line)
        (Colors.strip_colors_for_stderr output)
    else
      die "@{<error>%12s@} %a @{<error>(got signal %s)@}\n\
           @{<details>%s@}\n\
           %s"
        progname Fancy.pp_purpose purpose (Utils.signal_name n)
        (Ansi_color.strip command_line)
        output
  | WSTOPPED _ -> assert false

let run ?dir ?stdout_to ?stderr_to ~env ?(purpose=Internal_job) fail_mode
      prog args =
  map_result fail_mode
    (run_internal ?dir ?stdout_to ?stderr_to ~env ~purpose fail_mode prog args)
    ~f:ignore

let run_capture_gen ?dir ~env ?(purpose=Internal_job) fail_mode prog args ~f =
  let fn = Temp.create "jbuild" ".output" in
  map_result fail_mode
    (run_internal ?dir ~stdout_to:(File fn) ~env ~purpose fail_mode prog args)
    ~f:(fun () ->
      let x = f fn in
      Temp.destroy fn;
      x)

let run_capture       = run_capture_gen ~f:Io.read_file
let run_capture_lines = run_capture_gen ~f:Io.lines_of_file

let run_capture_line ?dir ~env ?(purpose=Internal_job) fail_mode prog args =
  run_capture_gen ?dir ~env ~purpose fail_mode prog args ~f:(fun fn ->
    match Io.lines_of_file fn with
    | [x] -> x
    | l ->
      let cmdline =
        let s = String.concat (prog :: args) ~sep:" " in
        match dir with
        | None -> s
        | Some dir -> sprintf "cd %s && %s" dir s
      in
      match l with
      | [] ->
        die "command returned nothing: %s" cmdline
      | _ ->
        die "command returned too many lines: %s\n%s"
          cmdline (String.concat l ~sep:"\n"))
# 10922 "boot.ml"
end
module Setup : sig
# 1 "src/setup.mli"
(** Setup of Jbuilder *)

(** These parameters are set by [ocaml configure.ml] or copied from
    [setup.defaults.ml]. During bootstrap, values from [setup.boot.ml]
    are used *)

(** Where to find installed libraries for the default context. If
    [None], auto-detect it using standard tools such as [ocamlfind]. *)
val library_path : string list option

(** Where to install libraries for the default context. *)
val library_destdir : string option
# 10938 "boot.ml"
end = struct
# 1 "src/setup.boot.ml"
let library_path    = Some []
let library_destdir = None
# 10943 "boot.ml"
end
module Workspace : sig
# 1 "src/workspace.mli"
(** Workspaces definitions *)

open! Import

module Context : sig
  module Target : sig
    type t =
      | Native
      | Named of string
  end
  module Opam : sig
    type t =
      { name   : string
      ; switch : string
      ; root   : string option
      ; merlin : bool
      ; targets : Target.t list
      }
  end

  type t = Default of Target.t list | Opam of Opam.t

  val name : t -> string
end

type t =
  { merlin_context : string option
  ; contexts       : Context.t list
  }

val load : ?x:string -> string -> t
# 10978 "boot.ml"
end = struct
# 1 "src/workspace.ml"
open Import
open Sexp.Of_sexp

module Context = struct
  module Target = struct
    type t =
      | Native
      | Named of string

    let t sexp =
      match string sexp with
      | "native" -> Native
      | s        -> Named s
  end

  module Opam = struct
    type t =
      { name    : string
      ; switch  : string
      ; root    : string option
      ; merlin  : bool
      ; targets : Target.t list
      }

    let t =
      field   "switch"  string                                    >>= fun switch ->
      field   "name"    string ~default:switch                    >>= fun name ->
      field   "targets" (list Target.t) ~default:[Target.Native]  >>= fun targets ->
      field_o "root"    string                                    >>= fun root ->
      field_b "merlin"                                            >>= fun merlin ->
      return { switch
             ; name
             ; root
             ; merlin
             ; targets
             }
  end

  type t = Default of Target.t list | Opam of Opam.t

  let t = function
    | Atom (_, A "default") -> Default [Native]
    | List (_, List _ :: _) as sexp -> Opam (record Opam.t sexp)
    | sexp ->
      sum
        [ cstr_record "default"
            (field "targets" (list Target.t) ~default:[Target.Native]
             >>= fun targets ->
             return (Default targets))
        ; cstr_record "opam"
            (Opam.t >>= fun x -> return (Opam x))
        ]
        sexp

  let name = function
    | Default _ -> "default"
    | Opam    o -> o.name

  let targets = function
    | Default l -> l
    | Opam    o -> o.targets

  let all_names t =
    let n = name t in
    n :: List.filter_map (targets t) ~f:(function
      | Native -> None
      | Named s -> Some (n ^ "." ^ s))
end

type t =
  { merlin_context : string option
  ; contexts       : Context.t list
  }

let t ?x sexps =
  let defined_names = ref String_set.empty in
  let merlin_ctx, contexts =
    List.fold_left sexps ~init:(None, []) ~f:(fun (merlin_ctx, ctxs) sexp ->
      let ctx =
        sum
          [ cstr "context" (Context.t @> nil) (fun x -> x) ]
          sexp
      in
      let ctx =
        match x with
        | None -> ctx
        | Some s ->
          let target = Context.Target.Named s in
          let add_target target targets =
            if List.mem target ~set:targets then
              targets
            else
              targets @ [target]
          in
          match ctx with
          | Default targets -> Default (add_target target targets)
          | Opam o -> Opam { o with targets = add_target target o.targets }
      in
      let name = Context.name ctx in
      if name = "" ||
         String.is_prefix name ~prefix:"." ||
         name = "log" ||
         name = "install" ||
         String.contains name '/' ||
         String.contains name '\\' then
        of_sexp_errorf sexp "%S is not allowed as a build context name" name;
      if String_set.mem !defined_names name then
        of_sexp_errorf sexp "second definition of build context %S" name;
      defined_names := String_set.union !defined_names
                         (String_set.of_list (Context.all_names ctx));
      match ctx, merlin_ctx with
      | Opam { merlin = true; _ }, Some _ ->
        of_sexp_errorf sexp "you can only have one context for merlin"
      | Opam { merlin = true; _ }, None ->
        (Some name, ctx :: ctxs)
      | _ ->
        (merlin_ctx, ctx :: ctxs))
  in
  let contexts =
    match contexts with
    | [] -> [Context.Default [Native]]
    | _  -> contexts
  in
  let merlin_ctx =
    match merlin_ctx with
    | Some _ -> merlin_ctx
    | None ->
      if List.exists contexts
           ~f:(function Context.Default _ -> true | _ -> false) then
        Some "default"
      else
        None
  in
  { merlin_context = merlin_ctx
  ; contexts       = List.rev contexts
  }

let load ?x fname = t ?x (Sexp.load ~fname ~mode:Many)
# 11119 "boot.ml"
end
module Context : sig
# 1 "src/context.mli"
(** Compilation contexts *)

(** jbuild supports two different kind of contexts:

    - the default context, which correspond to the environment jbuild is run, i.e. it
    takes [ocamlc] and other tools from the [PATH] and the ocamlfind configuration where
    it can find it

    - opam switch contexts, where one opam switch correspond to one context

    each context is built into a sub-directory of "_build":

    - _build/default for the default context
    - _build/<switch> for other contexts

    jbuild is able to build simultaneously against several contexts. In particular this
    allow for simple cross-compilation: when an executable running on the host is needed,
    it is obtained by looking in another context.
*)

open! Import

module Kind : sig
  module Opam : sig
    type t =
      { root   : string
      ; switch : string
      }
  end
  type t = Default | Opam of Opam.t
end

type t =
  { name : string
  ; kind : Kind.t

  ; (** [true] if this context is used for the .merlin files *)
    merlin : bool

  ; (** If this context is a cross-compilation context, you need another context for
        building tools used for the compilation that run on the host. *)
    for_host : t option

  ; (** [false] if a user explicitly listed this context in the workspace.
        Controls whether we add artifacts from this context @install *)
    implicit : bool

  ; (** Directory where artifact are stored, for instance "_build/default" *)
    build_dir : Path.t

  ; (** [PATH] *)
    path : Path.t list

  ; (** [OCAML_TOPLEVEL_PATH] *)
    toplevel_path : Path.t option

  ; (** Ocaml bin directory with all ocaml tools *)
    ocaml_bin  : Path.t
  ; ocaml      : Path.t
  ; ocamlc     : Path.t
  ; ocamlopt   : Path.t option
  ; ocamldep   : Path.t
  ; ocamlmklib : Path.t

  ; (** Environment variables *)
    env : Env.t

  ; findlib : Findlib.t
  ; findlib_toolchain : string option

  ; (** Misc *)
    arch_sixtyfour : bool

  ; opam_var_cache : (string, string) Hashtbl.t

  ; (** Native dynlink *)
    natdynlink_supported : bool

  ; ocaml_config            : Ocaml_config.t
  ; version_string          : string
  ; version                 : int * int * int
  ; stdlib_dir              : Path.t
  ; ccomp_type              : string
  ; c_compiler              : string
  ; ocamlc_cflags           : string list
  ; ocamlopt_cflags         : string list
  ; bytecomp_c_libraries    : string list
  ; native_c_libraries      : string list
  ; cc_profile              : string list
  ; architecture            : string
  ; system                  : string
  ; ext_obj                 : string
  ; ext_asm                 : string
  ; ext_lib                 : string
  ; ext_dll                 : string
  ; ext_exe                 : string
  ; os_type                 : string
  ; default_executable_name : string
  ; host                    : string
  ; target                  : string
  ; flambda                 : bool
  ; exec_magic_number       : string
  ; cmi_magic_number        : string
  ; cmo_magic_number        : string
  ; cma_magic_number        : string
  ; cmx_magic_number        : string
  ; cmxa_magic_number       : string
  ; ast_impl_magic_number   : string
  ; ast_intf_magic_number   : string
  ; cmxs_magic_number       : string
  ; cmt_magic_number        : string

  ; which_cache             : (string, Path.t option) Hashtbl.t
  }

val sexp_of_t : t -> Sexp.t

(** Compare the context names *)
val compare : t -> t -> Ordering.t

val create
  :  ?merlin:bool
  -> env:Env.t
  -> Workspace.Context.t
  -> t list Fiber.t

val which : t -> string -> Path.t option

val opam_config_var : t -> string -> string option Fiber.t

val install_prefix : t -> Path.t Fiber.t
val install_ocaml_libdir : t -> Path.t option Fiber.t

(** Return the compiler needed for this compilation mode *)
val compiler : t -> Mode.t -> Path.t option

(** The best compilation mode for this context *)
val best_mode : t -> Mode.t

(** [\["-g"\]] if [!Clflags.g] and [\[\]] otherwise *)
val cc_g : t -> string list
# 11264 "boot.ml"
end = struct
# 1 "src/context.ml"
open Import
open Fiber.O

module Kind = struct
  module Opam = struct
    type t =
      { root   : string
      ; switch : string
      }
  end
  type t = Default | Opam of Opam.t

  let sexp_of_t : t -> Sexp.t = function
    | Default -> Sexp.unsafe_atom_of_string "default"
    | Opam o  ->
      Sexp.To_sexp.(record [ "root"  , string o.root
                           ; "switch", string o.switch
                           ])
end

type t =
  { name                    : string
  ; kind                    : Kind.t
  ; merlin                  : bool
  ; for_host                : t option
  ; implicit                : bool
  ; build_dir               : Path.t
  ; path                    : Path.t list
  ; toplevel_path           : Path.t option
  ; ocaml_bin               : Path.t
  ; ocaml                   : Path.t
  ; ocamlc                  : Path.t
  ; ocamlopt                : Path.t option
  ; ocamldep                : Path.t
  ; ocamlmklib              : Path.t
  ; env                     : Env.t
  ; findlib                 : Findlib.t
  ; findlib_toolchain       : string option
  ; arch_sixtyfour          : bool
  ; opam_var_cache          : (string, string) Hashtbl.t
  ; natdynlink_supported    : bool
  ; ocaml_config            : Ocaml_config.t
  ; version_string          : string
  ; version                 : int * int * int
  ; stdlib_dir              : Path.t
  ; ccomp_type              : string
  ; c_compiler              : string
  ; ocamlc_cflags           : string list
  ; ocamlopt_cflags         : string list
  ; bytecomp_c_libraries    : string list
  ; native_c_libraries      : string list
  ; cc_profile              : string list
  ; architecture            : string
  ; system                  : string
  ; ext_obj                 : string
  ; ext_asm                 : string
  ; ext_lib                 : string
  ; ext_dll                 : string
  ; ext_exe                 : string
  ; os_type                 : string
  ; default_executable_name : string
  ; host                    : string
  ; target                  : string
  ; flambda                 : bool
  ; exec_magic_number       : string
  ; cmi_magic_number        : string
  ; cmo_magic_number        : string
  ; cma_magic_number        : string
  ; cmx_magic_number        : string
  ; cmxa_magic_number       : string
  ; ast_impl_magic_number   : string
  ; ast_intf_magic_number   : string
  ; cmxs_magic_number       : string
  ; cmt_magic_number        : string
  ; which_cache             : (string, Path.t option) Hashtbl.t
  }

let sexp_of_t t =
  let open Sexp.To_sexp in
  let path = Path.sexp_of_t in
  record
    [ "name", string t.name
    ; "kind", Kind.sexp_of_t t.kind
    ; "merlin", bool t.merlin
    ; "for_host", option string (Option.map t.for_host ~f:(fun t -> t.name))
    ; "build_dir", path t.build_dir
    ; "toplevel_path", option path t.toplevel_path
    ; "ocaml_bin", path t.ocaml_bin
    ; "ocaml", path t.ocaml
    ; "ocamlc", path t.ocamlc
    ; "ocamlopt", option path t.ocamlopt
    ; "ocamldep", path t.ocamldep
    ; "ocamlmklib", path t.ocamlmklib
    ; "env", Env.sexp_of_t (Env.diff t.env Env.initial)
    ; "findlib_path", list path (Findlib.path t.findlib)
    ; "arch_sixtyfour", bool t.arch_sixtyfour
    ; "natdynlink_supported", bool t.natdynlink_supported
    ; "opam_vars", string_hashtbl string t.opam_var_cache
    ; "ocaml_config", Ocaml_config.sexp_of_t t.ocaml_config
    ; "which", string_hashtbl (option path) t.which_cache
    ]

let compare a b = compare a.name b.name

let opam_config_var ~env ~cache var =
  match Hashtbl.find cache var with
  | Some _ as x -> Fiber.return x
  | None ->
    match Bin.opam with
    | None -> Fiber.return None
    | Some fn ->
      Process.run_capture (Accept All) (Path.to_string fn) ~env
        ["config"; "var"; var]
      >>| function
      | Ok s ->
        let s = String.trim s in
        Hashtbl.add cache var s;
        Some s
      | Error _ -> None

let which ~cache ~path x =
  Hashtbl.find_or_add cache x ~f:(Bin.which ~path)

let ocamlpath_sep =
  if Sys.cygwin then
    (* because that's what ocamlfind expects *)
    ';'
  else
    Bin.path_sep

let create ~(kind : Kind.t) ~path ~env ~name ~merlin ~targets () =
  let opam_var_cache = Hashtbl.create 128 in
  (match kind with
   | Opam { root; _ } ->
     Hashtbl.add opam_var_cache "root" root
   | Default -> ());
  let prog_not_found_in_path prog =
    Utils.program_not_found prog ~context:name
  in
  let which_cache = Hashtbl.create 128 in
  let which x = which ~cache:which_cache ~path x in
  let findlib_config_path = lazy (
    match which "ocamlfind" with
    | None -> prog_not_found_in_path "ocamlfind"
    | Some fn ->
      (* When OCAMLFIND_CONF is set, "ocamlfind printconf" does print
         the contents of the variable, but "ocamlfind printconf conf"
         still prints the configuration file set at the configuration
         time of ocamlfind, sigh... *)
      match Env.get env "OCAMLFIND_CONF" with
      | Some s -> Fiber.return (Path.absolute s)
      | None ->
        Process.run_capture_line ~env Strict
          (Path.to_string fn) ["printconf"; "conf"]
        >>| Path.absolute)
  in

  let create_one ~name ~implicit ?findlib_toolchain ?host ~merlin () =
    (match findlib_toolchain with
     | None -> Fiber.return None
     | Some toolchain ->
       Lazy.force findlib_config_path >>| fun path ->
       Some (Findlib.Config.load path ~toolchain ~context:name))
    >>= fun findlib_config ->

    let get_tool_using_findlib_config prog =
      Option.bind findlib_config ~f:(fun conf ->
        match Findlib.Config.get conf prog with
        | None -> None
        | Some s ->
          match Filename.analyze_program_name s with
          | In_path | Relative_to_current_dir -> which s
          | Absolute -> Some (Path.absolute s))
    in

    let ocamlc =
      match get_tool_using_findlib_config "ocamlc" with
      | Some x -> x
      | None ->
        match which "ocamlc" with
        | Some x -> x
        | None -> prog_not_found_in_path "ocamlc"
    in
    let dir = Path.parent ocamlc in
    let ocaml_tool_not_found prog =
      die "ocamlc found in %s, but %s/%s doesn't exist (context: %s)"
        (Path.to_string dir) (Path.to_string dir) prog name
    in
    let get_ocaml_tool prog =
      match get_tool_using_findlib_config prog with
      | None -> Bin.best_prog dir prog
      | Some _ as x -> x
    in
    let get_ocaml_tool_exn prog =
      match get_ocaml_tool prog with
      | None -> ocaml_tool_not_found prog
      | Some fn -> fn
    in

    let build_dir = Path.of_string (sprintf "_build/%s" name) in
    let ocamlpath =
      match
        let var = "OCAMLPATH" in
        match kind, findlib_toolchain with
        | Default, None -> Env.get env var
        | _ ->
          (* If we are not in the default context, we can only use the
             OCAMLPATH variable if it is specific to this build
             context *)
          (* CR-someday diml: maybe we should actually clear OCAMLPATH
             in other build contexts *)
          match Env.get env var, Env.get Env.initial var with
          | None  , None   -> None
          | Some s, None   -> Some s
          | None  , Some _ -> None
          | Some x, Some y -> Option.some_if (x <> y) x
      with
      | None -> []
      | Some s -> Bin.parse_path s ~sep:ocamlpath_sep
    in
    let findlib_path () =
      match kind, findlib_toolchain, Setup.library_path with
      | Default, None, Some l ->
        Fiber.return (ocamlpath @ List.map l ~f:Path.absolute)
      | _ ->
        (* If ocamlfind is present, it has precedence over everything else. *)
        match which "ocamlfind" with
        | Some fn ->
          let args =
            let args = ["printconf"; "path"] in
            match findlib_toolchain with
            | None -> args
            | Some s -> "-toolchain" :: s :: args
          in
          Process.run_capture_lines ~env Strict (Path.to_string fn) args
          >>| fun l ->
          (* Don't prepend the contents of [OCAMLPATH] since findlib
             does it already *)
          List.map l ~f:Path.absolute
        | None ->
          (* If there no ocamlfind in the PATH, check if we have opam
             and assume a standard opam setup *)
          opam_config_var ~env ~cache:opam_var_cache "lib"
          >>| function
          | Some s -> ocamlpath @ [Path.absolute s]
          | None ->
            (* If neither opam neither ocamlfind are present, assume
               that libraries are [dir ^ "/../lib"] *)
            ocamlpath @ [Path.relative (Path.parent dir) "lib"]
    in
    let ocaml_config_ok_exn = function
      | Ok x -> x
      | Error msg ->
        die "Failed to parse the output of '%s -config':@\n\
             %s"
          (Path.to_string ocamlc) msg
    in
    Fiber.fork_and_join
      findlib_path
      (fun () ->
         Process.run_capture_lines ~env Strict
           (Path.to_string ocamlc) ["-config"]
         >>| fun lines ->
         let open Result.O in
         ocaml_config_ok_exn
           (Ocaml_config.Vars.of_lines lines >>= Ocaml_config.make))
    >>= fun (findlib_path, ocfg) ->
    let version = Ocaml_config.version ocfg in
    let env =
      (* See comment in ansi_color.ml for setup_env_for_colors. For
         OCaml < 4.05, OCAML_COLOR is not supported so we use
         OCAMLPARAM. OCaml 4.02 doesn't support 'color' in OCAMLPARAM,
         so we just don't force colors with 4.02. *)
      if !Clflags.capture_outputs
      && Lazy.force Colors.stderr_supports_colors
      && version >= (4, 03, 0)
      && version <  (4, 05, 0) then
        let value =
          match Env.get env "OCAMLPARAM" with
          | None -> "color=always,_"
          | Some s -> "color=always," ^ s
        in
        Env.add env ~var:"OCAMLPARAM" ~value
      else
        env
    in
    let env =
      let cwd = Sys.getcwd () in
      let extend_var var ?(path_sep=Bin.path_sep) v =
        let v = Filename.concat cwd (Path.to_string v) in
        match Env.get env var with
        | None -> (var, v)
        | Some prev -> (var, sprintf "%s%c%s" v path_sep prev)
      in
      let vars =
        [ extend_var "CAML_LD_LIBRARY_PATH"
            (Path.relative
               (Config.local_install_dir ~context:name)
               "lib/stublibs")
        ; extend_var "OCAMLPATH" ~path_sep:ocamlpath_sep
            (Path.relative
               (Config.local_install_dir ~context:name)
               "lib")
        ; extend_var "MANPATH"
            (Config.local_install_man_dir ~context:name)
        ]
      in
      Env.extend env ~vars:(Env.Map.of_list_exn vars)
      |> Env.update ~var:"PATH" ~f:(fun _ ->
        match host with
        | None ->
          let _key, path =
            extend_var "PATH" (Config.local_install_bin_dir ~context:name) in
          Some path
        | Some host ->
          Env.get host.env "PATH"
      )
    in
    let stdlib_dir = Path.of_string (Ocaml_config.standard_library ocfg) in
    let natdynlink_supported = Ocaml_config.natdynlink_supported ocfg in
    let version        = Ocaml_config.version ocfg        in
    let version_string = Ocaml_config.version_string ocfg in
    let arch_sixtyfour = Ocaml_config.word_size ocfg = 64 in
    Fiber.return
      { name
      ; implicit
      ; kind
      ; merlin
      ; for_host = host
      ; build_dir
      ; path
      ; toplevel_path =
          Option.map (Env.get env "OCAML_TOPLEVEL_PATH") ~f:Path.absolute

      ; ocaml_bin  = dir
      ; ocaml      = (match which "ocaml" with Some p -> p | None -> prog_not_found_in_path "ocaml")
      ; ocamlc
      ; ocamlopt   = get_ocaml_tool     "ocamlopt"
      ; ocamldep   = get_ocaml_tool_exn "ocamldep"
      ; ocamlmklib = get_ocaml_tool_exn "ocamlmklib"

      ; env
      ; findlib = Findlib.create ~stdlib_dir ~path:findlib_path
      ; findlib_toolchain
      ; arch_sixtyfour

      ; opam_var_cache

      ; natdynlink_supported

      ; stdlib_dir
      ; ocaml_config = ocfg
      ; version_string
      ; version
      ; ccomp_type              = Ocaml_config.ccomp_type              ocfg
      ; c_compiler              = Ocaml_config.c_compiler              ocfg
      ; ocamlc_cflags           = Ocaml_config.ocamlc_cflags           ocfg
      ; ocamlopt_cflags         = Ocaml_config.ocamlopt_cflags         ocfg
      ; bytecomp_c_libraries    = Ocaml_config.bytecomp_c_libraries    ocfg
      ; native_c_libraries      = Ocaml_config.native_c_libraries      ocfg
      ; cc_profile              = Ocaml_config.cc_profile              ocfg
      ; architecture            = Ocaml_config.architecture            ocfg
      ; system                  = Ocaml_config.system                  ocfg
      ; ext_obj                 = Ocaml_config.ext_obj                 ocfg
      ; ext_asm                 = Ocaml_config.ext_asm                 ocfg
      ; ext_lib                 = Ocaml_config.ext_lib                 ocfg
      ; ext_dll                 = Ocaml_config.ext_dll                 ocfg
      ; ext_exe                 = Ocaml_config.ext_exe                 ocfg
      ; os_type                 = Ocaml_config.os_type                 ocfg
      ; default_executable_name = Ocaml_config.default_executable_name ocfg
      ; host                    = Ocaml_config.host                    ocfg
      ; target                  = Ocaml_config.target                  ocfg
      ; flambda                 = Ocaml_config.flambda                 ocfg
      ; exec_magic_number       = Ocaml_config.exec_magic_number       ocfg
      ; cmi_magic_number        = Ocaml_config.cmi_magic_number        ocfg
      ; cmo_magic_number        = Ocaml_config.cmo_magic_number        ocfg
      ; cma_magic_number        = Ocaml_config.cma_magic_number        ocfg
      ; cmx_magic_number        = Ocaml_config.cmx_magic_number        ocfg
      ; cmxa_magic_number       = Ocaml_config.cmxa_magic_number       ocfg
      ; ast_impl_magic_number   = Ocaml_config.ast_impl_magic_number   ocfg
      ; ast_intf_magic_number   = Ocaml_config.ast_intf_magic_number   ocfg
      ; cmxs_magic_number       = Ocaml_config.cmxs_magic_number       ocfg
      ; cmt_magic_number        = Ocaml_config.cmt_magic_number        ocfg

      ; which_cache
      }
  in

  let implicit = not (List.mem ~set:targets Workspace.Context.Target.Native) in
  create_one () ~implicit ~name ~merlin >>= fun native ->
  Fiber.parallel_map targets ~f:(function
    | Native -> Fiber.return None
    | Named findlib_toolchain ->
      let name = sprintf "%s.%s" name findlib_toolchain in
      create_one () ~implicit:false ~name ~findlib_toolchain ~host:native
        ~merlin:false
      >>| fun x -> Some x)
  >>| fun others ->
  native :: List.filter_map others ~f:(fun x -> x)

let opam_config_var t var = opam_config_var ~env:t.env ~cache:t.opam_var_cache var

let default ?(merlin=true) ~env ~targets () =
  create ~kind:Default ~path:Bin.path ~env ~name:"default" ~merlin ~targets ()

let create_for_opam ?root ~env ~targets ~switch ~name ?(merlin=false) () =
  match Bin.opam with
  | None -> Utils.program_not_found "opam"
  | Some fn ->
    (match root with
     | Some root -> Fiber.return root
     | None ->
       Process.run_capture_line Strict ~env
         (Path.to_string fn) ["config"; "var"; "root"])
    >>= fun root ->
    Process.run_capture ~env Strict (Path.to_string fn)
      ["config"; "env"; "--root"; root; "--switch"; switch; "--sexp"]
    >>= fun s ->
    let vars =
      Usexp.parse_string ~fname:"<opam output>" ~mode:Single s
      |> Sexp.Of_sexp.(list (pair string string))
      |> Env.Map.of_list_multi
      |> Env.Map.mapi ~f:(fun var values ->
        match List.rev values with
        | [] -> assert false
        | [x] -> x
        | x :: _ ->
          Format.eprintf
            "@{<warning>Warning@}: variable %S present multiple times in the output of:\n\
             @{<details>%s@}@."
            var
            (String.concat ~sep:" "
               (List.map ~f:quote_for_shell
                  [Path.to_string fn; "config"; "env"; "--root"; root;
                   "--switch"; switch; "--sexp"]));
          x)
    in
    let path =
      match Env.Map.find vars "PATH" with
      | None   -> Bin.path
      | Some s -> Bin.parse_path s
    in
    let env = Env.extend env ~vars in
    create ~kind:(Opam { root; switch }) ~targets ~path ~env ~name ~merlin ()

let create ?merlin ~env def =
  match (def : Workspace.Context.t) with
  | Default targets -> default ~env ~targets ?merlin ()
  | Opam { name; switch; root; targets; _ } ->
    create_for_opam ?root ~env ~switch ~name ?merlin ~targets ()

let which t s = which ~cache:t.which_cache ~path:t.path s

let install_prefix t =
  opam_config_var t "prefix" >>| function
  | Some x -> Path.absolute x
  | None   -> Path.parent t.ocaml_bin

let install_ocaml_libdir t =
  match t.kind, t.findlib_toolchain, Setup.library_destdir with
  | Default, None, Some d ->
    Fiber.return (Some (Path.absolute d))
  | _ ->
    (* If ocamlfind is present, it has precedence over everything else. *)
    match which t "ocamlfind" with
    | Some fn ->
      (Process.run_capture_line ~env:t.env Strict
         (Path.to_string fn) ["printconf"; "destdir"]
       >>| fun s ->
       Some (Path.absolute s))
    | None ->
      Fiber.return None

let compiler t (mode : Mode.t) =
  match mode with
  | Byte   -> Some t.ocamlc
  | Native -> t.ocamlopt

let best_mode t : Mode.t =
  match t.ocamlopt with
  | Some _ -> Native
  | None   -> Byte

let cc_g (ctx : t) =
  if !Clflags.g && ctx.ccomp_type <> "msvc" then
    ["-g"]
  else
    []
# 11755 "boot.ml"
end
module Print_diff : sig
# 1 "src/print_diff.mli"
(** Diff two files that are expected not to match. *)
val print : Path.t -> Path.t -> _ Fiber.t
# 11761 "boot.ml"
end = struct
# 1 "src/print_diff.ml"
open Import

open Fiber.O

let print path1 path2 =
  let dir, file1, file2 =
    match
      Path.extract_build_context_dir path1,
      Path.extract_build_context_dir path2
    with
    | Some (dir1, f1), Some (dir2, f2) when dir1 = dir2 ->
      (Path.to_string dir1, Path.to_string f1, Path.to_string f2)
    | _ ->
      (".", Path.to_string path1, Path.to_string path2)
  in
  let loc = Loc.in_file file1 in
  let fallback () =
    die "%aFiles %s and %s differ." Loc.print loc
      (Path.to_string_maybe_quoted path1)
      (Path.to_string_maybe_quoted path2)
  in
  let normal_diff () =
    match Bin.which "diff" with
    | None -> fallback ()
    | Some prog ->
      Format.eprintf "%a@?" Loc.print loc;
      Process.run ~dir ~env:Env.initial Strict (Path.to_string prog)
        ["-u"; file1; file2]
      >>= fun () ->
      fallback ()
  in
  match !Clflags.diff_command with
  | Some cmd ->
    let sh, arg = Utils.system_shell_exn ~needed_to:"print diffs" in
    let cmd =
      sprintf "%s %s %s" cmd (quote_for_shell file1) (quote_for_shell file2)
    in
    Process.run ~dir ~env:Env.initial Strict (Path.to_string sh) [arg; cmd]
    >>= fun () ->
    die "command reported no differences: %s"
      (if dir = "." then
         cmd
       else
         sprintf "cd %s && %s" (quote_for_shell dir) cmd)
  | None ->
    match Bin.which "patdiff" with
    | None -> normal_diff ()
    | Some prog ->
      Process.run ~dir ~env:Env.initial Strict (Path.to_string prog)
        [ "-keep-whitespace"
        ; "-location-style"; "omake"
        ; if Lazy.force Colors.stderr_supports_colors then
            "-unrefined"
          else
            "-ascii"
        ; file1
        ; file2
        ]
      >>= fun () ->
      (* Use "diff" if "patdiff" reported no differences *)
      normal_diff ()
# 11825 "boot.ml"
end
module String_with_vars : sig
# 1 "src/string_with_vars.mli"
(** String with variables of the form ${...} or $(...)

    Variables cannot contain "${", "$(", ")" or "}". For instance in "$(cat ${x})", only
    "${x}" will be considered a variable, the rest is text. *)

open Import

type t
(** A sequence of text and variables. *)

val t : t Sexp.Of_sexp.t
(** [t ast] takes an [ast] sexp and returns a string-with-vars.  This
   function distinguishes between unquoted variables — such as ${@} —
   and quoted variables — such as "${@}". *)

val loc : t -> Loc.t
(** [loc t] returns the location of [t] — typically, in the jbuild file. *)

val sexp_of_t : t -> Sexp.t

(** Same as [sexp_of_t] but the S-expression encodes the internal
    structure of [t]. *)
val sexp_of_ast : t -> Sexp.t

val to_string : t -> string

(** [t] generated by the OCaml code. The first argument should be
   [__POS__]. The second is either a string to parse, a variable name
   or plain text.  [quoted] says whether the string is quoted ([false]
   by default). *)
val virt       : ?quoted: bool -> (string * int * int * int) -> string -> t
val virt_var   : ?quoted: bool -> (string * int * int * int) -> string -> t
val virt_text  : (string * int * int * int) -> string -> t

val vars : t -> String_set.t
(** [vars t] returns the set of all variables in [t]. *)

val fold : t -> init:'a -> f:('a -> Loc.t -> string -> 'a) -> 'a
(** [fold t ~init ~f] fold [f] on all variables of [t], the text
   portions being ignored. *)

val iter : t -> f:(Loc.t -> string -> unit) -> unit
(** [iter t ~f] iterates [f] over all variables of [t], the text
   portions being ignored. *)

module type EXPANSION = sig
  type t
  (** The value to which variables are expanded. *)

  val is_multivalued : t -> bool
  (** Report whether the value is a multivalued one (such as for
     example ${@}) which much be in quoted strings to be concatenated
     to text or other variables. *)

  type context
  (** Context needed to expand values of type [t] to strings. *)

  val to_string : context -> t -> string
  (** When needing to expand with text portions or if the
     string-with-vars is quoted, the value is converted to a string
     using [to_string]. *)
end

module Expand_to(V : EXPANSION) : sig
  val expand : V.context -> t -> f:(Loc.t -> string -> V.t option) ->
               (V.t, string) Either.t
  (** [expand t ~f] return [t] where all variables have been expanded
     using [f].  If [f loc var] return [Some x], the variable [var] is
     replaced by [x]; otherwise, the variable is inserted as [${var}]
     or [$(var)] — depending on the original concrete syntax used.  *)

  val partial_expand :
    V.context -> t -> f:(Loc.t -> string -> V.t option) ->
    ((V.t, string) either, t) Either.t
  (** [partial_expand t ~f] is like [expand_generic] where all
     variables that could be expanded (i.e., those for which [f]
     returns [Some _]) are.  If all the variables of [t] were
     expanded, a string is returned.  If [f] returns [None] on at
     least a variable of [t], it returns a string-with-vars. *)
end

val expand :
  t -> f:(Loc.t -> string -> string option) -> string
(** Specialized version [Expand_to.expand] that returns a string (so
   variables are assumed to expand to a single value). *)

val partial_expand :
  t -> f:(Loc.t -> string -> string option) -> (string, t) Either.t
(** [partial_expand] is a specialized version of
   [Expand_to.partial_expand] that returns a string. *)
# 11919 "boot.ml"
end = struct
# 1 "src/string_with_vars.ml"
open! Import

type var_syntax = Parens | Braces

type item =
  | Text of string
  | Var of var_syntax * string

type t =
  { items : item list
  ; loc : Loc.t
  ; quoted : bool }

module Token = struct
  type t =
    | String of string
    | Open   of var_syntax
    | Close  of var_syntax

  let tokenise s =
    let len = String.length s in
    let sub i j = String.sub s ~pos:i ~len:(j - i) in
    let cons_str i j acc = if i = j then acc else String (sub i j) :: acc in
    let rec loop i j =
      if j = len
      then cons_str i j []
      else
        match s.[j] with
        | '}' -> cons_str i j (Close Braces :: loop (j + 1) (j + 1))
        | ')' -> cons_str i j (Close Parens :: loop (j + 1) (j + 1))
        | '$' when j + 1 < len -> begin
            match s.[j + 1] with
            | '{' -> cons_str i j (Open Braces :: loop (j + 2) (j + 2))
            | '(' -> cons_str i j (Open Parens :: loop (j + 2) (j + 2))
            | _   -> loop i (j + 1)
          end
        | _ -> loop i (j + 1)
    in
    loop 0 0

  let to_string = function
    | String s     -> s
    | Open  Braces -> "${"
    | Open  Parens -> "$("
    | Close Braces -> "}"
    | Close Parens -> ")"
end

(* Remark: Consecutive [Text] items are concatenated. *)
let rec of_tokens : Token.t list -> item list = function
  | [] -> []
  | Open a :: String s :: Close b :: rest when a = b ->
    Var (a, s) :: of_tokens rest
  | token :: rest ->
    let s = Token.to_string token in
    match of_tokens rest with
    | Text s' :: l -> Text (s ^ s') :: l
    | l -> Text s :: l

let items_of_string s = of_tokens (Token.tokenise s)

let t : Sexp.Of_sexp.ast -> t = function
  | Atom(loc, A s) -> { items = items_of_string s;  loc;  quoted = false }
  | Quoted_string (loc, s) ->
     { items = items_of_string s;  loc;  quoted = true }
  | List _ as sexp -> Sexp.Of_sexp.of_sexp_error sexp "Atom expected"

let loc t = t.loc

let virt ?(quoted=false) pos s =
  { items = items_of_string s;  loc = Loc.of_pos pos;  quoted }
let virt_var ?(quoted=false) pos s =
  { items = [Var (Braces, s)];  loc = Loc.of_pos pos;  quoted }
let virt_text pos s =
  { items = [Text s];  loc = Loc.of_pos pos;  quoted = true }

let sexp_of_var_syntax = function
  | Parens -> Sexp.unsafe_atom_of_string "parens"
  | Braces -> Sexp.unsafe_atom_of_string "braces"

let sexp_of_item =
  let open Sexp in function
    | Text s -> List [Sexp.unsafe_atom_of_string "text" ;
                      Sexp.atom_or_quoted_string s]
    | Var (vs, s) -> List [sexp_of_var_syntax vs ;
                           Sexp.atom_or_quoted_string s]

let sexp_of_ast t = Sexp.To_sexp.list sexp_of_item t.items

let fold t ~init ~f =
  List.fold_left t.items ~init ~f:(fun acc item ->
      match item with
      | Text _ -> acc
      | Var (_, v) -> f acc t.loc v)

let iter t ~f = List.iter t.items ~f:(function
                    | Text _ -> ()
                    | Var (_, v) -> f t.loc v)

let vars t = fold t ~init:String_set.empty ~f:(fun acc _ x -> String_set.add acc x)

let string_of_var syntax v =
  match syntax with
  | Parens -> sprintf "$(%s)" v
  | Braces -> sprintf "${%s}" v

module type EXPANSION = sig
  type t
  val is_multivalued : t -> bool
  type context
  val to_string : context -> t -> string
end

let concat_rev = function
  | [] -> ""
  | [s] -> s
  | l -> String.concat (List.rev l) ~sep:""

module Expand_to(V: EXPANSION) = struct

  let expand ctx t ~f =
    match t.items with
    | [Var (syntax, v)] when not t.quoted ->
      (* Unquoted single var *)
      (match f t.loc v with
       | Some e -> Left e
       | None -> Right (string_of_var syntax v))
    | _ ->
      Right (List.map t.items ~f:(function
        | Text s -> s
        | Var (syntax, v) ->
          match f t.loc v with
          | Some x ->
            if not t.quoted && V.is_multivalued x then
              Loc.fail t.loc "please quote the string \
                              containing the list variable %s"
                (string_of_var syntax v)
            else V.to_string ctx x
          | None -> string_of_var syntax v)
             |> String.concat ~sep:"")

  let partial_expand ctx t ~f =
    let commit_text acc_text acc =
      let s = concat_rev acc_text in
      if s = "" then acc else Text s :: acc
    in
    let rec loop acc_text acc items =
      match items with
      | [] -> begin
          match acc with
          | [] -> Left  (Right (concat_rev acc_text))
          | _  -> Right { t with items = List.rev (commit_text acc_text acc) }
        end
      | Text s :: items -> loop (s :: acc_text) acc items
      | Var (syntax, v) as it :: items ->
         match f t.loc v with
         | None -> loop [] (it :: commit_text acc_text acc) items
         | Some x ->
            if not t.quoted && V.is_multivalued x then
           Loc.fail t.loc "please quote the string containing the \
                           list variable %s" (string_of_var syntax v)
         else loop (V.to_string ctx x :: acc_text) acc items
    in
    match t.items with
    | [Var (_, v)] when not t.quoted ->
       (* Unquoted single var *)
       (match f t.loc v with
        | Some e -> Left (Left e)
        | None   -> Right t)
    | _ -> loop [] [] t.items
end

module String_expansion = struct
  type t = string
  let is_multivalued _ = false
  type context = unit
  let to_string () (s: string) = s
end

module S = Expand_to(String_expansion)

let expand t ~f =
  match S.expand () t ~f with Left s | Right s -> s

let partial_expand t ~f =
  match S.partial_expand () t ~f with
  | Left (Left s | Right s) -> Left s
  | Right _ as x -> x

let to_string t =
  match t.items with
  (* [to_string is only called from action.ml, always on [t]s of this form *)
  | [Var (syntax, v)] -> string_of_var syntax v
  | items ->
    List.map items ~f:(function
      | Text s -> s
      | Var (syntax, v) -> string_of_var syntax v)
    |> String.concat ~sep:""

let sexp_of_t t = Sexp.To_sexp.string (to_string t)
# 12122 "boot.ml"
end
module Action : sig
# 1 "src/action.mli"
open! Import

module Var_expansion : sig
  module Concat_or_split : sig
    type t =
      | Concat (** default *)
      | Split  (** the variable is a "split" list of items *)
  end

  type t =
    | Paths   of Path.t list * Concat_or_split.t
    | Strings of string list * Concat_or_split.t

  val to_string : Path.t -> t -> string
  (** [to_string dir v] convert the variable expansion to a string.
     If it is a path, the corresponding string will be relative to
     [dir]. *)
end

module Outputs : module type of struct include Action_intf.Outputs end

(** result of the lookup of a program, the path to it or information about the
    failure and possibly a hint how to fix it *)
module Prog : sig
  module Not_found : sig
    type t =
      { context : string
      ; program : string
      ; hint    : string option
      }

    val raise : t -> _
  end

  type t = (Path.t, Not_found.t) result
end

include Action_intf.Ast
  with type program := Prog.t
  with type path    := Path.t
  with type string  := string

include Action_intf.Helpers
  with type program := Prog.t
  with type path    := Path.t
  with type string  := string
  with type t       := t

val t : t Sexp.Of_sexp.t

module For_shell : sig
  include Action_intf.Ast
    with type program := string
    with type path    := string
    with type string  := string

  val sexp_of_t : t Sexp.To_sexp.t
end

(** Convert the action to a format suitable for printing *)
val for_shell : t -> For_shell.t

(** Return the list of directories the action chdirs to *)
val chdirs : t -> Path.Set.t

(** Ast where programs are not yet looked up in the PATH *)
module Unresolved : sig
  type action = t

  module Program : sig
    type t =
      | This   of Path.t
      | Search of string
  end

  include Action_intf.Ast
    with type program := Program.t
    with type path    := Path.t
    with type string  := string

  val resolve : t -> f:(string -> Path.t) -> action
end with type action := t

module Unexpanded : sig
  include Action_intf.Ast
    with type program := String_with_vars.t
    with type path    := String_with_vars.t
    with type string  := String_with_vars.t

  val t : t Sexp.Of_sexp.t
  val sexp_of_t : t Sexp.To_sexp.t

  module Partial : sig
    include Action_intf.Ast
      with type program = (Unresolved.Program.t, String_with_vars.t) either
      with type path    = (Path.t              , String_with_vars.t) either
      with type string  = (string              , String_with_vars.t) either

    val expand
      :  t
      -> dir:Path.t
      -> map_exe:(Path.t -> Path.t)
      -> f:(Loc.t -> String.t -> Var_expansion.t option)
      -> Unresolved.t
  end

  val partial_expand
    :  t
    -> dir:Path.t
    -> map_exe:(Path.t -> Path.t)
    -> f:(Loc.t -> string -> Var_expansion.t option)
    -> Partial.t
end

val exec : targets:Path.Set.t -> context:Context.t option -> t -> unit Fiber.t

(* Return a sandboxed version of an action *)
val sandbox
  :  t
  -> sandboxed:(Path.t -> Path.t)
  -> deps:Path.t list
  -> targets:Path.t list
  -> t

(** Infer dependencies and targets.

    This currently doesn't support well (rename ...) and (remove-tree ...). However these
    are not exposed in the DSL.
*)
module Infer : sig
  module Outcome : sig
    type t =
      { deps    : Path.Set.t
      ; targets : Path.Set.t
      }
  end

  val infer : t -> Outcome.t

  (** If [all_targets] is [true] and a target cannot be determined statically, fail *)
  val partial : all_targets:bool -> Unexpanded.Partial.t -> Outcome.t

  (** Return the list of targets of an unexpanded action. *)
  val unexpanded_targets : Unexpanded.t -> String_with_vars.t list
end

module Promotion : sig
  module File : sig
    type t =
      { src : Path.t
      ; dst : Path.t
      }

    (** Register a file to promote *)
    val register : t -> unit
  end

  (** Promote all registered files if [!Clflags.auto_promote]. Otherwise dump the list of
      registered files to [_build/.to-promote]. *)
  val finalize : unit -> unit

  val promote_files_registered_in_last_run : unit -> unit
end
# 12289 "boot.ml"
end = struct
# 1 "src/action.ml"
open Import
open Sexp.Of_sexp

module Outputs = struct
  include Action_intf.Outputs

  let to_string = function
    | Stdout -> "stdout"
    | Stderr -> "stderr"
    | Outputs -> "outputs"
end

module type Sexpable = sig
  type t
  val t : t Sexp.Of_sexp.t
  val sexp_of_t : t Sexp.To_sexp.t
end

module Make_ast
    (Program : Sexpable)
    (Path    : Sexpable)
    (String  : Sexpable)
    (Ast : Action_intf.Ast
     with type program := Program.t
     with type path    := Path.t
     with type string  := String.t) =
struct
  include Ast

  let rec t sexp =
    let path = Path.t and string = String.t in
    sum
      [ cstr_rest "run" (Program.t @> nil) string (fun prog args -> Run (prog, args))
      ; cstr "chdir"    (path @> t @> nil)        (fun dn t -> Chdir (dn, t))
      ; cstr "setenv"   (string @> string @> t @> nil)   (fun k v t -> Setenv (k, v, t))
      ; cstr "with-stdout-to"  (path @> t @> nil) (fun fn t -> Redirect (Stdout, fn, t))
      ; cstr "with-stderr-to"  (path @> t @> nil) (fun fn t -> Redirect (Stderr, fn, t))
      ; cstr "with-outputs-to" (path @> t @> nil) (fun fn t -> Redirect (Outputs, fn, t))
      ; cstr "ignore-stdout"   (t @> nil)      (fun t -> Ignore (Stdout, t))
      ; cstr "ignore-stderr"   (t @> nil)      (fun t -> Ignore (Stderr, t))
      ; cstr "ignore-outputs"  (t @> nil)      (fun t -> Ignore (Outputs, t))
      ; cstr_rest "progn"      nil t         (fun l -> Progn l)
      ; cstr "echo"           (string @> nil)         (fun x -> Echo x)
      ; cstr "cat"            (path @> nil)         (fun x -> Cat x)
      ; cstr "copy" (path @> path @> nil)              (fun src dst -> Copy (src, dst))
      (*
         (* We don't expose symlink to the user yet since this might complicate things *)
         ; cstr "symlink" (a @> a @> nil) (fun src dst -> Symlink (dst, Cat src))
      *)
      ; cstr "copy#" (path @> path @> nil) (fun src dst ->
          Copy_and_add_line_directive (src, dst))
      ; cstr_loc "copy-and-add-line-directive" (path @> path @> nil) (fun loc src dst ->
          Loc.warn loc "copy-and-add-line-directive is deprecated, use copy# instead";
          Copy_and_add_line_directive (src, dst))
      ; cstr "copy#" (path @> path @> nil) (fun src dst ->
          Copy_and_add_line_directive (src, dst))
      ; cstr "system" (string @> nil) (fun cmd -> System cmd)
      ; cstr "bash"   (string @> nil) (fun cmd -> Bash   cmd)
      ; cstr "write-file" (path @> string @> nil) (fun fn s -> Write_file (fn, s))
      ; cstr "diff" (path @> path @> nil)
          (fun file1 file2 -> Diff { optional = false; file1; file2 })
      ; cstr "diff?" (path @> path @> nil)
          (fun file1 file2 -> Diff { optional = true ; file1; file2 })
      ]
      sexp

  let rec sexp_of_t : _ -> Sexp.t =
    let path = Path.sexp_of_t and string = String.sexp_of_t in
    function
    | Run (a, xs) -> List (Sexp.unsafe_atom_of_string "run"
                           :: Program.sexp_of_t a :: List.map xs ~f:string)
    | Chdir (a, r) -> List [Sexp.unsafe_atom_of_string "chdir" ;
                            path a ; sexp_of_t r]
    | Setenv (k, v, r) -> List [Sexp.unsafe_atom_of_string "setenv" ;
                                string k ; string v ; sexp_of_t r]
    | Redirect (outputs, fn, r) ->
      List [ Sexp.atom (sprintf "with-%s-to" (Outputs.to_string outputs))
           ; path fn
           ; sexp_of_t r
           ]
    | Ignore (outputs, r) ->
      List [ Sexp.atom (sprintf "ignore-%s" (Outputs.to_string outputs))
           ; sexp_of_t r
           ]
    | Progn l -> List (Sexp.unsafe_atom_of_string "progn"
                       :: List.map l ~f:sexp_of_t)
    | Echo x -> List [Sexp.unsafe_atom_of_string "echo"; string x]
    | Cat x -> List [Sexp.unsafe_atom_of_string "cat"; path x]
    | Copy (x, y) ->
      List [Sexp.unsafe_atom_of_string "copy"; path x; path y]
    | Symlink (x, y) ->
      List [Sexp.unsafe_atom_of_string "symlink"; path x; path y]
    | Copy_and_add_line_directive (x, y) ->
      List [Sexp.unsafe_atom_of_string "copy#"; path x; path y]
    | System x -> List [Sexp.unsafe_atom_of_string "system"; string x]
    | Bash   x -> List [Sexp.unsafe_atom_of_string "bash"; string x]
    | Write_file (x, y) -> List [Sexp.unsafe_atom_of_string "write-file";
                                 path x; string y]
    | Rename (x, y) -> List [Sexp.unsafe_atom_of_string "rename";
                             path x; path y]
    | Remove_tree x -> List [Sexp.unsafe_atom_of_string "remove-tree"; path x]
    | Mkdir x       -> List [Sexp.unsafe_atom_of_string "mkdir"; path x]
    | Digest_files paths -> List [Sexp.unsafe_atom_of_string "digest-files";
                                  List (List.map paths ~f:path)]
    | Diff { optional = false; file1; file2 } ->
      List [Sexp.unsafe_atom_of_string "diff"; path file1; path file2]
    | Diff { optional = true; file1; file2 } ->
      List [Sexp.unsafe_atom_of_string "diff?"; path file1; path file2]

  let run prog args = Run (prog, args)
  let chdir path t = Chdir (path, t)
  let setenv var value t = Setenv (var, value, t)
  let with_stdout_to path t = Redirect (Stdout, path, t)
  let with_stderr_to path t = Redirect (Stderr, path, t)
  let with_outputs_to path t = Redirect (Outputs, path, t)
  let ignore_stdout t = Ignore (Stdout, t)
  let ignore_stderr t = Ignore (Stderr, t)
  let ignore_outputs t = Ignore (Outputs, t)
  let progn ts = Progn ts
  let echo s = Echo s
  let cat path = Cat path
  let copy a b = Copy (a, b)
  let symlink a b = Symlink (a, b)
  let copy_and_add_line_directive a b = Copy_and_add_line_directive (a, b)
  let system s = System s
  let bash s = Bash s
  let write_file p s = Write_file (p, s)
  let rename a b = Rename (a, b)
  let remove_tree path = Remove_tree path
  let mkdir path = Mkdir path
  let digest_files files = Digest_files files
  let diff ?(optional=false) file1 file2 = Diff { optional; file1; file2 }
end

module Make_mapper
    (Src : Action_intf.Ast)
    (Dst : Action_intf.Ast)
= struct
  let rec map (t : Src.t) ~dir ~f_program ~f_string ~f_path : Dst.t =
    match t with
    | Run (prog, args) ->
      Run (f_program ~dir prog, List.map args ~f:(f_string ~dir))
    | Chdir (fn, t) ->
      Chdir (f_path ~dir fn, map t ~dir:fn ~f_program ~f_string ~f_path)
    | Setenv (var, value, t) ->
      Setenv (f_string ~dir var, f_string ~dir value, map t ~dir ~f_program ~f_string ~f_path)
    | Redirect (outputs, fn, t) ->
      Redirect (outputs, f_path ~dir fn, map t ~dir ~f_program ~f_string ~f_path)
    | Ignore (outputs, t) ->
      Ignore (outputs, map t ~dir ~f_program ~f_string ~f_path)
    | Progn l -> Progn (List.map l ~f:(fun t -> map t ~dir ~f_program ~f_string ~f_path))
    | Echo x -> Echo (f_string ~dir x)
    | Cat x -> Cat (f_path ~dir x)
    | Copy (x, y) -> Copy (f_path ~dir x, f_path ~dir y)
    | Symlink (x, y) ->
      Symlink (f_path ~dir x, f_path ~dir y)
    | Copy_and_add_line_directive (x, y) ->
      Copy_and_add_line_directive (f_path ~dir x, f_path ~dir y)
    | System x -> System (f_string ~dir x)
    | Bash x -> Bash (f_string ~dir x)
    | Write_file (x, y) -> Write_file (f_path ~dir x, f_string ~dir y)
    | Rename (x, y) -> Rename (f_path ~dir x, f_path ~dir y)
    | Remove_tree x -> Remove_tree (f_path ~dir x)
    | Mkdir x -> Mkdir (f_path ~dir x)
    | Digest_files x -> Digest_files (List.map x ~f:(f_path ~dir))
    | Diff { optional; file1; file2 } ->
      Diff { optional; file1 = f_path ~dir file1; file2 = f_path ~dir file2 }
end

module Prog = struct
  module Not_found = struct
    type t =
      { context : string
      ; program : string
      ; hint    : string option
      }

    let raise { context ; program ; hint } =
      Utils.program_not_found ?hint ~context program
  end

  type t = (Path.t, Not_found.t) result

  let t sexp = Ok (Path.t sexp)

  let sexp_of_t = function
    | Ok s -> Path.sexp_of_t s
    | Error (e : Not_found.t) -> Sexp.To_sexp.string e.program
end

module type Ast = Action_intf.Ast
  with type program = Prog.t
  with type path    = Path.t
  with type string  = String.t
module rec Ast : Ast = Ast

module String_with_sexp = struct
  type t = string
  let t = Sexp.Of_sexp.string
  let sexp_of_t = Sexp.To_sexp.string
end

include Make_ast
    (Prog)
    (Path)
    (String_with_sexp)
    (Ast)

module For_shell = struct
  module type Ast = Action_intf.Ast
    with type program = string
    with type path    = string
    with type string  = string
  module rec Ast : Ast = Ast

  include Make_ast
      (String_with_sexp)
      (String_with_sexp)
      (String_with_sexp)
      (Ast)
end

module Relativise = Make_mapper(Ast)(For_shell.Ast)

let for_shell t =
  Relativise.map t
    ~dir:Path.root
    ~f_string:(fun ~dir:_ x -> x)
    ~f_path:(fun ~dir x -> Path.reach x ~from:dir)
    ~f_program:(fun ~dir x ->
      match x with
      | Ok p -> Path.reach p ~from:dir
      | Error e -> e.program)

module Unresolved = struct
  module Program = struct
    type t =
      | This   of Path.t
      | Search of string

    let of_string ~dir s =
      if String.contains s '/' then
        This (Path.relative dir s)
      else
        Search s
  end

  module type Uast = Action_intf.Ast
    with type program = Program.t
    with type path    = Path.t
    with type string  = String.t
  module rec Uast : Uast = Uast
  include Uast

  include Make_mapper(Uast)(Ast)

  let resolve t ~f =
    map t
      ~dir:Path.root
      ~f_path:(fun ~dir:_ x -> x)
      ~f_string:(fun ~dir:_ x -> x)
      ~f_program:(fun ~dir:_ -> function
        | This p -> Ok p
        | Search s -> Ok (f s))
end

module Var_expansion = struct
  module Concat_or_split = struct
    type t =
      | Concat (* default *)
      | Split  (* the variable is a "split" list of items *)
  end

  open Concat_or_split

  type t =
    | Paths   of Path.t list * Concat_or_split.t
    | Strings of string list * Concat_or_split.t

  let is_multivalued = function
    | Paths (_, Split) | Strings (_, Split) -> true
    | Paths (_, Concat) | Strings (_, Concat) -> false

  type context = Path.t (* For String_with_vars.Expand_to *)

  let concat = function
    | [s] -> s
    | l -> String.concat ~sep:" " l

  let string_of_path ~dir p = Path.reach ~from:dir p
  let path_of_string dir s = Path.relative dir s

  let to_strings dir = function
    | Strings (l, Split ) -> l
    | Strings (l, Concat) -> [concat l]
    | Paths   (l, Split ) -> List.map l ~f:(string_of_path ~dir)
    | Paths   (l, Concat) -> [concat (List.map l ~f:(string_of_path ~dir))]

  let to_string (dir: context) = function
    | Strings (l, _) -> concat l
    | Paths   (l, _) -> concat (List.map l ~f:(string_of_path ~dir))

  let to_path dir = function
    | Strings (l, _) -> path_of_string dir (concat l)
    | Paths ([p], _) -> p
    | Paths (l,   _) ->
      path_of_string dir (concat (List.map l ~f:(string_of_path ~dir)))

  let to_prog_and_args dir exp : Unresolved.Program.t * string list =
    let module P = Unresolved.Program in
    match exp with
    | Paths   ([p], _) -> (This p, [])
    | Strings ([s], _) -> (P.of_string ~dir s, [])
    | Paths ([], _) | Strings ([], _) -> (Search "", [])
    | Paths (l, Concat) ->
      (This
         (path_of_string dir
            (concat (List.map l ~f:(string_of_path ~dir)))),
       [])
    | Strings (l, Concat) ->
      (P.of_string ~dir (concat l), l)
    | Paths (p :: l, Split) ->
      (This p, List.map l ~f:(string_of_path ~dir))
    | Strings (s :: l, Split) ->
      (P.of_string ~dir s, l)
end

module VE = Var_expansion
module To_VE = String_with_vars.Expand_to(VE)
module SW = String_with_vars

module Unexpanded = struct
  module type Uast = Action_intf.Ast
    with type program = String_with_vars.t
    with type path    = String_with_vars.t
    with type string  = String_with_vars.t
  module rec Uast : Uast = Uast

  include Make_ast(String_with_vars)(String_with_vars)(String_with_vars)(Uast)

  let t sexp =
    match sexp with
    | Atom _ | Quoted_string _ ->
      of_sexp_errorf sexp
        "if you meant for this to be executed with bash, write (bash \"...\") instead"
    | List _ -> t sexp

  let check_mkdir loc path =
    if not (Path.is_local path) then
      Loc.fail loc
        "(mkdir ...) is not supported for paths outside of the workspace:\n\
        \  %a\n"
        Sexp.pp (List [Sexp.unsafe_atom_of_string "mkdir"; Path.sexp_of_t path])

  module Partial = struct
    module Program = Unresolved.Program

    module type Past = Action_intf.Ast
      with type program = (Program.t, String_with_vars.t) either
      with type path    = (Path.t   , String_with_vars.t) either
      with type string  = (string   , String_with_vars.t) either
    module rec Past : Past = Past

    include Past

    module E = struct
      let expand ~generic ~special ~map ~dir ~f = function
        | Left x -> map x
        | Right template ->
           match To_VE.expand dir template ~f with
           | Left  e -> special dir e
           | Right s -> generic dir s
      [@@inlined always]

      let string ~dir ~f x =
        expand ~dir ~f x
          ~generic:(fun _dir x -> x)
          ~special:VE.to_string
          ~map:(fun x -> x)

      let strings ~dir ~f x =
        expand ~dir ~f x
          ~generic:(fun _dir x -> [x])
          ~special:VE.to_strings
          ~map:(fun x -> [x])

      let path ~dir ~f x =
        expand ~dir ~f x
          ~generic:VE.path_of_string
          ~special:VE.to_path
          ~map:(fun x -> x)

      let prog_and_args ~dir ~f x =
        expand ~dir ~f x
          ~generic:(fun _dir s -> (Program.of_string ~dir s, []))
          ~special:VE.to_prog_and_args
          ~map:(fun x -> (x, []))
    end

    let rec expand t ~dir ~map_exe ~f : Unresolved.t =
      match t with
      | Run (prog, args) ->
        let args = List.concat_map args ~f:(E.strings ~dir ~f) in
        let prog, more_args = E.prog_and_args ~dir ~f prog in
        let prog =
          match prog with
          | Search _ -> prog
          | This path -> This (map_exe path)
        in
        Run (prog, more_args @ args)
      | Chdir (fn, t) ->
        let fn = E.path ~dir ~f fn in
        Chdir (fn, expand t ~dir:fn ~map_exe ~f)
      | Setenv (var, value, t) ->
        Setenv (E.string ~dir ~f var, E.string ~dir ~f value,
                expand t ~dir ~map_exe ~f)
      | Redirect (outputs, fn, t) ->
        Redirect (outputs, E.path ~dir ~f fn, expand t ~dir ~map_exe ~f)
      | Ignore (outputs, t) ->
        Ignore (outputs, expand t ~dir ~map_exe ~f)
      | Progn l -> Progn (List.map l ~f:(fun t -> expand t ~dir ~map_exe ~f))
      | Echo x -> Echo (E.string ~dir ~f x)
      | Cat x -> Cat (E.path ~dir ~f x)
      | Copy (x, y) ->
        Copy (E.path ~dir ~f x, E.path ~dir ~f y)
      | Symlink (x, y) ->
        Symlink (E.path ~dir ~f x, E.path ~dir ~f y)
      | Copy_and_add_line_directive (x, y) ->
        Copy_and_add_line_directive (E.path ~dir ~f x, E.path ~dir ~f y)
      | System x -> System (E.string ~dir ~f x)
      | Bash x -> Bash (E.string ~dir ~f x)
      | Write_file (x, y) -> Write_file (E.path ~dir ~f x, E.string ~dir ~f y)
      | Rename (x, y) ->
        Rename (E.path ~dir ~f x, E.path ~dir ~f y)
      | Remove_tree x ->
        Remove_tree (E.path ~dir ~f x)
      | Mkdir x -> begin
          match x with
          | Left  path -> Mkdir path
          | Right tmpl ->
            let path = E.path ~dir ~f x in
            check_mkdir (SW.loc tmpl) path;
            Mkdir path
        end
      | Digest_files x ->
        Digest_files (List.map x ~f:(E.path ~dir ~f))
      | Diff { optional; file1; file2 } ->
        Diff { optional
             ; file1 = E.path ~dir ~f file1
             ; file2 = E.path ~dir ~f file2
             }
  end

  module E = struct
    let expand ~generic ~special ~dir ~f template =
      match To_VE.partial_expand dir template ~f with
      | Left (Left  e) -> Left (special dir e)
      | Left (Right s) -> Left (generic dir s)
      | Right _ as x -> x

    let string ~dir ~f x =
      expand ~dir ~f x
        ~generic:(fun _dir x -> x)
        ~special:VE.to_string

    let strings ~dir ~f x =
      expand ~dir ~f x
        ~generic:(fun _dir x -> [x])
        ~special:VE.to_strings

    let path ~dir ~f x =
      expand ~dir ~f x
        ~generic:VE.path_of_string
        ~special:VE.to_path

    let prog_and_args ~dir ~f x =
      expand ~dir ~f x
        ~generic:(fun dir s -> (Unresolved.Program.of_string ~dir s, []))
        ~special:VE.to_prog_and_args
  end

  let rec partial_expand t ~dir ~map_exe ~f : Partial.t =
    match t with
    | Run (prog, args) ->
      let args =
        List.concat_map args ~f:(fun arg ->
          match E.strings ~dir ~f arg with
          | Left args -> List.map args ~f:(fun x -> Left x)
          | Right _ as x -> [x])
      in
      begin
        match E.prog_and_args ~dir ~f prog with
        | Left (prog, more_args) ->
          let more_args = List.map more_args ~f:(fun x -> Left x) in
          let prog =
            match prog with
            | Search _ -> prog
            | This path -> This (map_exe path)
          in
          Run (Left prog, more_args @ args)
        | Right _ as prog ->
          Run (prog, args)
      end
    | Chdir (fn, t) -> begin
        let res = E.path ~dir ~f fn in
        match res with
        | Left dir ->
          Chdir (res, partial_expand t ~dir ~map_exe ~f)
        | Right fn ->
          let loc = SW.loc fn in
          Loc.fail loc
            "This directory cannot be evaluated statically.\n\
             This is not allowed by jbuilder"
      end
    | Setenv (var, value, t) ->
      Setenv (E.string ~dir ~f var, E.string ~dir ~f value,
              partial_expand t ~dir ~map_exe ~f)
    | Redirect (outputs, fn, t) ->
      Redirect (outputs, E.path ~dir ~f fn, partial_expand t ~dir ~map_exe ~f)
    | Ignore (outputs, t) ->
      Ignore (outputs, partial_expand t ~dir ~map_exe ~f)
    | Progn l -> Progn (List.map l ~f:(fun t -> partial_expand t ~dir ~map_exe ~f))
    | Echo x -> Echo (E.string ~dir ~f x)
    | Cat x -> Cat (E.path ~dir ~f x)
    | Copy (x, y) ->
      Copy (E.path ~dir ~f x, E.path ~dir ~f y)
    | Symlink (x, y) ->
      Symlink (E.path ~dir ~f x, E.path ~dir ~f y)
    | Copy_and_add_line_directive (x, y) ->
      Copy_and_add_line_directive (E.path ~dir ~f x, E.path ~dir ~f y)
    | System x -> System (E.string ~dir ~f x)
    | Bash x -> Bash (E.string ~dir ~f x)
    | Write_file (x, y) -> Write_file (E.path ~dir ~f x, E.string ~dir ~f y)
    | Rename (x, y) ->
      Rename (E.path ~dir ~f x, E.path ~dir ~f y)
    | Remove_tree x ->
      Remove_tree (E.path ~dir ~f x)
    | Mkdir x ->
      let res = E.path ~dir ~f x in
      (match res with
       | Left path -> check_mkdir (SW.loc x) path
       | Right _   -> ());
      Mkdir res
    | Digest_files x ->
      Digest_files (List.map x ~f:(E.path ~dir ~f))
    | Diff { optional; file1; file2 } ->
      Diff { optional
           ; file1 = E.path ~dir ~f file1
           ; file2 = E.path ~dir ~f file2
           }
end

let fold_one_step t ~init:acc ~f =
  match t with
  | Chdir (_, t)
  | Setenv (_, _, t)
  | Redirect (_, _, t)
  | Ignore (_, t) -> f acc t
  | Progn l -> List.fold_left l ~init:acc ~f
  | Run _
  | Echo _
  | Cat _
  | Copy _
  | Symlink _
  | Copy_and_add_line_directive _
  | System _
  | Bash _
  | Write_file _
  | Rename _
  | Remove_tree _
  | Mkdir _
  | Digest_files _
  | Diff _ -> acc

include Make_mapper(Ast)(Ast)

let chdirs =
  let rec loop acc t =
    let acc =
      match t with
      | Chdir (dir, _) -> Path.Set.add acc dir
      | _ -> acc
    in
    fold_one_step t ~init:acc ~f:loop
  in
  fun t -> loop Path.Set.empty t

open Fiber.O

let get_std_output : _ -> Process.std_output_to = function
  | None          -> Terminal
  | Some (fn, oc) -> Opened_file { filename = fn; tail = false; desc = Channel oc }

module Promotion = struct
  module File = struct
    type t =
      { src : Path.t
      ; dst : Path.t
      }

    let t = function
      | Sexp.Ast.List (_, [src; Atom (_, A "as"); dst]) ->
        { src = Path.t src
        ; dst = Path.t dst
        }
      | sexp ->
        Sexp.Of_sexp.of_sexp_errorf sexp "(<file> as <file>) expected"

    let sexp_of_t { src; dst } =
      Sexp.List [Path.sexp_of_t src; Sexp.unsafe_atom_of_string "as";
                 Path.sexp_of_t dst]

    let db : t list ref = ref []

    let register t = db := t :: !db

    let promote { src; dst } =
      Format.eprintf "Promoting %s to %s.@."
        (Path.to_string_maybe_quoted src)
        (Path.to_string_maybe_quoted dst);
      Io.copy_file
        ~src:(Path.to_string src)
        ~dst:(Path.to_string dst)
  end

  let db_file = "_build/.to-promote"

  let dump_db db =
    if Sys.file_exists "_build" then begin
      match db with
      | [] -> if Sys.file_exists db_file then Sys.remove db_file
      | l ->
        Io.write_file db_file
          (String.concat ~sep:""
             (List.map l ~f:(fun x -> Sexp.to_string (File.sexp_of_t x) ^ "\n")))
    end

  let load_db () =
    if Sys.file_exists db_file then
      Sexp.load ~fname:db_file ~mode:Many
      |> List.map ~f:File.t
    else
      []

  let group_by_targets db =
    List.map db ~f:(fun { File. src; dst } ->
      (dst, src))
    |> Path.Map.of_list_multi
    (* Sort the list of possible sources for deterministic behavior *)
    |> Path.Map.map ~f:(List.sort ~compare:Path.compare)

  let do_promote db =
    let by_targets = group_by_targets db  in
    let potential_build_contexts =
      match Path.readdir Path.build_dir with
      | exception _ -> []
      | files ->
        List.filter_map files ~f:(fun fn ->
          if fn = "" || fn.[0] = '.' || fn = "install" then
            None
          else
            let path = Path.(relative build_dir) fn in
            Option.some_if (Path.is_directory path) path)
    in
    let dirs_to_clear_from_cache = Path.root :: potential_build_contexts in
    Path.Map.iteri by_targets ~f:(fun dst srcs ->
      match srcs with
      | [] -> assert false
      | src :: others ->
        (* We remove the files from the digest cache to force a rehash
           on the next run. We do this because on OSX [mtime] is not
           precise enough and if a file is modified and promoted
           quickly, it will look like it hasn't changed even though it
           might have. *)
        List.iter dirs_to_clear_from_cache ~f:(fun dir ->
          Utils.Cached_digest.remove (Path.append dir dst));
        File.promote { src; dst };
        List.iter others ~f:(fun path ->
          Format.eprintf " -> ignored %s.@."
            (Path.to_string_maybe_quoted path)))

  let finalize () =
    let db =
      if !Clflags.auto_promote then
        (do_promote !File.db; [])
      else
        !File.db
    in
    dump_db db

  let promote_files_registered_in_last_run () =
    let db = load_db () in
    do_promote db;
    dump_db []
end

type exec_context =
  { context : Context.t option
  ; purpose : Process.purpose
  }

let exec_run_direct ~ectx ~dir ~env ~stdout_to ~stderr_to prog args =
  begin match ectx.context with
   | None
   | Some { Context.for_host = None; _ } -> ()
   | Some ({ Context.for_host = Some host; _ } as target) ->
     let invalid_prefix prefix =
       match Path.descendant prog ~of_:(Path.of_string prefix) with
       | None -> ()
       | Some _ ->
         die "Context %s has a host %s.@.It's not possible to execute binary %a \
              in it.@.@.This is a bug and should be reported upstream."
           target.name host.name Path.pp prog in
     invalid_prefix ("_build/" ^ target.name);
     invalid_prefix ("_build/install/" ^ target.name);
  end;
  Process.run Strict ~dir:(Path.to_string dir) ~env
    ~stdout_to ~stderr_to
    ~purpose:ectx.purpose
    (Path.reach_for_running ~from:dir prog) args

let exec_run ~stdout_to ~stderr_to =
  let stdout_to = get_std_output stdout_to in
  let stderr_to = get_std_output stderr_to in
  exec_run_direct ~stdout_to ~stderr_to

let exec_echo stdout_to str =
  Fiber.return
    (match stdout_to with
     | None -> print_string str; flush stdout
     | Some (_, oc) -> output_string oc str)

let rec exec t ~ectx ~dir ~env ~stdout_to ~stderr_to =
  match t with
  | Run (Error e, _) ->
    Prog.Not_found.raise e
  | Run (Ok prog, args) ->
    exec_run ~ectx ~dir ~env ~stdout_to ~stderr_to prog args
  | Chdir (dir, t) ->
    exec t ~ectx ~dir ~env ~stdout_to ~stderr_to
  | Setenv (var, value, t) ->
    exec t ~ectx ~dir ~stdout_to ~stderr_to
      ~env:(Env.add env ~var ~value)
  | Redirect (Stdout, fn, Echo s) ->
    Io.write_file (Path.to_string fn) s;
    Fiber.return ()
  | Redirect (outputs, fn, Run (Ok prog, args)) ->
    let out = Process.File (Path.to_string fn) in
    let stdout_to, stderr_to =
      match outputs with
      | Stdout -> (out, get_std_output stderr_to)
      | Stderr -> (get_std_output stdout_to, out)
      | Outputs -> (out, out)
    in
    exec_run_direct ~ectx ~dir ~env ~stdout_to ~stderr_to prog args
  | Redirect (outputs, fn, t) ->
    redirect ~ectx ~dir outputs fn t ~env ~stdout_to ~stderr_to
  | Ignore (outputs, t) ->
    redirect ~ectx ~dir outputs Config.dev_null t ~env ~stdout_to ~stderr_to
  | Progn l ->
    exec_list l ~ectx ~dir ~env ~stdout_to ~stderr_to
  | Echo str -> exec_echo stdout_to str
  | Cat fn ->
    Io.with_file_in (Path.to_string fn) ~f:(fun ic ->
      let oc =
        match stdout_to with
        | None -> stdout
        | Some (_, oc) -> oc
      in
      Io.copy_channels ic oc);
    Fiber.return ()
  | Copy (src, dst) ->
    Io.copy_file ~src:(Path.to_string src) ~dst:(Path.to_string dst);
    Fiber.return ()
  | Symlink (src, dst) ->
    if Sys.win32 then
      Io.copy_file ~src:(Path.to_string src) ~dst:(Path.to_string dst)
    else begin
      let src =
        if Path.is_root dst then
          Path.to_string src
        else
          Path.reach ~from:(Path.parent dst) src
      in
      let dst = Path.to_string dst in
      match Unix.readlink dst with
      | target ->
        if target <> src then begin
          (* @@DRA Win32 remove read-only attribute needed when symlinking enabled *)
          Unix.unlink dst;
          Unix.symlink src dst
        end
      | exception _ ->
        Unix.symlink src dst
    end;
    Fiber.return ()
  | Copy_and_add_line_directive (src, dst) ->
    Io.with_file_in (Path.to_string src) ~f:(fun ic ->
      Io.with_file_out (Path.to_string dst) ~f:(fun oc ->
        let fn = Path.drop_optional_build_context src in
        let directive =
          if List.mem (Path.extension fn) ~set:[".c"; ".cpp"; ".h"] then
            "line"
          else
            ""
        in
        Printf.fprintf oc "#%s 1 %S\n" directive (Path.to_string fn);
        Io.copy_channels ic oc));
    Fiber.return ()
  | System cmd ->
    let path, arg =
      Utils.system_shell_exn ~needed_to:"interpret (system ...) actions"
    in
    exec_run ~ectx ~dir ~env ~stdout_to ~stderr_to path [arg; cmd]
  | Bash cmd ->
    exec_run ~ectx ~dir ~env ~stdout_to ~stderr_to
      (Utils.bash_exn ~needed_to:"interpret (bash ...) actions")
      ["-e"; "-u"; "-o"; "pipefail"; "-c"; cmd]
  | Write_file (fn, s) ->
    Io.write_file (Path.to_string fn) s;
    Fiber.return ()
  | Rename (src, dst) ->
    Unix.rename (Path.to_string src) (Path.to_string dst);
    Fiber.return ()
  | Remove_tree path ->
    Path.rm_rf path;
    Fiber.return ()
  | Mkdir path ->
    (match Path.kind path with
     | External _ ->
       (* Internally we make sure never to do that, and [Unexpanded.*expand] check that *)
       Sexp.code_error
         "(mkdir ...) is not supported for paths outside of the workspace"
         [ "mkdir", Path.sexp_of_t path ]
     | Local path ->
       Path.Local.mkdir_p path);
    Fiber.return ()
  | Digest_files paths ->
    let s =
      let data =
        List.map paths ~f:(fun fn ->
          (fn, Utils.Cached_digest.file fn))
      in
      Digest.string
        (Marshal.to_string data [])
    in
    exec_echo stdout_to s
  | Diff { optional; file1; file2 } ->
    if (optional && not (Path.exists file1 && Path.exists file2)) ||
       Io.compare_files (Path.to_string file1) (Path.to_string file2) = Eq then
      Fiber.return ()
    else begin
      let is_copied_from_source_tree file =
        match Path.drop_build_context file with
        | None -> false
        | Some file -> Path.exists file
      in
      if is_copied_from_source_tree file1 &&
         not (is_copied_from_source_tree file2) then begin
        Promotion.File.register
          { src = file2
          ; dst = Option.value_exn (Path.drop_build_context file1)
          }
      end;
      Print_diff.print file1 file2
    end

and redirect outputs fn t ~ectx ~dir ~env ~stdout_to ~stderr_to =
  let fn = Path.to_string fn in
  let oc = Io.open_out fn in
  let out = Some (fn, oc) in
  let stdout_to, stderr_to =
    match outputs with
    | Stdout -> (out, stderr_to)
    | Stderr -> (stdout_to, out)
    | Outputs -> (out, out)
  in
  exec t ~ectx ~dir ~env ~stdout_to ~stderr_to >>| fun () ->
  close_out oc

and exec_list l ~ectx ~dir ~env ~stdout_to ~stderr_to =
  match l with
  | [] ->
    Fiber.return ()
  | [t] ->
    exec t ~ectx ~dir ~env ~stdout_to ~stderr_to
  | t :: rest ->
    exec t ~ectx ~dir ~env ~stdout_to ~stderr_to >>= fun () ->
    exec_list rest ~ectx ~dir ~env ~stdout_to ~stderr_to

let exec ~targets ~context t =
  let env =
    match (context : Context.t option) with
    | None   -> Env.initial
    | Some c -> c.env
  in
  let targets = Path.Set.to_list targets in
  let purpose = Process.Build_job targets in
  let ectx = { purpose; context } in
  exec t ~ectx ~dir:Path.root ~env ~stdout_to:None ~stderr_to:None

let sandbox t ~sandboxed ~deps ~targets =
  Progn
    [ Progn (List.filter_map deps ~f:(fun path ->
        if Path.is_local path then
          Some (Ast.Symlink (path, sandboxed path))
        else
          None))
    ; map t
        ~dir:Path.root
        ~f_string:(fun ~dir:_ x -> x)
        ~f_path:(fun ~dir:_ p -> sandboxed p)
        ~f_program:(fun ~dir:_ x -> Result.map x ~f:sandboxed)
    ; Progn (List.filter_map targets ~f:(fun path ->
        if Path.is_local path then
          Some (Ast.Rename (sandboxed path, path))
        else
          None))
    ]

module Infer = struct
  module S = Path.Set
  module Outcome = struct
    type t =
      { deps    : S.t
      ; targets : S.t
      }
  end
  open Outcome

  module type Pset = sig
    type t
    val empty : t
    val diff : t -> t -> t
  end

  module type Outcome = sig
    type path_set
    type t =
      { deps    : path_set
      ; targets : path_set
      }
  end

  module type Primitives = sig
    type path
    type program
    type outcome
    val ( +@ ) : outcome -> path -> outcome
    val ( +< ) : outcome -> path -> outcome
    val ( +<! ) : outcome -> program -> outcome
  end

  module Make
      (Ast : Action_intf.Ast)
      (Pset : Pset)
      (Out : Outcome with type path_set := Pset.t)
      (Prim : Primitives
       with type path := Ast.path
       with type program := Ast.program
       with type outcome := Out.t) =
  struct
    open Ast
    open Out
    open Prim
    let rec infer acc t =
      match t with
      | Run (prog, _) -> acc +<! prog
      | Redirect (_, fn, t)  -> infer (acc +@ fn) t
      | Cat fn               -> acc +< fn
      | Write_file (fn, _)  -> acc +@ fn
      | Rename (src, dst)    -> acc +< src +@ dst
      | Copy (src, dst)
      | Copy_and_add_line_directive (src, dst)
      | Symlink (src, dst) -> acc +< src +@ dst
      | Chdir (_, t)
      | Setenv (_, _, t)
      | Ignore (_, t) -> infer acc t
      | Progn l -> List.fold_left l ~init:acc ~f:infer
      | Digest_files l -> List.fold_left l ~init:acc ~f:(+<)
      | Diff { optional; file1; file2 } ->
        if optional then acc else acc +< file1 +< file2
      | Echo _
      | System _
      | Bash _
      | Remove_tree _
      | Mkdir _ -> acc

    let infer t =
      let { deps; targets } =
        infer { deps = Pset.empty; targets = Pset.empty } t
      in
      (* A file can be inferred as both a dependency and a target,
         for instance:

         {[
           (progn (copy a b) (copy b c))
         ]}
      *)
      { deps = Pset.diff deps targets; targets }
  end [@@inline always]

  include Make(Ast)(S)(Outcome)(struct
      let ( +@ ) acc fn = { acc with targets = S.add acc.targets fn }
      let ( +< ) acc fn = { acc with deps    = S.add acc.deps    fn }
      let ( +<! ) acc prog =
        match prog with
        | Ok p -> acc +< p
        | Error _ -> acc
    end)

  module Partial = Make(Unexpanded.Partial.Past)(S)(Outcome)(struct
      let ( +@ ) acc fn =
        match fn with
        | Left  fn -> { acc with targets = S.add acc.targets fn }
        | Right _  -> acc
      let ( +< ) acc fn =
        match fn with
        | Left  fn -> { acc with deps    = S.add acc.deps fn }
        | Right _  -> acc
      let ( +<! ) acc fn =
        match (fn : Unexpanded.Partial.program) with
        | Left  (This fn) -> { acc with deps = S.add acc.deps fn }
        | Left  (Search _) | Right _ -> acc
    end)

  module Partial_with_all_targets = Make(Unexpanded.Partial.Past)(S)(Outcome)(struct
      let ( +@ ) acc fn =
        match fn with
        | Left  fn -> { acc with targets = S.add acc.targets fn }
        | Right sw ->
          Loc.fail (SW.loc sw) "Cannot determine this target statically."
      let ( +< ) acc fn =
        match fn with
        | Left  fn -> { acc with deps    = S.add acc.deps fn }
        | Right _  -> acc
      let ( +<! ) acc fn =
        match (fn : Unexpanded.Partial.program) with
        | Left  (This fn) -> { acc with deps = S.add acc.deps fn }
        | Left  (Search _) | Right _ -> acc
    end)

  let partial ~all_targets t =
    if all_targets then
      Partial_with_all_targets.infer t
    else
      Partial.infer t

  module S_unexp = struct
    type t = String_with_vars.t list
    let empty = []
    let diff a _ = a
  end

  module Outcome_unexp = struct
    type t =
      { deps    : S_unexp.t
      ; targets : S_unexp.t
      }
  end

  module Unexp = Make(Unexpanded.Uast)(S_unexp)(Outcome_unexp)(struct
      open Outcome_unexp
      let ( +@ ) acc fn = { acc with targets = fn :: acc.targets }
      let ( +< ) acc _ = acc
      let ( +<! )= ( +< )
    end)

  let unexpanded_targets t =
    (Unexp.infer t).targets
end
# 13363 "boot.ml"
end
module Binary_kind : sig
# 1 "src/binary_kind.mli"
(** Linking modes for binaries *)

type t =
  | Exe
  | Object
  | Shared_object

val t : t Sexp.Of_sexp.t

val all : t list
# 13377 "boot.ml"
end = struct
# 1 "src/binary_kind.ml"
type t =
  | Exe
  | Object
  | Shared_object

let t =
  let open Sexp.Of_sexp in
  enum
    [ "exe"           , Exe
    ; "object"        , Object
    ; "shared_object" , Shared_object
    ]

let all = [Exe; Object; Shared_object]
# 13394 "boot.ml"
end
module Install : sig
# 1 "src/install.mli"
(** Opam install file *)

module Section : sig
  type t =
    | Lib
    | Libexec
    | Bin
    | Sbin
    | Toplevel
    | Share
    | Share_root
    | Etc
    | Doc
    | Stublibs
    | Man
    | Misc

  val t : t Sexp.Of_sexp.t
end

module Entry : sig
  type t = private
    { src     : Path.t
    ; dst     : string option
    ; section : Section.t
    }

  val make : Section.t -> ?dst:string -> Path.t -> t
  val set_src : t -> Path.t -> t

  val relative_installed_path : t -> package:Package.Name.t -> Path.t
  val add_install_prefix : t -> package:Package.Name.t -> prefix:Path.t -> t
end

val files : Entry.t list -> Path.Set.t
val gen_install_file : Entry.t list -> string
# 13434 "boot.ml"
end = struct
# 1 "src/install.ml"
open Import

module Section = struct
  type t =
    | Lib
    | Libexec
    | Bin
    | Sbin
    | Toplevel
    | Share
    | Share_root
    | Etc
    | Doc
    | Stublibs
    | Man
    | Misc

  let compare : t -> t -> Ordering.t = compare

  let to_string = function
    | Lib        -> "lib"
    | Libexec    -> "libexec"
    | Bin        -> "bin"
    | Sbin       -> "sbin"
    | Toplevel   -> "toplevel"
    | Share      -> "share"
    | Share_root -> "share_root"
    | Etc        -> "etc"
    | Doc        -> "doc"
    | Stublibs   -> "stublibs"
    | Man        -> "man"
    | Misc       -> "misc"

  let t =
    let open Sexp.Of_sexp in
    enum
      [ "lib"        , Lib
      ; "libexec"    , Libexec
      ; "bin"        , Bin
      ; "sbin"       , Sbin
      ; "toplevel"   , Toplevel
      ; "share"      , Share
      ; "share_root" , Share_root
      ; "etc"        , Etc
      ; "doc"        , Doc
      ; "stublibs"   , Stublibs
      ; "man"        , Man
      ; "misc"       , Misc
      ]

  module Paths = struct
    let lib         = Path.(relative root) "lib"
    let libexec     = Path.(relative root) "lib"
    let bin         = Path.(relative root) "bin"
    let sbin        = Path.(relative root) "sbin"
    let toplevel    = Path.(relative root) "lib/toplevel"
    let share       = Path.(relative root) "share"
    let share_root  = Path.(relative root) "share_root"
    let etc         = Path.(relative root) "etc"
    let doc         = Path.(relative root) "doc"
    let stublibs    = Path.(relative root) "lib/stublibs"
    let man         = Path.(relative root) "man"
  end

  let install_dir t ~(package : Package.Name.t) =
    let package = Package.Name.to_string package in
    match t with
    | Bin        -> Paths.bin
    | Sbin       -> Paths.sbin
    | Toplevel   -> Paths.toplevel
    | Share_root -> Paths.share_root
    | Stublibs   -> Paths.stublibs
    | Man        -> Paths.man
    | Lib        -> Path.relative Paths.lib     package
    | Libexec    -> Path.relative Paths.libexec package
    | Share      -> Path.relative Paths.share   package
    | Etc        -> Path.relative Paths.etc     package
    | Doc        -> Path.relative Paths.doc     package
    | Misc       -> invalid_arg "Install.Section.install_dir"
end

module Entry = struct
  type t =
    { src     : Path.t
    ; dst     : string option
    ; section : Section.t
    }

  let make section ?dst src =
    let dst =
      if Sys.win32 then
        let src_base = Path.basename src in
        let dst' =
          match dst with
          | None -> src_base
          | Some s -> s
        in
        match Filename.extension src_base with
        | ".exe" | ".bc" ->
          if Filename.extension dst' <> ".exe" then
            Some (dst' ^ ".exe")
          else
            dst
        | _ -> dst
      else
        dst
    in
    { src
    ; dst
    ; section
    }

  let set_src t src = { t with src }

  let relative_installed_path t ~package =
    let main_dir = Section.install_dir t.section ~package in
    let dst =
      match t.dst with
      | Some x -> x
      | None ->
        let dst = Path.basename t.src in
        match t.section with
        | Man -> begin
            match String.rsplit2 dst ~on:'.' with
            | None -> dst
            | Some (_, sec) -> sprintf "man%s/%s" sec dst
          end
        | _ -> dst
    in
    Path.relative main_dir dst

  let add_install_prefix t ~package ~prefix =
    let opam_will_install_in_this_dir =
      Section.install_dir t.section ~package
    in
    let i_want_to_install_the_file_as =
      Path.append prefix (relative_installed_path t ~package)
    in
    let dst =
      Path.reach i_want_to_install_the_file_as ~from:opam_will_install_in_this_dir
    in
    { t with dst = Some dst }
end

module SMap = Map.Make(Section)

let files entries =
  List.fold_left entries ~init:Path.Set.empty ~f:(fun acc (entry : Entry.t) ->
    Path.Set.add acc entry.src)

let group entries =
  List.map entries ~f:(fun (entry : Entry.t) -> (entry.section, entry))
  |> SMap.of_list_multi
  |> SMap.to_list

let gen_install_file entries =
  let buf = Buffer.create 4096 in
  let pr fmt = Printf.bprintf buf (fmt ^^ "\n") in
  List.iter (group entries) ~f:(fun (section, entries) ->
    pr "%s: [" (Section.to_string section);
      List.iter entries ~f:(fun (e : Entry.t) ->
        let src = Path.to_string e.src in
        match e.dst with
        | None     -> pr "  %S"      src
        | Some dst -> pr "  %S {%S}" src dst);
    pr "]");
  Buffer.contents buf
# 13604 "boot.ml"
end
module Top_closure : sig
# 1 "src/top_closure.mli"
open Stdune

module type Keys = sig
  type t
  type elt
  val empty : t
  val add : t -> elt -> t
  val mem : t -> elt -> bool
end

module type S = sig
  type key

  (** Returns [Error cycle] in case the graph is not a DAG *)
  val top_closure
    :  key:('a -> key)
    -> deps:('a -> 'a list)
    -> 'a list
    -> ('a list, 'a list) result
end

module Int    : S with type key := int
module String : S with type key := string

module Make(Keys : Keys) : S with type key := Keys.elt
# 13633 "boot.ml"
end = struct
# 1 "src/top_closure.ml"
open Import

module type Keys = sig
  type t
  type elt
  val empty : t
  val add : t -> elt -> t
  val mem : t -> elt -> bool
end

module type S = sig
  type key
  val top_closure
    :  key:('a -> key)
    -> deps:('a -> 'a list)
    -> 'a list
    -> ('a list, 'a list) result
end

module Make(Keys : Keys) = struct
  let top_closure ~key ~deps elements =
    let visited = ref Keys.empty in
    let res = ref [] in
    let rec loop elt ~temporarily_marked =
      let key = key elt in
      if Keys.mem temporarily_marked key then
        Error [elt]
      else if not (Keys.mem !visited key) then begin
        visited := Keys.add !visited key;
        let temporarily_marked = Keys.add temporarily_marked key in
        match iter_elts (deps elt) ~temporarily_marked with
        | Ok () -> res := elt :: !res; Ok ()
        | Error l -> Error (elt :: l)
      end else
        Ok ()
    and iter_elts elts ~temporarily_marked =
      match elts with
      | [] -> Ok ()
      | elt :: elts ->
        match loop elt ~temporarily_marked with
        | Error _ as result -> result
        | Ok () -> iter_elts elts ~temporarily_marked
    in
    match iter_elts elements ~temporarily_marked:Keys.empty with
    | Ok () -> Ok (List.rev !res)
    | Error elts -> Error elts
end

module Int    = Make(Int_set)
module String = Make(String_set)
# 13686 "boot.ml"
end
module Module : sig
# 1 "src/module.mli"
open! Import

module Name : sig
  type t = private string

  val add_suffix : t -> string -> t

  val t : t Sexp.To_sexp.t
  val compare : t -> t -> Ordering.t
  val of_string : string -> t
  val to_string : t -> string

  val pp : Format.formatter -> t -> unit
  val pp_quote : Format.formatter -> t -> unit

  module Set : Set.S with type elt = t
  module Map : Map.S with type key = t

  module Top_closure : Top_closure.S with type key := t
end

module Syntax : sig
  type t = OCaml | Reason
end

module File : sig
  type t =
    { name : string
    ; syntax: Syntax.t
    }

  val to_ocaml : t -> t
end

type t =
  { name      : Name.t (** Name of the module. This is always the basename of the filename
                           without the extension. *)
  ; impl      : File.t option
  ; intf      : File.t option

  ; obj_name  : string (** Object name. It is different from [name] for wrapped
                           modules. *)
  }

val name : t -> Name.t

(** Real unit name once wrapped. This is always a valid module name. *)
val real_unit_name : t -> Name.t

val file      : t -> dir:    Path.t -> Ml_kind.t -> Path.t option
val cm_source : t -> dir:    Path.t -> Cm_kind.t -> Path.t option
val cm_file   : t -> obj_dir:Path.t -> Cm_kind.t -> Path.t option
val cmt_file  : t -> obj_dir:Path.t -> Ml_kind.t -> Path.t option

val obj_file : t -> obj_dir:Path.t -> ext:string -> Path.t

(** Same as [cm_file] but doesn't raise if [cm_kind] is [Cmo] or [Cmx]
    and the module has no implementation. *)
val cm_file_unsafe : t -> obj_dir:Path.t -> Cm_kind.t -> Path.t

val odoc_file : t -> doc_dir:Path.t -> Path.t

(** Either the .cmti, or .cmt if the module has no interface *)
val cmti_file : t -> obj_dir:Path.t -> Path.t

val iter : t -> f:(Ml_kind.t -> File.t -> unit) -> unit

val has_impl : t -> bool

(** Set the [obj_name] field of the module. [wrapper] might be a library name. *)
val set_obj_name : t -> wrapper:string option -> t
# 13761 "boot.ml"
end = struct
# 1 "src/module.ml"
open Import

module Name = struct
  type t = string

  let t = Sexp.atom

  let add_suffix = (^)

  let compare = compare
  let of_string = String.capitalize
  let to_string x = x

  let pp = Format.pp_print_string
  let pp_quote fmt x = Format.fprintf fmt "%S" x

  module Set = String_set
  module Map = String_map
  module Top_closure = Top_closure.String
end

module Syntax = struct
  type t = OCaml | Reason
end

module File = struct
  type t =
    { name : string
    ; syntax : Syntax.t
    }

  let to_ocaml t =
    match t.syntax with
    | OCaml -> code_errorf "to_ocaml: can only convert reason Files" ()
    | Reason ->
      { syntax = OCaml
      ; name =
          let base, ext = Filename.split_extension t.name in
          base ^ ".re" ^
          (match Filename.extension t.name with
           | ".re" -> ".ml"
           | ".rei" -> ".mli"
           | _ -> code_errorf "to_ocaml: unrecognized extension %s" ext ())
      }
end

type t =
  { name     : Name.t
  ; impl     : File.t option
  ; intf     : File.t option
  ; obj_name : string
  }

let name t = t.name

let real_unit_name t = Name.of_string (Filename.basename t.obj_name)

let has_impl t = Option.is_some t.impl

let file t ~dir (kind : Ml_kind.t) =
  let file =
    match kind with
    | Impl -> t.impl
    | Intf -> t.intf
  in
  Option.map file ~f:(fun f -> Path.relative dir f.name)

let obj_file t ~obj_dir ~ext = Path.relative obj_dir (t.obj_name ^ ext)

let cm_source t ~dir kind = file t ~dir (Cm_kind.source kind)

let cm_file_unsafe t ~obj_dir kind =
  obj_file t ~obj_dir ~ext:(Cm_kind.ext kind)

let cm_file t ~obj_dir (kind : Cm_kind.t) =
  match kind with
  | (Cmx | Cmo) when not (has_impl t) -> None
  | _ -> Some (cm_file_unsafe t ~obj_dir kind)

let cmt_file t ~obj_dir (kind : Ml_kind.t) =
  match kind with
  | Impl -> Option.map t.impl ~f:(fun _ -> obj_file t ~obj_dir ~ext:".cmt" )
  | Intf -> Option.map t.intf ~f:(fun _ -> obj_file t ~obj_dir ~ext:".cmti")

let odoc_file t ~doc_dir = obj_file t ~obj_dir:doc_dir~ext:".odoc"

let cmti_file t ~obj_dir =
  match t.intf with
  | None   -> obj_file t ~obj_dir ~ext:".cmt"
  | Some _ -> obj_file t ~obj_dir ~ext:".cmti"

let iter t ~f =
  Option.iter t.impl ~f:(f Ml_kind.Impl);
  Option.iter t.intf ~f:(f Ml_kind.Intf)

let set_obj_name t ~wrapper =
  match wrapper with
  | Some s -> { t with obj_name = sprintf "%s__%s" s t.name }
  | None ->
    let fn =
      match t.impl with
      | Some f -> f.name
      | None -> (Option.value_exn t.intf).name
    in
    let obj_name  =
      match String.index fn '.' with
      | None -> fn
      | Some i -> String.sub fn ~pos:0 ~len:i
    in
    { t with obj_name }
# 13874 "boot.ml"
end
module Ordered_set_lang : sig
# 1 "src/ordered_set_lang.mli"
(** [Ordered_set_lang.t] is a sexp-based representation for an ordered list of strings,
    with some set like operations. *)

open Import

type t
val t : t Sexp.Of_sexp.t

(** Return the location of the set. [loc standard] returns [None] *)
val loc : t -> Loc.t option

(** Value parsed from elements in the DSL *)
module type Value = sig
  type t
  type key
  val key : t -> key
end

module type Key = sig
  type t
  val compare : t -> t -> Ordering.t
  module Map : Map.S with type key = t
end

module type S = sig
  (** Evaluate an ordered set. [standard] is the interpretation of [:standard]
      inside the DSL. *)
  type value
  type 'a map

  val eval
    :  t
    -> parse:(loc:Loc.t -> string -> value)
    -> standard:value list
    -> value list

  (** Same as [eval] but the result is unordered *)
  val eval_unordered
    :  t
    -> parse:(loc:Loc.t -> string -> value)
    -> standard:value map
    -> value map
end

module Make(Key : Key)(Value : Value with type key = Key.t)
  : S with type value = Value.t
       and type 'a map = 'a Key.Map.t

val standard : t
val is_standard : t -> bool

module Unexpanded : sig
  type expanded = t
  type t
  val t : t Sexp.Of_sexp.t
  val sexp_of_t : t Sexp.To_sexp.t
  val standard : t

  val field : ?default:t -> string -> t Sexp.Of_sexp.record_parser

  (** List of files needed to expand this set *)
  val files : t -> f:(String_with_vars.t -> string) -> String_set.t

  (** Expand [t] using with the given file contents. [file_contents] is a map from
      filenames to their parsed contents. Every [(:include fn)] in [t] is replaced by
      [Map.find files_contents fn]. Every element is converted to a string using [f]. *)
  val expand
    :  t
    -> files_contents:Sexp.Ast.t String_map.t
    -> f:(String_with_vars.t -> string)
    -> expanded
end with type expanded := t

module String : S with type value = string and type 'a map = 'a String_map.t
# 13952 "boot.ml"
end = struct
# 1 "src/ordered_set_lang.ml"
open! Import

module Ast = struct
  [@@@warning "-37"]
  type expanded = Expanded
  type unexpanded = Unexpanded
  type ('a, _) t =
    | Element : 'a -> ('a, _) t
    | Special : Loc.t * string -> ('a, _) t
    | Union : ('a, 'b) t list -> ('a, 'b) t
    | Diff : ('a, 'b) t * ('a, 'b) t -> ('a, 'b) t
    | Include : String_with_vars.t -> ('a, unexpanded) t
end

type 'ast generic =
  { ast : 'ast
  ; loc : Loc.t option
  }

type ast_expanded = (Loc.t * string, Ast.expanded) Ast.t
type t = ast_expanded generic
let loc t = t.loc

let parse_general sexp ~f =
  let rec of_sexp : Sexp.Ast.t -> _ = function
    | Atom (loc, A "\\") -> Loc.fail loc "unexpected \\"
    | (Atom (_, A "") | Quoted_string (_, _)) as t -> Ast.Element (f t)
    | Atom (loc, A s) as t ->
      if s.[0] = ':' then
        Special (loc, String.sub s ~pos:1 ~len:(String.length s - 1))
      else
        Element (f t)
    | List (_, sexps) -> of_sexps [] sexps
  and of_sexps acc = function
    | Atom (_, A "\\") :: sexps ->
       Diff (Union (List.rev acc), of_sexps [] sexps)
    | elt :: sexps ->
      of_sexps (of_sexp elt :: acc) sexps
    | [] -> Union (List.rev acc)
  in
  of_sexp sexp

let t sexp : t =
  let ast =
    parse_general sexp ~f:(function
      | Atom (loc, A s) | Quoted_string (loc, s) -> (loc, s)
      | List _ -> assert false)
  in
  { ast
  ; loc = Some (Sexp.Ast.loc sexp)
  }

let is_standard t =
  match (t.ast : ast_expanded) with
  | Ast.Special (_, "standard") -> true
  | _ -> false

module type Value = sig
  type t
  type key
  val key : t -> key
end

module type Key = sig
  type t
  val compare : t -> t -> Ordering.t
  module Map : Map.S with type key = t
end

module type S = sig
  type value
  type 'a map

  val eval
    :  t
    -> parse:(loc:Loc.t -> string -> value)
    -> standard:value list
    -> value list

  val eval_unordered
    :  t
    -> parse:(loc:Loc.t -> string -> value)
    -> standard:value map
    -> value map
end

module Make(Key : Key)(Value : Value with type key = Key.t) = struct
  module type Named_values = sig
    type t

    val singleton : Value.t -> t
    val union : t list -> t
    val diff : t -> t -> t
  end

  module Make(M : Named_values) = struct
    let eval t ~parse ~special_values =
      let rec of_ast (t : ast_expanded) =
        let open Ast in
        match t with
        | Element (loc, s) ->
          let x = parse ~loc s in
          M.singleton x
        | Special (loc, name) -> begin
            match String_map.find special_values name with
            | Some x -> x
            | None   -> Loc.fail loc "undefined symbol %s" name
          end
        | Union elts -> M.union (List.map elts ~f:of_ast)
        | Diff (left, right) ->
          let left  = of_ast left  in
          let right = of_ast right in
          M.diff left right
      in
      of_ast t.ast
  end

  module Ordered = Make(struct
      type t = Value.t list

      let singleton x = [x]
      let union = List.flatten
      let diff a b =
        List.filter a ~f:(fun x ->
          List.for_all b ~f:(fun y ->
            Ordering.neq (Key.compare (Value.key x) (Value.key y))))
    end)

  module Unordered = Make(struct
      type t = Value.t Key.Map.t

      let singleton x = Key.Map.singleton (Value.key x) x

      let union l =
        List.fold_left l ~init:Key.Map.empty ~f:(fun acc t ->
          Key.Map.merge acc t ~f:(fun _name x y ->
            match x, y with
            | Some x, _ | _, Some x -> Some x
            | _ -> None))

      let diff a b =
        Key.Map.merge a b ~f:(fun _name x y ->
          match x, y with
          | Some _, None -> x
          | _ -> None)
    end)

  type value = Value.t
  type 'a map = 'a Key.Map.t

  let eval t ~parse ~standard =
    if is_standard t then
      standard (* inline common case *)
    else
      Ordered.eval t ~parse
        ~special_values:(String_map.singleton "standard" standard)

  let eval_unordered t ~parse ~standard =
    if is_standard t then
      standard (* inline common case *)
    else
      Unordered.eval t ~parse
        ~special_values:(String_map.singleton "standard" standard)
end

let standard =
  { ast = Ast.Special (Loc.none, "standard")
  ; loc = None
  }

module Unexpanded = struct
  type ast = (Sexp.Ast.t, Ast.unexpanded) Ast.t
  type t = ast generic
  let t sexp =
    let rec map (t : (Sexp.Ast.t, Ast.expanded) Ast.t) =
      let open Ast in
      match t with
      | Element x -> Element x
      | Union [Special (_, "include"); Element fn] ->
        Include (String_with_vars.t fn)
      | Union [Special (loc, "include"); _]
      | Special (loc, "include") ->
        Loc.fail loc "(:include expects a single element (do you need to quote the filename?)"
      | Special (l, s) -> Special (l, s)
      | Union l ->
        Union (List.map l ~f:map)
      | Diff (l, r) ->
        Diff (map l, map r)
    in
    { ast = map (parse_general sexp ~f:(fun x -> x))
    ; loc = Some (Sexp.Ast.loc sexp)
    }

  let sexp_of_t t =
    let open Ast in
    let rec loop : ast -> Sexp.t = function
      | Element sexp -> Usexp.Ast.remove_locs sexp
      | Special (_, s) -> Sexp.atom (":" ^ s)
      | Union l -> List (List.map l ~f:loop)
      | Diff (a, b) -> List [loop a; Sexp.unsafe_atom_of_string "\\"; loop b]
      | Include fn -> List [ Sexp.unsafe_atom_of_string ":include"
                           ; String_with_vars.sexp_of_t fn
                           ]
    in
    loop t.ast

  let standard = standard

  let field ?(default=standard) name = Sexp.Of_sexp.field name t ~default

  let files t ~f =
    let rec loop acc (t : ast) =
      let open Ast in
      match t with
      | Element _
      | Special _ -> acc
      | Include fn ->
        String_set.add acc (f fn)
      | Union l ->
        List.fold_left l ~init:acc ~f:loop
      | Diff (l, r) ->
        loop (loop acc l) r
    in
    loop String_set.empty t.ast

  let expand t ~files_contents ~f  =
    let rec expand (t : ast) : ast_expanded =
      let open Ast in
      match t with
      | Element s -> Element (Sexp.Ast.loc s, f (String_with_vars.t s))
      | Special (l, s) -> Special (l, s)
      | Include fn ->
        let sexp =
          let fn = f fn in
          match String_map.find files_contents fn with
          | Some x -> x
          | None ->
            Sexp.code_error
              "Ordered_set_lang.Unexpanded.expand"
              [ "included-file", Quoted_string fn
              ; "files", Sexp.To_sexp.(list string)
                           (String_map.keys files_contents)
              ]
        in
        parse_general sexp ~f:(fun sexp ->
          (Sexp.Ast.loc sexp, f (String_with_vars.t sexp)))
      | Union l -> Union (List.map l ~f:expand)
      | Diff (l, r) ->
        Diff (expand l, expand r)
    in
    { t with ast = expand t.ast }
end

module String = Make(struct
    type t = string
    let compare = String.compare
    module Map = String_map
  end)(struct
    type t = string
    type key = string
    let key x = x
  end)
# 14217 "boot.ml"
end
module Per_item : sig
# 1 "src/per_item.mli"
(** Module used to represent the [(per_xxx ...)] forms

    The main different between this module and a plain [Map] is that
    the [map] operation applies transformations only once per distinct
    value.
*)

open Import

module type S = sig
  type key

  type 'a t

  (** Create a mapping where all keys map to the same value *)
  val for_all : 'a -> 'a t

  (** Create a mapping from a list of bindings *)
  val of_mapping
    :  (key list * 'a) list
    -> default:'a
    -> ('a t, key * 'a * 'a) result

  (** Get the configuration for the given item *)
  val get : 'a t -> key -> 'a

  (** Returns [true] if the mapping returns the same value for all
      keys. Note that the mapping might still be constant if
      [is_constant] returns [false]. *)
  val is_constant : _ t -> bool

  val map : 'a t -> f:('a -> 'b) -> 'b t
  val fold : 'a t -> init:'acc -> f:('a -> 'acc -> 'acc) -> 'acc
end

module Make(Key : Comparable.S) : S with type key = Key.t
# 14257 "boot.ml"
end = struct
# 1 "src/per_item.ml"
open Import

module type S = sig
  type key

  type 'a t

  val for_all : 'a -> 'a t
  val of_mapping
    :  (key list * 'a) list
    -> default:'a
    -> ('a t, key * 'a * 'a) result
  val get : 'a t -> key -> 'a
  val is_constant : _ t -> bool
  val map : 'a t -> f:('a -> 'b) -> 'b t
  val fold : 'a t -> init:'acc -> f:('a -> 'acc -> 'acc) -> 'acc
end

module Make(Key : Comparable.S) : S with type key = Key.t = struct
  module Map = Map.Make(Key)

  type key = Key.t

  type 'a t =
    { map    : int Map.t
    ; values : 'a array
    }

  let for_all x =
    { map    = Map.empty
    ; values = [|x|]
    }

  let of_mapping l ~default =
    let values =
      Array.of_list (default :: List.map l ~f:snd)
    in
    List.mapi l ~f:(fun i (keys, _) ->
      List.map keys ~f:(fun key -> (key, i + 1)))
    |> List.concat
    |> Map.of_list
    |> function
    | Ok map ->
      Ok { map; values }
    | Error (key, x, y) ->
      Error (key, values.(x), values.(y))

  let get t key =
    let index =
      match Map.find t.map key with
      | None   -> 0
      | Some i -> i
    in
    t.values.(index)

  let map t ~f =
    { t with values = Array.map t.values ~f }

  let fold t ~init ~f =
    Array.fold_right t.values ~init ~f

  let is_constant t = Array.length t.values = 1
end
# 14323 "boot.ml"
end
module Sub_system_name : sig
# 1 "src/sub_system_name.mli"
include Interned.S
# 14328 "boot.ml"
end = struct
# 1 "src/sub_system_name.ml"
include Interned.Make ()
# 14332 "boot.ml"
end
module Syntax : sig
# 1 "src/syntax.mli"
(** Versioned syntaxes *)

module Version : sig
  (** A syntax version.

      It is always assumed that a parser with version [(X, Y)] can
      read the output produced by a printer at version [(X, Z)] for any
      [Z <= Y]. *)
  type t = int * int

  val sexp_of_t : t Sexp.To_sexp.t
  val t_of_sexp : t Sexp.Of_sexp.t

  (** Whether the parser can read the data or not *)
  val can_read : parser_version:t -> data_version:t -> bool
end

module Versioned_parser : sig
  (** Versioned parser *)
  type 'a t

  (** Create a versionned parser. There must be exactly one parser per
      major version. *)
  val make : (Version.t * 'a) list -> 'a t

  val last : 'a t -> Version.t * 'a

  (** Find a parser that can parse data of this version *)
  val find_exn : 'a t -> loc:Loc.t -> data_version:Version.t -> 'a
end
# 14366 "boot.ml"
end = struct
# 1 "src/syntax.ml"
open Import

module Version = struct
  type t = int * int

  let to_string (a, b) = sprintf "%u.%u" a b

  let sexp_of_t t = Sexp.unsafe_atom_of_string (to_string t)

  let t_of_sexp : t Sexp.Of_sexp.t = function
    | Atom (loc, A s) -> begin
        try
          Scanf.sscanf s "%u.%u" (fun a b -> (a, b))
        with _ ->
          Loc.fail loc "atom of the form NNN.NNN expected"
      end
    | sexp ->
      Sexp.Of_sexp.of_sexp_error sexp "atom expected"

  let can_read ~parser_version:(pa, pb) ~data_version:(da, db) =
    pa = da && db <= pb
end

module Versioned_parser = struct
  type 'a t = (int * 'a) Int_map.t

  let make l =
    if List.is_empty l then
      Sexp.code_error "Syntax.Versioned_parser.make got empty list" [];
    match
      List.map l ~f:(fun ((major, minor), p) -> (major, (minor, p)))
      |> Int_map.of_list
    with
    | Ok x -> x
    | Error _ ->
      Sexp.code_error
        "Syntax.Versioned_parser.make"
        [ "versions", Sexp.To_sexp.list Version.sexp_of_t (List.map l ~f:fst) ]

  let last t =
    let major, (minor, p) = Option.value_exn (Int_map.max_binding t) in
    ((major, minor), p)

  let find_exn t ~loc ~data_version:(major, minor) =
    match
      Option.bind (Int_map.find t major) ~f:(fun (minor', p) ->
        Option.some_if (minor' >= minor) p)
    with
    | None ->
      Loc.fail loc "Version %s is not supported.\n\
                    Supported versions:\n\
                    %s"
        (Version.to_string (major, minor))
        (String.concat ~sep:"\n"
           (Int_map.to_list t |> List.map ~f:(fun (major, (minor, _)) ->
              sprintf "- %u.0 to %u.%u" major major minor)))
    | Some p -> p
end
# 14427 "boot.ml"
end
module Jbuild : sig
# 1 "src/jbuild.mli"
(** Representation and parsing of jbuild files *)

open Import

module Jbuild_version : sig
  type t = V1
  val t : t Sexp.Of_sexp.t

  val latest_stable : t
end

module Scope_info : sig
  module Name : sig
    (* CR-someday diml: change to [private string] and encode [None]
       as [""] *)
    (** [None] is the for the {!anonymous} scope *)
    type t = string option

    val compare : t -> t -> Ordering.t

    val of_string : string -> t
    val to_string : t -> string
  end

  type t =
    { name     : string option (** First package name in alphabetical
                                   order. [None] for the global
                                   scope. *)
    ; packages : Package.t Package.Name.Map.t
    ; root     : Path.t
    }

  val make : Package.t list -> t

  (** The anonymous represent the scope at the root of the workspace
      when the root of the workspace contains no [<package>.opam]
      files. *)
  val anonymous : t

  (** [resolve t package_name] looks up [package_name] in [t] and returns the
      package description if it exists, otherwise it returns an error. *)
  val resolve : t -> Package.Name.t -> (Package.t, string) result
end

(** Ppx preprocessors  *)
module Pp : sig
  type t = private string
  val of_string : string -> t
  val to_string : t -> string
  val compare : t -> t -> Ordering.t
end

module Preprocess : sig
  type pps =
    { pps   : (Loc.t * Pp.t) list
    ; flags : string list
    }

  type t =
    | No_preprocessing
    | Action of Action.Unexpanded.t
    | Pps    of pps
end

module Per_module : Per_item.S with type key = Module.Name.t

module Preprocess_map : sig
  type t = Preprocess.t Per_module.t

  val no_preprocessing : t
  val default : t

  (** [find module_name] find the preprocessing specification for a
      given module *)
  val find : Module.Name.t -> t -> Preprocess.t

  val pps : t -> (Loc.t * Pp.t) list
end

module Lint : sig
  type t = Preprocess_map.t

  val no_lint : t
end


module Js_of_ocaml : sig
  type t =
    { flags            : Ordered_set_lang.Unexpanded.t
    ; javascript_files : string list
    }

  val default : t
end

module Lib_dep : sig
  type choice =
    { required  : String_set.t
    ; forbidden : String_set.t
    ; file      : string
    }

  type select =
    { result_fn : string
    ; choices   : choice list
    ; loc       : Loc.t
    }

  type t =
    | Direct of (Loc.t * string)
    | Select of select

  val to_lib_names : t -> string list
  val direct : Loc.t * string -> t
  val of_pp : Loc.t * Pp.t -> t
end

module Lib_deps : sig
  type t = Lib_dep.t list
  val of_pps : Pp.t list -> t
end

module Dep_conf : sig
  type t =
    | File of String_with_vars.t
    | Alias of String_with_vars.t
    | Alias_rec of String_with_vars.t
    | Glob_files of String_with_vars.t
    | Files_recursively_in of String_with_vars.t
    | Package of String_with_vars.t
    | Universe

  val t : t Sexp.Of_sexp.t
  val sexp_of_t : t -> Sexp.t
end

module Buildable : sig
  type t =
    { loc                      : Loc.t
    ; modules                  : Ordered_set_lang.t
    ; modules_without_implementation : Ordered_set_lang.t
    ; libraries                : Lib_dep.t list
    ; preprocess               : Preprocess_map.t
    ; preprocessor_deps        : Dep_conf.t list
    ; lint                     : Lint.t
    ; flags                    : Ordered_set_lang.Unexpanded.t
    ; ocamlc_flags             : Ordered_set_lang.Unexpanded.t
    ; ocamlopt_flags           : Ordered_set_lang.Unexpanded.t
    ; js_of_ocaml              : Js_of_ocaml.t
    ; allow_overlapping_dependencies : bool
    }

  (** Preprocessing specification used by all modules or [No_preprocessing] *)
  val single_preprocess : t -> Preprocess.t
end

module Public_lib : sig
  type t =
    { name    : string        (** Full public name *)
    ; package : Package.t     (** Package it is part of *)
    ; sub_dir : string option (** Subdirectory inside the installation
                                  directory *)
    }
end

module Sub_system_info : sig
  (** The type of all sub-systems informations. This type is what we
      get just after parsing a [jbuild] file. *)
  type t = ..
  type sub_system = t = ..

  type 'a parser =
    { short : (Loc.t -> 'a) option (** Value when the sub-system has
                                       no argument *)
    ; parse : 'a Sexp.Of_sexp.t    (** Parse the argument *)
    }

  (** What the user must provide in order to define the parsing part
      of a sub-system. *)
  module type S = sig
    type t
    type sub_system += T of t

    (** Name of the sub-system *)
    val name : Sub_system_name.t

    (** Location of the S-expression passed to [of_sexp] or [short]. *)
    val loc : t -> Loc.t

    val parsers : t parser Syntax.Versioned_parser.t
  end

  module Register(M : S) : sig end

  val get : Sub_system_name.t -> (module S)
end

module Mode_conf : sig
  type t =
    | Byte
    | Native
    | Best (** [Native] if available and [Byte] if not *)

  val t : t Sexp.Of_sexp.t
  val compare : t -> t -> Ordering.t

  module Set : sig
    include Set.S with type elt = t
    val t : t Sexp.Of_sexp.t

    (** Both Byte and Native *)
    val default : t

    val eval : t -> has_native:bool -> Mode.Dict.Set.t
  end
end

module Library : sig
  module Kind : sig
    type t =
      | Normal
      | Ppx_deriver
      | Ppx_rewriter
  end

  type t =
    { name                     : string
    ; public                   : Public_lib.t option
    ; synopsis                 : string option
    ; install_c_headers        : string list
    ; ppx_runtime_libraries    : (Loc.t * string) list
    ; modes                    : Mode_conf.Set.t
    ; kind                     : Kind.t
    ; c_flags                  : Ordered_set_lang.Unexpanded.t
    ; c_names                  : string list
    ; cxx_flags                : Ordered_set_lang.Unexpanded.t
    ; cxx_names                : string list
    ; library_flags            : Ordered_set_lang.Unexpanded.t
    ; c_library_flags          : Ordered_set_lang.Unexpanded.t
    ; self_build_stubs_archive : string option
    ; virtual_deps             : (Loc.t * string) list
    ; wrapped                  : bool
    ; optional                 : bool
    ; buildable                : Buildable.t
    ; dynlink                  : bool
    ; scope_name               : Scope_info.Name.t
    ; sub_systems              : Sub_system_info.t Sub_system_name.Map.t
    }

  val has_stubs : t -> bool
  val stubs_archive : t -> dir:Path.t -> ext_lib:string -> Path.t
  val best_name : t -> string
end

module Install_conf : sig
  type file =
    { src : string
    ; dst : string option
    }

  type t =
    { section : Install.Section.t
    ; files   : file list
    ; package : Package.t
    }
end

module Executables : sig
  module Link_mode : sig
    type t =
      { mode : Mode_conf.t
      ; kind : Binary_kind.t
      }

    val exe           : t
    val object_       : t
    val shared_object : t
    val byte          : t
    val native        : t

    val compare : t -> t -> Ordering.t

    module Set : Set.S with type elt = t
  end

  type t =
    { names            : (Loc.t * string) list
    ; link_executables : bool
    ; link_flags       : Ordered_set_lang.Unexpanded.t
    ; modes            : Link_mode.Set.t
    ; buildable        : Buildable.t
    }
end

module Rule : sig
  module Targets : sig
    type t =
      | Static of string list
      | Infer
  end

  module Mode : sig
    type t =
      | Standard
      (** Only use this rule if  the source files don't exist. *)
      | Fallback
      (** Silently promote the targets to the source tree. *)
      | Promote
      (** Same as [Promote] but [jbuilder clean] must delete the file *)
      | Promote_but_delete_on_clean
      (** Same as [Standard] however this is not a rule stanza, so it
          is not possible to add a [(fallback)] field to the rule. *)
      | Not_a_rule_stanza
      (** Just ignore the source files entirely. This is for cases
          where the targets are promoted only in a specific context,
          such as for .install files. *)
      | Ignore_source_files
  end

  type t =
    { targets  : Targets.t
    ; deps     : Dep_conf.t list
    ; action   : Action.Unexpanded.t
    ; mode     : Mode.t
    ; locks    : String_with_vars.t list
    ; loc      : Loc.t
    }
end

module Menhir : sig
  type t =
    { merge_into : string option
    ; flags      : Ordered_set_lang.Unexpanded.t
    ; modules    : string list
    ; mode       : Rule.Mode.t
    ; loc        : Loc.t
    }
end

module Provides : sig
  type t =
    { name : string
    ; file : string
    }
end

module Alias_conf : sig
  type t =
    { name    : string
    ; deps    : Dep_conf.t list
    ; action  : Action.Unexpanded.t option
    ; locks   : String_with_vars.t list
    ; package : Package.t option
    }
end

module Copy_files : sig
  type t =
    { add_line_directive : bool
    ; glob : String_with_vars.t
    }
end

module Documentation : sig
  type t =
    { package     : Package.t
    ; mld_files   : Ordered_set_lang.t
    }
end

module Stanza : sig
  type t =
    | Library     of Library.t
    | Executables of Executables.t
    | Rule        of Rule.t
    | Provides    of Provides.t
    | Install     of Install_conf.t
    | Alias       of Alias_conf.t
    | Copy_files  of Copy_files.t
    | Menhir      of Menhir.t
    | Documentation of Documentation.t
end

module Stanzas : sig
  type t = Stanza.t list

  type syntax = OCaml | Plain

  val parse
    :  ?default_version:Jbuild_version.t
    -> file:Path.t
    -> Scope_info.t
    -> Sexp.Ast.t list
    -> t
  val lib_names : (_ * _ * t) list -> String_set.t
end
# 14827 "boot.ml"
end = struct
# 1 "src/jbuild.ml"
open Import
open Sexp.Of_sexp

(* This file defines the jbuild types as well as the S-expression syntax for the various
   supported version of the specification.

   [vN] is for the version [N] of the specification and [vjs] is for the rolling
   [jane_street] version, when needed.
*)

module Jbuild_version = struct
  type t =
    | V1

  let t =
    enum
      [ "1", V1
      ]

  let latest_stable = V1
end

let invalid_module_name name sexp =
  of_sexp_error sexp (sprintf "invalid module name: %S" name)

let module_name sexp =
  let name = string sexp in
  match name with
  | "" -> invalid_module_name name sexp
  | s ->
    (match s.[0] with
     | 'A'..'Z' | 'a'..'z' -> ()
     | _ -> invalid_module_name name sexp);
    String.iter s ~f:(function
      | 'A'..'Z' | 'a'..'z' | '0'..'9' | '\'' | '_' -> ()
      | _ -> invalid_module_name name sexp);
    String.capitalize s

let module_names sexp = String_set.of_list (list module_name sexp)

let invalid_lib_name sexp =
  of_sexp_error sexp "invalid library name"

let library_name sexp =
  match string sexp with
  | "" -> invalid_lib_name sexp
  | s ->
    if s.[0] = '.' then invalid_lib_name sexp;
    String.iter s ~f:(function
      | 'A'..'Z' | 'a'..'z' | '_' | '.' | '0'..'9' -> ()
      | _ -> invalid_lib_name sexp);
    s

let file sexp =
  match string sexp with
  | "." | ".." ->
    of_sexp_error sexp "'.' and '..' are not valid filenames"
  | fn -> fn

let file_in_current_dir sexp =
  match string sexp with
  | "." | ".." ->
    of_sexp_error sexp "'.' and '..' are not valid filenames"
  | fn ->
    if Filename.dirname fn <> Filename.current_dir_name then
      of_sexp_error sexp "file in current directory expected";
    fn

let relative_file sexp =
  let fn = file sexp in
  if not (Filename.is_relative fn) then
    of_sexp_error sexp "relative filename expected";
  fn

module Scope_info = struct
  module Name = struct
    type t = string option

    let compare : t -> t -> Ordering.t = compare

    let of_string = function
      | "" -> None
      | s  -> Some s

    let to_string = function
      | None -> ""
      | Some "" -> assert false
      | Some s -> s
  end

  type t =
    { name     : Name.t
    ; packages : Package.t Package.Name.Map.t
    ; root     : Path.t
    }

  let anonymous =
    { name     = None
    ; packages = Package.Name.Map.empty
    ; root     = Path.root
    }

  let make = function
    | [] -> anonymous
    | pkg :: rest as pkgs ->
      let name =
        List.fold_left rest ~init:pkg.Package.name ~f:(fun acc pkg ->
          min acc pkg.Package.name)
      in
      let root = pkg.path in
      List.iter rest ~f:(fun pkg -> assert (pkg.Package.path = root));
      { name = Some (Package.Name.to_string name)
      ; packages =
          Package.Name.Map.of_list_exn (List.map pkgs ~f:(fun pkg ->
            pkg.Package.name, pkg))
      ; root
      }

  let package_listing packages =
    let longest_pkg =
      String.longest_map packages ~f:(fun p ->
        Package.Name.to_string p.Package.name)
    in
    String.concat ~sep:"\n"
      (List.map packages ~f:(fun pkg ->
         sprintf "- %-*s (because of %s)" longest_pkg
           (Package.Name.to_string pkg.Package.name)
           (Path.to_string (Package.opam_file pkg))))

  let default t =
    match Package.Name.Map.values t.packages with
    | [pkg] -> Ok pkg
    | [] ->
      Error
        "The current scope defines no packages.\n\
         What do you want me to do with this (install ...) stanzas?.\n\
         You need to add a <package>.opam file at the root \
         of your project so that\n\
         I know that you want to install things as part of package <package>."
    | _ :: _ :: _ ->
      Error
        (sprintf
           "I can't determine automatically which package this (install ...) \
            stanza is for. I have the choice between these ones:\n\
            %s\n\
            You need to add a (package ...) field in this (install ...) stanza"
           (package_listing (Package.Name.Map.values t.packages)))

  let resolve t name =
    match Package.Name.Map.find t.packages name with
    | Some pkg ->
      Ok pkg
    | None ->
      let name_s = Package.Name.to_string name in
      if Package.Name.Map.is_empty t.packages then
        Error (sprintf
                 "You cannot declare items to be installed without \
                  adding a <package>.opam file at the root of your project.\n\
                  To declare elements to be installed as part of package %S, \
                  add a %S file at the root of your project."
                 name_s (Package.Name.opam_fn name))
      else
        Error (sprintf
                 "The current scope doesn't define package %S.\n\
                  The only packages for which you can declare \
                  elements to be installed in this directory are:\n\
                  %s%s"
                 name_s
                 (package_listing (Package.Name.Map.values t.packages))
                 (hint name_s (Package.Name.Map.keys t.packages
                              |> List.map ~f:Package.Name.to_string)))

  let package t sexp =
    match resolve t (Package.Name.of_string (string sexp)) with
    | Ok p -> p
    | Error s -> Loc.fail (Sexp.Ast.loc sexp) "%s" s

  let package_field t =
    map_validate (field_o "package" string) ~f:(function
      | None -> default t
      | Some name -> resolve t (Package.Name.of_string name))
end


module Pp : sig
  type t = private string
  val of_string : string -> t
  val to_string : t -> string
  val compare : t -> t -> Ordering.t
end = struct
  type t = string

  let of_string s =
    assert (not (String.is_prefix s ~prefix:"-"));
    s

  let to_string t = t

  let compare = String.compare
end

module Pp_or_flags = struct
  type t =
    | PP of Loc.t * Pp.t
    | Flags of string list

  let of_string ~loc s =
    if String.is_prefix s ~prefix:"-" then
      Flags [s]
    else
      PP (loc, Pp.of_string s)

  let t = function
    | Atom (loc, A s) | Quoted_string (loc, s) -> of_string ~loc s
    | List (_, l) -> Flags (List.map l ~f:string)

  let split l =
    let pps, flags =
      List.partition_map l ~f:(function
        | PP (loc, pp) -> Left (loc, pp)
        | Flags s      -> Right s)
    in
    (pps, List.concat flags)
end

module Dep_conf = struct
  type t =
    | File of String_with_vars.t
    | Alias of String_with_vars.t
    | Alias_rec of String_with_vars.t
    | Glob_files of String_with_vars.t
    | Files_recursively_in of String_with_vars.t
    | Package of String_with_vars.t
    | Universe

  let t =
    let t =
      let cstr_sw name f =
        cstr name (String_with_vars.t @> nil) f
      in
      sum
        [ cstr_sw "file"                 (fun x -> File x)
        ; cstr_sw "alias"                (fun x -> Alias x)
        ; cstr_sw "alias_rec"            (fun x -> Alias_rec x)
        ; cstr_sw "glob_files"           (fun x -> Glob_files x)
        ; cstr_sw "files_recursively_in" (fun x -> Files_recursively_in x)
        ; cstr_sw "package"              (fun x -> Package x)
        ; cstr    "universe" nil         Universe
        ]
    in
    fun sexp ->
      match sexp with
      | Atom _ | Quoted_string _ -> File (String_with_vars.t sexp)
      | List _ -> t sexp

  open Sexp
  let sexp_of_t = function
    | File t ->
       List [Sexp.unsafe_atom_of_string "file" ; String_with_vars.sexp_of_t t]
    | Alias t ->
       List [Sexp.unsafe_atom_of_string "alias" ; String_with_vars.sexp_of_t t]
    | Alias_rec t ->
       List [Sexp.unsafe_atom_of_string "alias_rec" ;
             String_with_vars.sexp_of_t t]
    | Glob_files t ->
       List [Sexp.unsafe_atom_of_string "glob_files" ;
             String_with_vars.sexp_of_t t]
    | Files_recursively_in t ->
       List [Sexp.unsafe_atom_of_string "files_recursively_in" ;
             String_with_vars.sexp_of_t t]
    | Package t ->
      List [Sexp.unsafe_atom_of_string "package" ;
            String_with_vars.sexp_of_t t]
    | Universe ->
      Sexp.unsafe_atom_of_string "universe"
end

module Preprocess = struct
  type pps = { pps : (Loc.t * Pp.t) list; flags : string list }
  type t =
    | No_preprocessing
    | Action of Action.Unexpanded.t
    | Pps    of pps

  let t =
    sum
      [ cstr "no_preprocessing" nil No_preprocessing
      ; cstr "action" (Action.Unexpanded.t @> nil) (fun x -> Action x)
      ; cstr "pps" (list Pp_or_flags.t @> nil) (fun l ->
          let pps, flags = Pp_or_flags.split l in
          Pps { pps; flags })
      ]

  let pps = function
    | Pps { pps; _ } -> pps
    | _ -> []
end

module Per_module = struct
  include Per_item.Make(Module.Name)

  let t ~default a sexp =
    match sexp with
    | List (_, Atom (_, A "per_module") :: rest) -> begin
      List.map rest ~f:(fun sexp ->
        let pp, names = pair a module_names sexp in
        (List.map ~f:Module.Name.of_string (String_set.to_list names), pp))
      |> of_mapping ~default
      |> function
      | Ok t -> t
      | Error (name, _, _) ->
        of_sexp_error sexp (sprintf "module %s present in two different sets"
                              (Module.Name.to_string name))
    end
    | sexp -> for_all (a sexp)
end

module Preprocess_map = struct
  type t = Preprocess.t Per_module.t
  let t = Per_module.t Preprocess.t ~default:Preprocess.No_preprocessing

  let no_preprocessing = Per_module.for_all Preprocess.No_preprocessing

  let find module_name t = Per_module.get t module_name

  let default = Per_module.for_all Preprocess.No_preprocessing

  module Pp_map = Map.Make(Pp)

  let pps t =
    Per_module.fold t ~init:Pp_map.empty ~f:(fun pp acc ->
      List.fold_left (Preprocess.pps pp) ~init:acc ~f:(fun acc (loc, pp) ->
        Pp_map.add acc pp loc))
    |> Pp_map.foldi ~init:[] ~f:(fun pp loc acc -> (loc, pp) :: acc)
end

module Lint = struct
  type t = Preprocess_map.t

  let t = Preprocess_map.t

  let default = Preprocess_map.default
  let no_lint = default
end

let field_oslu name =
  field name Ordered_set_lang.Unexpanded.t
    ~default:Ordered_set_lang.Unexpanded.standard

module Js_of_ocaml = struct

  type t =
    { flags            : Ordered_set_lang.Unexpanded.t
    ; javascript_files : string list
    }

  let t =
    record
      (field_oslu "flags"                                     >>= fun flags ->
       field     "javascript_files" (list string) ~default:[] >>= fun javascript_files ->
       return { flags; javascript_files })

  let default =
    { flags = Ordered_set_lang.Unexpanded.standard
    ; javascript_files = [] }
end

module Lib_dep = struct
  type choice =
    { required  : String_set.t
    ; forbidden : String_set.t
    ; file      : string
    }

  type select =
    { result_fn : string
    ; choices   : choice list
    ; loc       : Loc.t (* For error messages *)
    }

  type t =
    | Direct of (Loc.t * string)
    | Select of select

  let choice = function
    | List (_, l) as sexp ->
      let rec loop required forbidden = function
        | [Atom (_, A "->"); fsexp] ->
          let common = String_set.inter required forbidden in
          Option.iter (String_set.choose common) ~f:(fun name ->
            of_sexp_errorf sexp
              "library %S is both required and forbidden in this clause"
              name);
          { required
          ; forbidden
          ; file = file fsexp
          }
        | Atom (_, A "->") :: _
        | List _ :: _ | [] ->
          of_sexp_error sexp "(<[!]libraries>... -> <file>) expected"
        | (Atom (_, A s) | Quoted_string (_, s)) :: l ->
          let len = String.length s in
          if len > 0 && s.[0] = '!' then
            let s = String.sub s ~pos:1 ~len:(len - 1) in
            loop required (String_set.add forbidden s) l
          else
            loop (String_set.add required s) forbidden l
      in
      loop String_set.empty String_set.empty l
    | sexp -> of_sexp_error sexp "(<library-name> <code>) expected"

  let t = function
    | Atom (loc, A s) | Quoted_string (loc, s) ->
      Direct (loc, s)
    | List (loc, Atom (_, A "select") :: m :: Atom (_, A "from") :: libs) ->
      Select { result_fn = file m
             ; choices   = List.map libs ~f:choice
             ; loc
             }
    | sexp ->
      of_sexp_error sexp "<library> or (select <module> from <libraries...>) expected"

  let to_lib_names = function
    | Direct (_, s) -> [s]
    | Select s ->
      List.fold_left s.choices ~init:String_set.empty ~f:(fun acc x ->
        String_set.union acc (String_set.union x.required x.forbidden))
      |> String_set.to_list

  let direct x = Direct x

  let of_pp (loc, pp) = Direct (loc, Pp.to_string pp)
end

module Lib_deps = struct
  type t = Lib_dep.t list

  type kind =
    | Required
    | Optional
    | Forbidden

  let t sexp =
    let t = list Lib_dep.t sexp in
    let add kind name acc =
      match String_map.find acc name with
      | None -> String_map.add acc name kind
      | Some kind' ->
        match kind, kind' with
        | Required, Required ->
          of_sexp_errorf sexp "library %S is present twice" name
        | (Optional|Forbidden), (Optional|Forbidden) ->
          acc
        | Optional, Required | Required, Optional ->
          of_sexp_errorf sexp
            "library %S is present both as an optional and required dependency"
            name
        | Forbidden, Required | Required, Forbidden ->
          of_sexp_errorf sexp
            "library %S is present both as a forbidden and required dependency"
            name
    in
    ignore (
      List.fold_left t ~init:String_map.empty ~f:(fun acc x ->
        match x with
        | Lib_dep.Direct (_, s) -> add Required s acc
        | Select { choices; _ } ->
          List.fold_left choices ~init:acc ~f:(fun acc c ->
            let acc = String_set.fold c.Lib_dep.required ~init:acc ~f:(add Optional) in
            String_set.fold c.forbidden ~init:acc ~f:(add Forbidden)))
      : kind String_map.t);
    t

  let of_pps pps =
    List.map pps ~f:(fun pp -> Lib_dep.of_pp (Loc.none, pp))
end

module Buildable = struct
  type t =
    { loc                      : Loc.t
    ; modules                  : Ordered_set_lang.t
    ; modules_without_implementation : Ordered_set_lang.t
    ; libraries                : Lib_dep.t list
    ; preprocess               : Preprocess_map.t
    ; preprocessor_deps        : Dep_conf.t list
    ; lint                     : Preprocess_map.t
    ; flags                    : Ordered_set_lang.Unexpanded.t
    ; ocamlc_flags             : Ordered_set_lang.Unexpanded.t
    ; ocamlopt_flags           : Ordered_set_lang.Unexpanded.t
    ; js_of_ocaml              : Js_of_ocaml.t
    ; allow_overlapping_dependencies : bool
    }

  let modules_field name =
    field name Ordered_set_lang.t ~default:Ordered_set_lang.standard

  let v1 =
    record_loc >>= fun loc ->
    field "preprocess" Preprocess_map.t ~default:Preprocess_map.default
    >>= fun preprocess ->
    field "preprocessor_deps" (list Dep_conf.t) ~default:[]
    >>= fun preprocessor_deps ->
    field "lint" Lint.t ~default:Lint.default
    >>= fun lint ->
    modules_field "modules"
    >>= fun modules ->
    modules_field "modules_without_implementation"
    >>= fun modules_without_implementation ->
    field "libraries" Lib_deps.t ~default:[]
    >>= fun libraries ->
    field_oslu "flags"          >>= fun flags          ->
    field_oslu "ocamlc_flags"   >>= fun ocamlc_flags   ->
    field_oslu "ocamlopt_flags" >>= fun ocamlopt_flags ->
    field "js_of_ocaml" (Js_of_ocaml.t) ~default:Js_of_ocaml.default
    >>= fun js_of_ocaml ->
    field_b "allow_overlapping_dependencies"
    >>= fun allow_overlapping_dependencies ->
    return
      { loc
      ; preprocess
      ; preprocessor_deps
      ; lint
      ; modules
      ; modules_without_implementation
      ; libraries
      ; flags
      ; ocamlc_flags
      ; ocamlopt_flags
      ; js_of_ocaml
      ; allow_overlapping_dependencies
      }

  let single_preprocess t =
    if Per_module.is_constant t.preprocess then
      Per_module.get t.preprocess (Module.Name.of_string "")
    else
      Preprocess.No_preprocessing
end

module Public_lib = struct
  type t =
    { name    : string
    ; package : Package.t
    ; sub_dir : string option
    }

  let public_name_field pkgs =
    map_validate (field_o "public_name" string) ~f:(function
      | None -> Ok None
      | Some s ->
        match String.split s ~on:'.' with
        | [] -> assert false
        | pkg :: rest ->
          match Scope_info.resolve pkgs (Package.Name.of_string pkg) with
          | Ok pkg ->
            Ok (Some
                  { package = pkg
                  ; sub_dir = if rest = [] then None else Some (String.concat rest ~sep:"/")
                  ; name    = s
                  })
          | Error _ as e -> e)
end

module Sub_system_info = struct
  type t = ..
  type sub_system = t = ..

  type 'a parser =
    { short : (Loc.t -> 'a) option
    ; parse : 'a Sexp.Of_sexp.t
    }

  module type S = sig
    type t
    type sub_system += T of t
    val name    : Sub_system_name.t
    val loc     : t -> Loc.t
    val parsers : t parser Syntax.Versioned_parser.t
  end

  let all = Sub_system_name.Table.create ~default_value:None

  (* For parsing config files in the workspace *)
  let record_parser = ref return

  module Register(M : S) : sig end = struct
    open M

    let { short; parse } = snd (Syntax.Versioned_parser.last M.parsers)

    let short =
      match short with
      | None -> Short_syntax.Not_allowed
      | Some f -> Located f

    let () =
      match Sub_system_name.Table.get all name with
      | Some _ ->
        Sexp.code_error "Sub_system_info.register: already registered"
          [ "name", Sexp.To_sexp.string (Sub_system_name.to_string name) ];
      | None ->
        Sub_system_name.Table.set all ~key:name ~data:(Some (module M : S));
        let p = !record_parser in
        let name_s = Sub_system_name.to_string name in
        record_parser := (fun acc ->
          field_o name_s ~short parse >>= function
          | None   -> p acc
          | Some x ->
            let acc = Sub_system_name.Map.add acc name (T x) in
            p acc)
  end

  let record_parser () = !record_parser Sub_system_name.Map.empty

  let get name = Option.value_exn (Sub_system_name.Table.get all name)
end

module Mode_conf = struct
  module T = struct
    type t =
      | Byte
      | Native
      | Best
    let compare (a : t) b = compare a b
  end
  include T

  let t =
    enum
      [ "byte"  , Byte
      ; "native", Native
      ; "best"  , Best
      ]

  module Set = struct
    include Set.Make(T)

    let t sexp = of_list (list t sexp)

    let default = of_list [Byte; Best]

    let eval t ~has_native =
      let best : Mode.t =
        if has_native then
          Native
        else
          Byte
      in
      let has_best = mem t Best in
      let byte = mem t Byte || (has_best && best = Byte) in
      let native = best = Native && (mem t Native || has_best) in
      { Mode.Dict.byte; native }
  end
end

module Library = struct
  module Kind = struct
    type t =
      | Normal
      | Ppx_deriver
      | Ppx_rewriter

    let t =
      enum
        [ "normal"       , Normal
        ; "ppx_deriver"  , Ppx_deriver
        ; "ppx_rewriter" , Ppx_rewriter
        ]
  end

  type t =
    { name                     : string
    ; public                   : Public_lib.t option
    ; synopsis                 : string option
    ; install_c_headers        : string list
    ; ppx_runtime_libraries    : (Loc.t * string) list
    ; modes                    : Mode_conf.Set.t
    ; kind                     : Kind.t
    ; c_flags                  : Ordered_set_lang.Unexpanded.t
    ; c_names                  : string list
    ; cxx_flags                : Ordered_set_lang.Unexpanded.t
    ; cxx_names                : string list
    ; library_flags            : Ordered_set_lang.Unexpanded.t
    ; c_library_flags          : Ordered_set_lang.Unexpanded.t
    ; self_build_stubs_archive : string option
    ; virtual_deps             : (Loc.t * string) list
    ; wrapped                  : bool
    ; optional                 : bool
    ; buildable                : Buildable.t
    ; dynlink                  : bool
    ; scope_name               : Scope_info.Name.t
    ; sub_systems              : Sub_system_info.t Sub_system_name.Map.t
    }

  let v1 pkgs =
    record
      (Buildable.v1 >>= fun buildable ->
       field      "name" library_name                                      >>= fun name                     ->
       Public_lib.public_name_field pkgs                                   >>= fun public                   ->
       field_o    "synopsis" string                                        >>= fun synopsis                 ->
       field      "install_c_headers" (list string) ~default:[]            >>= fun install_c_headers        ->
       field      "ppx_runtime_libraries" (list (located string)) ~default:[] >>= fun ppx_runtime_libraries    ->
       field_oslu "c_flags"                                                >>= fun c_flags                  ->
       field_oslu "cxx_flags"                                              >>= fun cxx_flags                ->
       field      "c_names" (list string) ~default:[]                      >>= fun c_names                  ->
       field      "cxx_names" (list string) ~default:[]                    >>= fun cxx_names                ->
       field_oslu "library_flags"                                          >>= fun library_flags            ->
       field_oslu "c_library_flags"                                        >>= fun c_library_flags          ->
       field      "virtual_deps" (list (located string)) ~default:[]       >>= fun virtual_deps             ->
       field      "modes" Mode_conf.Set.t ~default:Mode_conf.Set.default   >>= fun modes                    ->
       field      "kind" Kind.t ~default:Kind.Normal                       >>= fun kind                     ->
       field      "wrapped" bool ~default:true                             >>= fun wrapped                  ->
       field_b    "optional"                                               >>= fun optional                 ->
       field      "self_build_stubs_archive" (option string) ~default:None >>= fun self_build_stubs_archive ->
       field_b    "no_dynlink"                                             >>= fun no_dynlink               ->
       Sub_system_info.record_parser () >>= fun sub_systems ->
       field "ppx.driver" ignore ~default:() >>= fun () ->
       return
         { name
         ; public
         ; synopsis
         ; install_c_headers
         ; ppx_runtime_libraries
         ; modes
         ; kind
         ; c_names
         ; c_flags
         ; cxx_names
         ; cxx_flags
         ; library_flags
         ; c_library_flags
         ; self_build_stubs_archive
         ; virtual_deps
         ; wrapped
         ; optional
         ; buildable
         ; dynlink = not no_dynlink
         ; scope_name = pkgs.name
         ; sub_systems
         })

  let has_stubs t =
    match t.c_names, t.cxx_names, t.self_build_stubs_archive with
    | [], [], None -> false
    | _            -> true

  let stubs_archive t ~dir ~ext_lib =
    Path.relative dir (sprintf "lib%s_stubs%s" t.name ext_lib)

  let best_name t =
    match t.public with
    | None -> t.name
    | Some p -> p.name
end

module Install_conf = struct
  type file =
    { src : string
    ; dst : string option
    }

  let file sexp =
    match sexp with
    | Atom (_, A src) -> { src; dst = None }
    | List (_, [Atom (_, A src); Atom (_, A "as"); Atom (_, A dst)]) ->
      { src; dst = Some dst }
    | _ ->
      of_sexp_error sexp
        "invalid format, <name> or (<name> as <install-as>) expected"

  type t =
    { section : Install.Section.t
    ; files   : file list
    ; package : Package.t
    }

  let v1 pkgs =
    record
      (field   "section" Install.Section.t >>= fun section ->
       field   "files"   (list file)       >>= fun files ->
       Scope_info.package_field pkgs             >>= fun package ->
       return
         { section
         ; files
         ; package
         })
end

module Executables = struct

  module Link_mode = struct
    module T = struct
      type t =
        { mode : Mode_conf.t
        ; kind : Binary_kind.t
        }

      let compare a b =
        match compare a.mode b.mode with
        | Eq -> compare a.kind b.kind
        | ne -> ne
    end
    include T

    let make mode kind =
      { mode
      ; kind
      }

    let exe           = make Best Exe
    let object_       = make Best Object
    let shared_object = make Best Shared_object

    let byte_exe           = make Byte Exe

    let native_exe           = make Native Exe
    let native_object        = make Native Object
    let native_shared_object = make Native Shared_object

    let byte   = byte_exe
    let native = native_exe

    let simple =
      let open Sexp.Of_sexp in
      enum
        [ "exe"           , exe
        ; "object"        , object_
        ; "shared_object" , shared_object
        ; "byte"          , byte
        ; "native"        , native
        ]

    let t sexp =
      match sexp with
      | List _ ->
        let mode, kind = pair Mode_conf.t Binary_kind.t sexp in
        { mode; kind }
      | _ -> simple sexp

    module Set = struct
      include Set.Make(T)

      let t sexp : t =
        match list t sexp with
        | [] -> of_sexp_error sexp "No linking mode defined"
        | l ->
          let t = of_list l in
          if (mem t native_exe           && mem t exe          ) ||
             (mem t native_object        && mem t object_      ) ||
             (mem t native_shared_object && mem t shared_object) then
            of_sexp_error sexp
              "It is not allowed use both native and best \
               for the same binary kind."
          else
            t

      let default =
        of_list
          [ byte
          ; exe
          ]

      let best_install_mode t =
        if mem t exe then
          Some exe
        else if mem t native then
          Some native
        else if mem t byte then
          Some byte
        else
          None
    end
  end

  type t =
    { names            : (Loc.t * string) list
    ; link_executables : bool
    ; link_flags       : Ordered_set_lang.Unexpanded.t
    ; modes            : Link_mode.Set.t
    ; buildable        : Buildable.t
    }

  let common_v1 pkgs names public_names ~multi =
    Buildable.v1 >>= fun buildable ->
    field      "link_executables"   bool ~default:true >>= fun link_executables ->
    field_oslu "link_flags"                            >>= fun link_flags ->
    field "modes" Link_mode.Set.t ~default:Link_mode.Set.default
    >>= fun modes ->
    let t =
      { names
      ; link_executables
      ; link_flags
      ; modes
      ; buildable
      }
    in
    let to_install =
      match Link_mode.Set.best_install_mode t.modes with
      | None -> []
      | Some mode ->
        let ext =
          match mode.mode with
          | Native | Best -> ".exe"
          | Byte -> ".bc"
        in
        List.map2 names public_names
          ~f:(fun (_, name) pub ->
            match pub with
            | None -> None
            | Some pub -> Some ({ Install_conf.
                                  src = name ^ ext
                                ; dst = Some pub
                                }))
        |> List.filter_map ~f:(fun x -> x)
    in
    match to_install with
    | [] ->
      (field_o "package" Sexp.Ast.loc >>= function
       | None -> return (t, None)
       | Some loc ->
         Loc.warn loc
           "This field is useless without a (public_name%s ...) field."
           (if multi then "s" else "");
         return (t, None))
    | files ->
      Scope_info.package_field pkgs >>= fun package ->
      return (t, Some { Install_conf. section = Bin; files; package })

  let public_name sexp =
    match string sexp with
    | "-" -> None
    | s   -> Some s

  let v1_multi pkgs =
    record
      (field "names" (list (located string)) >>= fun names ->
       map_validate (field_o "public_names" (list public_name)) ~f:(function
         | None -> Ok (List.map names ~f:(fun _ -> None))
         | Some public_names ->
           if List.length public_names = List.length names then
             Ok public_names
           else
             Error "The list of public names must be of the same \
                    length as the list of names")
       >>= fun public_names ->
       common_v1 pkgs names public_names ~multi:true)

  let v1_single pkgs =
    record
      (field   "name" (located string) >>= fun name ->
       field_o "public_name" string >>= fun public_name ->
       common_v1 pkgs [name] [public_name] ~multi:false)
end

module Rule = struct
  module Targets = struct
    type t =
      | Static of string list (* List of files in the current directory *)
      | Infer
  end


  module Mode = struct
    type t =
      | Standard
      | Fallback
      | Promote
      | Promote_but_delete_on_clean
      | Not_a_rule_stanza
      | Ignore_source_files

    let t =
      enum
        [ "standard"           , Standard
        ; "fallback"           , Fallback
        ; "promote"            , Promote
        ; "promote-until-clean", Promote_but_delete_on_clean
        ]

    let field = field "mode" t ~default:Standard
  end

  type t =
    { targets  : Targets.t
    ; deps     : Dep_conf.t list
    ; action   : Action.Unexpanded.t
    ; mode     : Mode.t
    ; locks    : String_with_vars.t list
    ; loc      : Loc.t
    }

  let v1 sexp =
    let loc = Sexp.Ast.loc sexp in
    match sexp with
    | List (_, (Atom _ :: _)) ->
      { targets  = Infer
      ; deps     = []
      ; action   = Action.Unexpanded.t sexp
      ; mode     = Standard
      ; locks    = []
      ; loc      = loc
      }
    | _ ->
      record
        (field "targets" (list file_in_current_dir)    >>= fun targets ->
         field "deps"    (list Dep_conf.t) ~default:[] >>= fun deps ->
         field "action"  Action.Unexpanded.t           >>= fun action ->
         field "locks"   (list String_with_vars.t) ~default:[] >>= fun locks ->
         map_validate
           (field_b "fallback" >>= fun fallback ->
            field_o "mode" Mode.t >>= fun mode ->
            return (fallback, mode))
           ~f:(function
             | true, Some _ ->
               Error "Cannot use both (fallback) and (mode ...) at the \
                      same time.\n\
                      (fallback) is the same as (mode fallback), \
                      please use the latter in new code."
             | false, Some mode -> Ok mode
             | true, None -> Ok Fallback
             | false, None -> Ok Standard)
         >>= fun mode ->
         return { targets = Static targets
                ; deps
                ; action
                ; mode
                ; locks
                ; loc
                })
        sexp

  type lex_or_yacc =
    { modules : string list
    ; mode    : Mode.t
    }

  let ocamllex_v1 sexp =
    match sexp with
    | List (_, List (_, _) :: _) ->
      record
        (field "modules" (list string) >>= fun modules ->
         Mode.field >>= fun mode ->
         return { modules; mode })
        sexp
    | _ ->
      { modules = list string sexp
      ; mode    = Standard
      }

  let ocamlyacc_v1 = ocamllex_v1

  let ocamllex_to_rule loc { modules; mode } =
    let module S = String_with_vars in
    List.map modules ~f:(fun name ->
      let src = name ^ ".mll" in
      let dst = name ^ ".ml"  in
      { targets = Static [dst]
      ; deps    = [File (S.virt_text __POS__ src)]
      ; action  =
          Chdir
            (S.virt_var __POS__ "ROOT",
             Run (S.virt_text __POS__ "ocamllex",
                  [ S.virt_text __POS__ "-q"
                  ; S.virt_text __POS__ "-o"
                  ; S.virt_var __POS__ "@"
                  ; S.virt_var __POS__"<"
                  ]))
      ; mode
      ; locks = []
      ; loc
      })

  let ocamlyacc_to_rule loc { modules; mode } =
    let module S = String_with_vars in
    List.map modules ~f:(fun name ->
      let src = name ^ ".mly" in
      { targets = Static [name ^ ".ml"; name ^ ".mli"]
      ; deps    = [File (S.virt_text __POS__ src)]
      ; action  =
          Chdir
            (S.virt_var __POS__ "ROOT",
             Run (S.virt_text __POS__ "ocamlyacc",
                  [S.virt_var __POS__ "<"]))
      ; mode
      ; locks = []
      ; loc
      })
end

module Menhir = struct
  type t =
    { merge_into : string option
    ; flags      : Ordered_set_lang.Unexpanded.t
    ; modules    : string list
    ; mode       : Rule.Mode.t
    ; loc        :  Loc.t
    }

  let v1 =
    record
      (field_o "merge_into" string >>= fun merge_into ->
       field_oslu "flags" >>= fun flags ->
       field "modules" (list string) >>= fun modules ->
       Rule.Mode.field >>= fun mode ->
       return
         { merge_into
         ; flags
         ; modules
         ; mode
         ; loc = Loc.none
         }
      )
end

module Provides = struct
  type t =
    { name : string
    ; file : string
    }

(*  let v1 sexp =
    match sexp with
    | Atom (_, s) ->
      { name = s
      ; file =
          match String.lsplit2 s ~on:':' with
          | None        -> s
          | Some (_, s) -> s
      }
    | List (_, [Atom (_, s); List (_, [Atom (_, "file"); Atom (_, file)])]) ->
      { name = s
      ; file
      }
    | sexp ->
    of_sexp_error sexp "[<name>] or [<name> (file <file>)] expected"*)
end

module Alias_conf = struct
  type t =
    { name    : string
    ; deps    : Dep_conf.t list
    ; action  : Action.Unexpanded.t option
    ; locks   : String_with_vars.t list
    ; package : Package.t option
    }

  let v1 pkgs =
    record
      (field "name" string                              >>= fun name ->
       field "deps" (list Dep_conf.t) ~default:[]       >>= fun deps ->
       field_o "package" (Scope_info.package pkgs)      >>= fun package ->
       field_o "action" Action.Unexpanded.t  >>= fun action ->
       field "locks" (list String_with_vars.t) ~default:[] >>= fun locks ->
       return
         { name
         ; deps
         ; action
         ; package
         ; locks
         })
end

module Copy_files = struct
  type t = { add_line_directive : bool
           ; glob : String_with_vars.t
           }

  let v1 = String_with_vars.t
end

module Documentation = struct
  type t =
    { package: Package.t
    ; mld_files: Ordered_set_lang.t
    }

  let v1 pkgs =
    record
      (Scope_info.package_field pkgs >>= fun package ->
       field "mld_files" Ordered_set_lang.t ~default:Ordered_set_lang.standard
       >>= fun mld_files ->
       return
         { package
         ; mld_files
         }
      )
end

module Stanza = struct
  type t =
    | Library     of Library.t
    | Executables of Executables.t
    | Rule        of Rule.t
    | Provides    of Provides.t
    | Install     of Install_conf.t
    | Alias       of Alias_conf.t
    | Copy_files  of Copy_files.t
    | Menhir      of Menhir.t
    | Documentation of Documentation.t
end

module Stanzas = struct
  type t = Stanza.t list

  type syntax = OCaml | Plain

  open Stanza

  let rules l = List.map l ~f:(fun x -> Rule x)

  let execs (exe, install) =
    match install with
    | None -> [Executables exe]
    | Some i -> [Executables exe; Install i]

  exception Include_loop of Path.t * (Loc.t * Path.t) list

  let rec v1 pkgs ~file ~include_stack : Stanza.t list Sexp.Of_sexp.t =
    sum
      [ cstr "library"     (Library.v1 pkgs @> nil) (fun x -> [Library x])
      ; cstr "executable"  (Executables.v1_single pkgs @> nil) execs
      ; cstr "executables" (Executables.v1_multi  pkgs @> nil) execs
      ; cstr_loc "rule"      (Rule.v1     @> nil) (fun loc x -> [Rule { x with loc }])
      ; cstr_loc "ocamllex" (Rule.ocamllex_v1 @> nil)
          (fun loc x -> rules (Rule.ocamllex_to_rule loc x))
      ; cstr_loc "ocamlyacc" (Rule.ocamlyacc_v1 @> nil)
          (fun loc x -> rules (Rule.ocamlyacc_to_rule loc x))
      ; cstr_loc "menhir" (Menhir.v1 @> nil)
          (fun loc x -> [Menhir { x with loc }])
      ; cstr "install"     (Install_conf.v1 pkgs @> nil) (fun x -> [Install     x])
      ; cstr "alias"       (Alias_conf.v1 pkgs @> nil)   (fun x -> [Alias       x])
      ; cstr "copy_files" (Copy_files.v1 @> nil)
          (fun glob -> [Copy_files {add_line_directive = false; glob}])
      ; cstr "copy_files#" (Copy_files.v1 @> nil)
          (fun glob -> [Copy_files {add_line_directive = true; glob}])
      (* Just for validation and error messages *)
      ; cstr "jbuild_version" (Jbuild_version.t @> nil) (fun _ -> [])
      ; cstr_loc "include" (relative_file @> nil) (fun loc fn ->
          let include_stack = (loc, file) :: include_stack in
          let dir = Path.parent file in
          let file = Path.relative dir fn in
          if not (Path.exists file) then
            Loc.fail loc "File %s doesn't exist."
              (Path.to_string_maybe_quoted file);
          if List.exists include_stack ~f:(fun (_, f) -> f = file) then
            raise (Include_loop (file, include_stack));
          let sexps = Sexp.load ~fname:(Path.to_string file) ~mode:Many in
          parse pkgs sexps ~default_version:Jbuild_version.V1 ~file ~include_stack)
      ; cstr "documentation" (Documentation.v1 pkgs @> nil)
          (fun d -> [Documentation d])
      ]

  and select
    :  Jbuild_version.t
    -> Scope_info.t
    -> file:Path.t
    -> include_stack:(Loc.t * Path.t) list
    -> Stanza.t list Sexp.Of_sexp.t = function
    | V1  -> v1

  and parse ~default_version ~file ~include_stack pkgs sexps =
    let versions, sexps =
      List.partition_map sexps ~f:(function
        | List (loc, [Atom (_, A "jbuild_version"); ver]) ->
          Left (Jbuild_version.t ver, loc)
        | sexp -> Right sexp)
    in
    let version =
      match versions with
      | [] -> default_version
      | [(v, _)] -> v
      | _ :: (_, loc) :: _ ->
        Loc.fail loc "jbuild_version specified too many times"
    in
    List.concat_map sexps ~f:(select version pkgs ~file ~include_stack)

  let parse ?(default_version=Jbuild_version.latest_stable) ~file pkgs sexps =
    try
      parse pkgs sexps ~default_version ~include_stack:[] ~file
    with
    | Include_loop (_, []) -> assert false
    | Include_loop (file, last :: rest) ->
      let loc = fst (Option.value (List.last rest) ~default:last) in
      let line_loc (loc, file) =
        sprintf "%s:%d"
          (Path.to_string_maybe_quoted file)
          loc.Loc.start.pos_lnum
      in
      Loc.fail loc
        "Recursive inclusion of jbuild files detected:\n\
         File %s is included from %s%s"
        (Path.to_string_maybe_quoted file)
        (line_loc last)
        (String.concat ~sep:""
           (List.map rest ~f:(fun x ->
              sprintf
                "\n--> included from %s"
                (line_loc x))))

  let lib_names ts =
    List.fold_left ts ~init:String_set.empty ~f:(fun acc (_, _, stanzas) ->
      List.fold_left stanzas ~init:acc ~f:(fun acc -> function
        | Stanza.Library lib ->
          let acc =
            match lib.public with
             | None -> acc
             | Some { name; _ } -> String_set.add acc name
          in
          String_set.add acc lib.name
        | _ -> acc))
end
# 16141 "boot.ml"
end
module Installed_dune_file : sig
# 1 "src/installed_dune_file.mli"
(** Dune files that are installed on the system *)

val load : fname:string -> Jbuild.Sub_system_info.t Sub_system_name.Map.t
val gen : (Syntax.Version.t * Sexp.t) Sub_system_name.Map.t -> Sexp.t
# 16149 "boot.ml"
end = struct
# 1 "src/installed_dune_file.ml"
open Import

let parse_sub_systems sexps =
  List.filter_map sexps ~f:(fun sexp ->
    let name, ver, data =
      Sexp.Of_sexp.(triple string (located Syntax.Version.t_of_sexp) raw) sexp
    in
    match Sub_system_name.get name with
    | None ->
      (* We ignore sub-systems that are not internally known. These
         correspond to plugins that are not in use in the current
         workspace. *)
      None
    | Some name -> Some (name, (Sexp.Ast.loc sexp, ver, data)))
  |> Sub_system_name.Map.of_list
  |> (function
    | Ok x -> x
    | Error (name, _, (loc, _, _)) ->
      Loc.fail loc "%S present twice" (Sub_system_name.to_string name))
  |> Sub_system_name.Map.mapi ~f:(fun name (_, version, data) ->
    let (module M) = Jbuild.Sub_system_info.get name in
    let vloc, ver = version in
    let parser =
      Syntax.Versioned_parser.find_exn M.parsers ~loc:vloc
        ~data_version:ver
    in
    M.T (parser.parse data))

let of_sexp =
  let open Sexp.Of_sexp in
  let version sexp =
    match string sexp with
    | "1" -> ()
    | _  ->
      of_sexp_error sexp "Unsupported version, only version 1 is supported"
  in
  sum
    [ cstr "dune" (version @> list raw @> nil)
        (fun () l -> parse_sub_systems l)
    ]

let load ~fname = of_sexp (Sexp.load ~mode:Single ~fname)

let gen confs =
  let sexps =
    Sub_system_name.Map.to_list confs
    |> List.map ~f:(fun (name, (ver, conf)) ->
      let (module M) = Jbuild.Sub_system_info.get name in
      Sexp.List [ Sexp.atom (Sub_system_name.to_string name)
                ; Syntax.Version.sexp_of_t ver
                ; conf
                ])
  in
  Sexp.List
    [ Sexp.unsafe_atom_of_string "dune"
    ; Sexp.unsafe_atom_of_string "1"
    ; List sexps
    ]
# 16210 "boot.ml"
end
module Lib : sig
# 1 "src/lib.mli"
open Import

(** {1 Generals} *)

(** Representation of a library *)
type t

(** For libraries defined in the workspace, this is the [public_name] if
    present or the [name] if not. *)
val name : t -> string

(* CR-someday diml: this should be [Path.t list], since some libraries
   have multiple source directories because of [copy_files]. *)
(** Directory where the source files for the library are located. *)
val src_dir : t -> Path.t

(** Directory where the object files for the library are located. *)
val obj_dir : t -> Path.t

(** Same as [Path.is_local (obj_dir t)] *)
val is_local : t -> bool

val synopsis     : t -> string option
val kind         : t -> Jbuild.Library.Kind.t
val archives     : t -> Path.t list Mode.Dict.t
val plugins      : t -> Path.t list Mode.Dict.t
val jsoo_runtime : t -> Path.t list

(** A unique integer identifier. It is only unique for the duration of
    the process *)
val unique_id : t -> int

module Set : Set.S with type elt = t

module Map : Map.S with type key = t

module Status : sig
  type t =
    | Installed
    | Public  of Package.t
    | Private of Jbuild.Scope_info.Name.t

  val pp : t Fmt.t
end

val status : t -> Status.t

val package : t -> Package.Name.t option

(** Operations on list of libraries *)
module L : sig
  type nonrec t = t list

  val include_paths : t -> stdlib_dir:Path.t -> Path.Set.t
  val include_flags : t -> stdlib_dir:Path.t -> _ Arg_spec.t

  val c_include_flags : t -> stdlib_dir:Path.t -> _ Arg_spec.t

  val link_flags : t -> mode:Mode.t -> stdlib_dir:Path.t -> _ Arg_spec.t

  val compile_and_link_flags
    :  compile:t
    -> link:t
    -> mode:Mode.t
    -> stdlib_dir:Path.t
    -> _ Arg_spec.t

  (** All the library archive files (.a, .cmxa, _stubs.a, ...)  that
      should be linked in when linking an executable. *)
  val archive_files : t -> mode:Mode.t -> ext_lib:string -> Path.t list

  val jsoo_runtime_files : t -> Path.t list

  val remove_dups : t -> t
end

(** {1 Raw library descriptions} *)

(** Information about a library *)
module Info : sig
  module Deps : sig
    type t =
      | Simple  of (Loc.t * string) list
      | Complex of Jbuild.Lib_dep.t list
  end

  (** Raw description of a library, where dependencies are not
      resolved. *)
  type t =
    { loc              : Loc.t
    ; kind             : Jbuild.Library.Kind.t
    ; status           : Status.t
    ; src_dir          : Path.t
    ; obj_dir          : Path.t
    ; version          : string option
    ; synopsis         : string option
    ; archives         : Path.t list Mode.Dict.t
    ; plugins          : Path.t list Mode.Dict.t
    ; foreign_archives : Path.t list Mode.Dict.t (** [.a/.lib/...] files *)
    ; jsoo_runtime     : Path.t list
    ; requires         : Deps.t
    ; ppx_runtime_deps : (Loc.t * string) list
    ; pps              : (Loc.t * Jbuild.Pp.t) list
    ; optional         : bool
    ; virtual_deps     : (Loc.t * string) list
    ; sub_systems      : Jbuild.Sub_system_info.t Sub_system_name.Map.t
    }

  val of_library_stanza : dir:Path.t -> Jbuild.Library.t -> t
  val of_findlib_package : Findlib.Package.t -> t
end

(** {1 Errors} *)

module Error : sig
  module Library_not_available : sig
    module Reason : sig
      module Hidden : sig
        type t =
          { name   : string
          ; path   : Path.t
          ; reason : string
          }
      end

      type t =
        | Not_found
        | Hidden of Hidden.t

      val to_string : t -> string
      val pp : Format.formatter -> t -> unit
    end

    type nonrec t =
      { loc    : Loc.t (** For names coming from Jbuild files *)
      ; name   : string
      ; reason : Reason.t
      }
  end

  module No_solution_found_for_select : sig
    type t = { loc : Loc.t }
  end

  module Conflict : sig
    (** When two libraries in a transitive closure conflict *)
    type nonrec t =
      { lib1 : t * Dep_path.Entry.t list
      ; lib2 : t * Dep_path.Entry.t list
      }
  end

  module Overlap : sig
    (** A conflict that doesn't prevent compilation, but that we still
        consider as an error to avoid surprises. *)
    type nonrec t =
      { in_workspace : t
      ; installed    : t * Dep_path.Entry.t list
      }
  end

  module Private_deps_not_allowed : sig
    type nonrec t =
      { private_dep : t
      ; pd_loc      : Loc.t
      }
  end

  type t =
    | Library_not_available        of Library_not_available.t
    | No_solution_found_for_select of No_solution_found_for_select.t
    | Dependency_cycle             of (Path.t * string) list
    | Conflict                     of Conflict.t
    | Overlap                      of Overlap.t
    | Private_deps_not_allowed     of Private_deps_not_allowed.t
end

exception Error of Error.t

(** Raise a error about a library that is not available *)
val not_available
  :  loc:Loc.t
  -> Error.Library_not_available.Reason.t
  -> ('a, Format.formatter, unit, 'b) format4
  -> 'a

(** {1 Compilation contexts} *)

(** See {!Sub_system} *)
type sub_system = ..

(** For compiling a library or executable *)
module Compile : sig
  type t

  (** Return the list of dependencies needed for compiling this library *)
  val requires : t -> L.t Or_exn.t

  (** Dependencies listed by the user + runtime dependencies from ppx *)
  val direct_requires : t -> L.t Or_exn.t

  module Resolved_select : sig
    type t =
      { src_fn : (string, Error.No_solution_found_for_select.t) result
      ; dst_fn : string
      }
  end

  (** Resolved select forms *)
  val resolved_selects : t -> Resolved_select.t list

  (** Transitive closure of all used ppx rewriters *)
  val pps : t -> L.t Or_exn.t

  val optional          : t -> bool
  val user_written_deps : t -> Jbuild.Lib_deps.t

  (** Sub-systems used in this compilation context *)
  val sub_systems : t -> sub_system list
end

(** {1 Library name resolution} *)

(** Collection of libraries organized by names *)
module DB : sig
  type lib = t

  (** A database allow to resolve library names *)
  type t

  module Resolve_result : sig
    type nonrec t =
      | Not_found
      | Found    of Info.t
      | Hidden   of Info.t * string
      | Redirect of t option * string
  end

  (** Create a new library database. [resolve] is used to resolve
      library names in this database.

      When a library is not found, it is looked up in the parent
      database if any.

      [all] returns the list of names of libraries available in this database.
  *)
  val create
    :  ?parent:t
    -> resolve:(string -> Resolve_result.t)
    -> all:(unit -> string list)
    -> unit
    -> t

  (** Create a database from a list of library stanzas *)
  val create_from_library_stanzas
    :  ?parent:t
    -> (Path.t * Jbuild.Library.t) list
    -> t

  val create_from_findlib
    :  ?external_lib_deps_mode:bool
    -> Findlib.t
    -> t

  val find : t -> string -> (lib, Error.Library_not_available.Reason.t) result
  val find_many
    :  t
    -> string list
    -> lib list Or_exn.t

  val find_even_when_hidden : t -> string -> lib option

  val available : t -> string -> bool

  (** Retreive the compile informations for the given library. Works
      for libraries that are optional and not available as well. *)
  val get_compile_info : t -> ?allow_overlaps:bool -> string -> Compile.t

  val resolve : t -> Loc.t * string -> lib Or_exn.t

  (** Resolve libraries written by the user in a jbuild file. The
      resulting list of libraries is transitively closed and sorted by
      order of dependencies.

      This function is for executables stanzas.  *)
  val resolve_user_written_deps
    :  t
    -> ?allow_overlaps:bool
    -> Jbuild.Lib_dep.t list
    -> pps:(Loc.t * Jbuild.Pp.t) list
    -> Compile.t

  val resolve_pps
    :  t
    -> (Loc.t * Jbuild.Pp.t) list
    -> L.t Or_exn.t

  (** Return the list of all libraries in this database. If
      [recursive] is true, also include libraries in parent databases
      recursively. *)
  val all : ?recursive:bool -> t -> Set.t
end with type lib := t

(** {1 Transitive closure} *)

val closure : L.t -> L.t Or_exn.t

(** {1 Sub-systems} *)

module Sub_system : sig
  type lib = t

  type t = sub_system = ..

  module type S = sig
    module Info : Jbuild.Sub_system_info.S
    type t
    type sub_system += T of t
    val instantiate
      :  resolve:(Loc.t * string -> lib Or_exn.t)
      -> get:(lib -> t option)
      -> lib
      -> Info.t
      -> t
    val to_sexp : (t -> Syntax.Version.t * Sexp.t) option
  end

  module Register(M : S) : sig
    (** Get the instance of the subsystem for this library *)
    val get : lib -> M.t option
  end

  val dump_config : lib -> (Syntax.Version.t * Sexp.t) Sub_system_name.Map.t
end with type lib := t

(** {1 Dependencies for META files} *)

module Meta : sig
  val requires                               : t -> String_set.t
  val ppx_runtime_deps                       : t -> String_set.t
  val ppx_runtime_deps_for_deprecated_method : t -> String_set.t
end
# 16556 "boot.ml"
end = struct
# 1 "src/lib.ml"
open Import
open Result.O

(* +-----------------------------------------------------------------+
   | Raw library information                                         |
   +-----------------------------------------------------------------+ *)

module Status = struct
  type t =
    | Installed
    | Public  of Package.t
    | Private of Jbuild.Scope_info.Name.t

  let pp ppf t =
    Format.pp_print_string ppf
      (match t with
       | Installed -> "installed"
       | Public _ -> "public"
       | Private s ->
         sprintf "private (%s)" (Jbuild.Scope_info.Name.to_string s))

  let is_private = function
    | Private _ -> true
    | Installed | Public _ -> false
end

module Info = struct
  module Deps = struct
    type t =
      | Simple  of (Loc.t * string) list
      | Complex of Jbuild.Lib_dep.t list

    let of_lib_deps deps =
      let rec loop acc (deps : Jbuild.Lib_dep.t list) =
        match deps with
        | []               -> Some (List.rev acc)
        | Direct x :: deps -> loop (x :: acc) deps
        | Select _ :: _    -> None
      in
      match loop [] deps with
      | Some l -> Simple l
      | None   -> Complex deps

    let to_lib_deps = function
      | Simple  l -> List.map l ~f:Jbuild.Lib_dep.direct
      | Complex l -> l
  end

  type t =
    { loc              : Loc.t
    ; kind             : Jbuild.Library.Kind.t
    ; status           : Status.t
    ; src_dir          : Path.t
    ; obj_dir          : Path.t
    ; version          : string option
    ; synopsis         : string option
    ; archives         : Path.t list Mode.Dict.t
    ; plugins          : Path.t list Mode.Dict.t
    ; foreign_archives : Path.t list Mode.Dict.t
    ; jsoo_runtime     : Path.t list
    ; requires         : Deps.t
    ; ppx_runtime_deps : (Loc.t * string) list
    ; pps              : (Loc.t * Jbuild.Pp.t) list
    ; optional         : bool
    ; virtual_deps     : (Loc.t * string) list
    ; sub_systems      : Jbuild.Sub_system_info.t Sub_system_name.Map.t
    }

  let user_written_deps t =
    List.fold_left (t.virtual_deps @ t.ppx_runtime_deps)
      ~init:(Deps.to_lib_deps t.requires)
      ~f:(fun acc s -> Jbuild.Lib_dep.Direct s :: acc)

  let of_library_stanza ~dir (conf : Jbuild.Library.t) =
    let archive_file ext = Path.relative dir (conf.name ^ ext) in
    let archive_files ~f_ext =
      Mode.Dict.of_func (fun ~mode -> [archive_file (f_ext mode)])
    in
    let stubs =
      if Jbuild.Library.has_stubs conf then
        [Jbuild.Library.stubs_archive conf ~dir ~ext_lib:""]
      else
        []
    in
    let jsoo_runtime =
      List.map conf.buildable.js_of_ocaml.javascript_files
        ~f:(Path.relative dir)
    in
    let status =
      match conf.public with
      | None   -> Status.Private conf.scope_name
      | Some p -> Public p.package
    in
    let foreign_archives =
      { Mode.Dict.
        byte   = stubs
      ; native = Path.relative dir conf.name :: stubs
      }
    in
    { loc = conf.buildable.loc
    ; kind     = conf.kind
    ; src_dir  = dir
    ; obj_dir  = Utils.library_object_directory ~dir conf.name
    ; version  = None
    ; synopsis = conf.synopsis
    ; archives = archive_files ~f_ext:Mode.compiled_lib_ext
    ; plugins  = archive_files ~f_ext:Mode.plugin_ext
    ; optional = conf.optional
    ; foreign_archives
    ; jsoo_runtime
    ; status
    ; virtual_deps     = conf.virtual_deps
    ; requires         = Deps.of_lib_deps conf.buildable.libraries
    ; ppx_runtime_deps = conf.ppx_runtime_libraries
    ; pps = Jbuild.Preprocess_map.pps conf.buildable.preprocess
    ; sub_systems = conf.sub_systems
    }

  let of_findlib_package pkg =
    let module P = Findlib.Package in
    let loc = Loc.in_file (Path.to_string (P.meta_file pkg)) in
    let add_loc x = (loc, x) in
    let sub_systems =
      match P.dune_file pkg with
      | None -> Sub_system_name.Map.empty
      | Some fn ->
        Installed_dune_file.load ~fname:(Path.to_string fn)
    in
    { loc              = loc
    ; kind             = Normal
    ; src_dir          = P.dir pkg
    ; obj_dir          = P.dir pkg
    ; version          = P.version pkg
    ; synopsis         = P.description pkg
    ; archives         = P.archives pkg
    ; plugins          = P.plugins pkg
    ; jsoo_runtime     = P.jsoo_runtime pkg
    ; requires         = Simple (List.map (P.requires pkg) ~f:add_loc)
    ; ppx_runtime_deps = List.map (P.ppx_runtime_deps pkg) ~f:add_loc
    ; pps              = []
    ; virtual_deps     = []
    ; optional         = false
    ; status           = Installed
    ; (* We don't know how these are named for external libraries *)
      foreign_archives = Mode.Dict.make_both []
    ; sub_systems      = sub_systems
    }
end

(* +-----------------------------------------------------------------+
   | Types                                                           |
   +-----------------------------------------------------------------+ *)

module Error0 = struct
  module Library_not_available = struct
    module Reason = struct
      module Hidden = struct
        type t =
          { name   : string
          ; path   : Path.t
          ; reason : string
          }
      end

      type t =
        | Not_found
        | Hidden of Hidden.t

      let to_string = function
        | Not_found -> "not found"
        | Hidden { path; reason; _ } ->
          sprintf "in %s is hidden (%s)"
            (Path.to_string_maybe_quoted path) reason

      let pp ppf t = Format.pp_print_string ppf (to_string t)
    end

    type t =
      { loc    : Loc.t
      ; name   : string
      ; reason : Reason.t
      }
  end

  module No_solution_found_for_select = struct
    type t = { loc : Loc.t }
  end
end

module Resolved_select = struct
  type t =
    { src_fn : (string, Error0.No_solution_found_for_select.t) result
    ; dst_fn : string
    }
end

type sub_system = ..

module Sub_system0 = struct
  module type S = sig
    type t
    type sub_system += T of t
    val to_sexp : (t -> Syntax.Version.t * Sexp.t) option
  end

  type 'a s = (module S with type t = 'a)

  module Instance = struct
    type t = T : 'a s * 'a -> t
  end
end

module Id = struct
  type t =
    { unique_id : int
    ; path      : Path.t
    ; name      : string
    }
end

type t =
  { loc               : Loc.t
  ; name              : string
  ; unique_id         : int
  ; kind              : Jbuild.Library.Kind.t
  ; status            : Status.t
  ; src_dir           : Path.t
  ; obj_dir           : Path.t
  ; version           : string option
  ; synopsis          : string option
  ; archives          : Path.t list Mode.Dict.t
  ; plugins           : Path.t list Mode.Dict.t
  ; foreign_archives  : Path.t list Mode.Dict.t
  ; jsoo_runtime      : Path.t list
  ; requires          : t list Or_exn.t
  ; ppx_runtime_deps  : t list Or_exn.t
  ; pps               : t list Or_exn.t
  ; resolved_selects  : Resolved_select.t list
  ; optional          : bool
  ; user_written_deps : Jbuild.Lib_deps.t
  ; (* This is mutable to avoid this error:

       {[
         This kind of expression is not allowed as right-hand side of `let rec'
       }]
       *)
    mutable sub_systems : Sub_system0.Instance.t Lazy.t Sub_system_name.Map.t
  }

and db =
  { parent  : db option
  ; resolve : string -> resolve_result
  ; table   : (string, status) Hashtbl.t
  ; all     : string list Lazy.t
  }

and status =
  | St_initializing of Id.t (* To detect cycles *)
  | St_found        of t
  | St_not_found
  | St_hidden       of t * Error0.Library_not_available.Reason.Hidden.t

and error =
  | Library_not_available        of Error0.Library_not_available.t
  | No_solution_found_for_select of Error0.No_solution_found_for_select.t
  | Dependency_cycle             of (Path.t * string) list
  | Conflict                     of conflict
  | Overlap                      of overlap
  | Private_deps_not_allowed     of private_deps_not_allowed

and resolve_result =
  | Not_found
  | Found    of Info.t
  | Hidden   of Info.t * string
  | Redirect of db option * string

and conflict =
  { lib1 : t * Dep_path.Entry.t list
  ; lib2 : t * Dep_path.Entry.t list
  }

and overlap =
  { in_workspace : t
  ; installed    : t * Dep_path.Entry.t list
  }

and private_deps_not_allowed =
  { private_dep    : t
  ; pd_loc         : Loc.t
  }

type lib = t

module Error = struct
  include Error0

  module Conflict = struct
    type nonrec t = conflict =
      { lib1 : t * Dep_path.Entry.t list
      ; lib2 : t * Dep_path.Entry.t list
      }
  end

  module Overlap = struct
    type nonrec t = overlap =
      { in_workspace : t
      ; installed    : t * Dep_path.Entry.t list
      }
  end

  module Private_deps_not_allowed = struct
    type nonrec t = private_deps_not_allowed =
      { private_dep : t
      ; pd_loc      : Loc.t
      }
  end

  type t = error =
    | Library_not_available        of Library_not_available.t
    | No_solution_found_for_select of No_solution_found_for_select.t
    | Dependency_cycle             of (Path.t * string) list
    | Conflict                     of Conflict.t
    | Overlap                      of Overlap.t
    | Private_deps_not_allowed     of Private_deps_not_allowed.t
end

exception Error of Error.t

let not_available ~loc reason fmt =
  Errors.kerrf fmt ~f:(fun s ->
    Loc.fail loc "%s %a" s
      Error.Library_not_available.Reason.pp reason)

(* +-----------------------------------------------------------------+
   | Generals                                                        |
   +-----------------------------------------------------------------+ *)

let name t = t.name

let kind         t = t.kind
let synopsis     t = t.synopsis
let archives     t = t.archives
let plugins      t = t.plugins
let jsoo_runtime t = t.jsoo_runtime
let unique_id    t = t.unique_id

let src_dir t = t.src_dir
let obj_dir t = t.obj_dir

let is_local t = Path.is_local t.obj_dir

let status t = t.status

let package t =
  match t.status with
  | Installed ->
    Some (Findlib.root_package_name t.name
          |> Package.Name.of_string)
  | Public p -> Some p.name
  | Private _ ->
    None

let to_id t : Id.t =
  { unique_id = t.unique_id
  ; path      = t.src_dir
  ; name      = t.name
  }

module Set = Set.Make(
struct
  type nonrec t = t
  let compare x y = compare x.unique_id y.unique_id
end)

module Map = Map.Make(
struct
  type nonrec t = t
  let compare x y = compare x.unique_id y.unique_id
end)

module L = struct
  type nonrec t = t list

  let to_iflags dirs =
    Arg_spec.S
      (Path.Set.fold dirs ~init:[] ~f:(fun dir acc ->
         Arg_spec.Path dir :: A "-I" :: acc)
       |> List.rev)

  let include_paths ts ~stdlib_dir =
    let dirs =
      List.fold_left ts ~init:Path.Set.empty ~f:(fun acc t ->
        Path.Set.add acc (obj_dir t))
    in
    Path.Set.remove dirs stdlib_dir

  let include_flags ts ~stdlib_dir =
    to_iflags (include_paths ts ~stdlib_dir)

  let c_include_paths ts ~stdlib_dir =
    let dirs =
      List.fold_left ts ~init:Path.Set.empty ~f:(fun acc t ->
        Path.Set.add acc t.src_dir)
    in
    Path.Set.remove dirs stdlib_dir

  let c_include_flags ts ~stdlib_dir =
    to_iflags (c_include_paths ts ~stdlib_dir)

  let link_flags ts ~mode ~stdlib_dir =
    Arg_spec.S
      (c_include_flags ts ~stdlib_dir ::
       List.map ts ~f:(fun t -> Arg_spec.Deps (Mode.Dict.get t.archives mode)))

  let compile_and_link_flags ~compile ~link ~mode ~stdlib_dir =
    let dirs =
      Path.Set.union
        (  include_paths compile ~stdlib_dir)
        (c_include_paths link    ~stdlib_dir)
    in
    Arg_spec.S
      (to_iflags dirs ::
       List.map link ~f:(fun t ->
         Arg_spec.Deps (Mode.Dict.get t.archives mode)))

  let jsoo_runtime_files ts =
    List.concat_map ts ~f:(fun t -> t.jsoo_runtime)

  let archive_files ts ~mode ~ext_lib =
    List.concat_map ts ~f:(fun t ->
      Mode.Dict.get t.archives mode @
      List.map (Mode.Dict.get t.foreign_archives mode)
        ~f:(Path.extend_basename ~suffix:ext_lib))

  let remove_dups l =
    let rec loop acc l seen =
      match l with
      | [] -> acc
      | x :: l ->
        if Int_set.mem seen x.unique_id then
          loop acc l seen
        else
          loop (x :: acc) l (Int_set.add seen x.unique_id)
    in
    loop [] l Int_set.empty
end

(* +-----------------------------------------------------------------+
   | Sub-systems                                                     |
   +-----------------------------------------------------------------+ *)

module Sub_system = struct
  type t = sub_system = ..

  module type S = sig
    module Info : Jbuild.Sub_system_info.S
    type t
    type sub_system += T of t
    val instantiate
      :  resolve:(Loc.t * string -> (lib, exn) result)
      -> get:(lib -> t option)
      -> lib
      -> Info.t
      -> t
    val to_sexp : (t -> Syntax.Version.t * Sexp.t) option
  end

  module type S' = sig
    include S
    val for_instance : t Sub_system0.s
    val get : lib -> t option
  end

  let all = Sub_system_name.Table.create ~default_value:None

  module Register(M : S) = struct
    let get lib =
      Option.map (Sub_system_name.Map.find lib.sub_systems M.Info.name)
        ~f:(fun (lazy (Sub_system0.Instance.T ((module X), t))) ->
          match X.T t with
          | M.T t -> t
          | _   -> assert false)

    let () =
      let module M = struct
        include M
        let for_instance = (module M : Sub_system0.S with type t = t)
        let get = get
      end in
      Sub_system_name.Table.set all ~key:M.Info.name
        ~data:(Some (module M : S'))
  end

  let instantiate name info lib ~resolve =
    let impl = Option.value_exn (Sub_system_name.Table.get all name) in
    let (module M : S') = impl in
    match info with
    | M.Info.T info ->
      Sub_system0.Instance.T
        (M.for_instance, M.instantiate ~resolve ~get:M.get lib info)
    | _ -> assert false

  let dump_config lib =
    Sub_system_name.Map.filter_map lib.sub_systems ~f:(fun (lazy inst) ->
      let (Sub_system0.Instance.T ((module M), t)) = inst in
      match M.to_sexp with
      | None -> None
      | Some f -> Some (f t))
end

(* +-----------------------------------------------------------------+
   | Library name resolution and transitive closure                  |
   +-----------------------------------------------------------------+ *)

let gen_unique_id =
  let next = ref 0 in
  fun () ->
    let n = !next in
    next := n + 1;
    n

(* Dependency stack used while resolving the dependencies of a library
   that was just returned by the [resolve] callback *)
module Dep_stack = struct
  type t =
    { stack : Id.t list
    ; seen  : Int_set.t
    }

  let empty =
    { stack = []
    ; seen  = Int_set.empty
    }

  let to_required_by t ~stop_at =
    let stop_at = stop_at.stack in
    let rec loop acc l =
      if l == stop_at then
        List.rev acc
      else
        match l with
        | [] -> assert false
        | { Id.path; name; _ } :: l ->
          loop (Dep_path.Entry.Library (path, name) :: acc) l
    in
    loop [] t.stack

  let dependency_cycle t (last : Id.t) =
    assert (Int_set.mem t.seen last.unique_id);
    let rec build_loop acc stack =
      match stack with
      | [] -> assert false
      | (x : Id.t) :: stack ->
        let acc = (x.path, x.name) :: acc in
        if x.unique_id = last.unique_id then
          acc
        else
          build_loop acc stack
    in
    let loop = build_loop [(last.path, last.name)] t.stack in
    Error (Dependency_cycle loop)

  let create_and_push t name path =
    let unique_id = gen_unique_id () in
    let init = { Id. unique_id; name; path } in
    (init,
     { stack = init :: t.stack
     ; seen  = Int_set.add t.seen unique_id
     })

  let push t (x : Id.t) : (_, _) result =
    if Int_set.mem t.seen x.unique_id then
      Error (dependency_cycle t x)
    else
      Ok { stack = x :: t.stack
         ; seen  = Int_set.add t.seen x.unique_id
         }
end

let check_private_deps ~(lib : lib) ~loc ~allow_private_deps =
  if (not allow_private_deps) && Status.is_private lib.status then
    Result.Error (Error (
      Private_deps_not_allowed { private_dep = lib ; pd_loc = loc }))
  else
    Ok lib

let already_in_table (info : Info.t) name x =
  let to_sexp = Sexp.To_sexp.(pair Path.sexp_of_t string) in
  let sexp =
    match x with
    | St_initializing x ->
      Sexp.List [Sexp.unsafe_atom_of_string "Initializing";
                 Path.sexp_of_t x.path]
    | St_found t ->
      List [Sexp.unsafe_atom_of_string "Found";
            Path.sexp_of_t t.src_dir]
    | St_not_found ->
      Sexp.unsafe_atom_of_string "Not_found"
    | St_hidden (_, { path; reason; _ }) ->
      List [Sexp.unsafe_atom_of_string "Hidden";
            Path.sexp_of_t path; Sexp.atom reason]
  in
  Sexp.code_error
    "Lib_db.DB: resolver returned name that's already in the table"
    [ "name"            , Sexp.atom name
    ; "returned_lib"    , to_sexp (info.src_dir, name)
    ; "conflicting_with", sexp
    ]

let result_of_resolve_status = function
  | St_initializing _     -> assert false
  | St_found x            -> Ok x
  | St_not_found          -> Error Error.Library_not_available.Reason.Not_found
  | St_hidden (_, hidden) -> Error (Hidden hidden)

let rec instantiate db name (info : Info.t) ~stack ~hidden =
  let id, stack =
    Dep_stack.create_and_push stack name info.src_dir
  in
  Option.iter (Hashtbl.find db.table name) ~f:(fun x ->
    already_in_table info name x);
  (* Add [id] to the table, to detect loops *)
  Hashtbl.add db.table name (St_initializing id);

  let allow_private_deps = Status.is_private info.status in

  let requires, pps, resolved_selects =
    resolve_user_deps db info.requires ~allow_private_deps ~pps:info.pps ~stack
  in
  let ppx_runtime_deps =
    resolve_simple_deps db info.ppx_runtime_deps ~allow_private_deps ~stack
  in
  let map_error x =
    Result.map_error x ~f:(fun e ->
      Dep_path.prepend_exn e (Library (info.src_dir, name)))
  in
  let requires         = map_error requires         in
  let ppx_runtime_deps = map_error ppx_runtime_deps in
  let resolve (loc, name) =
    resolve_dep db name ~allow_private_deps ~loc ~stack in
  let t =
    { loc               = info.loc
    ; name              = name
    ; unique_id         = id.unique_id
    ; kind              = info.kind
    ; status            = info.status
    ; src_dir           = info.src_dir
    ; obj_dir           = info.obj_dir
    ; version           = info.version
    ; synopsis          = info.synopsis
    ; archives          = info.archives
    ; plugins           = info.plugins
    ; foreign_archives  = info.foreign_archives
    ; jsoo_runtime      = info.jsoo_runtime
    ; requires          = requires
    ; ppx_runtime_deps  = ppx_runtime_deps
    ; pps               = pps
    ; resolved_selects  = resolved_selects
    ; optional          = info.optional
    ; user_written_deps = Info.user_written_deps info
    ; sub_systems       = Sub_system_name.Map.empty
    }
  in
  t.sub_systems <-
    Sub_system_name.Map.mapi info.sub_systems ~f:(fun name info ->
      lazy (Sub_system.instantiate name info t ~resolve));

  let res =
    let hidden =
      match hidden with
      | None ->
        Option.some_if
          (info.optional &&
           not (Result.is_ok t.requires && Result.is_ok t.ppx_runtime_deps))
          "optional with unavailable dependencies"
      | Some _ -> hidden
    in
    match hidden with
    | None -> St_found t
    | Some reason ->
      St_hidden (t, { name; path = t.src_dir; reason })
  in
  Hashtbl.replace db.table ~key:name ~data:res;
  res

and find db name : (t, Error.Library_not_available.Reason.t) result =
  result_of_resolve_status (find_internal db name ~stack:Dep_stack.empty)

and find_even_when_hidden db name =
  match find_internal db name ~stack:Dep_stack.empty with
  | St_initializing _     -> assert false
  | St_found t            -> Some t
  | St_not_found          -> None
  | St_hidden (t, _)      -> Some t

and find_internal db name ~stack : status =
  match Hashtbl.find db.table name with
  | Some x -> x
  | None   -> resolve_name db name ~stack

and resolve_dep db name ~allow_private_deps ~loc ~stack : (t, exn) result =
  match find_internal db name ~stack with
  | St_initializing id ->
    Error (Dep_stack.dependency_cycle stack id)
  | St_found lib -> check_private_deps ~lib ~loc ~allow_private_deps
  | St_not_found ->
    Error (Error (Library_not_available { loc; name; reason = Not_found }))
  | St_hidden (_, hidden) ->
    Error (Error (Library_not_available { loc; name; reason = Hidden hidden }))

and resolve_name db name ~stack =
  match db.resolve name with
  | Redirect (db', name') -> begin
      let db' = Option.value db' ~default:db in
      match find_internal db' name' ~stack with
      | St_initializing _ as x -> x
      | x ->
        Hashtbl.add db.table name x;
        x
    end
  | Found info ->
    instantiate db name info ~stack ~hidden:None
  | Not_found ->
    let res =
      match db.parent with
      | None    -> St_not_found
      | Some db -> find_internal db name ~stack
    in
    Hashtbl.add db.table name res;
    res
  | Hidden (info, hidden) ->
    match
      match db.parent with
      | None    -> St_not_found
      | Some db -> find_internal db name ~stack
    with
    | St_found _ as x ->
      Hashtbl.add db.table name x;
      x
    | _ ->
      instantiate db name info ~stack ~hidden:(Some hidden)

and available_internal db name ~stack =
  match resolve_dep db name ~allow_private_deps:true ~loc:Loc.none ~stack with
  | Ok    _ -> true
  | Error _ -> false

and resolve_simple_deps db names ~allow_private_deps ~stack =
  let rec loop acc = function
    | [] -> Ok (List.rev acc)
    | (loc, name) :: names ->
      resolve_dep db name ~allow_private_deps ~loc ~stack >>= fun x ->
      loop (x :: acc) names
  in
  loop [] names

and resolve_complex_deps db deps ~allow_private_deps ~stack =
  let res, resolved_selects =
    List.fold_left deps ~init:(Ok [], []) ~f:(fun (acc_res, acc_selects) dep ->
      let res, acc_selects =
        match (dep : Jbuild.Lib_dep.t) with
        | Direct (loc, name) ->
          let res =
            resolve_dep db name ~allow_private_deps ~loc ~stack >>| fun x -> [x]
          in
          (res, acc_selects)
        | Select { result_fn; choices; loc } ->
          let res, src_fn =
            match
              List.find_map choices ~f:(fun { required; forbidden; file } ->
                if String_set.exists forbidden
                     ~f:(available_internal db ~stack) then
                  None
                else
                  match
                    let deps =
                      String_set.fold required ~init:[] ~f:(fun x acc ->
                        (Loc.none, x) :: acc)
                    in
                    resolve_simple_deps ~allow_private_deps db deps ~stack
                  with
                  | Ok ts -> Some (ts, file)
                  | Error _ -> None)
            with
            | Some (ts, file) ->
              (Ok ts, Ok file)
            | None ->
              let e = { Error.No_solution_found_for_select.loc } in
              (Error (Error (No_solution_found_for_select e)),
               Error e)
          in
          (res, { Resolved_select. src_fn; dst_fn = result_fn } :: acc_selects)
      in
      let res =
        match res, acc_res with
        | Ok l, Ok acc -> Ok (List.rev_append l acc)
        | (Error _ as res), _
        | _, (Error _ as res) -> res
      in
      (res, acc_selects))
  in
  let res =
    match res with
    | Ok    l -> Ok (List.rev l)
    | Error _ -> res
  in
  (res, resolved_selects)

and resolve_deps db deps ~allow_private_deps ~stack =
  match (deps : Info.Deps.t) with
  | Simple  names ->
    (resolve_simple_deps db names ~allow_private_deps ~stack, [])
  | Complex names ->
    resolve_complex_deps ~allow_private_deps db names ~stack

and resolve_user_deps db deps ~allow_private_deps ~pps ~stack =
  let deps, resolved_selects =
    resolve_deps db deps ~allow_private_deps ~stack in
  let deps, pps =
    match pps with
    | [] -> (deps, Ok [])
    | first :: others as pps ->
      (* Location of the list of ppx rewriters *)
      let loc =
        let last = Option.value (List.last others) ~default:first in
        { (fst first) with stop = (fst last).stop }
      in
      let pps =
        let pps = (pps : (Loc.t * Jbuild.Pp.t) list :> (Loc.t * string) list) in
        resolve_simple_deps db pps ~allow_private_deps:true ~stack
        >>= fun pps ->
        closure_with_overlap_checks None pps ~stack
      in
      let deps =
        let rec check_runtime_deps acc pps = function
          | [] -> loop acc pps
          | lib :: ppx_rts ->
            check_private_deps ~lib ~loc ~allow_private_deps >>= fun rt ->
            check_runtime_deps (rt :: acc) pps ppx_rts
        and loop acc = function
          | [] -> Ok acc
          | pp :: pps ->
            pp.ppx_runtime_deps >>= fun rt_deps ->
            check_runtime_deps acc pps rt_deps
        in
        deps >>= fun deps ->
        pps  >>= fun pps  ->
        loop deps pps
      in
      (deps, pps)
  in
  (deps, pps, resolved_selects)

and closure_with_overlap_checks db ts ~stack =
  let visited = ref String_map.empty in
  let res = ref [] in
  let orig_stack = stack in
  let rec loop t ~stack =
    match String_map.find !visited t.name with
    | Some (t', stack') ->
      if t.unique_id = t'.unique_id then
        Ok ()
      else
        let req_by = Dep_stack.to_required_by ~stop_at:orig_stack in
        Error
          (Error (Conflict { lib1 = (t', req_by stack')
                           ; lib2 = (t , req_by stack )
                           }))
    | None ->
      visited := String_map.add !visited t.name (t, stack);
      (match db with
       | None -> Ok ()
       | Some db ->
         match find_internal db t.name ~stack with
         | St_found t' ->
           if t.unique_id = t'.unique_id then
             Ok ()
           else begin
             let req_by = Dep_stack.to_required_by stack ~stop_at:orig_stack in
             Error
               (Error (Overlap
                         { in_workspace = t'
                         ; installed    = (t, req_by)
                         }))
           end
         | _ -> assert false)
      >>= fun () ->
      Dep_stack.push stack (to_id t) >>= fun stack ->
      t.requires >>= fun deps ->
      iter deps ~stack >>| fun () ->
      res := t :: !res
  and iter ts ~stack =
    match ts with
    | [] -> Ok ()
    | t :: ts ->
      loop t ~stack >>= fun () ->
      iter ts ~stack
  in
  iter ts ~stack >>| fun () ->
  List.rev !res

let closure_with_overlap_checks db l =
  closure_with_overlap_checks db l ~stack:Dep_stack.empty

let closure l = closure_with_overlap_checks None l

let to_exn res =
  match res with
  | Ok    x -> x
  | Error e -> raise e

let requires_exn         t = to_exn t.requires
let ppx_runtime_deps_exn t = to_exn t.ppx_runtime_deps
let closure_exn          l = to_exn (closure l)

module Compile = struct
  module Resolved_select = Resolved_select

  type nonrec t =
    { direct_requires   : t list Or_exn.t
    ; requires          : t list Or_exn.t
    ; pps               : t list Or_exn.t
    ; resolved_selects  : Resolved_select.t list
    ; optional          : bool
    ; user_written_deps : Jbuild.Lib_deps.t
    ; sub_systems       : Sub_system0.Instance.t Lazy.t Sub_system_name.Map.t
    }

  let for_lib db (t : lib) =
    { direct_requires   = t.requires
    ; requires          = t.requires >>= closure_with_overlap_checks db
    ; resolved_selects  = t.resolved_selects
    ; pps               = t.pps
    ; optional          = t.optional
    ; user_written_deps = t.user_written_deps
    ; sub_systems       = t.sub_systems
    }

  let direct_requires   t = t.direct_requires
  let requires          t = t.requires
  let resolved_selects  t = t.resolved_selects
  let pps               t = t.pps
  let optional          t = t.optional
  let user_written_deps t = t.user_written_deps
  let sub_systems t =
    Sub_system_name.Map.values t.sub_systems
    |> List.map ~f:(fun (lazy (Sub_system0.Instance.T ((module M), t))) ->
      M.T t)
end

(* +-----------------------------------------------------------------+
   | Databases                                                       |
   +-----------------------------------------------------------------+ *)

module DB = struct
  module Resolve_result = struct
    type t = resolve_result =
      | Not_found
      | Found    of Info.t
      | Hidden   of Info.t * string
      | Redirect of db option * string
  end

  type t = db

  let create ?parent ~resolve ~all () =
    { parent
    ; resolve
    ; table  = Hashtbl.create 1024
    ; all    = Lazy.from_fun all
    }

  let create_from_library_stanzas ?parent stanzas =
    let map =
      List.concat_map stanzas ~f:(fun (dir, (conf : Jbuild.Library.t)) ->
        let info = Info.of_library_stanza ~dir conf in
        match conf.public with
        | None ->
          [(conf.name, Resolve_result.Found info)]
        | Some p ->
          if p.name = conf.name then
            [(p.name, Found info)]
          else
            [ p.name   , Found info
            ; conf.name, Redirect (None, p.name)
            ])
      |> String_map.of_list
      |> function
      | Ok x -> x
      | Error (name, _, _) ->
        match
          List.filter_map stanzas ~f:(fun (_, (conf : Jbuild.Library.t)) ->
            if name = conf.name ||
               match conf.public with
               | None -> false
               | Some p -> name = p.name
            then Some conf.buildable.loc
            else None)
        with
        | [] | [_] -> assert false
        | loc1 :: loc2 :: _ ->
          die "Library %S is defined twice:\n\
               - %s\n\
               - %s"
            name
            (Loc.to_file_colon_line loc1)
            (Loc.to_file_colon_line loc2)
    in
    create () ?parent
      ~resolve:(fun name ->
        match String_map.find map name with
        | None   -> Not_found
        | Some x -> x)
      ~all:(fun () -> String_map.keys map)

  let create_from_findlib ?(external_lib_deps_mode=false) findlib =
    create ()
      ~resolve:(fun name ->
        match Findlib.find findlib name with
        | Ok pkg -> Found (Info.of_findlib_package pkg)
        | Error e ->
          match e with
          | Not_found ->
            if external_lib_deps_mode then
              Found
                (Info.of_findlib_package
                   (Findlib.dummy_package findlib ~name))
            else
              Not_found
          | Hidden pkg ->
            Hidden (Info.of_findlib_package pkg,
                    "unsatisfied 'exist_if'"))
      ~all:(fun () ->
        Findlib.all_packages findlib
        |> List.map ~f:Findlib.Package.name)

  let find = find
  let find_even_when_hidden = find_even_when_hidden

  let resolve t (loc, name) =
    match find t name with
    | Ok _ as res -> res
    | Error reason ->
      Error (Error (Library_not_available
                      { loc
                      ; name
                      ; reason
                      }))

  let find_many =
    let rec loop t acc = function
      | [] -> Ok (List.rev acc)
      | name :: names ->
        resolve t (Loc.none, name) >>= fun lib ->
        loop t (lib ::acc) names
    in
    fun t names -> loop t [] names

  let available t name = available_internal t name ~stack:Dep_stack.empty

  let get_compile_info t ?(allow_overlaps=false) name =
    match find_even_when_hidden t name with
    | None ->
      Sexp.code_error "Lib.DB.get_compile_info got library that doesn't exist"
        [ "name", Sexp.To_sexp.string name ]
    | Some lib ->
      let t = Option.some_if (not allow_overlaps) t in
      Compile.for_lib t lib

  let resolve_user_written_deps t ?(allow_overlaps=false) deps ~pps =
    let res, pps, resolved_selects =
      resolve_user_deps t (Info.Deps.of_lib_deps deps) ~pps
        ~stack:Dep_stack.empty ~allow_private_deps:true
    in
    let requires =
      res
      >>=
      closure_with_overlap_checks (Option.some_if (not allow_overlaps) t)
    in
    { Compile.
      direct_requires = res
    ; requires
    ; pps
    ; resolved_selects
    ; optional          = false
    ; user_written_deps = deps
    ; sub_systems       = Sub_system_name.Map.empty
    }

  let resolve_pps t pps =
    resolve_simple_deps t ~allow_private_deps:true
      (pps : (Loc.t *Jbuild.Pp.t) list :> (Loc.t * string) list)
      ~stack:Dep_stack.empty

  let rec all ?(recursive=false) t =
    let l =
      List.fold_left (Lazy.force t.all) ~f:(fun libs name ->
        match find t name with
        | Ok    x -> Set.add libs x
        | Error _ -> libs) ~init:Set.empty
    in
    match recursive, t.parent with
    | true, Some t -> Set.union (all ~recursive t) l
    | _ -> l
end

(* +-----------------------------------------------------------------+
   | META files                                                      |
   +-----------------------------------------------------------------+ *)

module Meta = struct
  let to_names ts =
    List.fold_left ts ~init:String_set.empty ~f:(fun acc t ->
      String_set.add acc t.name)

  (* For the deprecated method, we need to put all the runtime
     dependencies of the transitive closure.

     We need to do this because [ocamlfind ocamlc -package ppx_foo]
     will not look for the transitive dependencies of [foo], and the
     runtime dependencies might be attached to a dependency of [foo]
     rather than [foo] itself.

     Sigh... *)
  let ppx_runtime_deps_for_deprecated_method t =
    closure_exn [t]
    |> List.concat_map ~f:ppx_runtime_deps_exn
    |> to_names

  let requires         t = to_names (requires_exn         t)
  let ppx_runtime_deps t = to_names (ppx_runtime_deps_exn t)
end

(* +-----------------------------------------------------------------+
   | Error reporting                                                 |
   +-----------------------------------------------------------------+ *)

let report_lib_error ppf (e : Error.t) =
  match e with
  | Library_not_available { loc = _; name; reason } ->
    Format.fprintf ppf
      "@{<error>Error@}: Library %S %a.@\n"
      name
      Error.Library_not_available.Reason.pp reason
  | Conflict { lib1 = (lib1, rb1); lib2 = (lib2, rb2) } ->
    Format.fprintf ppf
      "@[<v>@{<error>Error@}: Conflict between the following libraries:@,\
       - %S in %s@,\
      \    %a@,\
       - %S in %s@,\
      \    %a@,\
       This cannot work.@\n"
      lib1.name (Path.to_string_maybe_quoted lib1.src_dir)
      Dep_path.Entries.pp rb1
      lib2.name (Path.to_string_maybe_quoted lib2.src_dir)
      Dep_path.Entries.pp rb2
  | Overlap { in_workspace = lib1; installed = (lib2, rb2) } ->
    Format.fprintf ppf
      "@[<v>@{<error>Error@}: Conflict between the following libraries:@,\
       - %S in %s@,\
       - %S in %s@,\
      \    %a@,\
       This is not allowed.@\n"
      lib1.name (Path.to_string_maybe_quoted lib1.src_dir)
      lib2.name (Path.to_string_maybe_quoted lib2.src_dir)
      Dep_path.Entries.pp rb2
  | No_solution_found_for_select { loc } ->
    Format.fprintf ppf
      "%a@{<error>Error@}: No solution found for this select form.\n"
      Loc.print loc
  | Dependency_cycle cycle ->
    Format.fprintf ppf
      "@{<error>Error@}: Dependency cycle detected between the \
       following libraries:@\n\
       @[<v>%a@]\n"
      (Format.pp_print_list (fun ppf (path, name) ->
         Format.fprintf ppf "-> %S in %s"
           name (Path.to_string_maybe_quoted path)))
      cycle
  | Private_deps_not_allowed t ->
    Format.fprintf ppf
      "@{<error>Error@}: Library %S is private, it cannot be a dependency of \
       a public library.\nYou need to give %S a public name.\n"
      t.private_dep.name
      t.private_dep.name

let () =
  Report_error.register (fun exn ->
    match exn with
    | Error e ->
      let loc, hint =
        match e with
        | Library_not_available { loc; _ } ->
          (Some loc,
           match !Clflags.external_lib_deps_hint with
           | [] -> (* during bootstrap *) None
           | l ->
             Some (List.map l ~f:quote_for_shell |> String.concat ~sep:" "))
        | Private_deps_not_allowed t ->
          (Some t.pd_loc, None)
        | _ -> (None, None)
      in
      Some
        { Report_error.
          loc
        ; hint
        ; pp = (fun ppf -> report_lib_error ppf e)
        ; backtrace = false
        }
    | _ -> None)
# 17769 "boot.ml"
end
module Artifacts : sig
# 1 "src/artifacts.mli"
open! Import

type t

val create
  :  Context.t
  -> public_libs:Lib.DB.t
  -> 'a list
  -> f:('a -> Jbuild.Stanza.t list)
  -> t

(** A named artifact that is looked up in the PATH if not found in the tree

    If the name is an absolute path, it is used as it.
*)
val binary
  :  t
  -> ?hint:string
  -> string
  -> Action.Prog.t

(** [file_of_lib t ~from ~lib ~file] returns the path to a file in the
    directory of the given library. *)
val file_of_lib
  :  t
  -> loc:Loc.t
  -> lib:string
  -> file:string
  -> (Path.t, fail) result
# 17802 "boot.ml"
end = struct
# 1 "src/artifacts.ml"
open Import
open Jbuild

type t =
  { context     : Context.t
  ; local_bins  : Path.t String_map.t
  ; public_libs : Lib.DB.t
  }

let create (context : Context.t) ~public_libs l ~f =
  let bin_dir = Config.local_install_bin_dir ~context:context.name in
  let local_bins =
    List.fold_left l ~init:String_map.empty ~f:(fun acc x ->
      List.fold_left (f x) ~init:acc ~f:(fun local_bins stanza ->
        match (stanza : Stanza.t) with
        | Install { section = Bin; files; _ } ->
          List.fold_left files ~init:local_bins
            ~f:(fun acc { Install_conf. src; dst } ->
              let name =
                match dst with
                | Some s -> s
                | None -> Filename.basename src
              in
              let key =
                if Sys.win32 && Filename.extension name = ".exe" then
                  String.sub name ~pos:0 ~len:(String.length name - 4)
                else
                  name
              in
              let in_bin_dir =
                let fn =
                  if Sys.win32 then
                    match Filename.extension src with
                    | ".exe" | ".bc" ->
                      if Filename.extension name <> ".exe" then
                        name ^ ".exe"
                      else
                        name
                    | _ -> name
                  else
                    name
                in
                Path.relative bin_dir fn
              in
              String_map.add acc key in_bin_dir)
        | _ ->
          local_bins))
  in
  { context
  ; local_bins
  ; public_libs
  }

let binary t ?hint name =
  if not (Filename.is_relative name) then
    Ok (Path.absolute name)
  else
    match String_map.find t.local_bins name with
    | Some path -> Ok path
    | None ->
      match Context.which t.context name with
      | Some p -> Ok p
      | None ->
        Error
          { Action.Prog.Not_found.
            program = name
          ; hint
          ; context = t.context.Context.name
          }

let file_of_lib t ~loc ~lib ~file =
  match Lib.DB.find t.public_libs lib with
  | Error reason ->
    Error { fail = fun () ->
      Lib.not_available ~loc reason "Public library %S" lib  }
  | Ok lib ->
    if Lib.is_local lib then begin
      match String.split (Lib.name lib) ~on:'.' with
      | [] -> assert false
      | package :: rest ->
        let lib_install_dir =
          Config.local_install_lib_dir ~context:t.context.name ~package
        in
        let lib_install_dir =
          match rest with
          | [] -> lib_install_dir
          | _  -> Path.relative lib_install_dir (String.concat rest ~sep:"/")
        in
        Ok (Path.relative lib_install_dir file)
    end else
      Ok (Path.relative (Lib.src_dir lib) file)
# 17896 "boot.ml"
end
module File_tree : sig
# 1 "src/file_tree.mli"
open! Import

module Dir : sig
  type t

  val path     : t -> Path.t
  val files    : t -> String_set.t
  val file_paths    : t -> Path.Set.t
  val sub_dirs : t -> t String_map.t
  val sub_dir_paths : t -> Path.Set.t
  val sub_dir_names : t -> String_set.t

  (** Whether this directory is ignored by a [jbuild-ignore] file in
      one of its ancestor directories. *)
  val ignored : t -> bool

  val fold
    :  t
    -> traverse_ignored_dirs:bool
    -> init:'a
    -> f:(t -> 'a -> 'a)
    -> 'a
end

type t

val load : ?extra_ignored_subtrees:Path.Set.t -> Path.t -> t

val fold
  :  t
  -> traverse_ignored_dirs:bool
  -> init:'a
  -> f:(Dir.t -> 'a -> 'a)
  -> 'a

val root : t -> Dir.t

val find_dir : t -> Path.t -> Dir.t option

val files_of : t -> Path.t -> Path.Set.t

val exists : t -> Path.t -> bool
val file_exists : t -> Path.t -> string -> bool

val files_recursively_in : t -> ?prefix_with:Path.t -> Path.t -> Path.Set.t
# 17945 "boot.ml"
end = struct
# 1 "src/file_tree.ml"
open! Import

module Dir = struct
  type t =
    { path     : Path.t
    ; files    : String_set.t
    ; sub_dirs : t String_map.t
    ; ignored  : bool
    }

  let path t = t.path
  let files t = t.files
  let sub_dirs t = t.sub_dirs
  let ignored t = t.ignored

  let file_paths t =
    Path.Set.of_string_set t.files ~f:(Path.relative t.path)

  let sub_dir_names t =
    String_map.foldi t.sub_dirs ~init:String_set.empty
      ~f:(fun s _ acc -> String_set.add acc s)

  let sub_dir_paths t =
    String_map.foldi t.sub_dirs ~init:Path.Set.empty
      ~f:(fun s _ acc -> Path.Set.add acc (Path.relative t.path s))

  let rec fold t ~traverse_ignored_dirs ~init:acc ~f =
    if not traverse_ignored_dirs && t.ignored then
      acc
    else
      let acc = f t acc in
      String_map.fold t.sub_dirs ~init:acc ~f:(fun t acc ->
        fold t ~traverse_ignored_dirs ~init:acc ~f)
end

type t =
  { root : Dir.t
  ; dirs : Dir.t Path.Map.t
  }

let root t = t.root

let ignore_file fn ~is_directory =
  fn = "" || fn = "." ||
  (is_directory && (fn.[0] = '.' || fn.[0] = '_')) ||
  (fn.[0] = '.' && fn.[1] = '#')

let load ?(extra_ignored_subtrees=Path.Set.empty) path =
  let rec walk path ~ignored : Dir.t =
    let files, sub_dirs =
      Path.readdir path
      |> List.filter_partition_map ~f:(fun fn ->
        let path = Path.relative path fn in
        let is_directory = Path.is_directory path in
        if ignore_file fn ~is_directory then
          Skip
        else if is_directory then
          Right (fn, path)
        else
          Left fn)
    in
    let files = String_set.of_list files in
    let ignored_sub_dirs =
      if not ignored && String_set.mem files "jbuild-ignore" then
        let ignore_file = Path.to_string (Path.relative path "jbuild-ignore") in
        let files =
          Io.lines_of_file ignore_file
        in
        let remove_subdirs index fn =
          if Filename.dirname fn = Filename.current_dir_name then
            true
          else begin
            Loc.(warn (of_pos (ignore_file, index + 1, 0, String.length fn))
            "subdirectory expression %s ignored" fn);
            false
          end
        in
        String_set.of_list (List.filteri ~f:remove_subdirs files)
      else
        String_set.empty
    in
    let sub_dirs =
      List.map sub_dirs ~f:(fun (fn, path) ->
        let ignored =
          ignored
          || String_set.mem ignored_sub_dirs fn
          || Path.Set.mem extra_ignored_subtrees path
        in
        (fn, walk path ~ignored))
      |> String_map.of_list_exn
    in
    { path
    ; files
    ; sub_dirs
    ; ignored
    }
  in
  let root = walk path ~ignored:false in
  let dirs =
    Dir.fold root ~init:Path.Map.empty ~traverse_ignored_dirs:true
      ~f:(fun dir acc ->
        Path.Map.add acc dir.path dir)
  in
  { root
  ; dirs
  }

let fold t ~traverse_ignored_dirs ~init ~f =
  Dir.fold t.root ~traverse_ignored_dirs ~init ~f

let find_dir t path = Path.Map.find t.dirs path

let files_of t path =
  match find_dir t path with
  | None -> Path.Set.empty
  | Some dir ->
    Path.Set.of_string_set (Dir.files dir) ~f:(Path.relative path)

let file_exists t path fn =
  match Path.Map.find t.dirs path with
  | None -> false
  | Some { files; _ } -> String_set.mem files fn

let exists t path =
  Path.Map.mem t.dirs path ||
  file_exists t (Path.parent path) (Path.basename path)

let files_recursively_in t ?(prefix_with=Path.root) path =
  match find_dir t path with
  | None -> Path.Set.empty
  | Some dir ->
    Dir.fold dir ~init:Path.Set.empty ~traverse_ignored_dirs:true
      ~f:(fun dir acc ->
        let path = Path.append prefix_with (Dir.path dir) in
        String_set.fold (Dir.files dir) ~init:acc ~f:(fun fn acc ->
          Path.Set.add acc (Path.relative path fn)))
# 18084 "boot.ml"
end
module Vfile_kind : sig
# 1 "src/vfile_kind.mli"
open Import

module Id : sig
  type 'a t

  val eq : 'a t -> 'b t -> ('a, 'b) eq option
end

module type S = sig
  type t

  val id : t Id.t

  val load : Path.t -> t
  val to_string : Path.t -> t -> string
end

type 'a t = (module S with type t = 'a)

val eq : 'a t -> 'b t -> ('a, 'b) eq option

module Make
    (T : sig type t end)
    (F : functor (C : Sexp.Combinators) -> sig val t : T.t C.t end)
  : S with type t = T.t

module Make_full
    (T : sig type t end)
    (To_sexp : sig val t : Path.t -> T.t -> Sexp.t end)
    (Of_sexp : sig val t : Path.t -> Sexp.Ast.t -> T.t end)
  : S with type t = T.t
# 18119 "boot.ml"
end = struct
# 1 "src/vfile_kind.ml"
open Import

module Id = struct
  type 'a tag = ..

  module type S = sig
    type t
    type 'a tag += X : t tag
  end

  type 'a t = (module S with type t = 'a)

  let create (type a) () =
    let module M = struct
      type t = a
      type 'a tag += X : t tag
    end in
    (module M : S with type t = a)

  let eq (type a) (type b)
        (module A : S with type t = a)
        (module B : S with type t = b)
    : (a, b) eq option =
    match A.X with
    | B.X -> Some Eq
    | _   -> None
end

module type S = sig
  type t

  val id : t Id.t

  val load : Path.t -> t
  val to_string : Path.t -> t -> string
end

type 'a t = (module S with type t = 'a)

let eq (type a) (type b)
      (module A : S with type t = a)
      (module B : S with type t = b) =
  Id.eq A.id B.id

module Make_full
    (T : sig type t end)
    (To_sexp : sig val t : Path.t -> T.t -> Sexp.t end)
    (Of_sexp : sig val t : Path.t -> Sexp.Ast.t -> T.t end)
  : S with type t = T.t =
struct
  type t = T.t

  let id = Id.create ()

  let to_string path x = To_sexp.t path x |> Sexp.to_string

  let load path =
    Of_sexp.t path (Sexp.load ~fname:(Path.to_string path) ~mode:Single)
end


module Make
    (T : sig type t end)
    (F : functor (C : Sexp.Combinators) -> sig val t : T.t C.t end)
  : S with type t = T.t =
struct
  module Of_sexp = struct
    include F(Sexp.Of_sexp)
    let t _ sexp = t sexp
  end
  module To_sexp = struct
    include F(Sexp.To_sexp)
    let t _ x = t x
  end

  include Make_full(T)(To_sexp)(Of_sexp)
end

# 18200 "boot.ml"
end
module Build : sig
# 1 "src/build.mli"
(** The build arrow *)

open! Import

type ('a, 'b) t

val arr : ('a -> 'b) -> ('a, 'b) t

val return : 'a -> (unit, 'a) t

module Vspec : sig
  type 'a t = T : Path.t * 'a Vfile_kind.t -> 'a t
end

val store_vfile : 'a Vspec.t -> ('a, Action.t) t

module O : sig
  val ( >>> ) : ('a, 'b) t -> ('b, 'c) t -> ('a, 'c) t
  val ( ^>> ) : ('a -> 'b) -> ('b, 'c) t -> ('a, 'c) t
  val ( >>^ ) : ('a, 'b) t -> ('b -> 'c) -> ('a, 'c) t
  val ( *** ) : ('a, 'b) t -> ('c, 'd) t -> ('a * 'c, 'b * 'd) t
  val ( &&& ) : ('a, 'b) t -> ('a, 'c) t -> ('a, 'b * 'c) t
end

val first  : ('a, 'b) t -> ('a * 'c, 'b * 'c) t
val second : ('a, 'b) t -> ('c * 'a, 'c * 'b) t

(** Same as [O.(&&&)]. Sends the input to both argument arrows and combine their output.

    The default definition may be overridden with a more efficient version if desired. *)
val fanout  : ('a, 'b) t -> ('a, 'c) t -> ('a, 'b * 'c) t
val fanout3 : ('a, 'b) t -> ('a, 'c) t -> ('a, 'd) t ->  ('a, 'b * 'c * 'd) t
val fanout4 : ('a, 'b) t -> ('a, 'c) t -> ('a, 'd) t -> ('a, 'e) t -> ('a, 'b * 'c * 'd * 'e) t

val all : ('a, 'b) t list -> ('a, 'b list) t

(* CR-someday diml: this API is not great, what about:

   {[
     module Action_with_deps : sig
       type t
       val add_file_dependency : t -> Path.t -> t
     end

     (** Same as [t >>> arr (fun x -> Action_with_deps.add_file_dependency x p)]
         but better as [p] is statically known *)
     val record_dependency
       :  Path.t
       -> ('a, Action_with_deps.t) t
       -> ('a, Action_with_deps.t) t
   ]}
*)
(** [path p] records [p] as a file that is read by the action produced by the
    build arrow. *)
val path  : Path.t -> ('a, 'a) t

val paths : Path.t list -> ('a, 'a) t
val path_set : Path.Set.t -> ('a, 'a) t

(** Evaluate a glob and record all the matched files as dependencies
    of the action produced by the build arrow. *)
val paths_glob : loc:Loc.t -> dir:Path.t -> Re.re -> ('a, Path.t list) t

(* CR-someday diml: rename to [source_files_recursively_in] *)
(** Compute the set of source of all files present in the sub-tree
    starting at [dir] and record them as dependencies. *)
val files_recursively_in
  :  dir:Path.t
  -> file_tree:File_tree.t
  -> ('a, Path.Set.t) t

(** Record dynamic dependencies *)
val dyn_paths : ('a, Path.t list) t -> ('a, 'a) t
val dyn_path_set : ('a, Path.Set.t) t -> ('a, 'a) t

val vpath : 'a Vspec.t  -> (unit, 'a) t

(** [catch t ~on_error] evaluates to [on_error exn] if exception [exn] is
    raised during the evaluation of [t]. *)
val catch : ('a, 'b) t -> on_error:(exn -> 'b) -> ('a, 'b) t

(** [contents path] returns an arrow that when run will return the contents of
    the file at [path]. *)
val contents : Path.t -> ('a, string) t

(** [lines_of path] returns an arrow that when run will return the contents of
    the file at [path] as a list of lines. *)
val lines_of : Path.t -> ('a, string list) t

(** [strings path] is like [lines_of path] except each line is unescaped using
    the OCaml conventions. *)
val strings : Path.t -> ('a, string list) t

(** Load an S-expression from a file *)
val read_sexp : Path.t -> (unit, Sexp.Ast.t) t

(** Evaluates to [true] if the file is present on the file system or is the target of a
    rule. *)
val file_exists : Path.t -> ('a, bool)  t

(** [if_file_exists p ~then ~else] is an arrow that behaves like [then_] if [file_exists
    p] evaluates to [true], and [else_] otherwise. *)
val if_file_exists : Path.t -> then_:('a, 'b) t -> else_:('a, 'b) t -> ('a, 'b) t

(** [file_exists_opt p t] is:

    {[
      if_file_exists p ~then_:(t >>^ fun x -> Some x) ~else_:(arr (fun _ -> None))
    ]}
*)
val file_exists_opt : Path.t -> ('a, 'b) t -> ('a, 'b option) t

(** Always fail when executed. We pass a function rather than an
    exception to get a proper backtrace *)
val fail : ?targets:Path.t list -> fail -> (_, _) t

val of_result
  :  ?targets:Path.t list
  -> ('a, 'b) t Or_exn.t
  -> ('a, 'b) t

val of_result_map
  : ?targets:Path.t list
  -> 'a Or_exn.t
  -> f:('a -> ('b, 'c) t)
  -> ('b, 'c) t

(** [memoize name t] is an arrow that behaves like [t] except that its
    result is computed only once. *)
val memoize : string -> (unit, 'a) t -> (unit, 'a) t

val run
  :  context:Context.t
  -> ?dir:Path.t (* default: [context.build_dir] *)
  -> ?stdout_to:Path.t
  -> ?extra_targets:Path.t list
  -> Action.Prog.t
  -> 'a Arg_spec.t list
  -> ('a, Action.t) t

val action
  :  ?dir:Path.t
  -> targets:Path.t list
  -> Action.t
  -> (_, Action.t) t

val action_dyn
  :  ?dir:Path.t
  -> targets:Path.t list
  -> unit
  -> (Action.t, Action.t) t

(** Create a file with the given contents. *)
val write_file : Path.t -> string -> (unit, Action.t) t
val write_file_dyn : Path.t -> (string, Action.t) t

val copy : src:Path.t -> dst:Path.t -> (unit, Action.t) t
val copy_and_add_line_directive : src:Path.t -> dst:Path.t -> (unit, Action.t) t

val symlink : src:Path.t -> dst:Path.t -> (unit, Action.t) t

val create_file : Path.t -> (_, Action.t) t
val remove_tree : Path.t -> (_, Action.t) t
val mkdir : Path.t -> (_, Action.t) t

(** Merge a list of actions *)
val progn : ('a, Action.t) t list -> ('a, Action.t) t

type lib_dep_kind =
  | Optional
  | Required

val record_lib_deps
  :  kind:lib_dep_kind
  -> Jbuild.Lib_dep.t list
  -> ('a, 'a) t

type lib_deps = lib_dep_kind String_map.t

val record_lib_deps_simple : lib_deps -> ('a, 'a) t

(**/**)


module Repr : sig
  type ('a, 'b) t =
    | Arr : ('a -> 'b) -> ('a, 'b) t
    | Targets : Path.t list -> ('a, 'a) t
    | Store_vfile : 'a Vspec.t -> ('a, Action.t) t
    | Compose : ('a, 'b) t * ('b, 'c) t -> ('a, 'c) t
    | First : ('a, 'b) t -> ('a * 'c, 'b * 'c) t
    | Second : ('a, 'b) t -> ('c * 'a, 'c * 'b) t
    | Split : ('a, 'b) t * ('c, 'd) t -> ('a * 'c, 'b * 'd) t
    | Fanout : ('a, 'b) t * ('a, 'c) t -> ('a, 'b * 'c) t
    | Paths : Path.Set.t -> ('a, 'a) t
    | Paths_for_rule : Path.Set.t -> ('a, 'a) t
    | Paths_glob : glob_state ref -> ('a, Path.t list) t
    | If_file_exists : Path.t * ('a, 'b) if_file_exists_state ref -> ('a, 'b) t
    | Contents : Path.t -> ('a, string) t
    | Lines_of : Path.t -> ('a, string list) t
    | Vpath : 'a Vspec.t -> (unit, 'a) t
    | Dyn_paths : ('a, Path.Set.t) t -> ('a, 'a) t
    | Record_lib_deps : lib_deps -> ('a, 'a) t
    | Fail : fail -> (_, _) t
    | Memo : 'a memo -> (unit, 'a) t
    | Catch : ('a, 'b) t * (exn -> 'b) -> ('a, 'b) t

  and 'a memo =
    { name          : string
    ; t             : (unit, 'a) t
    ; mutable state : 'a memo_state
    }

  and 'a memo_state =
    | Unevaluated
    | Evaluating
    | Evaluated of 'a * Path.Set.t (* dynamic dependencies *)

  and ('a, 'b) if_file_exists_state =
    | Undecided of ('a, 'b) t * ('a, 'b) t
    | Decided   of bool * ('a, 'b) t

  and glob_state =
    | G_unevaluated of Loc.t * Path.t * Re.re
    | G_evaluated   of Path.t list

  val get_if_file_exists_exn : ('a, 'b) if_file_exists_state ref -> ('a, 'b) t
  val get_glob_result_exn : glob_state ref -> Path.t list
end

val repr : ('a, 'b) t -> ('a, 'b) Repr.t

val merge_lib_deps : lib_deps -> lib_deps -> lib_deps

(**/**)
val paths_for_rule : Path.Set.t -> ('a, 'a) t
# 18440 "boot.ml"
end = struct
# 1 "src/build.ml"
open Import

module Pset = Path.Set

module Vspec = struct
  type 'a t = T : Path.t * 'a Vfile_kind.t -> 'a t
end

type lib_dep_kind =
  | Optional
  | Required
type lib_deps = lib_dep_kind String_map.t

let merge_lib_dep_kind a b =
  match a, b with
  | Optional, Optional -> Optional
  | _ -> Required

module Repr = struct
  type ('a, 'b) t =
    | Arr : ('a -> 'b) -> ('a, 'b) t
    | Targets : Path.t list -> ('a, 'a) t
    | Store_vfile : 'a Vspec.t -> ('a, Action.t) t
    | Compose : ('a, 'b) t * ('b, 'c) t -> ('a, 'c) t
    | First : ('a, 'b) t -> ('a * 'c, 'b * 'c) t
    | Second : ('a, 'b) t -> ('c * 'a, 'c * 'b) t
    | Split : ('a, 'b) t * ('c, 'd) t -> ('a * 'c, 'b * 'd) t
    | Fanout : ('a, 'b) t * ('a, 'c) t -> ('a, 'b * 'c) t
    | Paths : Pset.t -> ('a, 'a) t
    | Paths_for_rule : Path.Set.t -> ('a, 'a) t
    | Paths_glob : glob_state ref -> ('a, Path.t list) t
    (* The reference gets decided in Build_interpret.deps *)
    | If_file_exists : Path.t * ('a, 'b) if_file_exists_state ref -> ('a, 'b) t
    | Contents : Path.t -> ('a, string) t
    | Lines_of : Path.t -> ('a, string list) t
    | Vpath : 'a Vspec.t -> (unit, 'a) t
    | Dyn_paths : ('a, Path.Set.t) t -> ('a, 'a) t
    | Record_lib_deps : lib_deps -> ('a, 'a) t
    | Fail : fail -> (_, _) t
    | Memo : 'a memo -> (unit, 'a) t
    | Catch : ('a, 'b) t * (exn -> 'b) -> ('a, 'b) t

  and 'a memo =
    { name          : string
    ; t             : (unit, 'a) t
    ; mutable state : 'a memo_state
    }

  and 'a memo_state =
    | Unevaluated
    | Evaluating
    | Evaluated of 'a * Path.Set.t

  and ('a, 'b) if_file_exists_state =
    | Undecided of ('a, 'b) t * ('a, 'b) t
    | Decided   of bool * ('a, 'b) t

  and glob_state =
    | G_unevaluated of Loc.t * Path.t * Re.re
    | G_evaluated   of Path.t list

  let get_if_file_exists_exn state =
    match !state with
    | Decided (_, t) -> t
    | Undecided _ -> code_errorf "Build.get_if_file_exists_exn: got undecided"

  let get_glob_result_exn state =
    match !state with
    | G_evaluated l -> l
    | G_unevaluated _ -> code_errorf "Build.get_glob_result_exn: got unevaluated"
end
include Repr
let repr t = t

let merge_lib_deps a b =
  String_map.merge a b ~f:(fun _ a b ->
    match a, b with
    | None, None -> None
    | x, None | None, x -> x
    | Some a, Some b -> Some (merge_lib_dep_kind a b))

let arr f = Arr f
let return x = Arr (fun () -> x)

let record_lib_deps_simple lib_deps =
  Record_lib_deps lib_deps

let record_lib_deps ~kind lib_deps =
  Record_lib_deps
    (List.concat_map lib_deps ~f:(function
       | Jbuild.Lib_dep.Direct (_, s) -> [(s, kind)]
       | Select { choices; _ } ->
         List.concat_map choices ~f:(fun c ->
           String_set.to_list c.Jbuild.Lib_dep.required
           |> List.map ~f:(fun d -> (d, Optional))))
     |> String_map.of_list_reduce ~f:merge_lib_dep_kind)

module O = struct
  let ( >>> ) a b =
    match a, b with
    | Arr a, Arr b -> Arr (fun x -> (b (a x)))
    | _ -> Compose (a, b)

  let ( >>^ ) t f = t >>> arr f
  let ( ^>> ) f t = arr f >>> t

  let ( *** ) a b = Split (a, b)
  let ( &&& ) a b = Fanout (a, b)
end
open O

let first t = First t
let second t = Second t
let fanout a b = Fanout (a, b)
let fanout3 a b c =
  let open O in
  (a &&& (b &&& c))
  >>>
  arr (fun (a, (b, c)) -> (a, b, c))
let fanout4 a b c d =
  let open O in
  (a &&& (b &&& (c &&& d)))
  >>>
  arr (fun (a, (b, (c, d))) -> (a, b, c, d))

let rec all = function
  | [] -> arr (fun _ -> [])
  | t :: ts ->
    t &&& all ts
    >>>
    arr (fun (x, y) -> x :: y)

let path p = Paths (Pset.singleton p)
let paths ps = Paths (Pset.of_list ps)
let path_set ps = Paths ps
let paths_glob ~loc ~dir re = Paths_glob (ref (G_unevaluated (loc, dir, re)))
let vpath vp = Vpath vp
let dyn_paths t = Dyn_paths (t >>^ Path.Set.of_list)
let dyn_path_set t = Dyn_paths t
let paths_for_rule ps = Paths_for_rule ps

let catch t ~on_error = Catch (t, on_error)

let contents p = Contents p
let lines_of p = Lines_of p

let strings p =
  lines_of p
  >>^ fun l ->
  List.map l ~f:Scanf.unescaped

let read_sexp p =
  contents p
  >>^ fun s ->
  Usexp.parse_string s ~fname:(Path.to_string p) ~mode:Single

let if_file_exists p ~then_ ~else_ =
  If_file_exists (p, ref (Undecided (then_, else_)))

let file_exists p =
  if_file_exists p
    ~then_:(arr (fun _ -> true))
    ~else_:(arr (fun _ -> false))

let file_exists_opt p t =
  if_file_exists p
    ~then_:(t >>^ fun x -> Some x)
    ~else_:(arr (fun _ -> None))

let fail ?targets x =
  match targets with
  | None -> Fail x
  | Some l -> Targets l >>> Fail x

let of_result ?targets = function
  | Ok    x -> x
  | Error e -> fail ?targets { fail = fun () -> raise e }

let of_result_map ?targets res ~f =
  match res with
  | Ok    x -> f x
  | Error e -> fail ?targets { fail = fun () -> raise e }

let memoize name t =
  Memo { name; t; state = Unevaluated }

let files_recursively_in ~dir ~file_tree =
  let prefix_with, dir =
    match Path.extract_build_context_dir dir with
    | None -> (Path.root, dir)
    | Some (ctx_dir, src_dir) -> (ctx_dir, src_dir)
  in
  let paths = File_tree.files_recursively_in file_tree dir ~prefix_with in
  path_set paths >>^ fun _ -> paths

let store_vfile spec = Store_vfile spec

let get_prog = function
  | Ok p -> path p >>> arr (fun _ -> Ok p)
  | Error f ->
    arr (fun _ -> Error f)
    >>> dyn_paths (arr (function Error _ -> [] | Ok x -> [x]))

let prog_and_args ?(dir=Path.root) prog args =
  Paths (Arg_spec.add_deps args Pset.empty)
  >>>
  (get_prog prog &&&
   (arr (Arg_spec.expand ~dir args)
    >>>
    dyn_paths (arr (fun (_args, deps) -> Path.Set.to_list deps))
    >>>
    arr fst))

let run ~context ?(dir=context.Context.build_dir) ?stdout_to ?(extra_targets=[])
      prog args =
  let extra_targets =
    match stdout_to with
    | None -> extra_targets
    | Some fn -> fn :: extra_targets
  in
  let targets = Arg_spec.add_targets args extra_targets in
  prog_and_args ~dir prog args
  >>>
  Targets targets
  >>^ (fun (prog, args) ->
    let action : Action.t = Run (prog, args) in
    let action =
      match stdout_to with
      | None      -> action
      | Some path -> Redirect (Stdout, path, action)
    in
    Action.Chdir (dir, action))

let action ?dir ~targets action =
  Targets targets
  >>^ fun _ ->
  match dir with
  | None -> action
  | Some dir -> Action.Chdir (dir, action)

let action_dyn ?dir ~targets () =
  Targets targets
  >>^ fun action ->
  match dir with
  | None -> action
  | Some dir -> Action.Chdir (dir, action)

let write_file fn s =
  action ~targets:[fn] (Write_file (fn, s))

let write_file_dyn fn =
  Targets [fn]
  >>^ fun s ->
  Action.Write_file (fn, s)

let copy ~src ~dst =
  path src >>>
  action ~targets:[dst] (Copy (src, dst))

let copy_and_add_line_directive ~src ~dst =
  path src >>>
  action ~targets:[dst]
    (Copy_and_add_line_directive (src, dst))

let symlink ~src ~dst =
  path src >>>
  action ~targets:[dst] (Symlink (src, dst))

let create_file fn =
  action ~targets:[fn] (Redirect (Stdout, fn, Progn []))

let remove_tree dir =
  arr (fun _ -> Action.Remove_tree dir)

let mkdir dir =
  arr (fun _ -> Action.Mkdir dir)

let progn ts =
  all ts >>^ fun actions ->
  Action.Progn actions
# 18723 "boot.ml"
end
module Build_interpret : sig
# 1 "src/build_interpret.mli"
open! Import

module Target : sig
  type t =
    | Normal of Path.t
    | Vfile : _ Build.Vspec.t -> t

  val path : t -> Path.t
  val paths : t list -> Path.Set.t
end

module Rule : sig
  type t =
    { context  : Context.t option
    ; build    : (unit, Action.t) Build.t
    ; targets  : Target.t list
    ; sandbox  : bool
    ; mode     : Jbuild.Rule.Mode.t
    ; locks    : Path.t list
    ; loc      : Loc.t option
    ; (** Directory where all the targets are produced *)
      dir      : Path.t
    }

  val make
    :  ?sandbox:bool
    -> ?mode:Jbuild.Rule.Mode.t
    -> context:Context.t option
    -> ?locks:Path.t list
    -> ?loc:Loc.t
    -> (unit, Action.t) Build.t
    -> t
end

module Static_deps : sig
  type t =
    { rule_deps   : Path.Set.t
    ; action_deps : Path.Set.t
    }
end

(* must be called first *)
val static_deps
  :  (_, _) Build.t
  -> all_targets:(dir:Path.t -> Path.Set.t)
  -> file_tree:File_tree.t
  -> Static_deps.t

val lib_deps
  :  (_, _) Build.t
  -> Build.lib_deps

val targets
  :  (_, _) Build.t
  -> Target.t list
# 18782 "boot.ml"
end = struct
# 1 "src/build_interpret.ml"
open Import
open Build.Repr

module Pset = Path.Set
module Pmap = Path.Map
module Vspec = Build.Vspec

module Target = struct
  type t =
    | Normal of Path.t
    | Vfile : _ Vspec.t -> t

  let path = function
    | Normal p -> p
    | Vfile (Vspec.T (p, _)) -> p

  let paths ts =
    List.fold_left ts ~init:Pset.empty ~f:(fun acc t ->
      Pset.add acc (path t))
end

module Static_deps = struct
  type t =
    { rule_deps   : Path.Set.t
    ; action_deps : Path.Set.t
    }
end

type file_kind = Reg | Dir

let inspect_path file_tree path =
  match Path.drop_build_context path with
  | None ->
    if not (Path.exists path) then
      None
    else if Path.is_directory path then
      Some Dir
    else
      Some Reg
  | Some path ->
    match File_tree.find_dir file_tree path with
    | Some _ ->
      Some Dir
    | None ->
      if Path.is_root path then
        Some Dir
      else if File_tree.file_exists file_tree
                (Path.parent   path)
                (Path.basename path) then
        Some Reg
      else
        None

let static_deps t ~all_targets ~file_tree =
  let rec loop : type a b. (a, b) t -> Static_deps.t -> Static_deps.t = fun t acc ->
    match t with
    | Arr _ -> acc
    | Targets _ -> acc
    | Store_vfile _ -> acc
    | Compose (a, b) -> loop a (loop b acc)
    | First t -> loop t acc
    | Second t -> loop t acc
    | Split (a, b) -> loop a (loop b acc)
    | Fanout (a, b) -> loop a (loop b acc)
    | Paths fns -> { acc with action_deps = Pset.union fns acc.action_deps }
    | Paths_for_rule fns ->
      { acc with rule_deps = Pset.union fns acc.rule_deps }
    | Paths_glob state -> begin
        match !state with
        | G_evaluated l ->
          { acc with action_deps = Pset.union acc.action_deps (Pset.of_list l) }
        | G_unevaluated (loc, dir, re) ->
          let targets = all_targets ~dir in
          let result =
            Pset.filter targets ~f:(fun path ->
              Re.execp re (Path.basename path))
          in
          if Pset.is_empty result then begin
            match inspect_path file_tree dir with
            | None ->
              Loc.warn loc "Directory %s doesn't exist."
                (Path.to_string_maybe_quoted
                   (Path.drop_optional_build_context dir))
            | Some Reg ->
              Loc.warn loc "%s is not a directory."
                (Path.to_string_maybe_quoted
                   (Path.drop_optional_build_context dir))
            | Some Dir ->
              (* diml: we should probably warn in this case as well *)
              ()
          end;
          state := G_evaluated (Pset.to_list result);
          let action_deps = Pset.union result acc.action_deps in
          { acc with action_deps }
      end
    | If_file_exists (p, state) -> begin
        match !state with
        | Decided (_, t) -> loop t acc
        | Undecided (then_, else_) ->
          let dir = Path.parent p in
          let targets = all_targets ~dir in
          if Pset.mem targets p then begin
            state := Decided (true, then_);
            loop then_ acc
          end else begin
            state := Decided (false, else_);
            loop else_ acc
          end
      end
    | Dyn_paths t -> loop t acc
    | Vpath (Vspec.T (p, _)) -> { acc with rule_deps = Pset.add acc.rule_deps p }
    | Contents p -> { acc with rule_deps = Pset.add acc.rule_deps p }
    | Lines_of p -> { acc with rule_deps = Pset.add acc.rule_deps p }
    | Record_lib_deps _ -> acc
    | Fail _ -> acc
    | Memo m -> loop m.t acc
    | Catch (t, _) -> loop t acc
  in
  loop (Build.repr t) { rule_deps = Pset.empty; action_deps = Pset.empty }

let lib_deps =
  let rec loop : type a b. (a, b) t -> Build.lib_deps -> Build.lib_deps
    = fun t acc ->
      match t with
      | Arr _ -> acc
      | Targets _ -> acc
      | Store_vfile _ -> acc
      | Compose (a, b) -> loop a (loop b acc)
      | First t -> loop t acc
      | Second t -> loop t acc
      | Split (a, b) -> loop a (loop b acc)
      | Fanout (a, b) -> loop a (loop b acc)
      | Paths _ -> acc
      | Paths_for_rule _ -> acc
      | Vpath _ -> acc
      | Paths_glob _ -> acc
      | Dyn_paths t -> loop t acc
      | Contents _ -> acc
      | Lines_of _ -> acc
      | Record_lib_deps deps -> Build.merge_lib_deps deps acc
      | Fail _ -> acc
      | If_file_exists (_, state) ->
        loop (get_if_file_exists_exn state) acc
      | Memo m -> loop m.t acc
      | Catch (t, _) -> loop t acc
  in
  fun t -> loop (Build.repr t) String_map.empty

let targets =
  let rec loop : type a b. (a, b) t -> Target.t list -> Target.t list = fun t acc ->
    match t with
    | Arr _ -> acc
    | Targets targets ->
      List.fold_left targets ~init:acc ~f:(fun acc fn -> Target.Normal fn :: acc)
    | Store_vfile spec -> Vfile spec :: acc
    | Compose (a, b) -> loop a (loop b acc)
    | First t -> loop t acc
    | Second t -> loop t acc
    | Split (a, b) -> loop a (loop b acc)
    | Fanout (a, b) -> loop a (loop b acc)
    | Paths _ -> acc
    | Paths_for_rule _ -> acc
    | Vpath _ -> acc
    | Paths_glob _ -> acc
    | Dyn_paths t -> loop t acc
    | Contents _ -> acc
    | Lines_of _ -> acc
    | Record_lib_deps _ -> acc
    | Fail _ -> acc
    | If_file_exists (_, state) -> begin
        match !state with
        | Decided _ -> code_errorf "Build_interpret.targets got decided if_file_exists"
        | Undecided (a, b) ->
          match loop a [], loop b [] with
          | [], [] -> acc
          | _ ->
            code_errorf "Build_interpret.targets: cannot have targets \
                         under a [if_file_exists]"
      end
    | Memo m -> loop m.t acc
    | Catch (t, _) -> loop t acc
  in
  fun t -> loop (Build.repr t) []

module Rule = struct
  type t =
    { context  : Context.t option
    ; build    : (unit, Action.t) Build.t
    ; targets  : Target.t list
    ; sandbox  : bool
    ; mode     : Jbuild.Rule.Mode.t
    ; locks    : Path.t list
    ; loc      : Loc.t option
    ; dir      : Path.t
    }

  let make ?(sandbox=false) ?(mode=Jbuild.Rule.Mode.Not_a_rule_stanza)
        ~context ?(locks=[]) ?loc build =
    let targets = targets build in
    let dir =
      match targets with
      | [] ->
        invalid_arg "Build_interpret.Rule.make: rule has no targets"
      | x :: l ->
        let dir = Path.parent (Target.path x) in
        List.iter l ~f:(fun target ->
          let path = Target.path target in
          if Path.parent path <> dir then
            match loc with
            | None ->
              Sexp.code_error "rule has targets in different directories"
                [ "targets", Sexp.To_sexp.list Path.sexp_of_t
                               (List.map targets ~f:Target.path)
                ]
            | Some loc ->
              Loc.fail loc
                "Rule has targets in different directories.\nTargets:\n%s"
                (String.concat ~sep:"\n"
                   (List.map targets ~f:(fun t ->
                      sprintf "- %s"
                        (Target.path t |> Path.to_string_maybe_quoted)))));
        dir
    in
    { context
    ; build
    ; targets
    ; sandbox
    ; mode
    ; locks
    ; loc
    ; dir
    }
end
# 19018 "boot.ml"
end
module Build_system : sig
# 1 "src/build_system.mli"
(** Build rules *)

open! Import

type t

(** {1 Setup} *)

(** {2 Creation} *)

type hook =
  | Rule_started
  | Rule_completed

(** Create a new build system. [file_tree] represent the source
    tree. *)
val create
  :  contexts:Context.t list
  -> file_tree:File_tree.t
  -> hook:(hook -> unit)
  -> t

type extra_sub_directories_to_keep =
  | All
  | These of String_set.t

(** Set the rule generators callback. There must be one callback per
   build context name.

    Each callback is used to generate the rules for a given directory
    in the corresponding build context. It receive the directory for
    which to generate the rules and the splitted part of the path after
    the build context. It must return an additional list of
    sub-directories to keep. This is in addition to the ones that are
    present in the source tree and the ones that already contain rules.

    It is expected that [f] only generate rules whose targets are
   descendant of [dir]. *)
val set_rule_generators : t -> (dir:Path.t -> string list -> extra_sub_directories_to_keep) String_map.t -> unit

(** All other functions in this section must be called inside the rule generator
    callback. *)

(** {2 Primitive for rule generations} *)

(** Add a rule to the system. This function must be called from the [gen_rules]
    callback. All the target of the rule must be in the same directory.

    Assuming that [gen_rules ~dir:a] calls [add_rule r] where [r.dir] is [Some b], one of
    the following assumption must hold:

    - [a] and [b] are the same
    - [gen_rules ~dir:b] calls [load_dir ~dir:a]

    The call to [load_dir ~dir:a] from [gen_rules ~dir:b] declares a directory dependency
    from [b] to [a]. There must be no cyclic directory dependencies.
*)
val add_rule : t -> Build_interpret.Rule.t -> unit

(** [prefix_rules t prefix ~f] Runs [f] and adds [prefix] as a dependency to all
    the rules generated by [f] *)
val prefix_rules : t -> (unit, unit) Build.t -> f:(unit -> 'a) -> 'a

(** [eval_glob t ~dir re ~f] returns the list of files in [dir] that matches [re] to
    [f]. The list of files includes the list of targets. *)
val eval_glob : t -> dir:Path.t -> Re.re -> string list

(** Returns the set of targets in the given directory. *)
val targets_of : t -> dir:Path.t -> Path.Set.t

(** Load the rules for this directory. *)
val load_dir : t -> dir:Path.t -> unit

(** [on_load_dir ~dir ~f] remembers to run [f] when loading the rules for [dir]. *)
val on_load_dir : t -> dir:Path.t -> f:(unit -> unit) -> unit

(** Stamp file that depends on all files of [dir] with extension [ext]. *)
val stamp_file_for_files_of : t -> dir:Path.t -> ext:string -> Path.t

(** Sets the package this file is part of *)
val set_package : t -> Path.t -> Package.Name.t -> unit

(** Assuming [files] is the list of files in [_build/install] that
    belong to package [pkg], [package_deps t pkg files] is the set of
    direct package dependencies of [package]. *)
val package_deps
  :  t
  -> Package.Name.t
  -> Path.Set.t
  -> (unit, Package.Name.Set.t) Build.t

(** {2 Aliases} *)

module Alias : sig
  type build_system = t
  type t

  val pp : t Fmt.t

  val make : string -> dir:Path.t -> t

  val of_path : Path.t -> t

  (** The following always holds:

      {[
        make (name t) ~dir:(dir t) = t
      ]}
  *)
  val name : t -> string
  val dir  : t -> Path.t

  val fully_qualified_name : t -> Path.t

  val default     : dir:Path.t -> t
  val runtest     : dir:Path.t -> t
  val install     : dir:Path.t -> t
  val doc         : dir:Path.t -> t
  val private_doc : dir:Path.t -> t
  val lint        : dir:Path.t -> t

  (** Alias for all the files in [_build/install] that belong to this
      package *)
  val package_install : context:Context.t -> pkg:Package.Name.t -> t

  (** Return the underlying stamp file *)
  val stamp_file : t -> Path.t

  (** [dep t = Build.path (stamp_file t)] *)
  val dep : t -> ('a, 'a) Build.t

  (** Implements [(alias_rec ...)] in dependency specification *)
  val dep_rec
    :  t
    -> loc:Loc.t
    -> file_tree:File_tree.t
    -> (unit, unit) Build.t

  (** Implements [@alias] on the command line *)
  val dep_rec_multi_contexts
    :  dir:Path.t
    -> name:string
    -> file_tree:File_tree.t
    -> contexts:string list
    -> (unit, unit) Build.t

  (** [add_deps store alias ?dyn_deps deps] arrange things so that all
      [dyn_deps] and [deps] are built as part of the build of alias
      [alias]. *)
  val add_deps
    :  build_system
    -> t
    -> ?dyn_deps:(unit, Path.Set.t) Build.t
    -> Path.Set.t
    -> unit

  (** [add_action store alias ~stamp action] arrange things so that
      [action] is executed as part of the build of alias
      [alias]. [stamp] is any S-expression that is unique and
      persistent S-expression.
  *)
  val add_action
    :  build_system
    -> t
    -> context:Context.t
    -> ?locks:Path.t list
    -> stamp:Sexp.t
    -> (unit, Action.t) Build.t
    -> unit
end with type build_system := t

(** {1 Building} *)

(** ALl the functions in this section must be called outside the rule generator
    callback. *)

(** Do the actual build *)
val do_build
  :  t
  -> request:(unit, unit) Build.t
  -> unit Fiber.t

(** {2 Other queries} *)

(** File for the [(universe)] dependency. *)
val universe_file : Path.t

val is_target : t -> Path.t -> bool

(** Return all the library dependencies (as written by the user)
   needed to build this request *)
val all_lib_deps
  :  t
  -> request:(unit, unit) Build.t
  -> Build.lib_deps Path.Map.t

(** Return all the library dependencies required to build this
   request, by context name *)
val all_lib_deps_by_context
  :  t
  -> request:(unit, unit) Build.t
  -> Build.lib_deps String_map.t

(** List of all buildable targets *)
val all_targets : t -> Path.t list

(** Return the list of files that were created in the source tree and
    needs to be deleted *)
val files_in_source_tree_to_delete
  :  unit
  -> Path.t list

(** {2 Build rules} *)

(** A fully built rule *)
module Rule : sig
  module Id : sig
    type t
    val to_int : t -> int
    val compare : t -> t -> Ordering.t
  end

  type t =
    { id      : Id.t
    ; deps    : Path.Set.t
    ; targets : Path.Set.t
    ; context : Context.t option
    ; action  : Action.t
    }
end

(** Return the list of rules used to build the given targets. If
    [recursive] is [true], return all the rules needed to build the
    given targets and their transitive dependencies. *)
val build_rules
  :  ?recursive:bool (* default false *)
  -> t
  -> request:(unit, unit) Build.t
  -> Rule.t list Fiber.t

(** {1 Misc} *)

(** Dump various databases on disk *)
val finalize : t -> unit
# 19266 "boot.ml"
end = struct
# 1 "src/build_system.ml"
open Import
open Fiber.O

module Pset  = Path.Set
module Pmap  = Path.Map
module Vspec = Build.Vspec

(* Where we store stamp files for aliases *)
let alias_dir = Path.(relative build_dir) ".aliases"

(* Where we store stamp files for [stamp_file_for_files_of] *)
let misc_dir = Path.(relative build_dir) ".misc"

module Promoted_to_delete = struct
  let db = ref []

  let fn = "_build/.to-delete-in-source-tree"

  let add p = db := p :: !db

  let load () =
    if Sys.file_exists fn then
      Sexp.load ~fname:fn ~mode:Many
      |> List.map ~f:Path.t
    else
      []

  let dump () =
    let db = Pset.union (Pset.of_list !db) (Pset.of_list (load ())) in
    if Sys.file_exists "_build" then
      Io.write_file fn
        (String.concat ~sep:""
           (List.map (Pset.to_list db) ~f:(fun p ->
              Sexp.to_string (Path.sexp_of_t p) ^ "\n")))
end

let files_in_source_tree_to_delete () =
  Promoted_to_delete.load ()

module Dependency_path = struct
  type rule_info =
    { (* Reason why this rule was visited *)
      requested_file : Path.t
    ; (* All targets of the rule *)
      targets        : Pset.t
    }

  type t =
    { dependency_path : rule_info list
    ; (* Union of all [targets] fields in [dependency_path]. Depending on any of these
         means that there is a cycle. *)
      targets_seen    : Pset.t
    }

  let var = Fiber.Var.create ()

  let empty =
    { dependency_path = []
    ; targets_seen    = Pset.empty
    }

  let dependency_cycle last dep_path =
    let rec build_loop acc dep_path =
      match dep_path with
      | [] -> assert false
      |  { requested_file; targets } :: dep_path ->
        if Pset.mem targets last then
          last :: acc
        else
          build_loop (requested_file :: acc) dep_path
    in
    let loop = build_loop [last] dep_path in
    die "Dependency cycle between the following files:\n    %s"
      (String.concat ~sep:"\n--> "
         (List.map loop ~f:Path.to_string))

  let push requested_file ~targets ~f =
    Fiber.Var.get var >>= fun x ->
    let t = Option.value x ~default:empty in
    if Pset.mem t.targets_seen requested_file then
      dependency_cycle requested_file t.dependency_path;
    let dependency_path =
      { requested_file; targets } :: t.dependency_path
    in
    let t =
      { dependency_path
      ; targets_seen = Pset.union targets t.targets_seen
      }
    in
    let on_error exn =
      Dep_path.reraise exn (Path requested_file)
    in
    Fiber.Var.set var t (Fiber.with_error_handler f ~on_error)
end

module Exec_status = struct
  type rule_evaluation = (Action.t * Pset.t) Fiber.Future.t
  type rule_execution = unit Fiber.Future.t

  type eval_rule = unit -> (Action.t * Pset.t) Fiber.t
  type exec_rule = rule_evaluation -> unit Fiber.t

  module Evaluating_rule = struct
    type t =
      { rule_evaluation : rule_evaluation
      ; exec_rule       : exec_rule
      }
  end
  module Running = struct
    type t =
      { rule_evaluation : rule_evaluation
      ; rule_execution  : rule_execution
      }
  end
  module Not_started = struct
    type t =
      { eval_rule : eval_rule
      ; exec_rule : exec_rule
      }
  end
  type t =
    | Not_started     of Not_started.t
    | Evaluating_rule of Evaluating_rule.t
    | Running         of Running.t
end

let rule_loc ~loc ~dir =
  match loc with
  | Some loc -> loc
  | None ->
    Loc.in_file
      (Path.to_string
         (Path.drop_optional_build_context (Path.relative dir "jbuild")))

module Internal_rule = struct
  module Id : sig
    type t
    val to_int : t -> int
    val compare : t -> t -> Ordering.t
    val gen : unit -> t
    module Set : Set.S with type elt = t
    module Top_closure : Top_closure.S with type key := t
  end = struct
    module M = struct
      type t = int
      let compare (x : int) y = compare x y
    end
    include M
    module Set = Set.Make(M)
    module Top_closure = Top_closure.Make(Set)
    let to_int x = x

    let counter = ref 0
    let gen () =
      let n = !counter in
      counter := n + 1;
      n
  end

  type t =
    { id               : Id.t
    ; rule_deps        : Pset.t
    ; static_deps      : Pset.t
    ; targets          : Pset.t
    ; context          : Context.t option
    ; build            : (unit, Action.t) Build.t
    ; mode             : Jbuild.Rule.Mode.t
    ; loc              : Loc.t option
    ; dir              : Path.t
    ; mutable exec     : Exec_status.t
    }

  let compare a b = Id.compare a.id b.id

  let loc ~dir t = rule_loc ~dir ~loc:t.loc
end

module File_kind = struct
  type 'a t =
    | Ignore_contents : unit t
    | Sexp_file       : 'a Vfile_kind.t -> 'a t

  let eq : type a b. a t -> b t -> (a, b) eq option = fun a b ->
    match a, b with
    | Ignore_contents, Ignore_contents -> Some Eq
    | Sexp_file a    , Sexp_file b     -> Vfile_kind.eq a b
    | _                                -> None

  let eq_exn a b = Option.value_exn (eq a b)
end

module File_spec = struct
  type 'a t =
    { rule         : Internal_rule.t (* Rule which produces it *)
    ; mutable kind : 'a File_kind.t
    ; mutable data : 'a option
    }

  type packed = T : _ t -> packed

  let create rule kind =
    T { rule; kind; data = None }
end

module Alias0 = struct
  type t = { dir : Path.t; name : string }

  let pp fmt t = Path.pp fmt (Path.relative t.dir t.name)

  let suffix = "-" ^ String.make 32 '0'

  let of_path path =
    if not (Path.is_in_build_dir path) then
      die "Invalid alias!\nTried to reference alias %S"
        (Path.to_string_maybe_quoted path);
    { dir  = Path.parent path
    ; name = Path.basename path
    }

  let name t = t.name
  let dir  t = t.dir

  let fully_qualified_name t = Path.relative t.dir t.name

  let make name ~dir =
    assert (not (String.contains name '/'));
    { dir; name }

  let stamp_file t =
    Path.relative (Path.insert_after_build_dir_exn t.dir ".aliases") (t.name ^ suffix)

  let dep t = Build.path (stamp_file t)

  let is_standard = function
    | "runtest" | "install" | "doc" | "doc-private" | "lint" -> true
    | _ -> false

  open Build.O

  let dep_rec_internal ~name ~dir ~ctx_dir =
    File_tree.Dir.fold dir ~traverse_ignored_dirs:false ~init:(Build.return true)
      ~f:(fun dir acc ->
        let path = Path.append ctx_dir (File_tree.Dir.path dir) in
        let fn = stamp_file (make ~dir:path name) in
        acc
        >>>
        Build.if_file_exists fn
          ~then_:(Build.path fn >>^ fun _ -> false)
          ~else_:(Build.arr (fun x -> x)))

  let dep_rec t ~loc ~file_tree =
    let ctx_dir, src_dir = Path.extract_build_context_dir t.dir |> Option.value_exn in
    match File_tree.find_dir file_tree src_dir with
    | None -> Build.fail { fail = fun () ->
      Loc.fail loc "Don't know about directory %s!" (Path.to_string_maybe_quoted src_dir) }
    | Some dir ->
      dep_rec_internal ~name:t.name ~dir ~ctx_dir
      >>^ fun is_empty ->
      if is_empty && not (is_standard t.name) then
        Loc.fail loc "This alias is empty.\n\
                      Alias %S is not defined in %s or any of its descendants."
          t.name (Path.to_string_maybe_quoted src_dir)

  let dep_rec_multi_contexts ~dir:src_dir ~name ~file_tree ~contexts =
    match File_tree.find_dir file_tree src_dir with
    | None ->
      die "From the command line:\n\
           @{<error>Error@}: Don't know about directory %s!" (Path.to_string_maybe_quoted src_dir)
    | Some dir ->
      let open Build.O in
      Build.all (List.map contexts ~f:(fun ctx ->
        let ctx_dir = Path.(relative build_dir) ctx in
        dep_rec_internal ~name ~dir ~ctx_dir))
      >>^ fun is_empty_list ->
      let is_empty = List.for_all is_empty_list ~f:(fun x -> x) in
      if is_empty && not (is_standard name) then
        die "From the command line:\n\
             @{<error>Error@}: Alias %s is empty.\n\
             It is not defined in %s or any of its descendants."
          name (Path.to_string_maybe_quoted src_dir)

  let default     = make "DEFAULT"
  let runtest     = make "runtest"
  let install     = make "install"
  let doc         = make "doc"
  let private_doc = make "doc-private"
  let lint        = make "lint"

  let package_install ~(context : Context.t) ~pkg =
    make (sprintf ".%s-files" (Package.Name.to_string pkg))
      ~dir:context.build_dir
end

module Dir_status = struct
  type waiting_for_load_dir =
    { mutable lazy_generators : (unit -> unit) list }

  type collection_stage =
    | Loading
    | Pending of waiting_for_load_dir

  type alias_action =
    { stamp  : Digest.t
    ; action : (unit, Action.t) Build.t
    ; locks  : Path.t list
    ; context : Context.t
    }


  type alias =
    { mutable deps     : Pset.t
    ; mutable dyn_deps : (unit, Pset.t) Build.t
    ; mutable actions  : alias_action list
    }

  type rules_collector =
    { mutable rules   : Build_interpret.Rule.t list
    ; mutable aliases : alias String_map.t
    ; mutable stage   : collection_stage
    }

  type t =
    | Collecting_rules of rules_collector
    | Loaded  of Pset.t (* set of targets in the directory *)
    | Forward of Path.t (* Load this directory first       *)
    | Failed_to_load
end

module Files_of = struct
  type t =
    { files_by_ext   : Path.t list String_map.t
    ; dir_hash       : string
    ; mutable stamps : Path.t String_map.t
    }
end

type extra_sub_directories_to_keep =
  | All
  | These of String_set.t

type hook =
  | Rule_started
  | Rule_completed

type t =
  { (* File specification by targets *)
    files       : (Path.t, File_spec.packed) Hashtbl.t
  ; contexts    : Context.t String_map.t
  ; (* Table from target to digest of
       [(deps (filename + contents), targets (filename only), action)] *)
    trace       : (Path.t, Digest.t) Hashtbl.t
  ; file_tree   : File_tree.t
  ; mutable local_mkdirs : Path.Local.Set.t
  ; mutable dirs : (Path.t, Dir_status.t) Hashtbl.t
  ; mutable gen_rules :
      (dir:Path.t -> string list -> extra_sub_directories_to_keep) String_map.t
  ; mutable load_dir_stack : Path.t list
  ; (* Set of directories under _build that have at least one rule and
       all their ancestors. *)
    mutable build_dirs_to_keep : Path.Set.t
  ; files_of : (Path.t, Files_of.t) Hashtbl.t
  ; mutable prefix : (unit, unit) Build.t option
  ; hook : hook -> unit
  ; (* Package files are part of *)
    packages : (Path.t, Package.Name.t) Hashtbl.t
  }

let string_of_paths set =
  Pset.to_list set
  |> List.map ~f:(fun p -> sprintf "- %s"
                             (Path.to_string_maybe_quoted
                                (Path.drop_optional_build_context p)))
  |> String.concat ~sep:"\n"

let set_rule_generators t generators =
  assert (String_map.keys generators = String_map.keys t.contexts);
  t.gen_rules <- generators

let get_dir_status t ~dir =
  Hashtbl.find_or_add t.dirs dir ~f:(fun _ ->
    if Path.is_in_source_tree dir then
      Dir_status.Loaded (File_tree.files_of t.file_tree dir)
    else if dir = Path.build_dir then
      (* Not allowed to look here *)
      Dir_status.Loaded Pset.empty
    else if not (Path.is_local dir) then
      Dir_status.Loaded
        (match Path.readdir dir with
         | exception _ -> Path.Set.empty
         | files ->
           Pset.of_list (List.map files ~f:(Path.relative dir)))
    else begin
      let (ctx, sub_dir) = Option.value_exn (Path.extract_build_context dir) in
      if ctx = ".aliases" then
        Forward (Path.(append build_dir) sub_dir)
      else if ctx <> "install" && not (String_map.mem t.contexts ctx) then
        Dir_status.Loaded Pset.empty
      else
        Collecting_rules
          { rules   = []
          ; aliases = String_map.empty
          ; stage   = Pending { lazy_generators = [] }
          }
    end)

let entry_point t ~f =
  (match t.load_dir_stack with
   | [] ->
     ()
   | _ :: _ ->
     code_errorf
       "Build_system.entry_point: called inside the rule generator callback");
  f ()

module Target = Build_interpret.Target
module Pre_rule = Build_interpret.Rule

let get_file : type a. t -> Path.t -> a File_kind.t -> a File_spec.t = fun t fn kind ->
  match Hashtbl.find t.files fn with
  | None -> die "no rule found for %s" (Path.to_string fn)
  | Some (File_spec.T file) ->
    let Eq = File_kind.eq_exn kind file.kind in
    file

let vfile_to_string (type a) (module K : Vfile_kind.S with type t = a) fn x =
  K.to_string fn x

module Build_exec = struct
  open Build.Repr

  let exec bs t x =
    let rec exec
      : type a b. Pset.t ref -> (a, b) t -> a -> b = fun dyn_deps t x ->
      match t with
      | Arr f -> f x
      | Targets _ -> x
      | Store_vfile (Vspec.T (fn, kind)) ->
        let file = get_file bs fn (Sexp_file kind) in
        file.data <- Some x;
        Write_file (fn, vfile_to_string kind fn x)
      | Compose (a, b) ->
        exec dyn_deps a x |> exec dyn_deps b
      | First t ->
        let x, y = x in
        (exec dyn_deps t x, y)
      | Second t ->
        let x, y = x in
        (x, exec dyn_deps t y)
      | Split (a, b) ->
        let x, y = x in
        let x = exec dyn_deps a x in
        let y = exec dyn_deps b y in
        (x, y)
      | Fanout (a, b) ->
        let a = exec dyn_deps a x in
        let b = exec dyn_deps b x in
        (a, b)
      | Paths _ -> x
      | Paths_for_rule _ -> x
      | Paths_glob state -> get_glob_result_exn state
      | Contents p -> Io.read_file (Path.to_string p)
      | Lines_of p -> Io.lines_of_file (Path.to_string p)
      | Vpath (Vspec.T (fn, kind)) ->
        let file : b File_spec.t = get_file bs fn (Sexp_file kind) in
        Option.value_exn file.data
      | Dyn_paths t ->
        let fns = exec dyn_deps t x in
        dyn_deps := Pset.union !dyn_deps fns;
        x
      | Record_lib_deps _ -> x
      | Fail { fail } -> fail ()
      | If_file_exists (_, state) ->
        exec dyn_deps (get_if_file_exists_exn state) x
      | Catch (t, on_error) -> begin
          try
            exec dyn_deps t x
          with exn ->
            on_error exn
        end
      | Memo m ->
        match m.state with
        | Evaluated (x, deps) ->
          dyn_deps := Pset.union !dyn_deps deps;
          x
        | Evaluating ->
          die "Dependency cycle evaluating memoized build arrow %s" m.name
        | Unevaluated ->
          m.state <- Evaluating;
          let dyn_deps' = ref Pset.empty in
          match exec dyn_deps' m.t x with
          | x ->
            m.state <- Evaluated (x, !dyn_deps');
            dyn_deps := Pset.union !dyn_deps !dyn_deps';
            x
          | exception exn ->
            m.state <- Unevaluated;
            reraise exn
    in
    let dyn_deps = ref Pset.empty in
    let action = exec dyn_deps (Build.repr t) x in
    (action, !dyn_deps)

  let exec_nop bs t x =
    snd (exec bs (Build.O.(>>^) t (fun () -> Action.Progn [])) x)
end

(* [copy_source] is [true] for rules copying files from the source directory *)
let add_spec t fn spec ~copy_source =
  match Hashtbl.find t.files fn with
  | None ->
    Hashtbl.add t.files fn spec
  | Some (File_spec.T { rule; _ }) ->
    match copy_source, rule.mode with
    | true, (Standard | Not_a_rule_stanza) ->
      Loc.warn (Internal_rule.loc rule ~dir:(Path.parent fn))
        "File %s is both generated by a rule and present in the source tree.\n\
         As a result, the rule is currently ignored, however this will become an error \
         in the future.\n\
         %t"
        (maybe_quoted (Path.basename fn))
        (fun ppf ->
           match rule.mode with
           | Not_a_rule_stanza ->
             Format.fprintf ppf "Delete file %s to get rid of this warning."
               (Path.to_string_maybe_quoted (Path.drop_optional_build_context fn))
           | Standard ->
             Format.fprintf ppf
               "To keep the current behavior and get rid of this warning, add a field \
                (fallback) to the rule."
           | _ -> assert false);
      Hashtbl.add t.files fn spec
    | _ ->
      let (File_spec.T { rule = rule2; _ }) = spec in
      let string_of_loc = function
        | None -> "<internal location>"
        | Some { Loc.start; _ } ->
          start.pos_fname ^ ":" ^ string_of_int start.pos_lnum
      in
      die "Multiple rules generated for %s:\n\
           - %s\n\
           - %s"
        (Path.to_string_maybe_quoted fn)
        (if copy_source then
           "<internal copy rule>"
         else
           string_of_loc rule.loc)
        (string_of_loc rule2.loc)

let create_file_specs t targets rule ~copy_source =
  List.iter targets ~f:(function
    | Target.Normal fn ->
      add_spec t fn (File_spec.create rule Ignore_contents) ~copy_source
    | Target.Vfile (Vspec.T (fn, kind)) ->
      add_spec t fn (File_spec.create rule (Sexp_file kind)) ~copy_source)

(* This contains the targets of the actions that are being executed. On exit, we need to
   delete them as they might contain garbage *)
let pending_targets = ref Pset.empty

let () =
  at_exit (fun () ->
    let fns = !pending_targets in
    pending_targets := Pset.empty;
    Pset.iter fns ~f:Path.unlink_no_err)

let clear_targets_digests_after_rule_execution targets =
  let missing =
    List.fold_left targets ~init:Pset.empty ~f:(fun acc fn ->
      match Unix.lstat (Path.to_string fn) with
      | exception _ -> Pset.add acc fn
      | (_ : Unix.stats) ->
        Utils.Cached_digest.remove fn;
        acc)
  in
  if not (Pset.is_empty missing) then
    die "@{<error>Error@}: Rule failed to generate the following targets:\n%s"
      (string_of_paths missing)

let make_local_dirs t paths =
  Pset.iter paths ~f:(fun path ->
    match Path.kind path with
    | Local path ->
      if not (Path.Local.Set.mem t.local_mkdirs path) then begin
        Path.Local.mkdir_p path;
        t.local_mkdirs <- Path.Local.Set.add t.local_mkdirs path
      end
    | _ -> ())

let make_local_parent_dirs t paths ~map_path =
  Pset.iter paths ~f:(fun path ->
    match Path.kind (map_path path) with
    | Local path when not (Path.Local.is_root path) ->
      let parent = Path.Local.parent path in
      if not (Path.Local.Set.mem t.local_mkdirs parent) then begin
        Path.Local.mkdir_p parent;
        t.local_mkdirs <- Path.Local.Set.add t.local_mkdirs parent
      end
    | _ -> ())

let sandbox_dir = Path.of_string "_build/.sandbox"

let locks : (Path.t, Fiber.Mutex.t) Hashtbl.t = Hashtbl.create 32

let rec with_locks mutexes ~f =
  match mutexes with
  | [] -> f ()
  | m :: mutexes ->
    Fiber.Mutex.with_lock
      (Hashtbl.find_or_add locks m ~f:(fun _ -> Fiber.Mutex.create ()))
      (fun () -> with_locks mutexes ~f)

let remove_old_artifacts t ~dir ~subdirs_to_keep =
  if not (Path.is_in_build_dir dir) ||
     Hashtbl.mem t.files (Path.relative dir Config.jbuilder_keep_fname) then
    ()
  else
    match Path.readdir dir with
    | exception _ -> ()
    | files ->
      List.iter files ~f:(fun fn ->
        let path = Path.relative dir fn in
        match Unix.lstat (Path.to_string path) with
        | { st_kind = S_DIR; _ } -> begin
            match subdirs_to_keep with
            | All -> ()
            | These set ->
              if String_set.mem set fn ||
                 Pset.mem t.build_dirs_to_keep path then
                ()
              else
                Path.rm_rf path
          end
        | exception _ ->
          if not (Hashtbl.mem t.files path) then Path.unlink path
        | _ ->
          if not (Hashtbl.mem t.files path) then Path.unlink path)

let no_rule_found =
  let fail fn =
    die "No rule found for %s" (Utils.describe_target fn)
  in
  fun t fn ->
    match Path.extract_build_context fn with
    | None -> fail fn
    | Some (ctx, _) ->
      if String_map.mem t.contexts ctx then
        fail fn
      else
        die "Trying to build %s but build context %s doesn't exist.%s"
          (Path.to_string_maybe_quoted fn)
          ctx
          (hint ctx (String_map.keys t.contexts))

let rec compile_rule t ?(copy_source=false) pre_rule =
  let { Pre_rule.
        context
      ; build
      ; targets = target_specs
      ; sandbox
      ; mode
      ; locks
      ; loc
      ; dir
      } =
    pre_rule
  in
  let targets = Target.paths target_specs in
  let { Build_interpret.Static_deps.
        rule_deps
      ; action_deps = static_deps
      } = Build_interpret.static_deps build ~all_targets:(targets_of t)
            ~file_tree:t.file_tree
  in

  let eval_rule () =
    t.hook Rule_started;
    wait_for_deps t rule_deps
    >>| fun () ->
    Build_exec.exec t build ()
  in
  let exec_rule (rule_evaluation : Exec_status.rule_evaluation) =
    Fiber.fork_and_join_unit
      (fun () ->
         wait_for_deps t static_deps)
      (fun () ->
         Fiber.Future.wait rule_evaluation >>= fun (action, dyn_deps) ->
         wait_for_deps t (Pset.diff dyn_deps static_deps)
         >>| fun () ->
         (action, dyn_deps))
    >>= fun (action, dyn_deps) ->
    make_local_parent_dirs t targets ~map_path:(fun x -> x);
    let all_deps = Pset.union static_deps dyn_deps in
    let all_deps_as_list = Pset.to_list all_deps in
    let targets_as_list  = Pset.to_list targets  in
    let hash =
      let trace =
        (List.map all_deps_as_list ~f:(fun fn ->
           (fn, Utils.Cached_digest.file fn)),
         targets_as_list,
         Option.map context ~f:(fun c -> c.name),
         action)
      in
      Digest.string (Marshal.to_string trace [])
    in
    let sandbox_dir =
      if sandbox then
        Some (Path.relative sandbox_dir (Digest.to_hex hash))
      else
        None
    in
    let deps_or_rule_changed =
      List.fold_left targets_as_list ~init:false ~f:(fun acc fn ->
        match Hashtbl.find t.trace fn with
        | None ->
          Hashtbl.add t.trace fn hash;
          true
        | Some prev_hash ->
          Hashtbl.replace t.trace ~key:fn ~data:hash;
          acc || prev_hash <> hash)
    in
    let targets_missing =
      List.exists targets_as_list ~f:(fun fn ->
        match Unix.lstat (Path.to_string fn) with
        | exception _ -> true
        | (_ : Unix.stats) -> false)
    in
    let force =
      !Clflags.force &&
      List.exists targets_as_list ~f:Path.is_alias_stamp_file
    in
    if deps_or_rule_changed || targets_missing || force then begin
      List.iter targets_as_list ~f:Path.unlink_no_err;
      pending_targets := Pset.union targets !pending_targets;
      let action =
        match sandbox_dir with
        | Some sandbox_dir ->
          Path.rm_rf sandbox_dir;
          let sandboxed path =
            if Path.is_local path then
              Path.append sandbox_dir path
            else
              path
          in
          make_local_parent_dirs t all_deps ~map_path:sandboxed;
          make_local_parent_dirs t targets  ~map_path:sandboxed;
          Action.sandbox action
            ~sandboxed
            ~deps:all_deps_as_list
            ~targets:targets_as_list
        | None ->
          action
      in
      make_local_dirs t (Action.chdirs action);
      with_locks locks ~f:(fun () ->
        Action.exec ~context ~targets action) >>| fun () ->
      Option.iter sandbox_dir ~f:Path.rm_rf;
      (* All went well, these targets are no longer pending *)
      pending_targets := Pset.diff !pending_targets targets;
      clear_targets_digests_after_rule_execution targets_as_list;
      (match mode with
       | Standard | Fallback | Not_a_rule_stanza | Ignore_source_files -> ()
       | Promote | Promote_but_delete_on_clean ->
         Pset.iter targets ~f:(fun path ->
           let in_source_tree = Option.value_exn (Path.drop_build_context path) in
           if mode = Promote_but_delete_on_clean then
             Promoted_to_delete.add in_source_tree;
           Io.copy_file
             ~src:(Path.to_string path)
             ~dst:(Path.to_string in_source_tree)));
      t.hook Rule_completed
    end else begin
      t.hook Rule_completed;
      Fiber.return ()
    end
  in
  let rule =
    { Internal_rule.
      id = Internal_rule.Id.gen ()
    ; static_deps
    ; rule_deps
    ; targets
    ; build
    ; context
    ; exec = Not_started { eval_rule; exec_rule }
    ; mode
    ; loc
    ; dir
    }
  in
  create_file_specs t target_specs rule ~copy_source

and setup_copy_rules t ~ctx_dir ~non_target_source_files =
  Pset.iter non_target_source_files ~f:(fun path ->
    let ctx_path = Path.append ctx_dir path in
    let build = Build.copy ~src:path ~dst:ctx_path in
    (* We temporarily allow overrides while setting up copy rules from
       the source directory so that artifact that are already present
       in the source directory are not re-computed.

       This allows to keep generated files in tarballs. Maybe we
       should allow it on a case-by-case basis though. *)
    compile_rule t (Pre_rule.make build ~context:None) ~copy_source:true)

and load_dir   t ~dir = ignore (load_dir_and_get_targets t ~dir : Pset.t)
and targets_of t ~dir =         load_dir_and_get_targets t ~dir

and load_dir_and_get_targets t ~dir =
  match get_dir_status t ~dir with
  | Failed_to_load -> raise Already_reported

  | Loaded targets -> targets

  | Forward dir' ->
    load_dir t ~dir:dir';
    begin match get_dir_status t ~dir with
    | Loaded targets -> targets
    | _ -> assert false
    end

  | Collecting_rules collector ->
    let lazy_generators =
      match collector.stage with
      | Loading ->
        die "recursive dependency between directories:\n    %s"
          (String.concat ~sep:"\n--> "
             (List.map t.load_dir_stack ~f:Utils.describe_target))
      | Pending { lazy_generators } ->
        collector.stage <- Loading;
        lazy_generators
    in

    collector.stage <- Loading;
    t.load_dir_stack <- dir :: t.load_dir_stack;

    try
      load_dir_step2_exn t ~dir ~collector ~lazy_generators
    with exn ->
      (match Hashtbl.find t.dirs dir with
       | Some (Loaded _) -> ()
       | _ ->
         (match t.load_dir_stack with
          | [] -> assert false
          | x :: l ->
            t.load_dir_stack <- l;
            assert (x = dir)));
      Hashtbl.replace t.dirs ~key:dir ~data:Failed_to_load;
      reraise exn

and load_dir_step2_exn t ~dir ~collector ~lazy_generators =
  List.iter lazy_generators ~f:(fun f -> f ());

  let context_name, sub_dir = Option.value_exn (Path.extract_build_context dir) in

  (* Load all the rules *)
  let extra_subdirs_to_keep =
    if context_name = "install" then
      These String_set.empty
    else
      let gen_rules = Option.value_exn (String_map.find t.gen_rules context_name) in
      gen_rules ~dir (Option.value_exn (Path.explode sub_dir))
  in
  let rules = collector.rules in

  (* Compute alias rules *)
  let alias_dir = Path.append (Path.relative alias_dir context_name) sub_dir in
  let alias_rules, alias_stamp_files =
    let open Build.O in
    String_map.foldi collector.aliases ~init:([], Pset.empty)
      ~f:(fun name { Dir_status. deps; dyn_deps; actions } (rules, alias_stamp_files) ->
        let base_path = Path.relative alias_dir name in
        let rules, deps =
          List.fold_left actions ~init:(rules, deps)
            ~f:(fun (rules, deps)
                 { Dir_status. stamp; action; locks ; context } ->
                 let path =
                   Path.extend_basename base_path
                     ~suffix:("-" ^ Digest.to_hex stamp)
                 in
                 let rule =
                   Pre_rule.make ~locks ~context:(Some context)
                     (Build.progn [ action; Build.create_file path ])
                 in
                 (rule :: rules, Pset.add deps path))
        in
        let path = Path.extend_basename base_path ~suffix:Alias0.suffix in
        (Pre_rule.make
           ~context:None
           (Build.path_set deps >>>
            dyn_deps >>>
            Build.dyn_path_set (Build.arr (fun x -> x))
            >>^ (fun dyn_deps ->
              let deps = Pset.union deps dyn_deps in
              Action.with_stdout_to path
                (Action.digest_files (Pset.to_list deps)))
            >>>
            Build.action_dyn () ~targets:[path])
         :: rules,
         Pset.add alias_stamp_files path))
  in
  Hashtbl.replace t.dirs ~key:alias_dir ~data:(Loaded alias_stamp_files);

  (* Compute the set of targets and the set of source files that must not be copied *)
  let user_rule_targets, source_files_to_ignore =
    List.fold_left rules ~init:(Pset.empty, Pset.empty)
      ~f:(fun (acc_targets, acc_ignored) { Pre_rule.targets; mode; _ } ->
        let targets = Build_interpret.Target.paths targets in
        (Pset.union targets acc_targets,
         match mode with
         | Promote | Promote_but_delete_on_clean | Ignore_source_files ->
           Pset.union targets acc_ignored
         | _ ->
           acc_ignored))
  in
  let source_files_to_ignore =
    Pset.map source_files_to_ignore ~f:(fun p ->
      Option.value_exn (Path.drop_build_context p))
  in

  (* Take into account the source files *)
  let targets, to_copy, subdirs_to_keep =
    match context_name with
    | "install" ->
      (user_rule_targets,
       None,
       String_set.empty)
    | ctx_name ->
      (* This condition is [true] because of [get_dir_status] *)
      assert (String_map.mem t.contexts ctx_name);
      let files, subdirs =
        match File_tree.find_dir t.file_tree sub_dir with
        | None -> (Pset.empty, String_set.empty)
        | Some dir ->
          (File_tree.Dir.file_paths    dir,
           File_tree.Dir.sub_dir_names dir)
      in
      let files = Pset.diff files source_files_to_ignore in
      if Pset.is_empty files then
        (user_rule_targets, None, subdirs)
      else
        let ctx_path = Path.(relative build_dir) context_name in
        (Pset.union user_rule_targets
           (Pset.map files ~f:(Path.append ctx_path)),
         Some (ctx_path, files),
         subdirs)
  in
  let subdirs_to_keep =
    match extra_subdirs_to_keep with
    | All -> All
    | These set -> These (String_set.union subdirs_to_keep set)
  in

  (* Filter out fallback rules *)
  let rules =
    match to_copy with
    | None ->
      (* If there are no source files to copy, fallback rules are
         automatically kept *)
      rules
    | Some (_, to_copy) ->
      List.filter rules ~f:(fun (rule : Build_interpret.Rule.t) ->
        match rule.mode with
        | Standard | Promote | Promote_but_delete_on_clean
        | Not_a_rule_stanza | Ignore_source_files -> true
        | Fallback ->
          let source_files_for_targtes =
            List.fold_left rule.targets ~init:Pset.empty
              ~f:(fun acc target ->
                Pset.add acc
                  (Build_interpret.Target.path target
                   |> Path.drop_build_context
                   (* All targets are in [dir] and we know it
                      correspond to a directory of a build context
                      since there are source files to copy, so this
                      call can't fail. *)
                   |> Option.value_exn))
          in
          if Pset.is_subset source_files_for_targtes ~of_:to_copy then
            (* All targets are present *)
            false
          else begin
            if Pset.is_empty (Pset.inter source_files_for_targtes to_copy) then
              (* No target is present *)
              true
            else begin
              let absent_targets =
                Pset.diff source_files_for_targtes to_copy
              in
              let present_targets =
                Pset.diff source_files_for_targtes absent_targets
              in
              Loc.fail
                (rule_loc ~loc:rule.loc
                   ~dir:(Path.drop_optional_build_context dir))
                "\
Some of the targets of this fallback rule are present in the source tree,
and some are not. This is not allowed. Either none of the targets must
be present in the source tree, either they must all be.

The following targets are present:
%s

The following targets are not:
%s
"
                (string_of_paths present_targets)
                (string_of_paths absent_targets)
            end
          end)
  in

  (* Set the directory status to loaded *)
  Hashtbl.replace t.dirs ~key:dir ~data:(Loaded targets);
  (match t.load_dir_stack with
   | [] -> assert false
   | x :: l ->
     t.load_dir_stack <- l;
     assert (x = dir));

  (* Compile the rules and cleanup stale artifacts *)
  List.iter rules ~f:(compile_rule t ~copy_source:false);
  Option.iter to_copy ~f:(fun (ctx_dir, source_files) ->
    setup_copy_rules t ~ctx_dir ~non_target_source_files:source_files);
  remove_old_artifacts t ~dir ~subdirs_to_keep;

  List.iter alias_rules ~f:(compile_rule t ~copy_source:false);
  remove_old_artifacts t ~dir:alias_dir ~subdirs_to_keep;

  targets

and wait_for_file t fn =
  match Hashtbl.find t.files fn with
  | Some file -> wait_for_file_found fn file
  | None ->
    let dir = Path.parent fn in
    if Path.is_in_build_dir dir then begin
      load_dir t ~dir;
      match Hashtbl.find t.files fn with
      | Some file -> wait_for_file_found fn file
      | None -> no_rule_found t fn
    end else if Path.exists fn then
      Fiber.return ()
    else
      die "File unavailable: %s" (Path.to_string_maybe_quoted fn)

and wait_for_file_found fn (File_spec.T file) =
  Dependency_path.push fn ~targets:file.rule.targets ~f:(fun () ->
    match file.rule.exec with
    | Not_started { eval_rule; exec_rule } ->
      Fiber.fork eval_rule
      >>= fun rule_evaluation ->
      Fiber.fork (fun () -> exec_rule rule_evaluation)
      >>= fun rule_execution ->
      file.rule.exec <-
        Running { rule_evaluation
                ; rule_execution
                };
      Fiber.Future.wait rule_execution
    | Running { rule_execution; _ } ->
      Fiber.Future.wait rule_execution
    | Evaluating_rule { rule_evaluation; exec_rule } ->
      Fiber.fork (fun () -> exec_rule rule_evaluation)
      >>= fun rule_execution ->
      file.rule.exec <-
        Running { rule_evaluation
                ; rule_execution
                };
      Fiber.Future.wait rule_execution)

and wait_for_deps t deps =
  Fiber.parallel_iter (Pset.to_list deps) ~f:(wait_for_file t)

let stamp_file_for_files_of t ~dir ~ext =
  let files_of_dir =
    Hashtbl.find_or_add t.files_of dir ~f:(fun dir ->
      let files_by_ext =
        targets_of t ~dir
        |> Path.Set.to_list
        |> List.map ~f:(fun fn -> Filename.extension (Path.to_string fn), fn)
        |> String_map.of_list_multi
      in
      { files_by_ext
      ; dir_hash = Path.to_string dir |> Digest.string |> Digest.to_hex
      ; stamps = String_map.empty
      })
  in
  match String_map.find files_of_dir.stamps ext with
  | Some fn -> fn
  | None ->
    let stamp_file = Path.relative misc_dir (files_of_dir.dir_hash ^ ext) in
    let files =
      Option.value
        (String_map.find files_of_dir.files_by_ext ext)
        ~default:[]
    in
    compile_rule t
      (let open Build.O in
       Pre_rule.make
         ~context:None
         (Build.paths files >>>
          Build.action ~targets:[stamp_file]
            (Action.with_stdout_to stamp_file
               (Action.digest_files files))));
    files_of_dir.stamps <- String_map.add files_of_dir.stamps ext stamp_file;
    stamp_file

module Trace = struct
  type t = (Path.t, Digest.t) Hashtbl.t

  let file = "_build/.db"

  let dump (trace : t) =
    let sexp =
      Sexp.List (
        Hashtbl.foldi trace ~init:Pmap.empty ~f:(fun key data acc ->
          Pmap.add acc key data)
        |> Path.Map.to_list
        |> List.map ~f:(fun (path, hash) ->
               Sexp.List [ Path.sexp_of_t path;
                           Atom (Sexp.Atom.of_digest hash) ]))
    in
    if Sys.file_exists "_build" then
      Io.write_file file (Sexp.to_string sexp)

  let load () =
    let trace = Hashtbl.create 1024 in
    if Sys.file_exists file then begin
      let sexp = Sexp.load ~fname:file ~mode:Single in
      let bindings =
        let open Sexp.Of_sexp in
        list (pair Path.t (fun s -> Digest.from_hex (string s))) sexp
      in
      List.iter bindings ~f:(fun (path, hash) ->
        Hashtbl.add trace path hash);
    end;
    trace
end

let all_targets t =
  String_map.iter t.contexts ~f:(fun ctx ->
    File_tree.fold t.file_tree ~traverse_ignored_dirs:true ~init:()
      ~f:(fun dir () ->
        load_dir t
          ~dir:(Path.append ctx.Context.build_dir (File_tree.Dir.path dir))));
  Hashtbl.foldi t.files ~init:[] ~f:(fun key _ acc -> key :: acc)

let finalize t =
  (* Promotion must be handled before dumping the digest cache, as it
     might delete some entries. *)
  Action.Promotion.finalize ();
  Promoted_to_delete.dump ();
  Utils.Cached_digest.dump ();
  Trace.dump t.trace

let create ~contexts ~file_tree ~hook =
  Utils.Cached_digest.load ();
  let contexts =
    List.map contexts ~f:(fun c -> (c.Context.name, c))
    |> String_map.of_list_exn
  in
  let t =
    { contexts
    ; files      = Hashtbl.create 1024
    ; packages   = Hashtbl.create 1024
    ; trace      = Trace.load ()
    ; local_mkdirs = Path.Local.Set.empty
    ; dirs       = Hashtbl.create 1024
    ; load_dir_stack = []
    ; file_tree
    ; gen_rules = String_map.map contexts ~f:(fun _ ~dir:_ ->
        die "gen_rules called too early")
    ; build_dirs_to_keep = Pset.empty
    ; files_of = Hashtbl.create 1024
    ; prefix = None
    ; hook
    }
  in
  at_exit (fun () -> finalize t);
  t

let eval_request t ~request ~process_target =
  let { Build_interpret.Static_deps.
        rule_deps
      ; action_deps = static_deps
      } = Build_interpret.static_deps request ~all_targets:(targets_of t)
            ~file_tree:t.file_tree
  in

  let process_targets ts =
    Fiber.parallel_iter (Pset.to_list ts) ~f:process_target
  in

  Fiber.fork_and_join_unit
    (fun () -> process_targets static_deps)
    (fun () ->
       wait_for_deps t rule_deps
       >>= fun () ->
       let dyn_deps = Build_exec.exec_nop t request () in
       process_targets (Pset.diff dyn_deps static_deps))

let universe_file = Path.relative Path.build_dir ".universe-state"

let update_universe t =
  (* To workaround the fact that [mtime] is not precise enough on OSX *)
  Utils.Cached_digest.remove universe_file;
  let fname = Path.to_string universe_file in
  let n =
    if Sys.file_exists fname then
      Sexp.Of_sexp.int (Sexp.load ~mode:Single ~fname) + 1
    else
      0
  in
  make_local_dirs t (Pset.singleton Path.build_dir);
  Io.write_file fname (Sexp.to_string (Sexp.To_sexp.int n))

let do_build t ~request =
  entry_point t ~f:(fun () ->
    update_universe t;
    eval_request t ~request ~process_target:(wait_for_file t))

module Ir_set = Set.Make(Internal_rule)

let rules_for_files t paths =
  List.filter_map paths ~f:(fun path ->
    if Path.is_in_build_dir path then load_dir t ~dir:(Path.parent path);
    match Hashtbl.find t.files path with
    | None -> None
    | Some (File_spec.T { rule; _ }) -> Some rule)
  |> Ir_set.of_list
  |> Ir_set.to_list

let rules_for_targets t targets =
  match
    Internal_rule.Id.Top_closure.top_closure (rules_for_files t targets)
      ~key:(fun (r : Internal_rule.t) -> r.id)
      ~deps:(fun (r : Internal_rule.t) ->
        rules_for_files t
          (Pset.to_list
             (Pset.union
                r.static_deps
                r.rule_deps)))
  with
  | Ok l -> l
  | Error cycle ->
    die "dependency cycle detected:\n   %s"
      (List.map cycle ~f:(fun rule ->
         Path.to_string (Option.value_exn
                           (Pset.choose rule.Internal_rule.targets)))
       |> String.concat ~sep:"\n-> ")

let static_deps_of_request t request =
  let { Build_interpret.Static_deps.
        rule_deps
      ; action_deps
      } = Build_interpret.static_deps request ~all_targets:(targets_of t)
            ~file_tree:t.file_tree
  in
  Pset.to_list (Pset.union rule_deps action_deps)

let all_lib_deps t ~request =
  let targets = static_deps_of_request t request in
  List.fold_left (rules_for_targets t targets) ~init:Pmap.empty
    ~f:(fun acc (rule : Internal_rule.t) ->
      let deps = Build_interpret.lib_deps rule.build in
      if String_map.is_empty deps then
        acc
      else
        let deps =
          match Pmap.find acc rule.dir with
          | None -> deps
          | Some deps' -> Build.merge_lib_deps deps deps'
        in
        Pmap.add acc rule.dir deps)

let all_lib_deps_by_context t ~request =
  let targets = static_deps_of_request t request in
  let rules = rules_for_targets t targets in
  List.fold_left rules ~init:[] ~f:(fun acc (rule : Internal_rule.t) ->
    let deps = Build_interpret.lib_deps rule.build in
    if String_map.is_empty deps then
      acc
    else
      match Path.extract_build_context rule.dir with
      | None -> acc
      | Some (context, _) -> (context, deps) :: acc)
  |> String_map.of_list_multi
  |> String_map.map ~f:(function
    | [] -> String_map.empty
    | x :: l -> List.fold_left l ~init:x ~f:Build.merge_lib_deps)

module Rule = struct
  module Id = Internal_rule.Id

  type t =
    { id      : Id.t
    ; deps    : Path.Set.t
    ; targets : Path.Set.t
    ; context : Context.t option
    ; action  : Action.t
    }

  let compare a b = Id.compare a.id b.id
end

module Rule_set = Set.Make(Rule)

let rules_for_files rules paths =
  List.fold_left paths ~init:Rule_set.empty ~f:(fun acc path ->
    match Pmap.find rules path with
    | None -> acc
    | Some rule -> Rule_set.add acc rule)
  |> Rule_set.to_list

let build_rules_internal ?(recursive=false) t ~request =
  let rules_seen = ref Rule.Id.Set.empty in
  let rules = ref [] in
  let rec loop fn =
    let dir = Path.parent fn in
    if Path.is_in_build_dir dir then load_dir t ~dir;
    match Hashtbl.find t.files fn with
    | Some file ->
      file_found fn file
    | None ->
      Fiber.return ()
  and file_found fn (File_spec.T { rule = ir; _ }) =
    if Rule.Id.Set.mem !rules_seen ir.id then
      Fiber.return ()
    else begin
      rules_seen := Rule.Id.Set.add !rules_seen ir.id;
      (match ir.exec with
       | Running { rule_evaluation; _ }
       | Evaluating_rule { rule_evaluation; _ } ->
         Fiber.return rule_evaluation
       | Not_started { eval_rule; exec_rule } ->
         Fiber.fork (fun () ->
           Dependency_path.push fn ~targets:ir.targets ~f:eval_rule)
         >>| fun rule_evaluation ->
         ir.exec <- Evaluating_rule { rule_evaluation
                                    ; exec_rule
                                    };
         rule_evaluation)
      >>= fun rule_evaluation ->
      Fiber.fork (fun () ->
        Fiber.Future.wait rule_evaluation
        >>| fun (action, dyn_deps) ->
        { Rule.
          id      = ir.id
        ; deps    = Pset.union ir.static_deps dyn_deps
        ; targets = ir.targets
        ; context = ir.context
        ; action  = action
        })
      >>= fun rule ->
      rules := rule :: !rules;
      if not recursive then
        Fiber.return ()
      else
        Fiber.Future.wait rule >>= fun rule ->
        Fiber.parallel_iter (Pset.to_list rule.deps) ~f:loop
    end
  in
  let targets = ref Pset.empty in
  eval_request t ~request ~process_target:(fun fn ->
    targets := Pset.add !targets fn;
    loop fn)
  >>= fun () ->
  Fiber.all (List.map !rules ~f:Fiber.Future.wait)
  >>| fun rules ->
  let rules =
    List.fold_left rules ~init:Pmap.empty ~f:(fun acc (r : Rule.t) ->
      Pset.fold r.targets ~init:acc ~f:(fun fn acc ->
        Pmap.add acc fn r))
  in
  match
    Rule.Id.Top_closure.top_closure
      (rules_for_files rules (Pset.to_list !targets))
      ~key:(fun (r : Rule.t) -> r.id)
      ~deps:(fun (r : Rule.t) ->
        rules_for_files rules (Pset.to_list r.deps))
  with
  | Ok l -> l
  | Error cycle ->
    die "dependency cycle detected:\n   %s"
      (List.map cycle ~f:(fun rule ->
         Path.to_string (Option.value_exn (Pset.choose rule.Rule.targets)))
       |> String.concat ~sep:"\n-> ")

let build_rules ?recursive t ~request =
  entry_point t ~f:(fun () ->
    build_rules_internal ?recursive t ~request)

let set_package t file package =
  Hashtbl.add t.packages file package

let package_deps t pkg files =
  let rules_seen = ref Rule.Id.Set.empty in
  let rec loop fn acc =
    match Hashtbl.find_all t.packages fn with
    | [] -> loop_deps fn acc
    | [p] when p = pkg -> loop_deps fn acc
    | pkgs ->
      List.fold_left pkgs ~init:acc ~f:add_package
  and add_package acc p =
    if p = pkg then
      acc
    else
      Package.Name.Set.add acc p
  and loop_deps fn acc =
    match Hashtbl.find t.files fn with
    | None -> acc
    | Some (File_spec.T { rule = ir; _ }) ->
      if Rule.Id.Set.mem !rules_seen ir.id then
        acc
      else begin
        rules_seen := Rule.Id.Set.add !rules_seen ir.id;
        let _, dyn_deps =
          match ir.exec with
          | Running { rule_evaluation; _ }
          | Evaluating_rule { rule_evaluation; _ } ->
            Option.value_exn (Fiber.Future.peek rule_evaluation)
          | Not_started _ -> assert false
        in
        Pset.fold (Pset.union ir.static_deps dyn_deps) ~init:acc ~f:loop
      end
  in
  let open Build.O in
  Build.paths_for_rule files >>^ fun () ->
  (* We know that at this point of execution, all the relevant ivars
     have been filled *)
  Pset.fold files ~init:Package.Name.Set.empty ~f:loop_deps

(* +-----------------------------------------------------------------+
   | Adding rules to the system                                      |
   +-----------------------------------------------------------------+ *)

let rec add_build_dir_to_keep t ~dir =
  if not (Pset.mem t.build_dirs_to_keep dir) then begin
    t.build_dirs_to_keep <- Pset.add t.build_dirs_to_keep dir;
    let dir = Path.parent dir in
    if dir <> Path.root then
      add_build_dir_to_keep t ~dir
  end

let get_collector t ~dir =
  match get_dir_status t ~dir with
  | Collecting_rules collector ->
    if collector.rules = [] && String_map.is_empty collector.aliases then
      add_build_dir_to_keep t ~dir;
    collector
  | Failed_to_load -> raise Already_reported
  | Loaded _ | Forward _ ->
    Sexp.code_error
      (if Path.is_in_source_tree dir then
         "Build_system.get_collector called on source directory"
       else if dir = Path.build_dir then
         "Build_system.get_collector called on _build"
       else if not (Path.is_local dir) then
         "Build_system.get_collector called on external directory"
       else
         "Build_system.get_collector called on closed directory")
      [ "dir", Path.sexp_of_t dir
      ]

let add_rule t (rule : Build_interpret.Rule.t) =
  let rule =
    match t.prefix with
    | None -> rule
    | Some prefix -> { rule with build = Build.O.(>>>) prefix rule.build } in
  let collector = get_collector t ~dir:rule.dir in
  collector.rules <- rule :: collector.rules

let prefix_rules' t prefix ~f =
  let old_prefix = t.prefix in
  t.prefix <- prefix;
  protectx () ~f ~finally:(fun () -> t.prefix <- old_prefix)

let prefix_rules t prefix ~f =
  begin match Build_interpret.targets prefix with
  | [] -> ()
  | targets ->
    Sexp.code_error "Build_system.prefix_rules' prefix contains targets"
      ["targets", Path.Set.sexp_of_t (Build_interpret.Target.paths targets)]
  end;
  let prefix =
    match t.prefix with
    | None -> prefix
    | Some p -> Build.O.(>>>) p prefix
  in
  prefix_rules' t (Some prefix) ~f

let on_load_dir t ~dir ~f =
  let collector = get_collector t ~dir in
  let current_prefix = t.prefix in
  let f () = prefix_rules' t current_prefix ~f in
  match collector.stage with
  | Loading -> f ()
  | Pending p ->
    let lazy_generators = p.lazy_generators in
    if lazy_generators = [] &&
       collector.rules = [] &&
       String_map.is_empty collector.aliases then
      add_build_dir_to_keep t ~dir;
    p.lazy_generators <- f :: lazy_generators

let eval_glob t ~dir re =
  let targets = targets_of t ~dir |> Pset.to_list |> List.map ~f:Path.basename in
  let files =
    match File_tree.find_dir t.file_tree dir with
    | None -> targets
    | Some d ->
      String_set.union (String_set.of_list targets) (File_tree.Dir.files d)
      |> String_set.to_list
  in
  List.filter files ~f:(Re.execp re)

module Alias = struct
  include Alias0

  let get_alias_def build_system t =
    let collector = get_collector build_system ~dir:t.dir in
    match String_map.find collector.aliases t.name with
    | None ->
      let x =
        { Dir_status.
          deps     = Pset.empty
        ; dyn_deps = Build.return Pset.empty
        ; actions  = []
        }
      in
      collector.aliases <- String_map.add collector.aliases t.name x;
      x
    | Some x -> x

  let add_deps build_system t ?dyn_deps deps =
    let def = get_alias_def build_system t in
    def.deps <- Pset.union def.deps deps;
    match dyn_deps with
    | None -> ()
    | Some build ->
      let open Build.O in
      def.dyn_deps <-
        Build.fanout def.dyn_deps build >>^ fun (a, b) ->
        Pset.union a b

  let add_action build_system t ~context ?(locks=[]) ~stamp action =
    let def = get_alias_def build_system t in
    def.actions <- { stamp = Digest.string (Sexp.to_string stamp)
                   ; action
                   ; locks
                   ; context
                   } :: def.actions
end

let is_target t file = Pset.mem (targets_of t ~dir:(Path.parent file)) file
# 20822 "boot.ml"
end
module Glob_lexer : sig
# 1 "src/glob_lexer.mli"
open Import

val parse_string : string -> (Jbuilder_re.Re.t, int * string) result
# 20829 "boot.ml"
end = struct
# 1 "src/glob_lexer.boot.ml"
(* We don't use globs in the bootstrap path to avoid avoid to include
   ocaml-re. This speeds up the bootstrap. *)

let parse_string _ = failwith "globs are not available during bootstrap"
# 20836 "boot.ml"
end
module Scope : sig
# 1 "src/scope.mli"
(** Scopes *)

(** Representation of a Scope. It contain a library database for all
    the private libraries in the scope. *)
type t

val root : t -> Path.t
val name : t -> string option
val info : t -> Jbuild.Scope_info.t

(** Return the library database associated to this scope *)
val libs : t -> Lib.DB.t

(** Scope databases *)
module DB : sig
  type scope = t

  type t

  (** Return the new scope database as well as the public libraries
      database *)
  val create
    :  scopes:Jbuild.Scope_info.t list
    -> context:string
    -> installed_libs:Lib.DB.t
    -> (Path.t * Jbuild.Library.t) list
    -> t * Lib.DB.t

  val find_by_dir  : t -> Path.t        -> scope
  val find_by_name : t -> string option -> scope
end with type scope := t
# 20871 "boot.ml"
end = struct
# 1 "src/scope.ml"
open Import

type t =
  { info : Jbuild.Scope_info.t
  ; db   : Lib.DB.t
  }

let root t = t.info.root
let name t = t.info.name
let info t = t.info
let libs t = t.db

module DB = struct
  type scope = t

  module Scope_name_map = Map.Make(Jbuild.Scope_info.Name)

  type t =
    { by_dir  : (Path.t, scope) Hashtbl.t
    ; by_name : scope Scope_name_map.t
    ; context : string
    }

  let find_by_dir t dir =
    let rec loop d =
      match Hashtbl.find t.by_dir d with
      | Some scope -> scope
      | None ->
        if Path.is_root d || not (Path.is_local d) then
          Sexp.code_error "Scope.DB.find_by_dir got an invalid path"
            [ "dir"    , Path.sexp_of_t dir
            ; "context", Sexp.To_sexp.string t.context
            ];
        let scope = loop (Path.parent d) in
        Hashtbl.add t.by_dir d scope;
        scope
    in
    loop dir

  let find_by_name t name =
    match Scope_name_map.find t.by_name name with
    | Some x -> x
    | None ->
      Sexp.code_error "Scope.DB.find_by_name"
        [ "name"   , Sexp.To_sexp.(option string) name
        ; "context", Sexp.To_sexp.string t.context
        ; "names",
          Sexp.To_sexp.(list (option string)) (Scope_name_map.keys t.by_name)
        ]

  let create ~scopes ~context ~installed_libs internal_libs =
    let scopes_info_by_name =
      List.map scopes ~f:(fun (scope : Jbuild.Scope_info.t) ->
        (scope.name, scope))
      |> Scope_name_map.of_list
      |> function
      | Ok x -> x
      | Error (_name, scope1, scope2) ->
        let to_sexp (scope : Jbuild.Scope_info.t) =
          Sexp.To_sexp.(pair (option string) Path.sexp_of_t)
            (scope.name, scope.root)
        in
        Sexp.code_error "Scope.DB.create got two scopes with the same name"
          [ "scope1", to_sexp scope1
          ; "scope2", to_sexp scope2
          ]
    in
    let libs_by_scope_name =
      List.map internal_libs ~f:(fun (dir, (lib : Jbuild.Library.t)) ->
        (lib.scope_name, (dir, lib)))
      |> Scope_name_map.of_list_multi
    in
    let by_name_cell = ref Scope_name_map.empty in
    let public_libs =
      let public_libs =
        List.filter_map internal_libs ~f:(fun (_dir, lib) ->
          match lib.public with
          | None -> None
          | Some p -> Some (p.name, lib.scope_name))
        |> String_map.of_list
        |> function
        | Ok x -> x
        | Error (name, _, _) ->
          match
            List.filter_map internal_libs ~f:(fun (_dir, lib) ->
              match lib.public with
              | None   -> None
              | Some p -> Option.some_if (name = p.name) lib.buildable.loc)
          with
          | [] | [_] -> assert false
          | loc1 :: loc2 :: _ ->
            die "Public library %S is defined twice:\n\
                 - %s\n\
                 - %s"
              name
              (Loc.to_file_colon_line loc1)
              (Loc.to_file_colon_line loc2)
      in
      Lib.DB.create ()
        ~parent:installed_libs
        ~resolve:(fun name ->
          match String_map.find public_libs name with
          | None -> Not_found
          | Some scope_name ->
            let scope =
              Option.value_exn (Scope_name_map.find !by_name_cell scope_name)
            in
            Redirect (Some scope.db, name))
        ~all:(fun () -> String_map.keys public_libs)
    in
    let by_name =
      Scope_name_map.merge scopes_info_by_name libs_by_scope_name
        ~f:(fun _name info libs ->
          let info = Option.value_exn info         in
          let libs = Option.value libs ~default:[] in
          let db =
            Lib.DB.create_from_library_stanzas libs ~parent:public_libs
          in
          Some { info; db })
    in
    by_name_cell := by_name;
    let by_dir = Hashtbl.create 1024 in
    Scope_name_map.iter by_name ~f:(fun scope ->
      Hashtbl.add by_dir scope.info.root scope);
    ({ by_name; by_dir; context }, public_libs)
end
# 21000 "boot.ml"
end
module Super_context : sig
# 1 "src/super_context.mli"
(** A augmanted context *)

(** A context augmented with: a lib-db, ...

    Super context are used for generating rules.
*)

open Import
open Jbuild

(** A directory with a jbuild *)
module Dir_with_jbuild : sig
  type t =
    { src_dir : Path.t
    ; ctx_dir : Path.t (** [_build/context-name/src_dir] *)
    ; stanzas : Stanzas.t
    ; scope   : Scope.t
    }
end

type t

val create
  :  context:Context.t
  -> ?host:t
  -> scopes:Scope_info.t list
  -> file_tree:File_tree.t
  -> packages:Package.t Package.Name.Map.t
  -> stanzas:(Path.t * Scope_info.t * Stanzas.t) list
  -> external_lib_deps_mode:bool
  -> build_system:Build_system.t
  -> t

val context   : t -> Context.t
val stanzas   : t -> Dir_with_jbuild.t list
val packages  : t -> Package.t Package.Name.Map.t
val libs_by_package : t -> (Package.t * Lib.Set.t) Package.Name.Map.t
val file_tree : t -> File_tree.t
val artifacts : t -> Artifacts.t
val stanzas_to_consider_for_install : t -> (Path.t * Scope.t * Stanza.t) list
val cxx_flags : t -> string list
val build_dir : t -> Path.t
val host : t -> t
val build_system : t -> Build_system.t

(** All public libraries of the workspace *)
val public_libs : t -> Lib.DB.t

(** Installed libraries that are not part of the workspace *)
val installed_libs : t -> Lib.DB.t

val find_scope_by_dir  : t -> Path.t        -> Scope.t
val find_scope_by_name : t -> string option -> Scope.t

val expand_vars
  :  t
  -> scope:Scope.t
  -> dir:Path.t
  -> ?extra_vars:Action.Var_expansion.t String_map.t
  -> String_with_vars.t
  -> string

val expand_and_eval_set
  :  t
  -> scope:Scope.t
  -> dir:Path.t
  -> ?extra_vars:Action.Var_expansion.t String_map.t
  -> Ordered_set_lang.Unexpanded.t
  -> standard:string list
  -> (unit, string list) Build.t

val prefix_rules
  : t
  -> (unit, unit) Build.t
  -> f:(unit -> 'a)
  -> 'a
val add_rule
  :  t
  -> ?sandbox:bool
  -> ?mode:Jbuild.Rule.Mode.t
  -> ?locks:Path.t list
  -> ?loc:Loc.t
  -> (unit, Action.t) Build.t
  -> unit
val add_rule_get_targets
  :  t
  -> ?sandbox:bool
  -> ?mode:Jbuild.Rule.Mode.t
  -> ?locks:Path.t list
  -> ?loc:Loc.t
  -> (unit, Action.t) Build.t
  -> Path.t list
val add_rules
  :  t
  -> ?sandbox:bool
  -> (unit, Action.t) Build.t list
  -> unit
val add_alias_deps
  :  t
  -> Build_system.Alias.t
  -> ?dyn_deps:(unit, Path.Set.t) Build.t
  -> Path.Set.t
  -> unit
val add_alias_action
  :  t
  -> Build_system.Alias.t
  -> ?locks:Path.t list
  -> stamp:Sexp.t
  -> (unit, Action.t) Build.t
  -> unit

(** See [Build_system for details] *)
val eval_glob : t -> dir:Path.t -> Re.re -> string list
val load_dir : t -> dir:Path.t -> unit
val on_load_dir : t -> dir:Path.t -> f:(unit -> unit) -> unit

val source_files : t -> src_path:Path.t -> String_set.t

(** [prog_spec t ?hint name] resolve a program. [name] is looked up in the
    workspace, if it is not found in the tree is is looked up in the PATH. If it
    is not found at all, the resulting [Prog_spec.t] will either return the
    resolved path or a record with details about the error and possibly a hint.

    [hint] should tell the user what to install when the program is not found.
*)
val resolve_program
  :  t
  -> ?hint:string
  -> string
  -> Action.Prog.t

module Libs : sig
  (** Make sure all rules produces by [f] record the library
      dependencies for [jbuilder external-lib-deps] and depend on the
      generation of the .merlin file. *)
  val with_lib_deps
    :  t
    -> Lib.Compile.t
    -> dir:Path.t
    -> f:(unit -> 'a)
    -> 'a

  (** Generate the rules for the [(select ...)] forms in library dependencies *)
  val gen_select_rules : t -> dir:Path.t -> Lib.Compile.t -> unit

  (** [file_deps t libs ~ext] returns a list of path dependencies for
      all the files with extension [ext] of libraries [libs]. *)
  val file_deps : t -> Lib.L.t -> ext:string -> Path.t list

  (** Setup the alias that depends on all files with a given extension
      for a library *)
  val setup_file_deps_alias
    :  t
    -> dir:Path.t
    -> ext:string
    -> Library.t
    -> Path.Set.t
    -> unit

  (** Setup an alias that depend on all files with the given extensions.

      To depend on this alias, use [~ext:"ext1-and-ext2-...-extn"]
  *)
  val setup_file_deps_group_alias
    :  t
    -> dir:Path.t
    -> exts:string list
    -> Library.t
    -> unit
end

(** Interpret dependencies written in jbuild files *)
module Deps : sig
  (** Evaluates to the actual list of dependencies, ignoring aliases *)
  val interpret
    :  t
    -> scope:Scope.t
    -> dir:Path.t
    -> Dep_conf.t list
    -> (unit, Path.t list) Build.t
end

(** Interpret action written in jbuild files *)
module Action : sig
  type targets =
    | Static of Path.t list
    | Infer
    | Alias (** This action is for an alias *)

  (** The arrow takes as input the list of actual dependencies *)
  val run
    :  t
    -> ?extra_vars:Action.Var_expansion.t String_map.t
    -> Action.Unexpanded.t
    -> dir:Path.t
    -> dep_kind:Build.lib_dep_kind
    -> targets:targets
    -> scope:Scope.t
    -> (Path.t list, Action.t) Build.t
end

module Pkg_version : sig
  val set : t -> Package.t -> (unit, string option) Build.t -> (unit, string option) Build.t
end

module Scope_key : sig
  val of_string : t -> string -> string * Lib.DB.t

  val to_string : string -> Scope_info.Name.t -> string
end
# 21214 "boot.ml"
end = struct
# 1 "src/super_context.ml"
open Import
open Jbuild

module A = Action
module Pset = Path.Set
module Alias = Build_system.Alias

module Dir_with_jbuild = struct
  type t =
    { src_dir : Path.t
    ; ctx_dir : Path.t
    ; stanzas : Stanzas.t
    ; scope   : Scope.t
    }
end

type t =
  { context                          : Context.t
  ; build_system                     : Build_system.t
  ; scopes                           : Scope.DB.t
  ; public_libs                      : Lib.DB.t
  ; installed_libs                   : Lib.DB.t
  ; stanzas                          : Dir_with_jbuild.t list
  ; packages                         : Package.t Package.Name.Map.t
  ; file_tree                        : File_tree.t
  ; artifacts                        : Artifacts.t
  ; stanzas_to_consider_for_install  : (Path.t * Scope.t * Stanza.t) list
  ; cxx_flags                        : string list
  ; vars                             : Action.Var_expansion.t String_map.t
  ; chdir                            : (Action.t, Action.t) Build.t
  ; host                             : t option
  ; libs_by_package : (Package.t * Lib.Set.t) Package.Name.Map.t
  }

let context t = t.context
let stanzas t = t.stanzas
let packages t = t.packages
let libs_by_package t = t.libs_by_package
let artifacts t = t.artifacts
let file_tree t = t.file_tree
let stanzas_to_consider_for_install t = t.stanzas_to_consider_for_install
let cxx_flags t = t.cxx_flags
let build_dir t = t.context.build_dir
let build_system t = t.build_system

let host t = Option.value t.host ~default:t

let public_libs    t = t.public_libs
let installed_libs t = t.installed_libs

let find_scope_by_dir  t dir  = Scope.DB.find_by_dir  t.scopes dir
let find_scope_by_name t name = Scope.DB.find_by_name t.scopes name

let expand_var_no_root t var = String_map.find t.vars var

let expand_vars t ~scope ~dir ?(extra_vars=String_map.empty) s =
  String_with_vars.expand s ~f:(fun _loc -> function
    | "ROOT" -> Some (Path.reach ~from:dir t.context.build_dir)
    | "SCOPE_ROOT" ->
      Some (Path.reach ~from:dir (Scope.root scope))
    | var ->
      Option.map ~f:(fun e -> Action.Var_expansion.to_string dir e)
        (match expand_var_no_root t var with
         | Some _ as x -> x
         | None -> String_map.find extra_vars var))

let resolve_program t ?hint bin =
  Artifacts.binary ?hint t.artifacts bin

let create
      ~(context:Context.t)
      ?host
      ~scopes
      ~file_tree
      ~packages
      ~stanzas
      ~external_lib_deps_mode
      ~build_system
  =
  let installed_libs =
    Lib.DB.create_from_findlib context.findlib ~external_lib_deps_mode
  in
  let internal_libs =
    List.concat_map stanzas ~f:(fun (dir, _, stanzas) ->
      let ctx_dir = Path.append context.build_dir dir in
      List.filter_map stanzas ~f:(fun stanza ->
        match (stanza : Stanza.t) with
        | Library lib -> Some (ctx_dir, lib)
        | _ -> None))
  in
  let scopes, public_libs =
    let scopes =
      List.map scopes ~f:(fun (scope : Scope_info.t) ->
        { scope with root = Path.append context.build_dir scope.root })
    in
    Scope.DB.create
      ~scopes
      ~context:context.name
      ~installed_libs
      internal_libs
  in
  let stanzas =
    List.map stanzas
      ~f:(fun (dir, scope, stanzas) ->
        let ctx_dir = Path.append context.build_dir dir in
        { Dir_with_jbuild.
          src_dir = dir
        ; ctx_dir
        ; stanzas
        ; scope = Scope.DB.find_by_name scopes scope.Scope_info.name
        })
  in
  let stanzas_to_consider_for_install =
    if not external_lib_deps_mode then
      List.concat_map stanzas ~f:(fun { ctx_dir; stanzas; scope; _ } ->
        List.filter_map stanzas ~f:(fun stanza ->
          let keep =
            match (stanza : Stanza.t) with
            | Library lib -> Lib.DB.available (Scope.libs scope) lib.name
            | Documentation _
            | Install _   -> true
            | _           -> false
          in
          Option.some_if keep (ctx_dir, scope, stanza)))
    else
      List.concat_map stanzas ~f:(fun { ctx_dir; stanzas; scope; _ } ->
        List.map stanzas ~f:(fun s -> (ctx_dir, scope, s)))
  in
  let artifacts =
    Artifacts.create context ~public_libs stanzas
      ~f:(fun (d : Dir_with_jbuild.t) -> d.stanzas)
  in
  let cxx_flags =
    List.filter context.ocamlc_cflags
      ~f:(fun s -> not (String.is_prefix s ~prefix:"-std="))
  in
  let vars =
    let ocamlopt =
      match context.ocamlopt with
      | None -> Path.relative context.ocaml_bin "ocamlopt"
      | Some p -> p
    in
    let open Action.Var_expansion in
    let make =
      match Bin.make with
      | None   -> Strings (["make"], Split)
      | Some p -> Paths ([p], Split)
    in
    let cflags = context.ocamlc_cflags in
    let strings l = Strings (l  , Split)  in
    let string  s = Strings ([s], Concat) in
    let path    p = Paths   ([p], Split)  in
    let vars =
      [ "-verbose"       , Strings ([] (*"-verbose";*), Concat)
      ; "CPP"            , strings (context.c_compiler :: cflags @ ["-E"])
      ; "PA_CPP"         , strings (context.c_compiler :: cflags
                                    @ ["-undef"; "-traditional";
                                       "-x"; "c"; "-E"])
      ; "CC"             , strings (context.c_compiler :: cflags)
      ; "CXX"            , strings (context.c_compiler :: cxx_flags)
      ; "ocaml_bin"      , path context.ocaml_bin
      ; "OCAML"          , path context.ocaml
      ; "OCAMLC"         , path context.ocamlc
      ; "OCAMLOPT"       , path ocamlopt
      ; "ocaml_version"  , string context.version_string
      ; "ocaml_where"    , string (Path.to_string context.stdlib_dir)
      ; "ARCH_SIXTYFOUR" , string (string_of_bool context.arch_sixtyfour)
      ; "MAKE"           , make
      ; "null"           , string (Path.to_string Config.dev_null)
      ; "ext_obj"        , string context.ext_obj
      ; "ext_asm"        , string context.ext_asm
      ; "ext_lib"        , string context.ext_lib
      ; "ext_dll"        , string context.ext_dll
      ; "ext_exe"        , string context.ext_exe
      ]
    in
    let vars =
      vars @
      List.map (Ocaml_config.to_list context.ocaml_config) ~f:(fun (k, v) ->
        ("ocaml-config:" ^ k,
         match (v : Ocaml_config.Value.t) with
         | Bool   x -> string (string_of_bool x)
         | Int    x -> string (string_of_int x)
         | String x -> string x
         | Words  x -> strings x
         | Prog_and_args x -> strings (x.prog :: x.args)))
    in
    match String_map.of_list vars with
    | Ok    x -> x
    | Error _ -> assert false
  in
  { context
  ; host
  ; build_system
  ; scopes
  ; public_libs
  ; installed_libs
  ; stanzas
  ; packages
  ; file_tree
  ; stanzas_to_consider_for_install
  ; artifacts
  ; cxx_flags
  ; vars
  ; chdir = Build.arr (fun (action : Action.t) ->
      match action with
      | Chdir _ -> action
      | _ -> Chdir (context.build_dir, action))
  ; libs_by_package =
    Lib.DB.all public_libs
    |> Lib.Set.to_list
    |> List.map ~f:(fun lib ->
      (Option.value_exn (Lib.package lib), lib))
    |> Package.Name.Map.of_list_multi
    |> Package.Name.Map.merge packages ~f:(fun _name pkg libs ->
      let pkg  = Option.value_exn pkg          in
      let libs = Option.value libs ~default:[] in
      Some (pkg, Lib.Set.of_list libs))
  }

let prefix_rules t prefix ~f =
  Build_system.prefix_rules t.build_system prefix ~f

let add_rule t ?sandbox ?mode ?locks ?loc build =
  let build = Build.O.(>>>) build t.chdir in
  Build_system.add_rule t.build_system
    (Build_interpret.Rule.make ?sandbox ?mode ?locks ?loc
       ~context:(Some t.context) build)

let add_rule_get_targets t ?sandbox ?mode ?locks ?loc build =
  let build = Build.O.(>>>) build t.chdir in
  let rule =
    Build_interpret.Rule.make ?sandbox ?mode ?locks ?loc
      ~context:(Some t.context) build
  in
  Build_system.add_rule t.build_system rule;
  List.map rule.targets ~f:Build_interpret.Target.path

let add_rules t ?sandbox builds =
  List.iter builds ~f:(add_rule t ?sandbox)

let add_alias_deps t alias ?dyn_deps deps =
  Alias.add_deps t.build_system alias ?dyn_deps deps

let add_alias_action t alias ?locks ~stamp action =
  Alias.add_action t.build_system ~context:t.context alias ?locks ~stamp action

let eval_glob t ~dir re = Build_system.eval_glob t.build_system ~dir re
let load_dir t ~dir = Build_system.load_dir t.build_system ~dir
let on_load_dir t ~dir ~f = Build_system.on_load_dir t.build_system ~dir ~f

let source_files t ~src_path =
  match File_tree.find_dir t.file_tree src_path with
  | None -> String_set.empty
  | Some dir -> File_tree.Dir.files dir

module Libs = struct
  open Build.O

  let gen_select_rules t ~dir compile_info =
    List.iter (Lib.Compile.resolved_selects compile_info) ~f:(fun rs ->
      let { Lib.Compile.Resolved_select.dst_fn; src_fn } = rs in
      let dst = Path.relative dir dst_fn in
      add_rule t
        (match src_fn with
         | Ok src_fn ->
           let src = Path.relative dir src_fn in
           Build.copy_and_add_line_directive ~src ~dst
         | Error e ->
           Build.fail ~targets:[dst]
             { fail = fun () ->
                 raise (Lib.Error (No_solution_found_for_select e))
             }))

  let with_lib_deps t compile_info ~dir ~f =
    let prefix =
      Build.record_lib_deps (Lib.Compile.user_written_deps compile_info)
        ~kind:(if Lib.Compile.optional compile_info then
                 Optional
               else
                 Required)
    in
    let prefix =
      if t.context.merlin then
        Build.path (Path.relative dir ".merlin-exists")
        >>>
        prefix
      else
        prefix
    in
    prefix_rules t prefix ~f

  let lib_files_alias ~dir ~name ~ext =
    Alias.make (sprintf "lib-%s%s-all" name ext) ~dir

  let setup_file_deps_alias t ~dir ~ext lib files =
    add_alias_deps t
      (lib_files_alias ~dir ~name:(Library.best_name lib) ~ext) files

  let setup_file_deps_group_alias t ~dir ~exts lib =
    setup_file_deps_alias t lib ~dir
      ~ext:(String.concat exts ~sep:"-and-")
      (List.map exts ~f:(fun ext ->
         Alias.stamp_file
           (lib_files_alias ~dir ~name:(Library.best_name lib) ~ext))
       |> Path.Set.of_list)

  let file_deps t libs ~ext =
    List.rev_map libs ~f:(fun (lib : Lib.t) ->
      if Lib.is_local lib then
        Alias.stamp_file
          (lib_files_alias ~dir:(Lib.src_dir lib) ~name:(Lib.name lib) ~ext)
      else
        Build_system.stamp_file_for_files_of t.build_system
          ~dir:(Lib.obj_dir lib) ~ext)
end

module Deps = struct
  open Build.O
  open Dep_conf

  let make_alias t ~scope ~dir s =
    Alias.of_path (Path.relative dir (expand_vars t ~scope ~dir s))

  let dep t ~scope ~dir = function
    | File  s ->
      let path = Path.relative dir (expand_vars t ~scope ~dir s) in
      Build.path path
      >>^ fun () -> [path]
    | Alias s ->
      Alias.dep (make_alias t ~scope ~dir s)
      >>^ fun () -> []
    | Alias_rec s ->
      Alias.dep_rec ~loc:(String_with_vars.loc s) ~file_tree:t.file_tree
        (make_alias t ~scope ~dir s)
      >>^ fun () -> []
    | Glob_files s -> begin
        let path = Path.relative dir (expand_vars t ~scope ~dir s) in
        let loc = String_with_vars.loc s in
        match Glob_lexer.parse_string (Path.basename path) with
        | Ok re ->
          let dir = Path.parent path in
          Build.paths_glob ~loc ~dir (Re.compile re)
        | Error (_pos, msg) ->
          Loc.fail loc "invalid glob: %s" msg
      end
    | Files_recursively_in s ->
      let path = Path.relative dir (expand_vars t ~scope ~dir s) in
      Build.files_recursively_in ~dir:path ~file_tree:t.file_tree
      >>^ Pset.to_list
    | Package p ->
      let pkg = Package.Name.of_string (expand_vars t ~scope ~dir p) in
      Alias.dep (Alias.package_install ~context:t.context ~pkg)
      >>^ fun () -> []
    | Universe ->
      Build.path Build_system.universe_file
      >>^ fun () -> []

  let interpret t ~scope ~dir l =
    Build.all (List.map l ~f:(dep t ~scope ~dir))
    >>^ List.concat
end

module Pkg_version = struct
  open Build.O

  module V = Vfile_kind.Make(struct type t = string option end)
      (functor (C : Sexp.Combinators) -> struct
        let t = C.option C.string
      end)

  let spec sctx (p : Package.t) =
    let fn =
      Path.relative (Path.append sctx.context.build_dir p.path)
        (sprintf "%s.version.sexp" (Package.Name.to_string p.name))
    in
    Build.Vspec.T (fn, (module V))

  let read sctx p = Build.vpath (spec sctx p)

  let set sctx p get =
    let spec = spec sctx p in
    add_rule sctx (get >>> Build.store_vfile spec);
    Build.vpath spec
end

module Scope_key = struct
  let of_string sctx key =
    match String.rsplit2 key ~on:'@' with
    | None ->
      (key, public_libs sctx)
    | Some (key, scope) ->
      ( key
      , Scope.libs (find_scope_by_name sctx (Scope_info.Name.of_string scope)))

  let to_string key scope =
    sprintf "%s@%s" key (Scope_info.Name.to_string scope)
end

let parse_bang var : bool * string =
  let len = String.length var in
  if len > 0 && var.[0] = '!' then
    (true, String.sub var ~pos:1 ~len:(len - 1))
  else
    (false, var)

module Action = struct
  open Build.O
  module U = Action.Unexpanded

  type targets =
    | Static of Path.t list
    | Infer
    | Alias

  type resolved_forms =
    { (* Failed resolutions *)
      mutable failures  : fail list
    ; (* All "name" for ${lib:name:...}/${lib-available:name} forms *)
      mutable lib_deps  : Build.lib_deps
    ; (* Static deps from ${...} variables. For instance ${exe:...} *)
      mutable sdeps     : Pset.t
    ; (* Dynamic deps from ${...} variables. For instance ${read:...} *)
      mutable ddeps     : (unit, Action.Var_expansion.t) Build.t String_map.t
    }

  let add_lib_dep acc lib kind =
    acc.lib_deps <- String_map.add acc.lib_deps lib kind

  let add_fail acc fail =
    acc.failures <- fail :: acc.failures;
    None

  let add_ddep acc ~key dep =
    acc.ddeps <- String_map.add acc.ddeps key dep;
    None

  let path_exp path = Action.Var_expansion.Paths   ([path], Concat)
  let str_exp  path = Action.Var_expansion.Strings ([path], Concat)

  let map_exe sctx =
    match sctx.host with
    | None -> (fun exe -> exe)
    | Some host ->
      fun exe ->
        match Path.extract_build_context_dir exe with
        | Some (dir, exe) when dir = sctx.context.build_dir ->
          Path.append host.context.build_dir exe
        | _ -> exe

  let parse_lib_file ~loc s =
    match String.lsplit2 s ~on:':' with
    | None ->
      Loc.fail loc "invalid ${lib:...} form: %s" s
    | Some x -> x

  let expand_step1 sctx ~dir ~dep_kind ~scope ~targets_written_by_user
        ~map_exe ~extra_vars t =
    let acc =
      { failures  = []
      ; lib_deps  = String_map.empty
      ; sdeps     = Pset.empty
      ; ddeps     = String_map.empty
      }
    in
    let open Action.Var_expansion in
    let expand loc key var = function
      | Some ("exe"     , s) -> Some (path_exp (map_exe (Path.relative dir s)))
      | Some ("path"    , s) -> Some (path_exp          (Path.relative dir s) )
      | Some ("bin"     , s) -> begin
          let sctx = host sctx in
          match Artifacts.binary (artifacts sctx) s with
          | Ok path -> Some (path_exp path)
          | Error e ->
            add_fail acc ({ fail = fun () -> Action.Prog.Not_found.raise e })
        end
      (* "findlib" for compatibility with Jane Street packages which are not yet updated
         to convert "findlib" to "lib" *)
      | Some (("lib"|"findlib"), s) -> begin
          let lib_dep, file = parse_lib_file ~loc s in
          add_lib_dep acc lib_dep dep_kind;
          match
            Artifacts.file_of_lib (artifacts sctx) ~loc ~lib:lib_dep ~file
          with
          | Ok path -> Some (path_exp path)
          | Error fail -> add_fail acc fail
        end
      | Some ("libexec" , s) -> begin
          let sctx = host sctx in
          let lib_dep, file = parse_lib_file ~loc s in
          add_lib_dep acc lib_dep dep_kind;
          match
            Artifacts.file_of_lib (artifacts sctx) ~loc ~lib:lib_dep ~file
          with
          | Error fail -> add_fail acc fail
          | Ok path ->
            if not Sys.win32 || Filename.extension s = ".exe" then begin
              Some (path_exp path)
            end else begin
              let path_exe = Path.extend_basename path ~suffix:".exe" in
              let dep =
                Build.if_file_exists path_exe
                  ~then_:(Build.path path_exe >>^ fun _ -> path_exp path_exe)
                  ~else_:(Build.path path     >>^ fun _ -> path_exp path)
              in
              add_ddep acc ~key dep
            end
        end
      | Some ("lib-available", lib) ->
        add_lib_dep acc lib Optional;
        Some (str_exp (string_of_bool (
          Lib.DB.available (Scope.libs scope) lib)))
      | Some ("version", s) -> begin
          match Scope_info.resolve (Scope.info scope)
                  (Package.Name.of_string s) with
          | Ok p ->
            let x =
              Pkg_version.read sctx p >>^ function
              | None   -> Strings ([""], Concat)
              | Some s -> Strings ([s],  Concat)
            in
            add_ddep acc ~key x
          | Error s ->
            add_fail acc { fail = fun () -> Loc.fail loc "%s" s }
        end
      | Some ("read", s) -> begin
          let path = Path.relative dir s in
          let data =
            Build.contents path
            >>^ fun s -> Strings ([s], Concat)
          in
          add_ddep acc ~key data
        end
      | Some ("read-lines", s) -> begin
          let path = Path.relative dir s in
          let data =
            Build.lines_of path
            >>^ fun l -> Strings (l, Split)
          in
          add_ddep acc ~key data
        end
      | Some ("read-strings", s) -> begin
          let path = Path.relative dir s in
          let data =
            Build.strings path
            >>^ fun l -> Strings (l, Split)
          in
          add_ddep acc ~key data
        end
      | _ ->
        match expand_var_no_root sctx var with
        | Some _ as x -> x
        | None -> String_map.find extra_vars var
    in
    let t =
      U.partial_expand t ~dir ~map_exe ~f:(fun loc key ->
        let has_bang, var = parse_bang key in
        if has_bang then
          Loc.warn loc "The use of the variable prefix '!' is deprecated, \
                        simply use '${%s}'@." var;
        match var with
        | "ROOT" -> Some (path_exp sctx.context.build_dir)
        | "SCOPE_ROOT" -> Some (path_exp (Scope.root scope))
        | "@" -> begin
            match targets_written_by_user with
            | Infer -> Loc.fail loc "You cannot use ${@} with inferred rules."
            | Alias -> Loc.fail loc "You cannot use ${@} in aliases."
            | Static l -> Some (Paths (l, Split))
          end
        | _ ->
          match String.lsplit2 var ~on:':' with
          | Some ("path-no-dep", s) ->
            Some (path_exp (Path.relative dir s))
          | x ->
            let exp = expand loc key var x in
            (match exp with
             | Some (Paths (ps, _)) ->
               acc.sdeps <- Pset.union (Pset.of_list ps) acc.sdeps
             | _ -> ());
            exp)
    in
    (t, acc)

  let expand_step2 ~dir ~dynamic_expansions ~deps_written_by_user ~map_exe t =
    let open Action.Var_expansion in
    U.Partial.expand t ~dir ~map_exe ~f:(fun loc key ->
      match String_map.find dynamic_expansions key with
      | Some _ as opt -> opt
      | None ->
        let _, var = parse_bang key in
        match var with
        | "<" ->
          Some
            (match deps_written_by_user with
             | [] ->
                Loc.warn loc "Variable '<' used with no explicit \
                              dependencies@.";
                Strings ([""], Concat)
             | dep :: _ ->
               Paths ([dep], Concat))
        | "^" -> Some (Paths (deps_written_by_user, Split))
        | _ -> None)

  let run sctx ?(extra_vars=String_map.empty)
        t ~dir ~dep_kind ~targets:targets_written_by_user ~scope
    : (Path.t list, Action.t) Build.t =
    let map_exe = map_exe sctx in
    if targets_written_by_user = Alias then begin
      match Action.Infer.unexpanded_targets t with
      | [] -> ()
      | x :: _ ->
        let loc = String_with_vars.loc x in
        Loc.warn loc "Aliases must not have targets, this target will be ignored.\n\
                      This will become an error in the future.";
    end;
    let t, forms =
      expand_step1 sctx t ~dir ~dep_kind ~scope
        ~targets_written_by_user ~map_exe ~extra_vars
    in
    let { Action.Infer.Outcome. deps; targets } =
      match targets_written_by_user with
      | Infer -> Action.Infer.partial t ~all_targets:true
      | Static targets_written_by_user ->
        let targets_written_by_user = Pset.of_list targets_written_by_user in
        let { Action.Infer.Outcome. deps; targets } =
          Action.Infer.partial t ~all_targets:false
        in
        (* CR-someday jdimino: should this be an error or not?

           It's likely that what we get here is what the user thinks
           of as temporary files, even though they might conflict with
           actual targets. We need to tell jbuilder about such things,
           so that it can report better errors.

           {[
             let missing = Pset.diff targets targets_written_by_user in
             if not (Pset.is_empty missing) then
               Loc.warn (Loc.in_file (Utils.jbuild_name_in ~dir))
                 "Missing targets in user action:\n%s"
                 (List.map (Pset.elements missing) ~f:(fun target ->
                    sprintf "- %s" (Utils.describe_target target))
                  |> String.concat ~sep:"\n");
           ]}
        *)
        { deps; targets = Pset.union targets targets_written_by_user }
      | Alias ->
        let { Action.Infer.Outcome. deps; targets = _ } =
          Action.Infer.partial t ~all_targets:false
        in
        { deps; targets = Pset.empty }
    in
    let targets = Pset.to_list targets in
    List.iter targets ~f:(fun target ->
      if Path.parent target <> dir then
        Loc.fail (Loc.in_file (Utils.describe_target (Utils.jbuild_file_in ~dir)))
          "A rule in this jbuild has targets in a different directory \
           than the current one, this is not allowed by Jbuilder at the moment:\n%s"
          (List.map targets ~f:(fun target ->
             sprintf "- %s" (Utils.describe_target target))
           |> String.concat ~sep:"\n"));
    let build =
      Build.record_lib_deps_simple forms.lib_deps
      >>>
      Build.path_set deps
      >>>
      Build.path_set forms.sdeps
      >>>
      Build.arr (fun paths -> ((), paths))
      >>>
      let ddeps = String_map.to_list forms.ddeps in
      Build.first (Build.all (List.map ddeps ~f:snd))
      >>^ (fun (vals, deps_written_by_user) ->
        let dynamic_expansions =
          List.fold_left2 ddeps vals ~init:String_map.empty
            ~f:(fun acc (var, _) value -> String_map.add acc var value)
        in
        let unresolved =
          expand_step2 t ~dir ~dynamic_expansions ~deps_written_by_user ~map_exe
        in
        Action.Unresolved.resolve unresolved ~f:(fun prog ->
          let sctx = host sctx in
          match Artifacts.binary sctx.artifacts prog with
          | Ok path    -> path
          | Error fail -> Action.Prog.Not_found.raise fail))
      >>>
      Build.dyn_paths (Build.arr (fun action ->
        let { Action.Infer.Outcome.deps; targets = _ } =
          Action.Infer.infer action
        in
        Pset.to_list deps))
      >>>
      Build.action_dyn () ~dir ~targets
    in
    match forms.failures with
    | [] -> build
    | fail :: _ -> Build.fail fail >>> build
end

let expand_and_eval_set t ~scope ~dir ?extra_vars set ~standard =
  let open Build.O in
  let f = expand_vars t ~scope ~dir ?extra_vars in
  let parse ~loc:_ s = s in
  match Ordered_set_lang.Unexpanded.files set ~f |> String_set.to_list with
  | [] ->
    let set =
      Ordered_set_lang.Unexpanded.expand set ~files_contents:String_map.empty ~f
    in
    Build.return (Ordered_set_lang.String.eval set ~standard ~parse)
  | files ->
    let paths = List.map files ~f:(Path.relative dir) in
    Build.all (List.map paths ~f:Build.read_sexp)
    >>^ fun sexps ->
    let files_contents = List.combine files sexps |> String_map.of_list_exn in
    let set = Ordered_set_lang.Unexpanded.expand set ~files_contents ~f in
    Ordered_set_lang.String.eval set ~standard ~parse
# 21932 "boot.ml"
end
module Js_of_ocaml_rules : sig
# 1 "src/js_of_ocaml_rules.mli"
(** Generate rules for js_of_ocaml *)

open Import
open Jbuild

val build_cm
  :  Super_context.t
  -> scope:Scope.t
  -> dir:Path.t
  -> js_of_ocaml:Js_of_ocaml.t
  -> src:Path.t
  -> target:Path.t
  -> (unit, Action.t) Build.t list

val build_exe
  :  Super_context.t
  -> dir:Path.t
  -> js_of_ocaml:Js_of_ocaml.t
  -> src:Path.t
  -> requires:Lib.t list Or_exn.t
  -> (Path.t list * string list, Action.t) Build.t list

val setup_separate_compilation_rules
  :  Super_context.t
  -> string list
  -> unit

val standard : unit -> string list
# 21964 "boot.ml"
end = struct
# 1 "src/js_of_ocaml_rules.ml"
open Import
open! No_io
open Build.O

module SC = Super_context

let separate_compilation_enabled () = !Clflags.dev_mode

let pretty    () = if !Clflags.dev_mode then ["--pretty"           ] else []
let sourcemap () = if !Clflags.dev_mode then ["--source-map-inline"] else []

let standard () = pretty () @ sourcemap ()

let install_jsoo_hint = "opam install js_of_ocaml-compiler"

let in_build_dir ~ctx =
  let init = Path.relative ctx.Context.build_dir ".js" in
  List.fold_left ~init ~f:Path.relative

let runtime_file ~sctx ~dir fname =
  match
    Artifacts.file_of_lib (SC.artifacts sctx)
      ~loc:(Loc.in_file (Utils.jbuild_file_in ~dir |> Path.to_string))
      ~lib:"js_of_ocaml-compiler" ~file:fname
  with
  | Error _ ->
    Arg_spec.Dyn (fun _ ->
      Utils.library_not_found ~context:(SC.context sctx).name
        ~hint:install_jsoo_hint
        "js_of_ocaml-compiler")
  | Ok f -> Arg_spec.Dep f

let js_of_ocaml_rule ~sctx ~dir ~flags ~spec ~target =
  let jsoo = SC.resolve_program sctx ~hint:install_jsoo_hint "js_of_ocaml" in
  let runtime = runtime_file ~sctx ~dir "runtime.js" in
  Build.run ~context:(SC.context sctx) ~dir
    jsoo
    [ Arg_spec.Dyn flags
    ; Arg_spec.A "-o"; Target target
    ; Arg_spec.A "--no-runtime"; runtime
    ; spec
    ]

let standalone_runtime_rule ~sctx ~dir ~javascript_files ~target ~requires =
  let spec =
    Arg_spec.S
      [ Arg_spec.of_result_map requires ~f:(fun libs ->
          Arg_spec.Deps (Lib.L.jsoo_runtime_files libs))
      ; Arg_spec.Deps javascript_files
      ]
  in
  Build.arr (fun (cm_files, flags) -> (cm_files, "--runtime-only" :: flags))
  >>>
  js_of_ocaml_rule ~sctx ~dir ~flags:(fun (_,flags) -> As flags) ~target ~spec

let exe_rule ~sctx ~dir ~javascript_files ~src ~target ~requires =
  let spec =
    Arg_spec.S
      [ Arg_spec.of_result_map requires ~f:(fun libs ->
          Arg_spec.Deps (Lib.L.jsoo_runtime_files libs))
      ; Arg_spec.Deps javascript_files
      ; Arg_spec.Dep src
      ]
  in
  js_of_ocaml_rule ~sctx ~dir ~flags:(fun (_,flags) -> As flags) ~spec ~target

let jsoo_archives lib =
  List.map (Lib.archives lib).byte ~f:(Path.extend_basename ~suffix:".js")

let link_rule ~sctx ~dir ~runtime ~target ~requires =
  let ctx = SC.context sctx in
  let get_all (cm, _) =
    Arg_spec.of_result_map requires ~f:(fun libs ->
      let all_libs =
        List.concat_map libs ~f:(fun (lib : Lib.t) ->
          let jsoo_archives = jsoo_archives lib in
          if Lib.is_local lib then (
            jsoo_archives
          ) else (
            let lib_name = Lib.name lib in
            List.map ~f:(fun js ->
              in_build_dir ~ctx [lib_name ; Path.basename js]) jsoo_archives
          )
        )
      in
      (* Special case for the stdlib because it is not referenced in the META *)
      let all_libs = in_build_dir ~ctx ["stdlib"; "stdlib.cma.js"] :: all_libs in
      let all_other_modules =
        List.map cm ~f:(fun m -> Path.extend_basename m ~suffix:".js")
      in
      Arg_spec.Deps (List.concat [all_libs;all_other_modules]))
  in
  let jsoo_link = SC.resolve_program sctx ~hint:install_jsoo_hint "jsoo_link" in
  Build.run ~context:(SC.context sctx) ~dir
    jsoo_link
    [ Arg_spec.A "-o"; Target target
    ; Arg_spec.Dep runtime
    ; Arg_spec.As (sourcemap ())
    ; Arg_spec.Dyn get_all
    ]

let build_cm sctx ~scope ~dir ~(js_of_ocaml:Jbuild.Js_of_ocaml.t) ~src ~target =
  if separate_compilation_enabled ()
  then
    let itarget = Path.extend_basename src ~suffix:".js" in
    let spec = Arg_spec.Dep src in
    let flags =
      SC.expand_and_eval_set sctx ~scope ~dir js_of_ocaml.flags
        ~standard:(standard ())
    in
    [ flags
      >>>
      js_of_ocaml_rule ~sctx ~dir ~flags:(fun flags ->
        As flags) ~spec ~target:itarget ]
    @ (if target = itarget then
         []
       else
         [Build.symlink ~src:itarget ~dst:target])
  else []

let setup_separate_compilation_rules sctx components =
  if separate_compilation_enabled ()
  then
    match components with
    | [] | _ :: _ :: _ -> ()
    | [pkg] ->
      let ctx = SC.context sctx in
      match Lib.DB.find (SC.installed_libs sctx) pkg with
      | Error _ -> ()
      | Ok pkg ->
        let archives = (Lib.archives pkg).byte in
        let archives =
          (* Special case for the stdlib because it is not referenced
             in the META *)
          match Lib.name pkg with
          | "stdlib" -> Path.relative ctx.stdlib_dir "stdlib.cma" :: archives
          | _ -> archives
        in
        List.iter archives ~f:(fun fn ->
          let name = Path.basename fn in
          let src = Path.relative (Lib.src_dir pkg) name in
          let target =
            in_build_dir ~ctx [ Lib.name pkg; sprintf "%s.js" name]
          in
          let dir = in_build_dir ~ctx [ Lib.name pkg ] in
          let spec = Arg_spec.Dep src in
          SC.add_rule sctx
            (Build.return (standard ())
             >>>
             js_of_ocaml_rule ~sctx ~dir ~flags:(fun flags ->
               As flags) ~spec ~target)
        )

let build_exe sctx ~dir ~js_of_ocaml ~src ~requires =
  let {Jbuild.Js_of_ocaml.javascript_files; _} = js_of_ocaml in
  let javascript_files = List.map javascript_files ~f:(Path.relative dir) in
  let mk_target ext = Path.extend_basename src ~suffix:ext in
  let target = mk_target ".js" in
  let standalone_runtime = mk_target ".runtime.js" in
  if separate_compilation_enabled () then
    [ link_rule ~sctx ~dir ~runtime:standalone_runtime ~target ~requires
    ; standalone_runtime_rule ~sctx ~dir ~javascript_files
        ~target:standalone_runtime ~requires
    ]
  else
    [ exe_rule ~sctx ~dir ~javascript_files ~src ~target ~requires ]
# 22133 "boot.ml"
end
module Ocaml_flags : sig
# 1 "src/ocaml_flags.mli"
(** OCaml flags *)

type t

val make
  :  Jbuild.Buildable.t
  -> Super_context.t
  -> scope:Scope.t
  -> dir:Path.t
  -> t

val default : unit -> t

val empty : t

val of_list : string list -> t

val get : t -> Mode.t -> (unit, string list) Build.t
val get_for_cm : t -> cm_kind:Cm_kind.t -> (unit, string list) Build.t

val append_common : t -> string list -> t
val prepend_common : string list -> t -> t

val common : t -> (unit, string list) Build.t
# 22161 "boot.ml"
end = struct
# 1 "src/ocaml_flags.ml"
open Import
open Build.O

let default_ocamlc_flags   = Utils.g
let default_ocamlopt_flags = Utils.g

let dev_mode_warnings =
  "@a" ^
  String.concat ~sep:""
    (List.map ~f:(sprintf "-%d")
       [ 4
       ; 29
       ; 40
       ; 41
       ; 42
       ; 44
       ; 45
       ; 48
       ; 58
       ; 59
       ; 60
       ])

let default_flags () =
  if !Clflags.dev_mode then
    [ "-w"; dev_mode_warnings ^ !Clflags.warnings
    ; "-strict-sequence"
    ; "-strict-formats"
    ; "-short-paths"
    ; "-keep-locs"
    ]
  else
    [ "-w"; !Clflags.warnings ]

type t =
  { common     : (unit, string list) Build.t
  ; specific   : (unit, string list) Build.t Mode.Dict.t
  }

let empty =
  let build = Build.arr (fun () -> []) in
  { common   = build
  ; specific = Mode.Dict.make_both build
  }

let of_list l =
  { empty with common = Build.arr (fun () -> l) }

let make { Jbuild.Buildable. flags; ocamlc_flags; ocamlopt_flags; _ } ctx ~scope ~dir =
  let eval = Super_context.expand_and_eval_set ctx ~scope ~dir in
  { common   = Build.memoize "common flags" (eval flags ~standard:(default_flags ()))
  ; specific =
      { byte   = Build.memoize "ocamlc flags" (eval ocamlc_flags   ~standard:(default_ocamlc_flags ()))
      ; native = Build.memoize "ocamlopt flags" (eval ocamlopt_flags ~standard:(default_ocamlopt_flags ()))
      }
  }

let get t mode =
  t.common
  &&&
  (Mode.Dict.get t.specific mode)
  >>^ fun (common, specific) ->
  common @ specific

let get_for_cm t ~cm_kind = get t (Mode.of_cm_kind cm_kind)

let default () =
  { common = Build.return (default_flags ())
  ; specific =
      { byte   = Build.return (default_ocamlc_flags   ())
      ; native = Build.return (default_ocamlopt_flags ())
      }
  }

let append_common t flags = {t with common = t.common >>^ fun l -> l @ flags}

let prepend_common flags t = {t with common = t.common >>^ fun l -> flags @ l}

let common t = t.common
# 22243 "boot.ml"
end
module Ocamldep : sig
# 1 "src/ocamldep.mli"
(** ocamldep management *)

module Dep_graph : sig
  type t

  val deps_of
    :  t
    -> Module.t
    -> (unit, Module.t list) Build.t

  val top_closed_implementations
    :  t
    -> Module.t list
    -> (unit, Module.t list) Build.t
end

module Dep_graphs : sig
  type t = Dep_graph.t Ml_kind.Dict.t

  val dummy : Module.t -> t
end

(** Generate ocamldep rules for the given modules. [item] is either
    the internal name of a library of the first name of a list of
    executables.

    For wrapped libraries, [lib_interface_module] is the main module
    of the library.

    [already_used] represents the modules that are used by another
    stanzas in the same directory. No [.d] rule will be generated for
    such modules.

    Return arrows that evaluate to the dependency graphs.  *)
val rules
  :  dir:Path.t
  -> modules:Module.t Module.Name.Map.t
  -> ?already_used:Module.Name.Set.t
  -> alias_module:Module.t option
  -> lib_interface_module:Module.t option
  -> Super_context.t
  -> Dep_graphs.t
# 22289 "boot.ml"
end = struct
# 1 "src/ocamldep.ml"
open Import
open Build.O

module SC = Super_context

module Dep_graph = struct
  type t =
    { dir        : Path.t
    ; per_module : (unit, Module.t list) Build.t Module.Name.Map.t
    }

  let deps_of t (m : Module.t) =
    match Module.Name.Map.find t.per_module m.name with
    | Some x -> x
    | None ->
      Sexp.code_error "Ocamldep.Dep_graph.deps_of"
        [ "dir", Path.sexp_of_t t.dir
        ; "modules", Sexp.To_sexp.(list Module.Name.t)
                       (Module.Name.Map.keys t.per_module)
        ; "module", Module.Name.t m.name
        ]

  let top_closed t modules =
    Build.all
      (List.map (Module.Name.Map.to_list t.per_module) ~f:(fun (unit, deps) ->
         deps >>^ fun deps -> (unit, deps)))
    >>^ fun per_module ->
    let per_module = Module.Name.Map.of_list_exn per_module in
    match
      Module.Name.Top_closure.top_closure modules
        ~key:Module.name
        ~deps:(fun m ->
          Option.value_exn (Module.Name.Map.find per_module (Module.name m)))
    with
    | Ok modules -> modules
    | Error cycle ->
      die "dependency cycle between modules in %s:\n   %a"
        (Path.to_string t.dir)
        (Fmt.list ~pp_sep:Fmt.nl (Fmt.prefix (Fmt.string "-> ") Module.Name.pp))
        (List.map cycle ~f:Module.name)

  let top_closed_implementations t modules =
    Build.memoize "top sorted implementations" (
      let filter_out_intf_only = List.filter ~f:Module.has_impl in
      top_closed t (filter_out_intf_only modules)
      >>^ filter_out_intf_only)

  let dummy (m : Module.t) =
    { dir = Path.root
    ; per_module = Module.Name.Map.singleton m.name (Build.return [])
    }
end

module Dep_graphs = struct
  type t = Dep_graph.t Ml_kind.Dict.t

  let dummy m =
    Ml_kind.Dict.make_both (Dep_graph.dummy m)
end

let parse_deps ~dir ~file ~(unit : Module.t)
      ~modules ~alias_module ~lib_interface_module lines =
  let invalid () =
    die "ocamldep returned unexpected output for %s:\n\
         %s"
      (Path.to_string_maybe_quoted file)
      (String.concat ~sep:"\n"
         (List.map lines ~f:(sprintf "> %s")))
  in
  match lines with
  | [] | _ :: _ :: _ -> invalid ()
  | [line] ->
    match String.index line ':' with
    | None -> invalid ()
    | Some i ->
      let basename =
        String.sub line ~pos:0 ~len:i
        |> Filename.basename
      in
      if basename <> Path.basename file then invalid ();
      let deps =
        String.extract_blank_separated_words
          (String.sub line ~pos:(i + 1) ~len:(String.length line - (i + 1)))
        |> List.filter_map ~f:(fun m ->
          let m = Module.Name.of_string m in
          if m = unit.name then
            None
          else
            Module.Name.Map.find modules m)
      in
      (match lib_interface_module with
       | None -> ()
       | Some (m : Module.t) ->
         let is_alias_module =
           match alias_module with
           | None -> false
           | Some (m : Module.t) -> unit.name = m.name
         in
         if unit.name <> m.name && not is_alias_module &&
            List.exists deps ~f:(fun x -> Module.name x = m.name) then
           die "Module %a in directory %s depends on %a.\n\
                This doesn't make sense to me.\n\
                \n\
                %a is the main module of the library and is \
                the only module exposed \n\
                outside of the library. Consequently, it should \
                be the one depending \n\
                on all the other modules in the library."
             Module.Name.pp unit.name (Path.to_string dir)
             Module.Name.pp m.name
             Module.Name.pp m.name);
      let deps =
        match alias_module with
        | None -> deps
        | Some m -> m :: deps
      in
      deps

let rules ~(ml_kind:Ml_kind.t) ~dir ~modules
      ?(already_used=Module.Name.Set.empty)
      ~alias_module ~lib_interface_module sctx =
  let per_module =
    Module.Name.Map.map modules ~f:(fun unit ->
      match Module.file ~dir unit ml_kind with
      | None -> Build.return []
      | Some file ->
        let ocamldep_output = Path.extend_basename file ~suffix:".d" in
        let context = SC.context sctx in
        if not (Module.Name.Set.mem already_used unit.name) then
          SC.add_rule sctx
            (Build.run ~context (Ok context.ocamldep)
               [A "-modules"; Ml_kind.flag ml_kind; Dep file]
               ~stdout_to:ocamldep_output);
        Build.memoize (Path.to_string ocamldep_output)
          (Build.lines_of ocamldep_output
           >>^ parse_deps ~dir ~file ~unit ~modules ~alias_module
                 ~lib_interface_module))
  in
  let per_module =
    match alias_module with
    | None -> per_module
    | Some m -> Module.Name.Map.add per_module m.name (Build.return [])
  in
  { Dep_graph.
    dir
  ; per_module
  }

let rules ~dir ~modules ?already_used ~alias_module ~lib_interface_module sctx =
  Ml_kind.Dict.of_func (rules sctx ~dir ~modules ?already_used ~alias_module
                          ~lib_interface_module)
# 22443 "boot.ml"
end
module Module_compilation : sig
# 1 "src/module_compilation.mli"
(** OCaml module compilation *)

open Import

(** Setup rules to build a single module. *)
val build_module
  :  Super_context.t
  -> ?sandbox:bool
  -> dynlink:bool
  -> js_of_ocaml:Jbuild.Js_of_ocaml.t
  -> flags:Ocaml_flags.t
  -> Module.t
  -> scope:Scope.t
  -> dir:Path.t
  -> obj_dir:Path.t
  -> dep_graphs:Ocamldep.Dep_graphs.t
  -> includes:string list Arg_spec.t Cm_kind.Dict.t
  -> alias_module:Module.t option
  -> unit

(** Setup rules to build all of [modules] *)
val build_modules
  :  Super_context.t
  -> dynlink:bool
  -> js_of_ocaml:Jbuild.Js_of_ocaml.t
  -> flags:Ocaml_flags.t
  -> scope:Scope.t
  -> dir:Path.t
  -> obj_dir:Path.t
  -> dep_graphs:Ocamldep.Dep_graphs.t
  -> modules:Module.t Module.Name.Map.t
  -> requires:Lib.t list Or_exn.t
  -> alias_module:Module.t option
  -> unit
# 22481 "boot.ml"
end = struct
# 1 "src/module_compilation.ml"
open Import
open Build.O
open! No_io

module SC = Super_context

module Target : sig
  type t
  val cm : Module.t -> Cm_kind.t -> t
  val obj : Module.t -> ext:string -> t
  val cmt : Module.t -> Ml_kind.t -> t option
  val file : Path.t -> t -> Path.t
end = struct
  type t = Path.t
  let cm m cm_kind = Module.cm_file_unsafe m ~obj_dir:Path.root cm_kind
  let obj m ~ext = Module.obj_file m ~obj_dir:Path.root ~ext
  let cmt m ml_kind = Module.cmt_file m ~obj_dir:Path.root ml_kind
  let file dir t = Path.append dir t
end

let build_cm sctx ?sandbox ~dynlink ~flags ~cm_kind ~dep_graphs
      ~includes ~dir ~obj_dir ~alias_module (m : Module.t) =
  let ctx = SC.context sctx in
  Option.iter (Mode.of_cm_kind cm_kind |> Context.compiler ctx) ~f:(fun compiler ->
    Option.iter (Module.cm_source ~dir m cm_kind) ~f:(fun src ->
      let ml_kind = Cm_kind.source cm_kind in
      let dst = Module.cm_file_unsafe m ~obj_dir cm_kind in
      let extra_args, extra_deps, other_targets =
        match cm_kind, m.intf with
        (* If there is no mli, [ocamlY -c file.ml] produces both the
           .cmY and .cmi. We choose to use ocamlc to produce the cmi
           and to produce the cmx we have to wait to avoid race
           conditions. *)
        | Cmo, None -> [], [], [Target.cm m Cmi]
        | Cmx, None ->
          (* Change [-intf-suffix] so that the compiler thinks the
             cmi exists and reads it instead of re-creating it, which
             could create a race condition. *)
          [ "-intf-suffix"
          ; Filename.extension (Option.value_exn m.impl).name
          ],
          [Module.cm_file_unsafe m ~obj_dir Cmi],
          []
        | Cmi, None -> assert false
        | Cmi, Some _ -> [], [], []
        (* We need the .cmi to build either the .cmo or .cmx *)
        | (Cmo | Cmx), Some _ -> [], [Module.cm_file_unsafe m ~obj_dir Cmi], []
      in
      let other_targets =
        match cm_kind with
        | Cmx -> Target.obj m ~ext:ctx.ext_obj :: other_targets
        | Cmi | Cmo -> other_targets
      in
      let dep_graph = Ml_kind.Dict.get dep_graphs ml_kind in
      let other_cm_files =
        Build.dyn_paths
          (Ocamldep.Dep_graph.deps_of dep_graph m >>^ fun deps ->
           List.concat_map deps
             ~f:(fun m ->
               let deps = [Module.cm_file_unsafe m ~obj_dir Cmi] in
               if Module.has_impl m && cm_kind = Cmx then
                 Module.cm_file_unsafe m ~obj_dir Cmx :: deps
               else
                 deps))
      in
      let other_targets, cmt_args =
        match cm_kind with
        | Cmx -> (other_targets, Arg_spec.S [])
        | Cmi | Cmo ->
          let fn = Option.value_exn (Target.cmt m ml_kind) in
          (fn :: other_targets, A "-bin-annot")
      in
      let extra_targets = List.map other_targets ~f:(Target.file obj_dir) in
      if obj_dir <> dir then begin
        (* Symlink the object files in the original directory for
           backward compatibility *)
        let old_dst = Module.cm_file_unsafe m ~obj_dir:dir cm_kind in
        SC.add_rule sctx (Build.symlink ~src:dst ~dst:old_dst) ;
        List.iter2 extra_targets other_targets ~f:(fun in_obj_dir target ->
          let in_dir = Target.file dir target in
          SC.add_rule sctx (Build.symlink ~src:in_obj_dir ~dst:in_dir))
      end;
      let opaque =
        if cm_kind = Cmi && not (Module.has_impl m) && ctx.version >= (4, 03, 0) then
          Arg_spec.A "-opaque"
        else
          As []
      in
      SC.add_rule sctx ?sandbox
        (Build.paths extra_deps >>>
         other_cm_files >>>
         Ocaml_flags.get_for_cm flags ~cm_kind >>>
         Build.run ~context:ctx (Ok compiler)
           ~extra_targets
           [ Dyn (fun ocaml_flags -> As ocaml_flags)
           ; cmt_args
           ; A "-I"; Path obj_dir
           ; includes
           ; As extra_args
           ; if dynlink || cm_kind <> Cmx then As [] else A "-nodynlink"
           ; A "-no-alias-deps"; opaque
           ; (match alias_module with
              | None -> S []
              | Some (m : Module.t) ->
                As ["-open"; Module.Name.to_string m.name])
           ; A "-o"; Target dst
           ; A "-c"; Ml_kind.flag ml_kind; Dep src
           ])))

let build_module sctx ?sandbox ~dynlink ~js_of_ocaml ~flags m ~scope ~dir
      ~obj_dir ~dep_graphs ~includes ~alias_module =
  List.iter Cm_kind.all ~f:(fun cm_kind ->
    let includes = Cm_kind.Dict.get includes cm_kind in
    build_cm sctx ?sandbox ~dynlink ~flags ~dir ~obj_dir ~dep_graphs m ~cm_kind
      ~includes ~alias_module);
  (* Build *.cmo.js *)
  let src = Module.cm_file_unsafe m ~obj_dir Cm_kind.Cmo in
  let target =
    Path.extend_basename (Module.cm_file_unsafe m ~obj_dir:dir Cm_kind.Cmo)
      ~suffix:".js"
  in
  SC.add_rules sctx
    (Js_of_ocaml_rules.build_cm sctx ~scope ~dir ~js_of_ocaml ~src ~target)

let build_modules sctx ~dynlink ~js_of_ocaml ~flags ~scope ~dir ~obj_dir
      ~dep_graphs ~modules ~requires ~alias_module =
  let includes : _ Cm_kind.Dict.t =
    match requires with
    | Error exn -> Cm_kind.Dict.make_all (Arg_spec.Dyn (fun _ -> raise exn))
    | Ok libs ->
      let iflags =
        Lib.L.include_flags libs ~stdlib_dir:(SC.context sctx).stdlib_dir
      in
      let cmi_includes =
        Arg_spec.S [ iflags
                   ; Hidden_deps
                       (SC.Libs.file_deps sctx libs ~ext:".cmi")
                   ]
      in
      let cmi_and_cmx_includes =
        Arg_spec.S [ iflags
                   ; Hidden_deps
                       (SC.Libs.file_deps sctx libs ~ext:".cmi-and-.cmx")
                   ]
      in
      { cmi = cmi_includes
      ; cmo = cmi_includes
      ; cmx = cmi_and_cmx_includes
      }
  in
  Module.Name.Map.iter
    (match alias_module with
     | None -> modules
     | Some (m : Module.t) -> Module.Name.Map.remove modules m.name)
    ~f:(fun m ->
      build_module sctx m ~dynlink ~js_of_ocaml ~flags ~scope ~dir ~obj_dir
        ~dep_graphs ~includes ~alias_module)
# 22641 "boot.ml"
end
module Exe : sig
# 1 "src/exe.mli"
(** Compilation and linking of executables *)

open Import

module Program : sig
  type t =
    { name             : string
    ; main_module_name : Module.Name.t
    }
end

module Linkage : sig
  type t

  (** Byte compilation, exetension [.bc] *)
  val byte : t

  (** Native compilation, extension [.exe] *)
  val native : t

  (** Byte compilation, link with [-custom], extension [.exe] *)
  val custom : t

  (** [native] if supported, [custom] if not *)
  val native_or_custom : Context.t -> t

  val make
    :  mode:Mode.t
    -> ext:string
    -> ?flags:string list
    -> unit
    -> t

  val of_user_config : Context.t -> Jbuild.Executables.Link_mode.t -> t
end

(** {1 High-level functions} *)

(** Build and link one or more executables *)

val build_and_link
  :  dir:Path.t
  -> obj_dir:Path.t
  -> program:Program.t
  -> modules:Module.t Module.Name.Map.t
  -> scope:Scope.t
  -> linkages:Linkage.t list
  -> ?requires:Lib.t list Or_exn.t
  -> ?already_used:Module.Name.Set.t
  -> ?flags:Ocaml_flags.t
  -> ?link_flags:(unit, string list) Build.t
  -> ?js_of_ocaml:Jbuild.Js_of_ocaml.t
  -> Super_context.t
  -> unit

val build_and_link_many
  :  dir:Path.t
  -> obj_dir:Path.t
  -> programs:Program.t list
  -> modules:Module.t Module.Name.Map.t
  -> scope:Scope.t
  -> linkages:Linkage.t list
  -> ?requires:Lib.t list Or_exn.t
  -> ?already_used:Module.Name.Set.t
  -> ?flags:Ocaml_flags.t
  -> ?link_flags:(unit, string list) Build.t
  -> ?js_of_ocaml:Jbuild.Js_of_ocaml.t
  -> Super_context.t
  -> unit

(** {1 Low-level functions} *)

(** Link a single executable *)
val link_exe
  :  dir:Path.t
  -> obj_dir:Path.t
  -> scope:Scope.t
  -> requires:Lib.t list Or_exn.t
  -> name:string
  -> linkage:Linkage.t
  -> top_sorted_modules:(unit, Module.t list) Build.t
  -> ?flags:Ocaml_flags.t
  -> ?link_flags:(unit, string list) Build.t
  -> ?js_of_ocaml:Jbuild.Js_of_ocaml.t
  -> Super_context.t
  -> unit
# 22731 "boot.ml"
end = struct
# 1 "src/exe.ml"
open Import
open Build.O

module SC = Super_context

module Program = struct
  type t =
    { name             : string
    ; main_module_name : Module.Name.t
    }
end

module Linkage = struct
  type t =
    { mode  : Mode.t
    ; ext   : string
    ; flags : string list
    }

  let byte =
    { mode  = Byte
    ; ext   = ".bc"
    ; flags = []
    }

  let native =
    { mode  = Native
    ; ext   = ".exe"
    ; flags = []
    }

  let custom =
    { mode  = Byte
    ; ext   = ".exe"
    ; flags = ["-custom"]
    }

  let native_or_custom (context : Context.t) =
    match context.ocamlopt with
    | None   -> custom
    | Some _ -> native

  let make ~mode ~ext ?(flags=[]) () =
    { mode
    ; ext
    ; flags
    }

  let  o_flags = ["-output-complete-obj"]
  let so_flags_windows = o_flags
  let so_flags_unix    = ["-output-complete-obj"; "-runtime-variant"; "_pic"]

  let of_user_config (ctx : Context.t) (m : Jbuild.Executables.Link_mode.t) =
    let wanted_mode : Mode.t =
      match m.mode with
      | Byte   -> Byte
      | Native -> Native
      | Best   -> Native
    in
    let real_mode : Mode.t =
      match m.mode with
      | Byte   -> Byte
      | Native -> Native
      | Best   -> if Option.is_some ctx.ocamlopt then Native else Byte
    in
    let ext =
      match wanted_mode, m.kind with
      | Byte   , Exe           -> ".bc"
      | Native , Exe           -> ".exe"
      | Byte   , Object        -> ".bc"  ^ ctx.ext_obj
      | Native , Object        -> ".exe" ^ ctx.ext_obj
      | Byte   , Shared_object -> ".bc"  ^ ctx.ext_dll
      | Native , Shared_object ->          ctx.ext_dll
    in
    let flags =
      match m.kind with
      | Exe ->
        if wanted_mode = Native && real_mode = Byte then
          ["-custom"]
        else
          []
      | Object -> o_flags
      | Shared_object ->
        let so_flags =
          if ctx.os_type = "Win32" then
            so_flags_windows
          else
            so_flags_unix
        in
        if real_mode = Native then
          (* The compiler doesn't pass these flags in native mode. This
             looks like a bug in the compiler. *)
          List.concat_map ctx.native_c_libraries ~f:(fun flag ->
            ["-cclib"; flag])
          @ so_flags
        else
          so_flags
    in
    { ext
    ; mode = real_mode
    ; flags
    }
end

let link_exe
      ~dir
      ~obj_dir
      ~scope
      ~requires
      ~name
      ~(linkage:Linkage.t)
      ~top_sorted_modules
      ?(flags=Ocaml_flags.empty)
      ?(link_flags=Build.arr (fun _ -> []))
      ?(js_of_ocaml=Jbuild.Js_of_ocaml.default)
      sctx
  =
  let ctx = SC.context sctx in
  let mode = linkage.mode in
  let exe = Path.relative dir (name ^ linkage.ext) in
  let compiler = Option.value_exn (Context.compiler ctx mode) in
  let artifacts ~ext modules =
    List.map modules ~f:(Module.obj_file ~obj_dir ~ext)
  in
  let modules_and_cm_files =
    Build.memoize "cm files"
      (top_sorted_modules >>^ fun modules ->
       (modules,
        artifacts modules ~ext:(Cm_kind.ext (Mode.cm_kind mode))))
  in
  let register_native_objs_deps build =
    match mode with
    | Byte -> build
    | Native ->
      build >>>
      Build.dyn_paths (Build.arr (fun (modules, _) ->
        artifacts modules ~ext:ctx.ext_obj))
  in
  SC.add_rule sctx
    (Build.fanout3
       (register_native_objs_deps modules_and_cm_files >>^ snd)
       (Ocaml_flags.get flags mode)
       link_flags
     >>>
     Build.of_result_map requires ~f:(fun libs ->
       Build.paths (Lib.L.archive_files libs ~mode ~ext_lib:ctx.ext_lib))
     >>>
     Build.run ~context:ctx
       (Ok compiler)
       [ Dyn (fun (_, flags,_) -> As flags)
       ; A "-o"; Target exe
       ; As linkage.flags
       ; Dyn (fun (_, _, link_flags) -> As link_flags)
       ; Arg_spec.of_result_map requires ~f:(fun libs ->
           Lib.L.link_flags libs ~mode ~stdlib_dir:ctx.stdlib_dir)
       ; Dyn (fun (cm_files, _, _) -> Deps cm_files)
       ]);
  if linkage.ext = ".bc" then
    let rules =
      Js_of_ocaml_rules.build_exe sctx ~dir ~js_of_ocaml ~src:exe ~requires
    in
    let cm_and_flags =
      Build.fanout
        (modules_and_cm_files >>^ snd)
        (SC.expand_and_eval_set sctx ~scope ~dir js_of_ocaml.flags
           ~standard:(Js_of_ocaml_rules.standard ()))
    in
    SC.add_rules sctx (List.map rules ~f:(fun r -> cm_and_flags >>> r))

let build_and_link_many
      ~dir ~obj_dir ~programs ~modules
      ~scope
      ~linkages
      ?(requires=Ok [])
      ?already_used
      ?(flags=Ocaml_flags.empty)
      ?link_flags
      ?(js_of_ocaml=Jbuild.Js_of_ocaml.default)
      sctx
  =
  let modules =
    Module.Name.Map.map modules ~f:(Module.set_obj_name ~wrapper:None)
  in

  let dep_graphs =
    Ocamldep.rules sctx ~dir ~modules ?already_used
      ~alias_module:None ~lib_interface_module:None
  in

  (* CR-someday jdimino: this should probably say [~dynlink:false] *)
  Module_compilation.build_modules sctx
    ~js_of_ocaml
    ~dynlink:true ~flags ~scope ~dir ~obj_dir ~dep_graphs ~modules
    ~requires ~alias_module:None;

  List.iter programs ~f:(fun { Program.name; main_module_name } ->
    let top_sorted_modules =
      let main = Option.value_exn
                   (Module.Name.Map.find modules main_module_name) in
      Ocamldep.Dep_graph.top_closed_implementations dep_graphs.impl
        [main]
    in
    List.iter linkages ~f:(fun linkage ->
      link_exe sctx
        ~dir
        ~obj_dir
        ~scope
        ~requires
        ~name
        ~linkage
        ~top_sorted_modules
        ~js_of_ocaml
        ~flags
        ?link_flags))

let build_and_link ~dir ~obj_dir ~program =
  build_and_link_many ~dir ~obj_dir ~programs:[program]
# 22951 "boot.ml"
end
module Gen_meta : sig
# 1 "src/gen_meta.mli"
(** Generate a META file *)

open! Import

val gen
  :  package:string
  -> version:string option
  -> Lib.t list
  -> Meta.t
# 22964 "boot.ml"
end = struct
# 1 "src/gen_meta.ml"
open Import
open Meta

module Pub_name = struct
  type t =
    | Dot of t * string
    | Id  of string

  let parse s =
    match String.split s ~on:'.' with
    | [] -> assert false
    | x :: l ->
      let rec loop acc l =
        match l with
        | [] -> acc
        | x :: l -> loop (Dot (acc, x)) l
      in
      loop (Id x) l

  let rec root = function
    | Dot (t, _) -> root t
    | Id n       -> n

  let to_list =
    let rec loop acc = function
      | Dot (t, n) -> loop (n :: acc) t
      | Id n       -> n :: acc
    in
    fun t -> loop [] t

  let to_string t = String.concat ~sep:"." (to_list t)
end

let string_of_deps deps = String_set.to_list deps |> String.concat ~sep:" "

let rule var predicates action value =
  Rule { var; predicates; action; value }
let requires ?(preds=[]) pkgs =
  rule "requires" preds Set (string_of_deps pkgs)
let ppx_runtime_deps ?(preds=[]) pkgs =
  rule "ppx_runtime_deps" preds Set (string_of_deps pkgs)
let description s = rule "description" []      Set s
let directory   s = rule "directory"   []      Set s
let archive preds s = rule "archive"   preds Set s
let plugin preds  s = rule "plugin"    preds Set s
let archives ?(preds=[]) lib =
  let archives = Lib.archives lib in
  let plugins  = Lib.plugins  lib in
  let make ps =
    String.concat ~sep:" " (List.map ps ~f:Path.basename)
  in
  [ archive (preds @ [Pos "byte"  ]) (make archives.byte  )
  ; archive (preds @ [Pos "native"]) (make archives.native)
  ; plugin  (preds @ [Pos "byte"  ]) (make plugins .byte  )
  ; plugin  (preds @ [Pos "native"]) (make plugins .native)
  ]

let gen_lib pub_name lib ~version =
  let desc =
    match Lib.synopsis lib with
    | Some s -> s
    | None ->
      (* CR-someday jdimino: wut? this looks old *)
      match (pub_name : Pub_name.t) with
      | Dot (p, "runtime-lib") ->
        sprintf "Runtime library for %s" (Pub_name.to_string p)
      | Dot (p, "expander") ->
        sprintf "Expander for %s" (Pub_name.to_string p)
      | _ -> ""
  in
  let preds =
    match Lib.kind lib with
    | Normal -> []
    | Ppx_rewriter | Ppx_deriver -> [Pos "ppx_driver"]
  in
  let lib_deps    = Lib.Meta.requires lib in
  let ppx_rt_deps = Lib.Meta.ppx_runtime_deps lib in
  List.concat
    [ version
    ; [ description desc
      ; requires ~preds lib_deps
      ]
    ; archives ~preds lib
    ; if String_set.is_empty ppx_rt_deps then
        []
      else
        [ Comment "This is what jbuilder uses to find out the runtime \
                   dependencies of"
        ; Comment "a preprocessor"
        ; ppx_runtime_deps ppx_rt_deps
        ]
    ; (match Lib.kind lib with
       | Normal -> []
       | Ppx_rewriter | Ppx_deriver ->
         (* Deprecated ppx method support *)
         let no_ppx_driver = Neg "ppx_driver" and no_custom_ppx = Neg "custom_ppx" in
         List.concat
           [ [ Comment "This line makes things transparent for people mixing \
                        preprocessors"
             ; Comment "and normal dependencies"
             ; requires ~preds:[no_ppx_driver]
                 (Lib.Meta.ppx_runtime_deps_for_deprecated_method lib)
             ]
           ; match Lib.kind lib with
           | Normal -> assert false
           | Ppx_rewriter ->
             [ rule "ppx" [no_ppx_driver; no_custom_ppx]
                 Set "./ppx.exe --as-ppx" ]
           | Ppx_deriver ->
             [ rule "requires" [no_ppx_driver; no_custom_ppx] Add
                 "ppx_deriving"
             ; rule "ppxopt" [no_ppx_driver; no_custom_ppx] Set
                 ("ppx_deriving,package:" ^ Pub_name.to_string pub_name)
             ]
           ]
      )
    ; (match Lib.jsoo_runtime lib with
       | [] -> []
       | l  ->
         let root = Pub_name.root pub_name in
         let l = List.map l ~f:Path.basename in
         [ rule "linkopts" [Pos "javascript"] Set
             (List.map l ~f:(sprintf "+%s/%s" root) |> String.concat ~sep:" ")
         ; rule "jsoo_runtime" [] Set
             (String.concat l ~sep:" ")
         ]
      )
    ]

let gen ~package ~version libs =
  let version =
    match version with
    | None -> []
    | Some s -> [rule "version" [] Set s]
  in
  let pkgs =
    List.map libs ~f:(fun lib ->
      let pub_name = Pub_name.parse (Lib.name lib) in
      (pub_name,
       gen_lib pub_name lib ~version))
  in
  let pkgs =
    List.map pkgs ~f:(fun (pn, meta) ->
      match Pub_name.to_list pn with
      | [] -> assert false
      | _package :: path -> (path, meta))
  in
  let pkgs = List.sort pkgs ~compare:(fun (a, _) (b, _) -> compare a b) in
  let rec loop name pkgs =
    let entries, sub_pkgs =
      List.partition_map pkgs ~f:(function
        | ([]    , entries) -> Left  entries
        | (x :: p, entries) -> Right (x, (p, entries)))
    in
    let entries = List.concat entries in
    let subs =
      String_map.of_list_multi sub_pkgs
      |> String_map.to_list
      |> List.map ~f:(fun (name, pkgs) ->
        let pkg = loop name pkgs in
        Package { pkg with
                  entries = directory name :: pkg.entries
                })
    in
    { name
    ; entries = entries @ subs
    }
  in
  loop package pkgs
# 23136 "boot.ml"
end
module Preprocessing : sig
# 1 "src/preprocessing.mli"
(** Preprocessing of OCaml source files *)

open! Import

(** Setup pre-processing and linting rules and return the list of
    pre-processed modules *)
val pp_and_lint_modules
  :  Super_context.t
  -> dir:Path.t
  -> dep_kind:Build.lib_dep_kind
  -> modules:Module.t Module.Name.Map.t
  -> lint:Jbuild.Preprocess_map.t
  -> preprocess:Jbuild.Preprocess_map.t
  -> preprocessor_deps:(unit, Path.t list) Build.t
  -> lib_name:string option
  -> scope:Scope.t
  -> Module.t Module.Name.Map.t 

(** Get a path to a cached ppx driver *)
val get_ppx_driver
  : Super_context.t
  -> scope:Scope.t
  -> (Loc.t * Jbuild.Pp.t) list
  -> Path.t

(** [cookie_library_name lib_name] is ["--cookie"; lib_name] if [lib_name] is not
    [None] *)
val cookie_library_name : string option -> string list

val gen_rules : Super_context.t -> string list -> unit
# 23170 "boot.ml"
end = struct
# 1 "src/preprocessing.ml"
open Import
open Build.O
open Jbuild

module SC = Super_context

let pp_fname fn =
  let fn, ext = Filename.split_extension fn in
  (* We need to to put the .pp before the .ml so that the compiler realises that
     [foo.pp.mli] is the interface for [foo.pp.ml] *)
  fn ^ ".pp" ^ ext

let pped_module ~dir (m : Module.t) ~f =
  let pped_file (kind : Ml_kind.t) (file : Module.File.t) =
    let pp_fname = pp_fname file.name in
    f kind (Path.relative dir file.name) (Path.relative dir pp_fname);
    {file with name = pp_fname}
  in
  { m with
    impl = Option.map m.impl ~f:(pped_file Impl)
  ; intf = Option.map m.intf ~f:(pped_file Intf)
  }

let migrate_driver_main = "ocaml-migrate-parsetree.driver-main"

let ppx_exe sctx ~key =
  Path.relative (SC.build_dir sctx) (".ppx/" ^ key ^ "/ppx.exe")

let build_ppx_driver sctx ~lib_db ~dep_kind ~target pps =
  let ctx = SC.context sctx in
  let mode = Context.best_mode ctx in
  let compiler = Option.value_exn (Context.compiler ctx mode) in
  let pps = pps @ [Pp.of_string migrate_driver_main] in
  let driver, libs =
    let resolved_pps =
      Lib.DB.resolve_pps lib_db
        (List.map pps ~f:(fun x -> (Loc.none, x)))
      (* Extend the dependency stack as we don't have locations at
         this point *)
      |> Result.map_error ~f:(fun e ->
        Dep_path.prepend_exn e
          (Preprocess (pps : Jbuild.Pp.t list :> string list)))
    in
    let driver =
      match resolved_pps with
      | Ok    l -> List.last l
      | Error _ -> None
    in
    (driver,
     Result.bind resolved_pps ~f:Lib.closure
     |> Result.map ~f:Build.return
     |> Build.of_result)
  in
  let libs =
    Build.record_lib_deps ~kind:dep_kind
      (List.map pps ~f:(fun pp -> Lib_dep.of_pp (Loc.none, pp)))
    >>>
    libs
  in
  let libs =
    (* Put the driver back at the end, just before migrate_driver_main *)
    match driver with
    | None -> libs
    | Some driver ->
      libs >>^ fun libs ->
      let libs, drivers =
        List.partition_map libs ~f:(fun lib ->
          if lib == driver || Lib.name lib = migrate_driver_main then
            Right lib
          else
            Left lib)
      in
      let user_driver, migrate_driver =
        List.partition_map drivers ~f:(fun lib ->
          if Lib.name lib = migrate_driver_main then
            Right lib
          else
            Left lib)
      in
      libs @ user_driver @ migrate_driver
  in
  (* Provide a better error for migrate_driver_main given that this
     is an implicit dependency *)
  let libs =
    match Lib.DB.available lib_db migrate_driver_main with
    | false ->
      Build.fail { fail = fun () ->
        die "@{<error>Error@}: I couldn't find '%s'.\n\
             I need this library in order to use ppx rewriters.\n\
             See the manual for details.\n\
             Hint: opam install ocaml-migrate-parsetree"
          migrate_driver_main
      }
      >>>
      libs
    | true ->
      libs
  in
  SC.add_rule sctx
    (libs
     >>>
     Build.dyn_paths
       (Build.arr
          (Lib.L.archive_files ~mode ~ext_lib:ctx.ext_lib))
     >>>
     Build.run ~context:ctx (Ok compiler)
       [ A "-o" ; Target target
       ; Dyn (Lib.L.link_flags ~mode ~stdlib_dir:ctx.stdlib_dir)
       ])

let gen_rules sctx components =
  match components with
  | [key] ->
    let exe = ppx_exe sctx ~key in
    let (key, lib_db) = SC.Scope_key.of_string sctx key in
    let names =
      match key with
      | "+none+" -> []
      | _ -> String.split key ~on:'+'
    in
    let names =
      match List.rev names with
      | [] -> []
      | driver :: rest -> List.sort rest ~compare:String.compare @ [driver]
    in
    let pps = List.map names ~f:Jbuild.Pp.of_string in
    build_ppx_driver sctx pps ~lib_db ~dep_kind:Required ~target:exe
  | _ -> ()

let get_ppx_driver sctx ~scope pps =
  let driver, names =
    match List.rev_map pps ~f:(fun (_loc, pp) -> Pp.to_string pp) with
    | [] -> (None, [])
    | driver :: rest -> (Some driver, rest)
  in
  let sctx = SC.host sctx in
  let name_and_scope_for_key name =
    match Lib.DB.find (Scope.libs scope) name with
    | Error _ ->
      (* XXX unknown but assume it's public *)
      (name, None)
    | Ok lib ->
      (Lib.name lib,
       match Lib.status lib with
       | Private scope_name   -> Some scope_name
       | Public _ | Installed -> None)
  in
  let driver, scope_for_key =
    match driver with
    | None -> (None, None)
    | Some driver ->
      let name, scope_for_key = name_and_scope_for_key driver in
      (Some name, scope_for_key)
  in
  let names, scope_for_key =
    List.fold_left names ~init:([], scope_for_key)
      ~f:(fun (names, scope_for_key) lib ->
        let name, scope_for_key' = name_and_scope_for_key lib in
        (name :: names,
         match scope_for_key, scope_for_key' with
         | Some a, Some b -> assert (a = b); scope_for_key
         | Some _, None   -> scope_for_key
         | None  , Some _ -> scope_for_key'
         | None  , None   -> None))
  in
  let names = List.sort ~compare:String.compare names in
  let names =
    match driver with
    | None        -> names
    | Some driver -> names @ [driver]
  in
  let key =
    match names with
    | [] -> "+none+"
    | _  -> String.concat names ~sep:"+"
  in
  let key =
    match scope_for_key with
    | None            -> key
    | Some scope_name -> SC.Scope_key.to_string key scope_name
  in
  let sctx = SC.host sctx in
  ppx_exe sctx ~key

let target_var = String_with_vars.virt_var __POS__ "@"
let root_var   = String_with_vars.virt_var __POS__ "ROOT"

let cookie_library_name lib_name =
  match lib_name with
  | None -> []
  | Some name -> ["--cookie"; sprintf "library-name=%S" name]

(* Generate rules for the reason modules in [modules] and return a
   a new module with only OCaml sources *)
let setup_reason_rules sctx ~dir (m : Module.t) =
  let ctx = SC.context sctx in
  let refmt =
    Artifacts.binary (SC.artifacts sctx) "refmt" ~hint:"opam install reason" in
  let rule src target =
    let src_path = Path.relative dir src in
    Build.run ~context:ctx refmt
      [ A "--print"
      ; A "binary"
      ; Dep src_path ]
      ~stdout_to:(Path.relative dir target) in
  let to_ml (f : Module.File.t) =
    match f.syntax with
    | OCaml  -> f
    | Reason ->
      let ml = Module.File.to_ocaml f in
      SC.add_rule sctx (rule f.name ml.name);
      ml
  in
  { m with
    impl = Option.map m.impl ~f:to_ml
  ; intf = Option.map m.intf ~f:to_ml
  }

let uses_ppx_driver ~pps =
  match (List.last pps : (_ * Pp.t) option :> (_ * string) option) with
  | Some (_, ("ppx_driver.runner" | "ppxlib.runner")) -> true
  | Some _ | None -> false

let promote_correction ~uses_ppx_driver fn build =
  if not uses_ppx_driver then
    build
  else
    Build.progn
      [ build
      ; Build.return
          (Action.diff ~optional:true
             fn
             (Path.extend_basename fn ~suffix:".ppx-corrected"))
      ]

let lint_module sctx ~dir ~dep_kind ~lint ~lib_name ~scope = Staged.stage (
  let alias = Build_system.Alias.lint ~dir in
  let add_alias fn build =
    SC.add_alias_action sctx alias build
      ~stamp:(List [ Sexp.unsafe_atom_of_string "lint"
                   ; Sexp.To_sexp.(option string) lib_name
                   ; Sexp.atom fn
                   ])
  in
  let lint =
    Per_module.map lint ~f:(function
      | Preprocess.No_preprocessing ->
        (fun ~source:_ ~ast:_ -> ())
      | Action action ->
        (fun ~source ~ast:_ ->
           let action = Action.Unexpanded.Chdir (root_var, action) in
           Module.iter source ~f:(fun _ (src : Module.File.t) ->
             let src_path = Path.relative dir src.name in
             add_alias src.name
               (Build.path src_path
                >>^ (fun _ -> [src_path])
                >>> SC.Action.run sctx
                      action
                      ~dir
                      ~dep_kind
                      ~targets:(Static [])
                      ~scope)))
      | Pps { pps; flags } ->
        let ppx_exe = get_ppx_driver sctx ~scope pps in
        let uses_ppx_driver = uses_ppx_driver ~pps in
        let args : _ Arg_spec.t =
          S [ As flags
            ; As (cookie_library_name lib_name)
            (* This hack is needed until -null is standard:
               https://github.com/ocaml-ppx/ocaml-migrate-parsetree/issues/35
            *)
            ; As (if uses_ppx_driver then
                    [ "-null"; "-diff-cmd"; "-" ]
                  else
                    [])
            ]
        in
        (fun ~source ~ast ->
           Module.iter ast ~f:(fun kind src ->
             let args =
               [ args
               ; Ml_kind.ppx_driver_flag kind
               ; Dep (Path.relative dir src.name)
               ]
             in
             add_alias src.name
               (promote_correction ~uses_ppx_driver
                  (Option.value_exn (Module.file ~dir source kind))
                  (Build.run ~context:(SC.context sctx) (Ok ppx_exe) args))
           )))
  in
  fun ~(source : Module.t) ~ast ->
    Per_module.get lint source.name ~source ~ast)

(* Generate rules to build the .pp files and return a new module map
   where all filenames point to the .pp files *)
let pp_and_lint_modules sctx ~dir ~dep_kind ~modules ~lint ~preprocess
      ~preprocessor_deps ~lib_name ~scope =
  let preprocessor_deps =
    Build.memoize "preprocessor deps" preprocessor_deps
  in
  let lint_module =
    Staged.unstage (lint_module sctx ~dir ~dep_kind ~lint ~lib_name ~scope)
  in
  let preprocess =
    Per_module.map preprocess ~f:(function
      | Preprocess.No_preprocessing ->
        (fun m ->
           let ast = setup_reason_rules sctx ~dir m in
           lint_module ~ast ~source:m;
           ast)
      | Action action ->
        (fun m ->
           let ast =
             pped_module m ~dir ~f:(fun _kind src dst ->
               SC.add_rule sctx
                 (preprocessor_deps
                  >>>
                  Build.path src
                  >>^ (fun _ -> [src])
                  >>>
                  SC.Action.run sctx
                    (Redirect
                       (Stdout,
                        target_var,
                        Chdir (root_var,
                               action)))
                    ~dir
                    ~dep_kind
                    ~targets:(Static [dst])
                    ~scope))
             |> setup_reason_rules sctx ~dir in
           lint_module ~ast ~source:m;
           ast)
      | Pps { pps; flags } ->
        let ppx_exe = get_ppx_driver sctx ~scope pps in
        let uses_ppx_driver = uses_ppx_driver ~pps in
        let args : _ Arg_spec.t =
          S [ As flags
            ; A "--dump-ast"
            ; As (cookie_library_name lib_name)
            ; As (if uses_ppx_driver then ["-diff-cmd"; "-"] else [])
            ]
        in
        (fun m ->
           let ast = setup_reason_rules sctx ~dir m in
           lint_module ~ast ~source:m;
           pped_module ast ~dir ~f:(fun kind src dst ->
             SC.add_rule sctx
               (promote_correction ~uses_ppx_driver
                  (Option.value_exn (Module.file m ~dir kind))
                  (preprocessor_deps
                   >>>
                   Build.run ~context:(SC.context sctx)
                     (Ok ppx_exe)
                     [ args
                     ; A "-o"; Target dst
                     ; Ml_kind.ppx_driver_flag kind; Dep src
                     ])))))
  in
 Module.Name.Map.map modules ~f:(fun (m : Module.t) ->
    Per_module.get preprocess m.name m)
# 23535 "boot.ml"
end
module Install_rules : sig
# 1 "src/install_rules.mli"
module type Params = sig
  val sctx : Super_context.t
end

module Archives (P : Params) : sig
  val lib_archive : Jbuild.Library.t -> dir:Path.t -> ext:string -> Path.t
  val stubs_archive : Jbuild.Library.t -> dir:Path.t -> Path.t
  val dll : Jbuild.Library.t -> dir:Path.t -> Path.t
end

module type Install_params = sig
  include Params
  val module_names_of_lib : Jbuild.Library.t -> dir:Path.t -> Module.t list
  val mlds_of_dir : Jbuild.Documentation.t -> dir:Path.t -> Path.t list
end

(** Generate install rules for META and .install files *)
module Gen (P : Install_params) : sig
  val init : unit -> unit
end
# 23559 "boot.ml"
end = struct
# 1 "src/install_rules.ml"
open Import
open Jbuild
open Build.O
open! No_io

module type Params = sig
  val sctx : Super_context.t
end

module type Install_params = sig
  include Params
  val module_names_of_lib : Library.t -> dir:Path.t -> Module.t list
  val mlds_of_dir : Documentation.t -> dir:Path.t -> Path.t list
end

module Archives(P : Params) = struct
  let ctx = Super_context.context P.sctx

  let lib_archive (lib : Library.t) ~dir ~ext =
    Path.relative dir (lib.name ^ ext)

  let stubs_archive lib ~dir =
    Library.stubs_archive lib ~dir ~ext_lib:ctx.ext_lib

  let dll (lib : Library.t) ~dir =
    Path.relative dir (sprintf "dll%s_stubs%s" lib.name ctx.ext_dll)
end

module Gen(P : Install_params) = struct
  module Alias = Build_system.Alias
  module SC = Super_context
  open P

  include (Archives(P))

  let lib_dune_file ~dir ~name =
    Path.relative dir (name ^ ".dune")

  let gen_lib_dune_file lib =
    SC.add_rule sctx
      (Build.arr (fun () ->
         Format.asprintf "%a@." Sexp.pp
           (Lib.Sub_system.dump_config lib |> Installed_dune_file.gen))
       >>> Build.write_file_dyn
             (lib_dune_file ~dir:(Lib.src_dir lib) ~name:(Lib.name lib)))

  let init_meta () =
    SC.libs_by_package sctx
    |> Package.Name.Map.iter ~f:(fun ((pkg : Package.t), libs) ->
        Lib.Set.iter libs ~f:gen_lib_dune_file;
      let path = Path.append ctx.build_dir pkg.path in
      SC.on_load_dir sctx ~dir:path ~f:(fun () ->
        let meta_fn = "META." ^ (Package.Name.to_string pkg.name) in

        let meta_template = Path.relative path (meta_fn ^ ".template"     ) in
        let meta          = Path.relative path  meta_fn                     in

        let version =
          let get =
            match pkg.version_from_opam_file with
            | Some s -> Build.return (Some s)
            | None ->
              let rec loop = function
                | [] -> Build.return None
                | candidate :: rest ->
                  let p = Path.relative path candidate in
                  Build.if_file_exists p
                    ~then_:(Build.lines_of p
                            >>^ function
                            | ver :: _ -> Some ver
                            | _ -> Some "")
                    ~else_:(loop rest)
              in
              loop
                [ (Package.Name.to_string pkg.name) ^ ".version"
                ; "version"
                ; "VERSION"
                ]
          in
          Super_context.Pkg_version.set sctx pkg get
        in

        let template =
          Build.if_file_exists meta_template
            ~then_:(Build.lines_of meta_template)
            ~else_:(Build.return ["# JBUILDER_GEN"])
        in
        let meta_contents =
          version >>^ fun version ->
          Gen_meta.gen
            ~package:(Package.Name.to_string pkg.name)
            ~version
            (Lib.Set.to_list libs)
        in
        SC.add_rule sctx
          (Build.fanout meta_contents template
           >>^ (fun ((meta : Meta.t), template) ->
             let buf = Buffer.create 1024 in
             let ppf = Format.formatter_of_buffer buf in
             Format.pp_open_vbox ppf 0;
             List.iter template ~f:(fun s ->
               if String.is_prefix s ~prefix:"#" then
                 match
                   String.extract_blank_separated_words
                     (String.sub s ~pos:1 ~len:(String.length s - 1))
                 with
                 | ["JBUILDER_GEN"] -> Format.fprintf ppf "%a@," Meta.pp meta.entries
                 | _ -> Format.fprintf ppf "%s@," s
               else
                 Format.fprintf ppf "%s@," s);
             Format.pp_close_box ppf ();
             Format.pp_print_flush ppf ();
             Buffer.contents buf)
           >>>
           Build.write_file_dyn meta)))

  let lib_install_files ~dir ~sub_dir ~scope ~name (lib : Library.t) =
    let obj_dir = Utils.library_object_directory ~dir lib.name in
    let make_entry section ?dst fn =
      Install.Entry.make section fn
        ~dst:(
          let dst =
            match dst with
            | Some s -> s
            | None   -> Path.basename fn
          in
          match sub_dir with
          | None -> dst
          | Some dir -> sprintf "%s/%s" dir dst)
    in
    let { Mode.Dict.byte; native } =
      Mode_conf.Set.eval lib.modes
        ~has_native:(Option.is_some ctx.ocamlopt)
    in
    let if_ cond l = if cond then l else [] in
    let files =
      let modules = module_names_of_lib lib ~dir in
      List.concat
        [ List.concat_map modules ~f:(fun m ->
            List.concat
              [ [ Module.cm_file_unsafe m ~obj_dir Cmi ]
              ; if_ (native && Module.has_impl m)
                  [ Module.cm_file_unsafe m ~obj_dir Cmx ]
              ; List.filter_map Ml_kind.all ~f:(Module.cmt_file m ~obj_dir)
              ; List.filter_map [m.intf;m.impl] ~f:(function
                  | None -> None
                  | Some f -> Some (Path.relative dir f.name))
              ])
        ; if_ byte [ lib_archive ~dir lib ~ext:".cma" ]
        ; if_ (Library.has_stubs lib) [ stubs_archive ~dir lib ]
        ; if_ native
            (let files =
               [ lib_archive ~dir lib ~ext:".cmxa"
               ; lib_archive ~dir lib ~ext:ctx.ext_lib
               ]
             in
             if ctx.natdynlink_supported && lib.dynlink then
               files @ [ lib_archive ~dir lib ~ext:".cmxs" ]
             else
               files)
        ; List.map lib.buildable.js_of_ocaml.javascript_files ~f:(Path.relative dir)
        ; List.map lib.install_c_headers ~f:(fun fn ->
            Path.relative dir (fn ^ ".h"))
        ]
    in
    let dlls  = if_ (byte && Library.has_stubs lib && lib.dynlink) [dll ~dir lib] in
    let execs =
      match lib.kind with
      | Normal | Ppx_deriver -> []
      | Ppx_rewriter ->
        let pps = [(lib.buildable.loc, Pp.of_string lib.name)] in
        let pps =
          (* This is a temporary hack until we get a standard driver *)
          let deps =
            List.concat_map lib.buildable.libraries ~f:Lib_dep.to_lib_names
          in
          if List.exists deps ~f:(function
            | "ppx_driver" | "ppx_type_conv" -> true
            | _ -> false) then
            pps @ [match Scope.name scope with
              | Some "ppxlib" ->
                Loc.none, Pp.of_string "ppxlib.runner"
              | _ ->
                Loc.none, Pp.of_string "ppx_driver.runner"]
          else
            pps
        in
        let ppx_exe = Preprocessing.get_ppx_driver sctx ~scope pps in
        [ppx_exe]
    in
    List.concat
      [ List.map files ~f:(make_entry Lib    )
      ; List.map execs ~f:(make_entry Libexec)
      ; List.map dlls  ~f:(Install.Entry.make Stublibs)
      ; [make_entry Lib (lib_dune_file ~dir ~name)]
      ]

  let is_odig_doc_file fn =
    List.exists [ "README"; "LICENSE"; "CHANGE"; "HISTORY"]
      ~f:(fun prefix -> String.is_prefix fn ~prefix)

  let local_install_rules (entries : Install.Entry.t list) ~package =
    let install_dir = Config.local_install_dir ~context:ctx.name in
    List.map entries ~f:(fun entry ->
      let dst =
        Path.append install_dir
          (Install.Entry.relative_installed_path entry ~package)
      in
      Build_system.set_package (SC.build_system sctx) entry.src package;
      SC.add_rule sctx (Build.symlink ~src:entry.src ~dst);
      Install.Entry.set_src entry dst)

  let promote_install_file =
    not ctx.implicit &&
    match ctx.kind with
    | Default -> true
    | Opam _  -> false

  let install_file package_path package entries =
    let entries =
      let files = SC.source_files sctx ~src_path:Path.root in
      String_set.fold files ~init:entries ~f:(fun fn acc ->
        if is_odig_doc_file fn then
          Install.Entry.make Doc (Path.relative ctx.build_dir fn) :: acc
        else
          acc)
    in
    let entries =
      let opam = Path.relative package_path (Package.Name.opam_fn package) in
      Install.Entry.make Lib opam ~dst:"opam" :: entries
    in
    let entries =
      let meta_fn = "META." ^ (Package.Name.to_string package) in
      let meta = Path.append ctx.build_dir (Path.relative package_path meta_fn) in
      Install.Entry.make Lib meta ~dst:"META" :: entries
    in
    let fn =
      Path.relative (Path.append ctx.build_dir package_path)
        (Utils.install_file ~package ~findlib_toolchain:ctx.findlib_toolchain)
    in
    let entries = local_install_rules entries ~package in
    let files = Install.files entries in
    SC.add_alias_deps sctx
      (Alias.package_install ~context:ctx ~pkg:package)
      files
      ~dyn_deps:
        (Build_system.package_deps (SC.build_system sctx) package files
         >>^ fun packages ->
         Package.Name.Set.to_list packages
         |> List.map ~f:(fun pkg ->
           Build_system.Alias.package_install
             ~context:(SC.context sctx) ~pkg
           |> Build_system.Alias.stamp_file)
         |> Path.Set.of_list);
    SC.add_rule sctx
      ~mode:(if promote_install_file then
               Promote_but_delete_on_clean
             else
               (* We must ignore the source file since it might be
                  copied to the source tree by another context. *)
               Ignore_source_files)
      (Build.path_set files
       >>^ (fun () ->
         let entries =
           match ctx.findlib_toolchain with
           | None -> entries
           | Some toolchain ->
             let prefix = Path.of_string (toolchain ^ "-sysroot") in
             List.map entries
               ~f:(Install.Entry.add_install_prefix ~prefix ~package)
         in
         Install.gen_install_file entries)
       >>>
       Build.write_file_dyn fn)

  let init_install () =
    let entries_per_package =
      List.concat_map (SC.stanzas_to_consider_for_install sctx)
        ~f:(fun (dir, scope, stanza) ->
          match stanza with
          | Library ({ public = Some { package; sub_dir; name; _ }; _ } as lib) ->
            List.map (lib_install_files ~dir ~sub_dir ~scope ~name lib)
              ~f:(fun x -> package.name, x)
          | Install { section; files; package}->
            List.map files ~f:(fun { Install_conf. src; dst } ->
              (package.name,
               Install.Entry.make section (Path.relative dir src) ?dst))
          | Documentation ({ package; _ } as d) ->
            List.map ~f:(fun mld ->
              (package.name,
               (Install.Entry.make
                  ~dst:(sprintf "odoc-pages/%s" (Path.basename mld))
                  Install.Section.Doc mld))
            ) (mlds_of_dir d ~dir)
          | _ -> [])
      |> Package.Name.Map.of_list_multi
    in
    Package.Name.Map.iter (SC.packages sctx) ~f:(fun (pkg : Package.t) ->
      let stanzas =
        Option.value (Package.Name.Map.find entries_per_package pkg.name)
          ~default:[]
      in
      install_file pkg.path pkg.name stanzas)

  let init_install_files () =
    if not ctx.implicit then
      Package.Name.Map.iteri (SC.packages sctx)
        ~f:(fun pkg { Package.path = src_path; _ } ->
          let install_fn =
            Utils.install_file ~package:pkg
              ~findlib_toolchain:ctx.findlib_toolchain
          in

          let path = Path.append ctx.build_dir src_path in
          let install_alias = Alias.install ~dir:path in
          let install_file = Path.relative path install_fn in
          SC.add_alias_deps sctx install_alias (Path.Set.singleton install_file))

  let init () =
    init_meta ();
    init_install ();
    init_install_files ()
end
# 23885 "boot.ml"
end
module Jbuilder_opam_file_format = struct
# 1 "vendor/boot/jbuilder_opam_file_format.ml"
module OpamParserTypes = struct
  type value =
    | String of unit * string
    | List of unit * value list
    | Other

  type opamfile_item =
    | Variable of unit * string * value
    | Other

  type opamfile =
    { file_contents : opamfile_item list
    ; file_name     : string
    }
end
module OpamBaseParser = struct
  open OpamParserTypes
  let main _lex _lexbuf fn =
    assert (fn = "jbuilder.opam");
    { file_contents = []
    ; file_name     = fn
    }
end
module OpamLexer = struct
  exception Error of string
  let token _ = assert false
end
# 23916 "boot.ml"
end
module Opam_file : sig
# 1 "src/opam_file.mli"
(** Parsing and interpretation of opam files *)

open Jbuilder_opam_file_format.OpamParserTypes

(** Type of opam files *)
type t = opamfile

(** Load a file *)
val load : string -> t

(** Extracts a field *)
val get_field : t -> string -> value option
# 23932 "boot.ml"
end = struct
# 1 "src/opam_file.ml"
open Import
open Jbuilder_opam_file_format
open OpamParserTypes

type t = opamfile

let load fn =
  Io.with_lexbuf_from_file fn ~f:(fun lb ->
    try
      OpamBaseParser.main OpamLexer.token lb fn
    with
    | OpamLexer.Error msg ->
      Loc.fail_lex lb "%s" msg
    | Parsing.Parse_error ->
      Loc.fail_lex lb "Parse error")

let get_field t name =
  List.find_map t.file_contents
    ~f:(function
      | Variable (_, var, value) when name = var ->
        Some value
      | _ -> None)
# 23957 "boot.ml"
end
module Jbuild_load : sig
# 1 "src/jbuild_load.mli"
open Jbuild

module Jbuilds : sig
  type t

  val eval
    :  t
    -> context:Context.t
    -> (Path.t * Scope_info.t * Stanzas.t) list Fiber.t
end

type conf =
  { file_tree : File_tree.t
  ; jbuilds   : Jbuilds.t
  ; packages  : Package.t Package.Name.Map.t
  ; scopes    : Scope_info.t list
  }

val load
  :  ?extra_ignored_subtrees:Path.Set.t
  -> ?ignore_promoted_rules:bool
  -> unit
  -> conf
# 23984 "boot.ml"
end = struct
# 1 "src/jbuild_load.ml"
open Import
open Jbuild

let filter_stanzas ~ignore_promoted_rules stanzas =
  if ignore_promoted_rules then
    List.filter stanzas ~f:(function
      | Stanza.Rule { mode = Promote; _ } -> false
      | _ -> true)
  else
    stanzas

module Jbuilds = struct
  type script =
    { dir   : Path.t
    ; scope : Scope_info.t
    }

  type one =
    | Literal of (Path.t * Scope_info.t * Stanza.t list)
    | Script of script

  type t =
    { jbuilds               : one list
    ; ignore_promoted_rules : bool
    }

  let generated_jbuilds_dir = Path.(relative root) "_build/.jbuilds"

  let ensure_parent_dir_exists path =
    match Path.kind path with
    | Local path -> Path.Local.ensure_parent_directory_exists path
    | External _ -> ()

  type requires = No_requires | Unix

  let extract_requires ~fname str =
    let rec loop n lines acc =
      match lines with
      | [] -> acc
      | line :: lines ->
        let acc =
          match Scanf.sscanf line "#require %S" (fun x -> x) with
          | exception _ -> acc
          | s ->
            match String.split s ~on:',' with
            | [] -> acc
            | ["unix"] -> Unix
            | _ ->
              let start =
                { Lexing.
                  pos_fname = fname
                ; pos_lnum  = n
                ; pos_cnum  = 0
                ; pos_bol   = 0
                }
              in
              Loc.fail
                { start; stop = { start with pos_cnum = String.length line } }
                "Using libraries other that \"unix\" is not supported.\n\
                 See the manual for details.";
        in
        loop (n + 1) lines acc
    in
    loop 1 (String.split str ~on:'\n') No_requires

  let create_plugin_wrapper (context : Context.t) ~exec_dir ~plugin ~wrapper ~target =
    let plugin = Path.to_string plugin in
    let plugin_contents = Io.read_file plugin in
    Io.with_file_out (Path.to_string wrapper) ~f:(fun oc ->
      let ocamlc_config =
        let vars =
          Ocaml_config.to_list context.ocaml_config
          |> List.map ~f:(fun (k, v) -> k, Ocaml_config.Value.to_string v)
        in
        let longest = String.longest_map vars ~f:fst in
        List.map vars ~f:(fun (k, v) -> sprintf "%-*S , %S" (longest + 2) k v)
        |> String.concat ~sep:"\n      ; "
      in
      Printf.fprintf oc {|
let () =
  Hashtbl.add Toploop.directive_table "require" (Toploop.Directive_string ignore);
  Hashtbl.add Toploop.directive_table "use" (Toploop.Directive_string (fun _ ->
    failwith "#use is not allowed inside jbuild in OCaml syntax"));
  Hashtbl.add Toploop.directive_table "use_mod" (Toploop.Directive_string (fun _ ->
    failwith "#use is not allowed inside jbuild in OCaml syntax"))

module Jbuild_plugin = struct
  module V1 = struct
    let context       = %S
    let ocaml_version = %S

    let ocamlc_config =
      [ %s
      ]

    let send s =
      let oc = open_out_bin %S in
      output_string oc s;
      close_out oc
  end
end
# 1 %S
%s|}
        context.name
        context.version_string
        ocamlc_config
        (Path.reach ~from:exec_dir target)
        plugin plugin_contents);
    extract_requires ~fname:plugin plugin_contents

  let eval { jbuilds; ignore_promoted_rules } ~(context : Context.t) =
    let open Fiber.O in
    let static, dynamic =
      List.partition_map jbuilds ~f:(function
        | Literal x -> Left  x
        | Script  x -> Right x)
    in
    Fiber.parallel_map dynamic ~f:(fun { dir; scope } ->
      let file = Path.relative dir "jbuild" in
      let generated_jbuild =
        Path.append (Path.relative generated_jbuilds_dir context.name) file
      in
      let wrapper = Path.extend_basename generated_jbuild ~suffix:".ml" in
      ensure_parent_dir_exists generated_jbuild;
      let requires =
        create_plugin_wrapper context ~exec_dir:dir ~plugin:file ~wrapper
          ~target:generated_jbuild
      in
      let context = Option.value context.for_host ~default:context in
      let cmas =
        match requires with
        | No_requires -> []
        | Unix        -> ["unix.cma"]
      in
      let args =
        List.concat
          [ [ "-I"; "+compiler-libs" ]
          ; cmas
          ; [ Path.to_absolute_filename wrapper ]
          ]
      in
      (* CR-someday jdimino: if we want to allow plugins to use findlib:
         {[
           let args =
             match context.toplevel_path with
             | None -> args
             | Some path -> "-I" :: Path.reach ~from:dir path :: args
           in
         ]}
      *)
      Process.run Strict ~dir:(Path.to_string dir)
        ~env:context.env
        (Path.to_string context.ocaml)
        args
      >>= fun () ->
      if not (Path.exists generated_jbuild) then
        die "@{<error>Error:@} %s failed to produce a valid jbuild file.\n\
             Did you forgot to call [Jbuild_plugin.V*.send]?"
          (Path.to_string file);
      let sexps = Sexp.load ~fname:(Path.to_string generated_jbuild) ~mode:Many in
      Fiber.return (dir, scope, Stanzas.parse scope sexps ~file:generated_jbuild
                                |> filter_stanzas ~ignore_promoted_rules))
    >>| fun dynamic ->
    static @ dynamic
end

type conf =
  { file_tree : File_tree.t
  ; jbuilds   : Jbuilds.t
  ; packages  : Package.t Package.Name.Map.t
  ; scopes    : Scope_info.t list
  }

let load ~dir ~scope ~ignore_promoted_rules =
  let file = Path.relative dir "jbuild" in
  match Sexp.load_many_or_ocaml_script (Path.to_string file) with
  | Sexps sexps ->
    Jbuilds.Literal (dir, scope,
                     Stanzas.parse scope sexps ~file
                     |> filter_stanzas ~ignore_promoted_rules)
  | Ocaml_script ->
    Script { dir; scope }

let load ?extra_ignored_subtrees ?(ignore_promoted_rules=false) () =
  let ftree = File_tree.load Path.root ?extra_ignored_subtrees in
  let packages =
    File_tree.fold ftree ~traverse_ignored_dirs:false ~init:[] ~f:(fun dir pkgs ->
      let path = File_tree.Dir.path dir in
      let files = File_tree.Dir.files dir in
      String_set.fold files ~init:pkgs ~f:(fun fn acc ->
        match Filename.split_extension fn with
        | (pkg, ".opam") when pkg <> "" ->
          let version_from_opam_file =
            let opam = Opam_file.load (Path.relative path fn |> Path.to_string) in
            match Opam_file.get_field opam "version" with
            | Some (String (_, s)) -> Some s
            | _ -> None
          in
          let name = Package.Name.of_string pkg in
          (name,
           { Package. name
           ; path
           ; version_from_opam_file
           }) :: acc
        | _ -> acc))
  in
  let packages =
    Package.Name.Map.of_list_multi packages
    |> Package.Name.Map.mapi ~f:(fun name pkgs ->
      match pkgs with
      | [pkg] -> pkg
      | _ ->
        die "Too many opam files for package %S:\n%s"
          (Package.Name.to_string name)
          (String.concat ~sep:"\n"
             (List.map pkgs ~f:(fun pkg ->
                sprintf "- %s" (Path.to_string (Package.opam_file pkg))))))
  in
  let scopes =
    Package.Name.Map.values packages
    |> List.map ~f:(fun pkg -> (pkg.Package.path, pkg))
    |> Path.Map.of_list_multi
    |> Path.Map.map ~f:Scope_info.make
  in
  let scopes =
    if Path.Map.mem scopes Path.root then
      scopes
    else
      Path.Map.add scopes Path.root Scope_info.anonymous
  in
  let rec walk dir jbuilds scope =
    if File_tree.Dir.ignored dir then
      jbuilds
    else begin
      let path = File_tree.Dir.path dir in
      let files = File_tree.Dir.files dir in
      let sub_dirs = File_tree.Dir.sub_dirs dir in
      let scope = Option.value (Path.Map.find scopes path) ~default:scope in
      let jbuilds =
        if String_set.mem files "jbuild" then
          let jbuild = load ~dir:path ~scope ~ignore_promoted_rules in
          jbuild :: jbuilds
        else
          jbuilds
      in
      String_map.fold sub_dirs ~init:jbuilds
        ~f:(fun dir jbuilds -> walk dir jbuilds scope)
    end
  in
  let jbuilds = walk (File_tree.root ftree) [] Scope_info.anonymous in
  { file_tree = ftree
  ; jbuilds = { jbuilds; ignore_promoted_rules }
  ; packages
  ; scopes = Path.Map.values scopes
  }
# 24242 "boot.ml"
end
module Menhir : sig
# 1 "src/menhir.mli"
(** Menhir rules *)

(** Generate the rules for a [(menhir ...)] stanza. Return the list of
    targets that are generated by these rules. This list of targets is
    used by the code that computes the list of modules in the
    directory. *)
val gen_rules
  :  Super_context.t
  -> dir:Path.t
  -> scope:Scope.t
  -> Jbuild.Menhir.t
  -> Path.t list
# 24258 "boot.ml"
end = struct
# 1 "src/menhir.ml"
open Import
open Build.O
open! No_io

module SC = Super_context

(* - [sctx]: super context. Stores all the informations about the
   current build context. The current compiler can be obtained via:

   {[
     (SC.context sctx).ocamlc
   ]}

   - [dir]: directory inside [_build/<context>/...] where the build happens.
   If the [(menhir ...)] appears in [src/jbuild], then [dir] is of the form
   [_build/<context>/src], for instance [_build/default/src].

   - [scope]: represent the scope this stanza is part of. Jbuilder allows to
   build multiple projects at once and splits the source tree into one
   scope per project

   - [t]: the parsed [(menhir ...)] stanza
*)
let gen_rules sctx ~dir ~scope (t : Jbuild.Menhir.t) =
  let targets n = List.map ~f:(Path.relative dir) [n ^ ".ml"; n ^ ".mli"] in
  (* This expands special variables such as ${ROOT} in the flags *)
  let flags = SC.expand_and_eval_set sctx ~scope ~dir t.flags ~standard:[] in
  let menhir_binary =
    SC.resolve_program sctx "menhir" ~hint:"opam install menhir"
  in
  (* [extra_targets] is to tell Jbuilder about generated files that do
     not appear in the menhir command line. *)
  let menhir ~extra_targets args =
    flags
    >>>
    Build.run ~extra_targets
      menhir_binary
      ~dir
      ~context:(SC.context sctx)
      args
  in
  let add_rule_get_targets =
    SC.add_rule_get_targets sctx ~mode:t.mode ~loc:t.loc
  in
  let mly name = Path.relative dir (name ^ ".mly") in
  match t.merge_into with
  | None ->
    List.concat_map t.modules ~f:(fun name ->
      add_rule_get_targets (
        menhir
          ~extra_targets:(targets name)
          [ Dyn (fun x -> As x)
          ; Dep (mly name)
          ]))
  | Some merge_into ->
    add_rule_get_targets (
      menhir
        ~extra_targets:(targets merge_into)
        [ A "--base" ; A merge_into
        ; Dyn (fun x -> As x)
        ; Deps (List.map ~f:mly t.modules)
        ]
    )
# 24324 "boot.ml"
end
module Merlin : sig
# 1 "src/merlin.mli"
(** Merlin rules *)

open Import

type t

val make
  :  ?requires:(Lib.t list, exn) result
  -> ?flags:(unit, string list) Build.t
  -> ?preprocess:Jbuild.Preprocess.t
  -> ?libname:string
  -> ?source_dirs: Path.Set.t
  -> ?objs_dirs:Path.Set.t
  -> unit
  -> t

val add_source_dir : t -> Path.t -> t

val merge_all : t list -> t option

(** Add rules for generating the .merlin in a directory *)
val add_rules
  : Super_context.t
  -> dir:Path.t
  -> scope:Scope.t
  -> t
  -> unit
# 24355 "boot.ml"
end = struct
# 1 "src/merlin.ml"
open Import
open Build.O
open! No_io

module SC = Super_context

module Preprocess = struct
  type t =
    | Pps of Jbuild.Preprocess.pps
    | Other

  let make : Jbuild.Preprocess.t -> t = function
    | Pps pps -> Pps pps
    | _       -> Other

  let merge a b =
    match a, b with
    | Other, Other -> Other
    | Pps _, Other -> a
    | Other, Pps _ -> b
    | Pps { pps = pps1; flags = flags1 },
      Pps { pps = pps2; flags = flags2 } ->
      match
        match List.compare flags1 flags2 ~compare:String.compare with
        | Eq ->
          List.compare pps1 pps2 ~compare:(fun (_, a) (_, b) ->
            Jbuild.Pp.compare a b)
        | ne -> ne
      with
      | Eq -> a
      | _  -> Other
end

type t =
  { requires   : Lib.Set.t
  ; flags      : (unit, string list) Build.t
  ; preprocess : Preprocess.t
  ; libname    : string option
  ; source_dirs: Path.Set.t
  ; objs_dirs  : Path.Set.t
  }

let make
      ?(requires=Ok [])
      ?(flags=Build.return [])
      ?(preprocess=Jbuild.Preprocess.No_preprocessing)
      ?libname
      ?(source_dirs=Path.Set.empty)
      ?(objs_dirs=Path.Set.empty)
      () =
  (* Merlin shouldn't cause the build to fail, so we just ignore errors *)
  let requires =
    match requires with
    | Ok    l -> Lib.Set.of_list l
    | Error _ -> Lib.Set.empty
  in
  { requires
  ; flags      = Build.catch flags    ~on_error:(fun _ -> [])
  ; preprocess = Preprocess.make preprocess
  ; libname
  ; source_dirs
  ; objs_dirs
  }

let add_source_dir t dir =
  { t with source_dirs = Path.Set.add t.source_dirs dir }

let ppx_flags sctx ~dir:_ ~scope ~src_dir:_ { preprocess; libname; _ } =
  match preprocess with
  | Pps { pps; flags } ->
    let exe = Preprocessing.get_ppx_driver sctx ~scope pps in
    let command =
      List.map (Path.to_absolute_filename exe
                :: "--as-ppx"
                :: Preprocessing.cookie_library_name libname
                @ flags)
        ~f:quote_for_shell
      |> String.concat ~sep:" "
    in
    [sprintf "FLG -ppx %s" (Filename.quote command)]
  | Other -> []

let dot_merlin sctx ~dir ~scope ({ requires; flags; _ } as t) =
  match Path.drop_build_context dir with
  | Some remaindir ->
    let merlin_file = Path.relative dir ".merlin" in
    (* We make the compilation of .ml/.mli files depend on the
       existence of .merlin so that they are always generated, however
       the command themselves don't read the merlin file, so we don't
       want to declare a dependency on the contents of the .merlin
       file.

       Currently jbuilder doesn't support declaring a dependency only
       on the existence of a file, so we have to use this trick. *)
    SC.add_rule sctx
      (Build.path merlin_file
       >>>
       Build.create_file (Path.relative dir ".merlin-exists"));
    SC.add_rule sctx ~mode:Promote_but_delete_on_clean (
      flags
      >>^ (fun flags ->
        let ppx_flags = ppx_flags sctx ~dir ~scope ~src_dir:remaindir t in
        let libs =
          Lib.Set.fold requires ~init:[] ~f:(fun (lib : Lib.t) acc ->
            let serialize_path = Path.reach ~from:remaindir in
            let bpath = serialize_path (Lib.obj_dir lib) in
            let spath =
              Lib.src_dir lib
              |> Path.drop_optional_build_context
              |> serialize_path
            in
            ("B " ^ bpath) :: ("S " ^ spath) :: acc
          )
        in
        let source_dirs =
          Path.Set.fold t.source_dirs ~init:[] ~f:(fun path acc ->
            let path = Path.reach path ~from:remaindir in
            ("S " ^ path)::acc
          )
        in
        let objs_dirs =
          Path.Set.fold t.objs_dirs ~init:[] ~f:(fun path acc ->
            let path = Path.reach path ~from:remaindir in
            ("B " ^ path)::acc
          )
        in
        let flags =
          match flags with
          | [] -> []
          | _  ->
            let escaped_flags = List.map ~f:quote_for_shell flags in
            ["FLG " ^ String.concat escaped_flags ~sep:" "]
        in
        let dot_merlin =
          List.concat
            [ source_dirs
            ; objs_dirs
            ; libs
            ; flags
            ; ppx_flags
            ]
        in
        dot_merlin
        |> String_set.of_list
        |> String_set.to_list
        |> List.map ~f:(Printf.sprintf "%s\n")
        |> String.concat ~sep:"")
      >>>
      Build.write_file_dyn merlin_file
    )
  | _ ->
    ()

let merge_two a b =
  { requires = Lib.Set.union a.requires b.requires
  ; flags = a.flags &&& b.flags >>^ (fun (a, b) -> a @ b)
  ; preprocess = Preprocess.merge a.preprocess b.preprocess
  ; libname =
      (match a.libname with
       | Some _ as x -> x
       | None -> b.libname)
  ; source_dirs = Path.Set.union a.source_dirs b.source_dirs
  ; objs_dirs = Path.Set.union a.objs_dirs b.objs_dirs
  }

let merge_all = function
  | [] -> None
  | init::ts -> Some (List.fold_left ~init ~f:merge_two ts)

let add_rules sctx ~dir ~scope merlin =
  if (SC.context sctx).merlin then
    dot_merlin sctx ~dir ~scope merlin
# 24530 "boot.ml"
end
module Modules_partitioner : sig
# 1 "src/modules_partitioner.mli"
(** Checks modules partitioning inside a directory *)

type t

val create
  :  dir:Path.t
  -> all_modules:Module.t Module.Name.Map.t
  -> t

(** [acknowledge t ~loc ~modules] registers the fact that [modules]
    are associated with [loc].

    Returns the set of modules that are already used at another
    location.
*)
val acknowledge
  :  t
  -> loc:Loc.t
  -> modules:Module.t Module.Name.Map.t
  -> Module.Name.Set.t

(** To be called after processing a directory. Emit warnings about
    detected problems *)
val emit_warnings : t -> unit
# 24558 "boot.ml"
end = struct
# 1 "src/modules_partitioner.ml"
open Import

type t =
  { dir          : Path.t
  ; all_modules  : Module.t Module.Name.Map.t
  ; mutable used : Loc.t list Module.Name.Map.t
  }

let create ~dir ~all_modules =
  { dir
  ; all_modules
  ; used = Module.Name.Map.empty
  }

let acknowledge t ~loc ~modules =
  let already_used =
    Module.Name.Map.merge modules t.used ~f:(fun _name x l ->
      Option.some_if (Option.is_some x && Option.is_some l) ())
    |> Module.Name.Map.keys
    |> Module.Name.Set.of_list
  in
  t.used <-
    Module.Name.Map.merge modules t.used ~f:(fun _name x l ->
      match x with
      | None -> l
      | Some _ -> Some (loc :: Option.value l ~default:[]));
  already_used

let emit_warnings t =
  let loc =
    Utils.jbuild_file_in ~dir:t.dir
    |> Path.to_string
    |> Loc.in_file
  in
  Module.Name.Map.iteri t.used ~f:(fun name locs ->
    if List.length locs > 1 then
      Loc.warn loc
        "Module %a is used in several stanzas:@\n\
         @[<v>%a@]@\n\
         @[%a@]@\n\
         This warning will become an error in the future."
        Module.Name.pp_quote name
        (Fmt.list (Fmt.prefix (Fmt.string "- ") Loc.pp_file_colon_line))
        locs
        Format.pp_print_text
        "To remove this warning, you must specify an explicit \"modules\" \
         field in every library, executable, and executables stanzas in \
         this jbuild file. Note that each module cannot appear in more \
         than one \"modules\" field - it must belong to a single library \
         or executable.")
# 24611 "boot.ml"
end
module Odoc : sig
# 1 "src/odoc.mli"
(** Odoc rules *)

open Import
open Jbuild

module Gen (S : sig val sctx : Super_context.t end) : sig

  val setup_library_odoc_rules
    :  Library.t
    -> scope:Scope.t
    -> modules:Module.t Module.Name.Map.t
    -> requires:Lib.t list Or_exn.t
    -> dep_graphs:Ocamldep.Dep_graphs.t
    -> unit

  val init
    :  modules_by_lib:(dir:Path.t -> Library.t -> Module.t list)
    -> mlds_of_dir:(Documentation.t -> dir:Path.t -> Path.t list)
    -> unit

  val gen_rules : dir:Path.t -> string list -> unit
end
# 24637 "boot.ml"
end = struct
# 1 "src/odoc.boot.ml"

module Gen (S : sig val sctx : Super_context.t end) = struct

  let setup_library_odoc_rules _ ~scope:_ ~modules:_ ~requires:_
        ~dep_graphs:_ = ()

  let init ~modules_by_lib:_ ~mlds_of_dir:_ = ()

  let gen_rules ~dir:_ _ = ()
end
# 24650 "boot.ml"
end
module Sub_system_intf = struct
# 1 "src/sub_system_intf.ml"
open! Import

module type Info = Jbuild.Sub_system_info.S

module type S = sig
  module Info : Info

  (** Instantiated representation of the sub-system. I.e. with names
      resolved using a library database. *)
  type t

  (** Create an instance of the sub-system *)
  val instantiate
    :  resolve:(Loc.t * string -> Lib.t Or_exn.t)
    -> get:(Lib.t -> t option)
    -> Lib.t
    -> Info.t
    -> t
end

(** A backend, for implementors of the sub-system *)
module type Backend = sig
  include S

  (** Description of a backend, such as "inline tests framework" or
      "ppx driver". *)
  val desc : plural:bool -> string

  (** "a" or "an" *)
  val desc_article : string

  (** Library the backend is attached to *)
  val lib : t -> Lib.t

  (** Dump the sub-system configuration. This is used to generate META
      files. *)
  val to_sexp : t -> Syntax.Version.t * Sexp.t
end

module type Registered_backend = sig
  type t

  val get : Lib.t -> t option

  (** Resolve a backend name *)
  val resolve : Lib.DB.t -> Loc.t * string -> t Or_exn.t

  (** Choose a backend by either using the ones written by the user or
      by scanning the dependencies.

      The returned list is sorted by order of dependencies. It is not
      allowed to have two different backend that are completely
      independant, i.e. none of them is in the transitive closure of
      the other one. *)
  val select_extensible_backends
    :  loc:Loc.t
    -> ?written_by_user:t list
    -> extends:(t -> t list Or_exn.t)
    -> Lib.t list
    -> t list Or_exn.t

  (** Choose a backend by either using the ones written by the user or
      by scanning the dependencies.

      A backend can replace other backends *)
  val select_replaceable_backend
    :  loc:Loc.t
    -> ?written_by_user:t list
    -> replaces:(t -> t list Or_exn.t)
    -> Lib.t list
    -> t Or_exn.t
end

(* This is probably what we'll give to plugins *)
module Library_compilation_context = struct
  type t =
    { super_context  : Super_context.t
    ; dir            : Path.t
    ; stanza         : Jbuild.Library.t
    ; scope          : Scope.t
    ; source_modules : Module.t Module.Name.Map.t
    ; compile_info   : Lib.Compile.t
    }
end

(** An end-point, for users of the systems *)
module type End_point = sig
  module Backend : sig
    include Registered_backend

    (** Backends that this backends extends *)
    val extends : t -> t list Or_exn.t
  end

  module Info : sig
    include Info

    (** Additional backends specified by the user at use-site *)
    val backends : t -> (Loc.t * string) list option
  end

  val gen_rules
    :  Library_compilation_context.t
    -> info:Info.t
    -> backends:Backend.t list
    -> unit
end
# 24761 "boot.ml"
end
module Sub_system : sig
# 1 "src/sub_system.mli"
(** Dune sub-systems *)

(** This module allows to define sub-systems. The aim is to define
    everything related to the sub-system, such as the parser for jbuild
    files, the metadata attached to libraries and the specific rules in
    one place.

    A sub-system is generally split into two sub-systems:

    - a backend part, which will be used by libraries that provide an
      implementation of the sub-system
    - an end point, for users of the sub-system

    For instance, for inline tests, the backend is what defines the inline
    tests framework. "ppx_inline_test" and "qtest" are examples of backends.
    An end point is any library that contains inline tests.
*)

open! Import

include module type of struct include Sub_system_intf end

(** Register a sub-system backend:
    - connect the parser to the jbuild files parser
    - connect the metatada generator [M.to_sexp] so that metadata are
      included in installed [<lib>.dune] files
*)
module Register_backend(M : Backend) : Registered_backend with type t := M.t

(** Register a sub-system backend:
    - connect the parser to the jbuild files parser
    - connect the rule generator to the rule generator for libraries
*)
module Register_end_point(M : End_point) : sig end

(** Scan the sub-systems used by the library and generate rules for
    all of the ones that needs it. *)
val gen_rules : Library_compilation_context.t -> unit
# 24803 "boot.ml"
end = struct
# 1 "src/sub_system.ml"
open! Import

include Sub_system_intf

module Register_backend(M : Backend) = struct
  include Jbuild.Sub_system_info.Register(M.Info)
  include Lib.Sub_system.Register(struct
      include M
      type Lib.Sub_system.t += T of t
      let to_sexp = Some to_sexp
    end)

  let top_closure l ~deps =
    match
      Top_closure.Int.top_closure l
        ~key:(fun t -> Lib.unique_id (M.lib t))
        ~deps:(fun t ->
          match deps t with
          | Ok l    -> l
          | Error e -> raise_notrace e)
    with
    | Ok _ as res -> res
    | Error _ ->
      (* Lib.t values can't be cyclic, so we can't have cycles here *)
      assert false
    | exception exn -> Error exn

  module Set =
    Set.Make(struct
      type t = M.t
      let compare a b =
        compare
          (Lib.unique_id (M.lib a))
          (Lib.unique_id (M.lib b))
    end)

  let resolve db (loc, name) =
    let open Result.O in
    Lib.DB.resolve db (loc, name) >>= fun lib ->
    match get lib with
    | None ->
      Error (Loc.exnf loc "%S is not %s %s" name M.desc_article
               (M.desc ~plural:false))
    | Some t -> Ok t

  let written_by_user_or_scan ~loc ~written_by_user ~to_scan =
    match
      match written_by_user with
      | Some l -> l
      | None   -> List.filter_map to_scan ~f:get
    with
    | [] ->
      Error
        (Loc.exnf loc "No %s found." (M.desc ~plural:false))
    | l -> Ok l

  let too_many_backends ~loc backends =
    Loc.exnf loc
      "Too many independant %s found:\n%s"
      (M.desc ~plural:true)
      (String.concat ~sep:"\n"
         (List.map backends ~f:(fun t ->
            let lib = M.lib t in
            sprintf "- %S in %s"
              (Lib.name lib)
              (Path.to_string_maybe_quoted (Lib.src_dir lib)))))

  let select_extensible_backends ~loc ?written_by_user ~extends to_scan =
    let open Result.O in
    written_by_user_or_scan ~loc ~written_by_user ~to_scan
    >>= fun backends ->
    top_closure backends ~deps:extends
    >>= fun backends ->
    let roots =
      let all = Set.of_list backends in
      List.fold_left backends ~init:all ~f:(fun acc t ->
        Set.diff acc (Set.of_list (Result.ok_exn (extends t))))
      |> Set.to_list
    in
    if List.length roots = 1 then
      Ok backends
    else
      Error (too_many_backends ~loc roots)

  let select_replaceable_backend ~loc ?written_by_user ~replaces to_scan =
    let open Result.O in
    written_by_user_or_scan ~loc ~written_by_user ~to_scan
    >>= fun backends ->
    Result.concat_map backends ~f:replaces
    >>= fun replaced_backends ->
    match
      Set.diff (Set.of_list backends) (Set.of_list replaced_backends)
      |> Set.to_list
    with
    | [b] -> Ok b
    | l   -> Error (too_many_backends ~loc l)
end

type Lib.Sub_system.t +=
    Gen of (Library_compilation_context.t -> unit)

module Register_end_point(M : End_point) = struct
  include Jbuild.Sub_system_info.Register(M.Info)

  let gen info (c : Library_compilation_context.t) =
    let open Result.O in
    let backends =
      Lib.Compile.direct_requires c.compile_info >>= fun deps ->
      Lib.Compile.pps             c.compile_info >>= fun pps  ->
      (match M.Info.backends info with
       | None -> Ok None
       | Some l ->
         Result.all (List.map l ~f:(M.Backend.resolve (Scope.libs c.scope)))
         >>| Option.some)
      >>= fun written_by_user ->
      M.Backend.select_extensible_backends
        ~loc:(M.Info.loc info)
        ?written_by_user
        ~extends:M.Backend.extends
        (deps @ pps)
    in
    let fail, backends =
      match backends with
      | Ok backends -> (None, backends)
      | Error e ->
        (Some { fail = fun () -> raise e },
         [])
    in
    match fail with
    | None -> M.gen_rules c ~info ~backends
    | Some fail ->
      Super_context.prefix_rules c.super_context (Build.fail fail)
        ~f:(fun () -> M.gen_rules c ~info ~backends)

  include
    Lib.Sub_system.Register
      (struct
        module Info = M.Info
        type t = Library_compilation_context.t -> unit
        type Lib.Sub_system.t += T = Gen
        let instantiate ~resolve:_ ~get:_ _id info = gen info
        let to_sexp = None
      end)
end

let gen_rules (c : Library_compilation_context.t) =
  List.iter (Lib.Compile.sub_systems c.compile_info) ~f:(function
    | Gen gen -> gen c
    | _ -> ())
# 24955 "boot.ml"
end
module Utop : sig
# 1 "src/utop.mli"
(** Utop rules *)

val utop_exe : Path.t -> Path.t
(** Return the path of the utop bytecode binary inside a directory where
    some libraries are defined. *)

val setup
  : Super_context.t
  -> dir:Path.t
  -> libs:Jbuild.Library.t list
  -> scope:Scope.t
  -> unit
# 24971 "boot.ml"
end = struct
# 1 "src/utop.ml"
open Import
open Jbuild
open Build.O
open! No_io

let exe_name = "utop"
let main_module_name = Module.Name.of_string exe_name
let main_module_filename = exe_name ^ ".ml"

let pp_ml fmt include_dirs =
  let pp_include fmt =
    let pp_sep fmt () = Format.fprintf fmt "@ ; " in
    Format.pp_print_list ~pp_sep (fun fmt p ->
      Format.fprintf fmt "%S" (Path.to_absolute_filename p)
    ) fmt
  in
  Format.fprintf fmt "@[<v 2>Clflags.include_dirs :=@ [ %a@ ]@];@."
    pp_include include_dirs;
  Format.fprintf fmt "@.UTop_main.main ();@."

let add_module_rules sctx ~dir lib_requires =
  let path = Path.relative dir main_module_filename in
  let utop_ml =
    Build.of_result_map lib_requires ~f:(fun libs ->
      Build.arr (fun () ->
        let include_paths =
          let ctx = Super_context.context sctx in
          Path.Set.to_list
            (Lib.L.include_paths libs ~stdlib_dir:ctx.stdlib_dir)
        in
        let b = Buffer.create 64 in
        let fmt = Format.formatter_of_buffer b in
        pp_ml fmt include_paths;
        Format.pp_print_flush fmt ();
        Buffer.contents b))
    >>> Build.write_file_dyn path
  in
  Super_context.add_rule sctx utop_ml

let utop_exe_dir ~dir = Path.relative dir ".utop"

let utop_exe dir =
  Path.relative (utop_exe_dir ~dir) exe_name
  (* Use the [.exe] version. As the utop executable is declared with
     [(modes (byte))], the [.exe] correspond the bytecode linked in
     custom mode. We do that so that it works without hassle when
     generating a utop for a library with C stubs. *)
  |> Path.extend_basename ~suffix:(Mode.exe_ext Mode.Native)

let setup sctx ~dir ~(libs : Library.t list) ~scope =
  match libs with
  | [] -> ()
  | _ :: _ ->
    let modules =
      Module.Name.Map.singleton
        main_module_name
        { Module.
          name = main_module_name
        ; impl = Some { Module.File.
                        name = main_module_filename
                      ; syntax = Module.Syntax.OCaml
                      }
        ; intf = None
        ; obj_name = "" } in
    let utop_exe_dir = utop_exe_dir ~dir in
    let requires =
      let open Result.O in
      Lib.DB.find_many (Scope.libs scope)
        ("utop" :: List.map libs ~f:(fun (lib : Library.t) -> lib.name))
      >>= Lib.closure
    in
    Exe.build_and_link sctx
      ~dir:utop_exe_dir
      ~obj_dir:utop_exe_dir
      ~program:{ name = exe_name ; main_module_name }
      ~modules
      ~scope
      ~linkages:[Exe.Linkage.custom]
      ~requires
      ~flags:(Ocaml_flags.append_common (Ocaml_flags.default ()) ["-w"; "-24"])
      ~link_flags:(Build.return ["-linkall"; "-warn-error"; "-31"]);
    add_module_rules sctx ~dir:utop_exe_dir requires
# 25056 "boot.ml"
end
module Gen_rules : sig
# 1 "src/gen_rules.mli"
open! Import
open Jbuild

(* Generate rules. Returns evaluated jbuilds per context names. *)
val gen
  :  contexts:Context.t list
  -> build_system:Build_system.t
  -> ?external_lib_deps_mode:bool (* default: false *)
  -> ?only_packages:Package.Name.Set.t
  -> Jbuild_load.conf
  -> (Path.t * Scope_info.t * Stanzas.t) list String_map.t Fiber.t
# 25071 "boot.ml"
end = struct
# 1 "src/gen_rules.ml"
open Import
module Menhir_rules = Menhir
open Jbuild
open Build.O
open! No_io

(* +-----------------------------------------------------------------+
   | Utils                                                           |
   +-----------------------------------------------------------------+ *)

module Gen(P : Install_rules.Params) = struct
  module Alias = Build_system.Alias
  module SC = Super_context
  module Odoc = Odoc.Gen(P)

  open P

  let ctx = SC.context sctx

  let stanzas_per_dir =
    List.map (SC.stanzas sctx) ~f:(fun stanzas ->
      (stanzas.SC.Dir_with_jbuild.ctx_dir, stanzas))
    |> Path.Map.of_list_exn

  (* +-----------------------------------------------------------------+
     | Interpretation of [modules] fields                              |
     +-----------------------------------------------------------------+ *)

  module Eval_modules = Ordered_set_lang.Make(Module.Name)(struct
      type t = (Module.t, Module.Name.t * Loc.t) result

      type key = Module.Name.t

      let key = function
        | Error (s, _) -> s
        | Ok m -> Module.name m
    end)

  let parse_modules ~(all_modules : Module.t Module.Name.Map.t)
        ~buildable:(conf : Buildable.t) =
    let fake_modules = ref Module.Name.Map.empty in
    let parse ~loc s =
      let name = Module.Name.of_string s in
      match Module.Name.Map.find all_modules name with
      | Some m -> Ok m
      | None ->
        fake_modules := Module.Name.Map.add !fake_modules name loc;
        Error (name, loc)
    in
    let modules =
      Eval_modules.eval_unordered
        conf.modules
        ~parse
        ~standard:(Module.Name.Map.map all_modules ~f:(fun m -> Ok m))
    in
    let only_present_modules modules =
      Module.Name.Map.filter_map ~f:(function
        | Ok m -> Some m
        | Error (s, loc) ->
          Loc.fail loc "Module %a doesn't exist." Module.Name.pp s
      ) modules
    in
    let modules = only_present_modules modules in
    let intf_only =
      Eval_modules.eval_unordered
        conf.modules_without_implementation
        ~parse
        ~standard:Module.Name.Map.empty
    in
    let intf_only = only_present_modules intf_only in
    Module.Name.Map.iteri !fake_modules ~f:(fun m loc ->
      Loc.warn loc "Module %a is excluded but it doesn't exist."
        Module.Name.pp m
    );
    let real_intf_only =
      Module.Name.Map.filter modules
        ~f:(fun (m : Module.t) -> Option.is_none m.impl)
    in
    if Module.Name.Map.equal intf_only real_intf_only
         ~equal:(fun a b -> Module.name a = Module.name b) then
      modules
    else begin
      let should_be_listed, shouldn't_be_listed =
        Module.Name.Map.merge intf_only real_intf_only ~f:(fun name x y ->
          match x, y with
          | Some _, Some _ -> None
          | None  , Some _ ->
            Some (Left  (String.uncapitalize (Module.Name.to_string name)))
          | Some _, None   ->
            Some (Right (String.uncapitalize (Module.Name.to_string name)))
          | None  , None   -> assert false)
        |> Module.Name.Map.values
        |> List.partition_map ~f:(fun x -> x)
      in
      let list_modules l =
        String.concat ~sep:"\n" (List.map l ~f:(sprintf "- %s"))
      in
      if should_be_listed <> [] then begin
        match Ordered_set_lang.loc conf.modules_without_implementation with
        | None ->
          Loc.warn conf.loc
            "Some modules don't have an implementation.\
             \nYou need to add the following field to this stanza:\
             \n\
             \n  %s\
             \n\
             \nThis will become an error in the future."
            (let tag = Sexp.unsafe_atom_of_string
                         "modules_without_implementation" in
             Sexp.to_string (List [ tag
                                  ; Sexp.To_sexp.(list string) should_be_listed
                                  ]))
        | Some loc ->
          Loc.warn loc
            "The following modules must be listed here as they don't \
             have an implementation:\n\
             %s\n\
             This will become an error in the future."
            (list_modules should_be_listed)
      end;
      if shouldn't_be_listed <> [] then begin
        (* Re-evaluate conf.modules_without_implementation but this
           time keep locations *)
        let module Eval =
          Ordered_set_lang.Make(Module.Name)(struct
            type t = Loc.t * Module.t
            type key = Module.Name.t
            let key (_, m) = Module.name m
          end)
        in
        let parse ~loc s =
          let name = Module.Name.of_string s in
          match Module.Name.Map.find all_modules name with
          | Some m -> m
          | None -> Loc.fail loc "Module %s doesn't exist." s
        in
        let parse ~loc s = (loc, parse ~loc s) in
        let shouldn't_be_listed =
          Eval.eval_unordered conf.modules_without_implementation
            ~parse
            ~standard:(Module.Name.Map.map all_modules ~f:(fun m -> (Loc.none, m)))
          |> Module.Name.Map.values
          |> List.filter ~f:(fun (_, (m : Module.t)) ->
            Option.is_some m.impl)
        in
        (* CR-soon jdimino for jdimino: report all errors *)
        let loc, m = List.hd shouldn't_be_listed in
        Loc.fail loc
          "Module %a has an implementation, it cannot be listed here"
          Module.Name.pp m.name
      end;
      modules
    end

  let parse_mlds ~dir ~(all_mlds : string String_map.t) ~mlds_written_by_user =
    if Ordered_set_lang.is_standard mlds_written_by_user then
      all_mlds
    else
      let mlds =
        Ordered_set_lang.String.eval_unordered
          mlds_written_by_user
          ~parse:(fun ~loc s ->
            match String_map.find all_mlds s with
            | Some s ->
              s
            | None ->
              Loc.fail loc "%s.mld doesn't exist in %s" s
                (Path.to_string_maybe_quoted
                   (Path.drop_optional_build_context dir))
          )
          ~standard:all_mlds in
      mlds

  (* +-----------------------------------------------------------------+
     | User rules & copy files                                         |
     +-----------------------------------------------------------------+ *)

  let interpret_locks ~dir ~scope locks =
    List.map locks ~f:(fun s ->
      Path.relative dir (SC.expand_vars sctx ~dir ~scope s))

  let user_rule (rule : Rule.t) ~dir ~scope =
    let targets : SC.Action.targets =
      match rule.targets with
      | Infer -> Infer
      | Static fns -> Static (List.map fns ~f:(Path.relative dir))
    in
    SC.add_rule_get_targets sctx ~mode:rule.mode ~loc:rule.loc
      ~locks:(interpret_locks ~dir ~scope rule.locks)
      (SC.Deps.interpret sctx ~scope ~dir rule.deps
       >>>
       SC.Action.run
         sctx
         rule.action
         ~dir
         ~dep_kind:Required
         ~targets
         ~scope)

  let copy_files_rules (def: Copy_files.t) ~src_dir ~dir ~scope =
    let loc = String_with_vars.loc def.glob in
    let glob_in_src =
      let src_glob = SC.expand_vars sctx ~dir def.glob ~scope in
      Path.relative src_dir src_glob ~error_loc:loc
    in
    (* The following condition is required for merlin to work.
       Additionally, the order in which the rules are evaluated only
       ensures that [sources_and_targets_known_so_far] returns the
       right answer for sub-directories only. *)
    if not (Path.is_descendant glob_in_src ~of_:src_dir) then
      Loc.fail loc "%s is not a sub-directory of %s"
        (Path.to_string_maybe_quoted glob_in_src) (Path.to_string_maybe_quoted src_dir);
    let glob = Path.basename glob_in_src in
    let src_in_src = Path.parent glob_in_src in
    let re =
      match Glob_lexer.parse_string glob with
      | Ok re ->
        Re.compile re
      | Error (_pos, msg) ->
        Loc.fail (String_with_vars.loc def.glob) "invalid glob: %s" msg
    in
    (* add rules *)
    let src_in_build = Path.append ctx.build_dir src_in_src in
    let files = SC.eval_glob sctx ~dir:src_in_build re in
    List.map files ~f:(fun basename ->
      let file_src = Path.relative src_in_build basename in
      let file_dst = Path.relative dir basename in
      SC.add_rule sctx
        ((if def.add_line_directive
          then Build.copy_and_add_line_directive
          else Build.copy)
           ~src:file_src
           ~dst:file_dst);
      file_dst)

  (* +-----------------------------------------------------------------+
     | "text" file listing                                             |
     +-----------------------------------------------------------------+ *)

  (* Compute the list of "text" files (.ml, .c, ...). This is the list
     of source files + user generated ones. As a side-effect, setup
     user rules and copy_files rules. *)
  let text_files =
    let cache = Hashtbl.create 32 in
    fun ~dir ->
      Hashtbl.find_or_add cache dir ~f:(fun dir ->
        match Path.Map.find stanzas_per_dir dir with
        | None -> String_set.empty
        | Some { stanzas; src_dir; scope; _ } ->
          (* Interpret a few stanzas in order to determine the list of
             files generated by the user. *)
          let generated_files =
            List.concat_map stanzas ~f:(fun stanza ->
              match (stanza : Stanza.t) with
              | Menhir menhir ->
                Menhir_rules.gen_rules sctx ~dir ~scope menhir
                |> List.map ~f:Path.basename
              | Rule rule ->
                List.map (user_rule rule  ~dir ~scope) ~f:Path.basename
              | Copy_files def ->
                List.map (copy_files_rules def ~src_dir ~dir ~scope)
                  ~f:Path.basename
              | Library { buildable; _ } | Executables { buildable; _ } ->
                (* Manually add files generated by the (select ...)
                   dependencies *)
                List.filter_map buildable.libraries ~f:(fun dep ->
                  match (dep : Jbuild.Lib_dep.t) with
                  | Direct _ -> None
                  | Select s -> Some s.result_fn)
              | Documentation _ | Alias _ | Provides _ | Install _ -> [])
            |> String_set.of_list
          in
          String_set.union generated_files
            (SC.source_files sctx ~src_path:src_dir))

  (* +-----------------------------------------------------------------+
     | Modules listing                                                 |
     +-----------------------------------------------------------------+ *)

  let guess_modules ~dir ~files =
    let impl_files, intf_files =
      String_set.to_list files
      |> List.filter_partition_map ~f:(fun fn ->
        (* we aren't using Filename.extension because we want to handle
           filenames such as foo.cppo.ml *)
        match String.lsplit2 fn ~on:'.' with
        | Some (_, "ml") -> Left { Module.File.syntax=OCaml ; name=fn }
        | Some (_, "re") -> Left { Module.File.syntax=Reason ; name=fn }
        | Some (_, "mli") -> Right { Module.File.syntax=OCaml ; name=fn }
        | Some (_, "rei") -> Right { Module.File.syntax=Reason ; name=fn }
        | _ -> Skip)
    in
    let parse_one_set files =
      List.map files ~f:(fun (f : Module.File.t) ->
        (Module.Name.of_string (Filename.chop_extension f.name), f))
      |> Module.Name.Map.of_list
      |> function
      | Ok x -> x
      | Error (name, f1, f2) ->
        let src_dir = Path.drop_build_context_exn dir in
        die "too many files for module %a in %s: %s and %s"
          Module.Name.pp name (Path.to_string src_dir)
          f1.name f2.name
    in
    let impls = parse_one_set impl_files in
    let intfs = parse_one_set intf_files in
    Module.Name.Map.merge impls intfs ~f:(fun name impl intf ->
      Some
        { Module.name
        ; impl
        ; intf
        ; obj_name = ""
        }
    )

  let guess_mlds ~files =
    String_set.to_list files
    |> List.filter_map ~f:(fun fn ->
      match String.lsplit2 fn ~on:'.' with
      | Some (s, "mld") -> Some (s, fn)
      | _ -> None)
    |> String_map.of_list_exn

  let mlds_by_dir =
    let cache = Hashtbl.create 32 in
    fun ~dir ->
      Hashtbl.find_or_add cache dir ~f:(fun dir ->
        let files = text_files ~dir in
        guess_mlds ~files)

  let mlds_of_dir (doc : Documentation.t) ~dir =
    parse_mlds ~dir
      ~all_mlds:(mlds_by_dir ~dir)
      ~mlds_written_by_user:doc.mld_files
    |> String_map.values
    |> List.map ~f:(Path.relative dir)

  let modules_by_dir =
    let cache = Hashtbl.create 32 in
    fun ~dir ->
      Hashtbl.find_or_add cache dir ~f:(fun dir ->
        let files = text_files ~dir in
        guess_modules ~dir ~files)

  type modules_by_lib =
    { modules          : Module.t Module.Name.Map.t
    ; alias_module     : Module.t option
    ; main_module_name : Module.Name.t
    }

  let modules_by_lib =
    let cache = Hashtbl.create 32 in
    fun (lib : Library.t) ~dir ->
      Hashtbl.find_or_add cache (dir, lib.name) ~f:(fun _ ->
        let all_modules = modules_by_dir ~dir in
        let modules =
          parse_modules ~all_modules ~buildable:lib.buildable
        in
        let main_module_name = Module.Name.of_string lib.name in
        let modules =
          Module.Name.Map.map modules ~f:(fun (m : Module.t) ->
            let wrapper =
              if not lib.wrapped || m.name = main_module_name then
                None
              else
                Some lib.name
            in
            Module.set_obj_name m ~wrapper)
        in
        let alias_module =
          if not lib.wrapped ||
             (Module.Name.Map.cardinal modules = 1 &&
              Module.Name.Map.mem modules main_module_name) then
            None
          else if Module.Name.Map.mem modules main_module_name then
            (* This module needs an implementaion for non-jbuilder
               users of the library:

               https://github.com/ocaml/dune/issues/567 *)
            Some
              { Module.name = Module.Name.add_suffix main_module_name "__"
              ; intf = None
              ; impl = Some { name   = sprintf "%s__.ml-gen" lib.name
                            ; syntax = OCaml
                            }
              ; obj_name = lib.name ^ "__"
              }
          else
            Some
              { Module.name = main_module_name
              ; impl = Some { name   = lib.name ^ ".ml-gen"
                            ; syntax = OCaml
                            }
              ; intf = None
              ; obj_name = lib.name
              }
        in
        { modules; alias_module; main_module_name })

  let module_names_of_lib lib ~dir =
    let { modules; alias_module; _ } = modules_by_lib lib ~dir in
    let modules =
      match alias_module with
      | None -> modules
      | Some m -> Module.Name.Map.add modules m.name m
    in
    Module.Name.Map.values modules

  (* +-----------------------------------------------------------------+
     | Library stuff                                                   |
     +-----------------------------------------------------------------+ *)

  include Install_rules.Archives(P)

  let msvc_hack_cclibs cclibs =
    let f lib =
      if String.is_prefix lib ~prefix:"-l" then
        String.sub lib ~pos:2 ~len:(String.length lib - 2) ^ ".lib"
      else
        lib
    in
    let cclibs = List.map cclibs ~f in
    let f lib =
      if String.is_prefix lib ~prefix:"-l" then
        String.sub lib ~pos:2 ~len:(String.length lib - 2)
      else
        lib
    in
    List.map cclibs ~f

  let build_lib (lib : Library.t) ~scope ~flags ~dir ~obj_dir ~mode
        ~top_sorted_modules ~modules =
    Option.iter (Context.compiler ctx mode) ~f:(fun compiler ->
      let target = lib_archive lib ~dir ~ext:(Mode.compiled_lib_ext mode) in
      let stubs_flags =
        if not (Library.has_stubs lib) then
          []
        else
          let stubs_name = lib.name ^ "_stubs" in
          match mode with
          | Byte -> ["-dllib"; "-l" ^ stubs_name; "-cclib"; "-l" ^ stubs_name]
          | Native -> ["-cclib"; "-l" ^ stubs_name]
      in
      let map_cclibs =
        (* https://github.com/ocaml/dune/issues/119 *)
        if ctx.ccomp_type = "msvc" then
          msvc_hack_cclibs
        else
          fun x -> x
      in
      let artifacts ~ext modules =
        List.map modules ~f:(Module.obj_file ~obj_dir ~ext)
      in
      let obj_deps =
        Build.paths (artifacts modules ~ext:(Cm_kind.ext (Mode.cm_kind mode)))
      in
      let obj_deps =
        match mode with
        | Byte   -> obj_deps
        | Native ->
          obj_deps >>>
          Build.paths (artifacts modules ~ext:ctx.ext_obj)
      in
      SC.add_rule sctx
        (obj_deps
         >>>
         Build.fanout4
           (top_sorted_modules >>^ artifacts ~ext:(Cm_kind.ext (Mode.cm_kind mode)))
           (SC.expand_and_eval_set sctx ~scope ~dir lib.c_library_flags ~standard:[])
           (Ocaml_flags.get flags mode)
           (SC.expand_and_eval_set sctx ~scope ~dir lib.library_flags ~standard:[])
         >>>
         Build.run ~context:ctx (Ok compiler)
           ~extra_targets:(
             match mode with
             | Byte -> []
             | Native -> [lib_archive lib ~dir ~ext:ctx.ext_lib])
           [ Dyn (fun (_, _, flags, _) -> As flags)
           ; A "-a"; A "-o"; Target target
           ; As stubs_flags
           ; Dyn (fun (_, cclibs, _, _) -> Arg_spec.quote_args "-cclib" (map_cclibs cclibs))
           ; Dyn (fun (_, _, _, library_flags) -> As library_flags)
           ; As (match lib.kind with
               | Normal -> []
               | Ppx_deriver | Ppx_rewriter -> ["-linkall"])
           ; Dyn (fun (cm_files, _, _, _) -> Deps cm_files)
           ]))

  let build_c_file (lib : Library.t) ~scope ~dir ~includes c_name =
    let src = Path.relative dir (c_name ^ ".c") in
    let dst = Path.relative dir (c_name ^ ctx.ext_obj) in
    SC.add_rule sctx
      (SC.expand_and_eval_set sctx ~scope ~dir lib.c_flags
         ~standard:(Context.cc_g ctx)
       >>>
       Build.run ~context:ctx
         (* We have to execute the rule in the library directory as
            the .o is produced in the current directory *)
         ~dir
         (Ok ctx.ocamlc)
         [ As (Utils.g ())
         ; includes
         ; Dyn (fun c_flags -> Arg_spec.quote_args "-ccopt" c_flags)
         ; A "-o"; Target dst
         ; Dep src
         ]);
    dst

  let build_cxx_file (lib : Library.t) ~scope ~dir ~includes c_name =
    let src = Path.relative dir (c_name ^ ".cpp") in
    let dst = Path.relative dir (c_name ^ ctx.ext_obj) in
    let open Arg_spec in
    let output_param =
      if ctx.ccomp_type = "msvc" then
        [Concat ("", [A "/Fo"; Target dst])]
      else
        [A "-o"; Target dst]
    in
    SC.add_rule sctx
      (SC.expand_and_eval_set sctx ~scope ~dir lib.cxx_flags
         ~standard:(Context.cc_g ctx)
       >>>
       Build.run ~context:ctx
         (* We have to execute the rule in the library directory as
            the .o is produced in the current directory *)
         ~dir
         (SC.resolve_program sctx ctx.c_compiler)
         ([ S [A "-I"; Path ctx.stdlib_dir]
          ; As (SC.cxx_flags sctx)
          ; includes
          ; Dyn (fun cxx_flags -> As cxx_flags)
          ] @ output_param @
          [ A "-c"; Dep src
          ]));
    dst

  (* In 4.02, the compiler reads the cmi for module alias even with
     [-w -49 -no-alias-deps], so we must sandbox the build of the
     alias module since the modules it references are built after. *)
  let alias_module_build_sandbox = ctx.version < (4, 03, 0)

  let library_rules (lib : Library.t) ~modules_partitioner ~dir ~files ~scope
        ~compile_info =
    let obj_dir = Utils.library_object_directory ~dir lib.name in
    let requires = Lib.Compile.requires compile_info in
    let dep_kind = if lib.optional then Build.Optional else Required in
    let flags = Ocaml_flags.make lib.buildable sctx ~scope ~dir in
    let { modules; main_module_name; alias_module } = modules_by_lib ~dir lib in
    let source_modules = modules in
    let already_used =
      Modules_partitioner.acknowledge modules_partitioner
        ~loc:lib.buildable.loc ~modules
    in
    (* Preprocess before adding the alias module as it doesn't need
       preprocessing *)
    let modules =
      Preprocessing.pp_and_lint_modules sctx ~dir ~dep_kind ~modules ~scope
        ~preprocess:lib.buildable.preprocess
        ~preprocessor_deps:
          (SC.Deps.interpret sctx ~scope ~dir
             lib.buildable.preprocessor_deps)
        ~lint:lib.buildable.lint
        ~lib_name:(Some lib.name)
    in

    let modules =
      match alias_module with
      | None -> modules
      | Some m -> Module.Name.Map.add modules m.name m
    in

    let dep_graphs =
      Ocamldep.rules sctx ~dir ~modules ~already_used ~alias_module
        ~lib_interface_module:(
          if lib.wrapped then
            Module.Name.Map.find modules main_module_name
          else
            None)
    in

    Option.iter alias_module ~f:(fun m ->
      let file =
        match m.impl with
        | Some f -> f
        | None -> Option.value_exn m.intf
      in
      SC.add_rule sctx
        (Build.return
           (Module.Name.Map.values (Module.Name.Map.remove modules m.name)
            |> List.map ~f:(fun (m : Module.t) ->
              sprintf "(** @canonical %s.%s *)\n\
                       module %s = %s\n"
                (Module.Name.to_string main_module_name)
                (Module.Name.to_string m.name)
                (Module.Name.to_string m.name)
                (Module.Name.to_string (Module.real_unit_name m))
            )
            |> String.concat ~sep:"\n")
         >>> Build.write_file_dyn (Path.relative dir file.name)));


    let dynlink = lib.dynlink in
    let js_of_ocaml = lib.buildable.js_of_ocaml in
    Module_compilation.build_modules sctx
      ~js_of_ocaml ~dynlink ~flags ~scope ~dir ~obj_dir ~dep_graphs
      ~modules ~requires ~alias_module;
    Option.iter alias_module ~f:(fun m ->
      let flags = Ocaml_flags.default () in
      Module_compilation.build_module sctx m
        ~js_of_ocaml
        ~dynlink
        ~sandbox:alias_module_build_sandbox
        ~flags:(Ocaml_flags.append_common flags ["-w"; "-49"])
        ~scope
        ~dir
        ~obj_dir
        ~dep_graphs:(Ocamldep.Dep_graphs.dummy m)
        ~includes:(Cm_kind.Dict.make_all (Arg_spec.As []))
        ~alias_module:None);

    if Library.has_stubs lib then begin
      let h_files =
        String_set.to_list files
        |> List.filter_map ~f:(fun fn ->
          if String.is_suffix fn ~suffix:".h" then
            Some (Path.relative dir fn)
          else
            None)
      in
      let o_files =
        let includes =
          Arg_spec.S
            [ Hidden_deps h_files
            ; Arg_spec.of_result_map requires ~f:(fun libs ->
                S [ Lib.L.c_include_flags libs ~stdlib_dir:ctx.stdlib_dir
                  ; Hidden_deps (SC.Libs.file_deps sctx libs ~ext:".h")
                  ])
            ]
        in
        List.map lib.c_names ~f:(
          build_c_file   lib ~scope ~dir ~includes
        ) @ List.map lib.cxx_names ~f:(
          build_cxx_file lib ~scope ~dir ~includes
        )
      in
      match lib.self_build_stubs_archive with
      | Some _ -> ()
      | None ->
        let ocamlmklib ~sandbox ~custom ~targets =
          SC.add_rule sctx ~sandbox
            (SC.expand_and_eval_set sctx ~scope ~dir
               lib.c_library_flags ~standard:[]
             >>>
             Build.run ~context:ctx
               ~extra_targets:targets
               (Ok ctx.ocamlmklib)
               [ As (Utils.g ())
               ; if custom then A "-custom" else As []
               ; A "-o"
               ; Path (Path.relative dir (sprintf "%s_stubs" lib.name))
               ; Deps o_files
               ; Dyn (fun cclibs ->
                   (* https://github.com/ocaml/dune/issues/119 *)
                   if ctx.ccomp_type = "msvc" then
                     let cclibs = msvc_hack_cclibs cclibs in
                     Arg_spec.quote_args "-ldopt" cclibs
                   else
                     As cclibs
                 )
               ])
        in
        let static = stubs_archive lib ~dir in
        let dynamic = dll lib ~dir in
        let modes =
          Mode_conf.Set.eval lib.modes
            ~has_native:(Option.is_some ctx.ocamlopt)
        in
        if modes.native &&
           modes.byte   &&
           lib.dynlink
        then begin
          (* If we build for both modes and support dynlink, use a
             single invocation to build both the static and dynamic
             libraries *)
          ocamlmklib ~sandbox:false ~custom:false ~targets:[static; dynamic]
        end else begin
          ocamlmklib ~sandbox:false ~custom:true ~targets:[static];
          (* We can't tell ocamlmklib to build only the dll, so we
             sandbox the action to avoid overriding the static archive *)
          ocamlmklib ~sandbox:true ~custom:false ~targets:[dynamic]
        end
    end;

    List.iter Cm_kind.all ~f:(fun cm_kind ->
      let files =
        Module.Name.Map.fold modules ~init:Path.Set.empty ~f:(fun m acc ->
          match Module.cm_file m ~obj_dir cm_kind with
          | None -> acc
          | Some fn -> Path.Set.add acc fn)
      in
      SC.Libs.setup_file_deps_alias sctx ~dir lib ~ext:(Cm_kind.ext cm_kind)
        files);
    SC.Libs.setup_file_deps_group_alias sctx ~dir lib ~exts:[".cmi"; ".cmx"];
    SC.Libs.setup_file_deps_alias sctx ~dir lib ~ext:".h"
      (List.map lib.install_c_headers ~f:(fun header ->
         Path.relative dir (header ^ ".h"))
       |> Path.Set.of_list);

    (let modules =
       Module.Name.Map.fold modules ~init:[] ~f:(fun m acc ->
         if Module.has_impl m then
           m :: acc
         else
           acc)
     in
     let top_sorted_modules =
       Ocamldep.Dep_graph.top_closed_implementations dep_graphs.impl modules
     in
     List.iter Mode.all ~f:(fun mode ->
       build_lib lib ~scope ~flags ~dir ~obj_dir ~mode ~top_sorted_modules
         ~modules));
    (* Build *.cma.js *)
    SC.add_rules sctx (
      let src = lib_archive lib ~dir ~ext:(Mode.compiled_lib_ext Mode.Byte) in
      let target = Path.extend_basename src ~suffix:".js" in
      Js_of_ocaml_rules.build_cm sctx ~scope ~dir
        ~js_of_ocaml:lib.buildable.js_of_ocaml ~src ~target);

    if ctx.natdynlink_supported then
      Option.iter ctx.ocamlopt ~f:(fun ocamlopt ->
        let src = lib_archive lib ~dir ~ext:(Mode.compiled_lib_ext Native) in
        let dst = lib_archive lib ~dir ~ext:".cmxs" in
        let build =
          Build.dyn_paths (Build.arr (fun () -> [lib_archive lib ~dir ~ext:ctx.ext_lib]))
          >>>
          Ocaml_flags.get flags Native
          >>>
          Build.run ~context:ctx
            (Ok ocamlopt)
            [ Dyn (fun flags -> As flags)
            ; A "-shared"; A "-linkall"
            ; A "-I"; Path dir
            ; A "-o"; Target dst
            ; Dep src
            ]
        in
        let build =
          if Library.has_stubs lib then
            Build.path (stubs_archive ~dir lib)
            >>>
            build
          else
            build
        in
        SC.add_rule sctx build
      );

    Odoc.setup_library_odoc_rules lib ~requires ~modules ~dep_graphs ~scope;

    let flags =
      match alias_module with
      | None -> Ocaml_flags.common flags
      | Some m ->
        Ocaml_flags.prepend_common ["-open"; Module.Name.to_string m.name] flags
        |> Ocaml_flags.common
    in

    Sub_system.gen_rules
      { super_context = sctx
      ; dir
      ; stanza = lib
      ; scope
      ; source_modules
      ; compile_info
      };

    Merlin.make ()
      ~requires:(Lib.Compile.requires compile_info)
      ~flags
      ~preprocess:(Buildable.single_preprocess lib.buildable)
      ~libname:lib.name
      ~objs_dirs:(Path.Set.singleton obj_dir)

  let library_rules (lib : Library.t) ~modules_partitioner ~dir ~files ~scope =
    let compile_info =
      Lib.DB.get_compile_info (Scope.libs scope) lib.name
        ~allow_overlaps:lib.buildable.allow_overlapping_dependencies
    in
    SC.Libs.gen_select_rules sctx compile_info ~dir;
    SC.Libs.with_lib_deps sctx compile_info ~dir
      ~f:(fun () ->
        library_rules lib ~modules_partitioner ~dir ~files ~scope ~compile_info)

  (* +-----------------------------------------------------------------+
     | Executables stuff                                               |
     +-----------------------------------------------------------------+ *)

  let executables_rules ~dir ~all_modules
        ?modules_partitioner ~scope ~compile_info
        (exes : Executables.t) =
    let requires = Lib.Compile.requires compile_info in
    let modules =
      parse_modules ~all_modules ~buildable:exes.buildable
    in
    let already_used =
      match modules_partitioner with
      | None -> Module.Name.Set.empty
      | Some mp ->
        Modules_partitioner.acknowledge mp
          ~loc:exes.buildable.loc ~modules
    in

    let modules =
      let preprocessor_deps =
        SC.Deps.interpret sctx exes.buildable.preprocessor_deps
          ~scope ~dir
      in
      Preprocessing.pp_and_lint_modules sctx ~dir ~dep_kind:Required ~modules
        ~scope
        ~preprocess:exes.buildable.preprocess
        ~preprocessor_deps
        ~lint:exes.buildable.lint
        ~lib_name:None
    in

    let programs =
      List.map exes.names ~f:(fun (loc, name) ->
        let mod_name = Module.Name.of_string name in
        match Module.Name.Map.find modules mod_name with
        | Some m ->
          if not (Module.has_impl m) then
            Loc.fail loc "Module %a has no implementation."
              Module.Name.pp mod_name
          else
            { Exe.Program.name; main_module_name = mod_name }
        | None -> Loc.fail loc "Module %a doesn't exist."
                    Module.Name.pp mod_name)
    in

    let linkages =
      let module L = Executables.Link_mode in
      let l =
        let has_native = Option.is_some ctx.ocamlopt in
        List.filter_map (L.Set.to_list exes.modes) ~f:(fun (mode : L.t) ->
          if not has_native && mode.mode = Native then
            None
          else
            Some (Exe.Linkage.of_user_config ctx mode))
      in
      (* If bytecode was requested but not native or best version,
         add custom linking *)
      if L.Set.mem exes.modes L.byte         &&
         not (L.Set.mem exes.modes L.native) &&
         not (L.Set.mem exes.modes L.exe) then
        Exe.Linkage.custom :: l
      else
        l
    in

    let flags =
      Ocaml_flags.make exes.buildable sctx ~scope ~dir
    in
    let link_flags =
      SC.expand_and_eval_set sctx exes.link_flags
        ~scope
        ~dir
        ~standard:[]
    in

    (* Use "eobjs" rather than "objs" to avoid a potential conflict
       with a library of the same name *)
    let obj_dir =
      Path.relative dir ("." ^ (List.hd programs).name ^ ".eobjs")
    in
    Exe.build_and_link_many sctx
      ~dir
      ~obj_dir
      ~programs
      ~modules
      ~already_used
      ~scope
      ~linkages
      ~requires
      ~flags
      ~link_flags
      ~js_of_ocaml:exes.buildable.js_of_ocaml;

    Merlin.make ()
      ~requires:(Lib.Compile.requires compile_info)
      ~flags:(Ocaml_flags.common flags)
      ~preprocess:(Buildable.single_preprocess exes.buildable)
      ~objs_dirs:(Path.Set.singleton obj_dir)

  let executables_rules ~dir ~all_modules
        ?modules_partitioner ~scope (exes : Executables.t) =
    let compile_info =
      Lib.DB.resolve_user_written_deps (Scope.libs scope)
        exes.buildable.libraries
        ~pps:(Jbuild.Preprocess_map.pps exes.buildable.preprocess)
        ~allow_overlaps:exes.buildable.allow_overlapping_dependencies
    in
    SC.Libs.gen_select_rules sctx compile_info ~dir;
    SC.Libs.with_lib_deps sctx compile_info ~dir
      ~f:(fun () ->
        executables_rules exes ~dir ~all_modules
          ?modules_partitioner ~scope ~compile_info)

  (* +-----------------------------------------------------------------+
     | Aliases                                                         |
     +-----------------------------------------------------------------+ *)

  let add_alias ~dir ~name ~stamp ?(locks=[]) build =
    let alias = Build_system.Alias.make name ~dir in
    SC.add_alias_action sctx alias ~locks ~stamp build

  let alias_rules (alias_conf : Alias_conf.t) ~dir ~scope =
    let stamp =
      let module S = Sexp.To_sexp in
      Sexp.List
        [ Sexp.unsafe_atom_of_string "user-alias"
        ; S.list   Jbuild.Dep_conf.sexp_of_t   alias_conf.deps
        ; S.option Action.Unexpanded.sexp_of_t alias_conf.action
        ]
    in
    add_alias
      ~dir
      ~name:alias_conf.name
      ~stamp
      ~locks:(interpret_locks ~dir ~scope alias_conf.locks)
      (SC.Deps.interpret sctx ~scope ~dir alias_conf.deps
       >>>
       match alias_conf.action with
       | None -> Build.progn []
       | Some action ->
         SC.Action.run
           sctx
           action
           ~dir
           ~dep_kind:Required
           ~targets:Alias
           ~scope)

  (* +-----------------------------------------------------------------+
     | Stanza                                                          |
     +-----------------------------------------------------------------+ *)

  let gen_rules { SC.Dir_with_jbuild. src_dir; ctx_dir; stanzas; scope } =
    (* This interprets "rule" and "copy_files" stanzas. *)
    let files = text_files ~dir:ctx_dir in
    let all_modules = modules_by_dir ~dir:ctx_dir in
    let modules_partitioner =
      Modules_partitioner.create ~dir:src_dir ~all_modules
    in
    List.filter_map stanzas ~f:(fun stanza ->
      let dir = ctx_dir in
      match (stanza : Stanza.t) with
      | Library lib ->
        Some (library_rules lib ~dir ~files ~scope ~modules_partitioner)
      | Executables exes ->
        Some (executables_rules exes ~dir ~all_modules ~scope
                ~modules_partitioner)
      | Alias alias ->
        alias_rules alias ~dir ~scope;
        None
      | Copy_files { glob; _ } ->
        let src_dir =
          let loc = String_with_vars.loc glob in
          let src_glob = SC.expand_vars sctx ~dir glob ~scope in
          Path.parent (Path.relative src_dir src_glob ~error_loc:loc)
        in
        Some
          (Merlin.make ()
             ~source_dirs:(Path.Set.singleton src_dir))
      | _ -> None)
    |> Merlin.merge_all
    |> Option.map ~f:(fun m -> Merlin.add_source_dir m src_dir)
    |> Option.iter ~f:(Merlin.add_rules sctx ~dir:ctx_dir ~scope);
    Utop.setup sctx ~dir:ctx_dir ~libs:(
      List.filter_map stanzas ~f:(function
        | Stanza.Library lib -> Some lib
        | _ -> None)
    ) ~scope;
    Modules_partitioner.emit_warnings modules_partitioner

  let gen_rules ~dir components : Build_system.extra_sub_directories_to_keep =
    (match components with
     | ".js"  :: rest -> Js_of_ocaml_rules.setup_separate_compilation_rules
                           sctx rest
     | "_doc" :: rest -> Odoc.gen_rules rest ~dir
     | ".ppx"  :: rest -> Preprocessing.gen_rules sctx rest
     | _ ->
       match Path.Map.find stanzas_per_dir dir with
       | Some x -> gen_rules x
       | None ->
         if components <> [] &&
            Option.is_none
              (File_tree.find_dir (SC.file_tree sctx)
                 (Path.drop_build_context_exn dir)) then
           SC.load_dir sctx ~dir:(Path.parent dir));
    match components with
    | [] -> These (String_set.of_list [".js"; "_doc"; ".ppx"])
    | [(".js"|"_doc"|".ppx")] -> All
    | _  -> These String_set.empty

  let init () =
    let module Install_rules =
      Install_rules.Gen(struct
        include P
        let module_names_of_lib = module_names_of_lib
        let mlds_of_dir = mlds_of_dir
      end) in
    Install_rules.init ();
    Odoc.init ~modules_by_lib:(fun ~dir lib ->
      let m = modules_by_lib ~dir lib in
      match m.alias_module with
      | Some alias_mod ->
        [Option.value ~default:alias_mod
           (Module.Name.Map.find m.modules m.main_module_name)]
      | None -> Module.Name.Map.values m.modules
    ) ~mlds_of_dir
end

module type Gen = sig
  val gen_rules
    :  dir:Path.t
    -> string list
    -> Build_system.extra_sub_directories_to_keep
  val init : unit -> unit
end

let gen ~contexts ~build_system
      ?(external_lib_deps_mode=false)
      ?only_packages conf =
  let open Fiber.O in
  let { Jbuild_load. file_tree; jbuilds; packages; scopes } = conf in
  let packages =
    match only_packages with
    | None -> packages
    | Some pkgs ->
      Package.Name.Map.filter packages ~f:(fun { Package.name; _ } ->
        Package.Name.Set.mem pkgs name)
  in
  let sctxs = Hashtbl.create 4 in
  List.iter contexts ~f:(fun c ->
    Hashtbl.add sctxs c.Context.name (Fiber.Ivar.create ()));
  let make_sctx (context : Context.t) : _ Fiber.t =
    let host () =
      match context.for_host with
      | None -> Fiber.return None
      | Some h ->
        Fiber.Ivar.read (Option.value_exn (Hashtbl.find sctxs h.name))
        >>| fun x -> Some x
    in
    let stanzas () =
      Jbuild_load.Jbuilds.eval ~context jbuilds >>| fun stanzas ->
      match only_packages with
      | None -> stanzas
      | Some pkgs ->
        List.map stanzas ~f:(fun (dir, pkgs_ctx, stanzas) ->
          (dir,
           pkgs_ctx,
           List.filter stanzas ~f:(fun stanza ->
             match (stanza : Stanza.t) with
             | Library { public = Some { package; _ }; _ }
             | Alias { package = Some package ;  _ }
             | Install { package; _ }
             | Documentation { package; _ } ->
               Package.Name.Set.mem pkgs package.name
             | _ -> true)))
    in
    Fiber.fork_and_join host stanzas >>= fun (host, stanzas) ->
    let sctx =
      Super_context.create
        ?host
        ~build_system
        ~context
        ~scopes
        ~file_tree
        ~packages
        ~external_lib_deps_mode
        ~stanzas
    in
    let module M = Gen(struct let sctx = sctx end) in
    Fiber.Ivar.fill (Option.value_exn (Hashtbl.find sctxs context.name)) sctx
    >>| fun () ->
    (context.name, ((module M : Gen), stanzas))
  in
  Fiber.parallel_map contexts ~f:make_sctx >>| fun l ->
  let map = String_map.of_list_exn l in
  Build_system.set_rule_generators build_system
    (String_map.map map ~f:(fun ((module M : Gen), _) -> M.gen_rules));
  String_map.iter map ~f:(fun ((module M : Gen), _) -> M.init ());
  String_map.map map ~f:snd
# 26167 "boot.ml"
end
module Inline_tests : sig
# 1 "src/inline_tests.mli"
val linkme : unit
# 26172 "boot.ml"
end = struct
# 1 "src/inline_tests.ml"
open Import
open Jbuild
open Build.O
open! No_io

module SC = Super_context

module Backend = struct
  module M = struct
    module Info = struct
      let name = Sub_system_name.make "inline_tests.backend"

      type t =
        { loc              : Loc.t
        ; runner_libraries : (Loc.t * string) list
        ; flags            : Ordered_set_lang.Unexpanded.t
        ; generate_runner  : Action.Unexpanded.t option
        ; extends          : (Loc.t * string) list
        }

      type Jbuild.Sub_system_info.t += T of t

      let loc t = t.loc

      open Sexp.Of_sexp

      let short = None
      let parse =
        record
          (record_loc >>= fun loc ->
           field "runner_libraries" (list (located string)) ~default:[]
           >>= fun runner_libraries ->
           Ordered_set_lang.Unexpanded.field "flags" >>= fun flags ->
           field_o "generate_runner" Action.Unexpanded.t
           >>= fun generate_runner ->
           field "extends" (list (located string)) ~default:[]
           >>= fun extends ->
           return
             { loc
             ; runner_libraries
             ; flags
             ; generate_runner
             ; extends
             })

      let parsers =
        Syntax.Versioned_parser.make
          [ (1, 0),
            { Jbuild.Sub_system_info.
              short
            ; parse
            }
          ]
    end

    type t =
      { info             : Info.t
      ; lib              : Lib.t
      ; runner_libraries : Lib.t list Or_exn.t
      ; extends          :     t list Or_exn.t
      }

    let desc ~plural = "inline tests backend" ^ if plural then "s" else ""
    let desc_article = "an"

    let lib  t = t.lib
    let extends t = t.extends

    let instantiate ~resolve ~get lib (info : Info.t) =
      { info
      ; lib
      ; runner_libraries =
          Result.all (List.map info.runner_libraries ~f:resolve)
      ; extends =
          let open Result.O in
          Result.all
            (List.map info.extends
               ~f:(fun ((loc, name) as x) ->
                 resolve x >>= fun lib ->
                 match get lib with
                 | None ->
                   Error (Loc.exnf loc "%S is not an %s" name
                            (desc ~plural:false))
                 | Some t -> Ok t))
      }

    let to_sexp t =
      let open Sexp.To_sexp in
      let lib x = string (Lib.name x) in
      let f x = string (Lib.name x.lib) in
      ((1, 0),
       record_fields
         [ field "runner_libraries" (list lib)
             (Result.ok_exn t.runner_libraries)
         ; field "flags" Ordered_set_lang.Unexpanded.sexp_of_t t.info.flags
         ; field_o "generate_runner" Action.Unexpanded.sexp_of_t
             t.info.generate_runner
         ; field "extends" (list f) (Result.ok_exn t.extends) ~default:[]
         ])
  end
  include M
  include Sub_system.Register_backend(M)
end

include Sub_system.Register_end_point(
  struct
    module Backend = Backend

    module Info = struct
      let name = Sub_system_name.make "inline_tests"

      type t =
        { loc       : Loc.t
        ; deps      : Dep_conf.t list
        ; flags     : Ordered_set_lang.Unexpanded.t
        ; backend   : (Loc.t * string) option
        ; libraries : (Loc.t * string) list
        }

      type Jbuild.Sub_system_info.t += T of t

      let empty loc =
        { loc
        ; deps      = []
        ; flags     = Ordered_set_lang.Unexpanded.standard
        ; backend   = None
        ; libraries = []
        }

    let loc      t = t.loc
    let backends t = Option.map t.backend ~f:(fun x -> [x])

    open Sexp.Of_sexp

    let short = Some empty
    let parse =
      record
        (record_loc >>= fun loc ->
         field "deps" (list Dep_conf.t) ~default:[] >>= fun deps ->
         Ordered_set_lang.Unexpanded.field "flags" >>= fun flags ->
         field_o "backend" (located string) >>= fun backend ->
         field "libraries" (list (located string)) ~default:[]
         >>= fun libraries ->
         return
           { loc
           ; deps
           ; flags
           ; backend
           ; libraries
           })

    let parsers =
      Syntax.Versioned_parser.make
        [ (1, 0),
          { Jbuild.Sub_system_info.
            short
          ; parse
          }
        ]
  end

  let gen_rules c ~(info:Info.t) ~backends =
    let { Sub_system.Library_compilation_context.
          super_context = sctx
        ; dir
        ; stanza = lib
        ; scope
        ; source_modules
        ; _
        } = c
    in

    let inline_test_dir =
      Path.relative dir (sprintf ".%s.inline-tests" lib.name)
    in

    let name = "run" in
    let main_module_filename = name ^ ".ml" in
    let main_module_name = Module.Name.of_string name in
    let modules =
      Module.Name.Map.singleton main_module_name
        { Module.
          name = main_module_name
        ; impl = Some { name   = main_module_filename
                      ; syntax = OCaml
                      }
        ; intf = None
        ; obj_name = ""
        }
    in

    let extra_vars =
      String_map.singleton "library-name"
        (Action.Var_expansion.Strings ([lib.name], Concat))
    in

    let runner_libs =
      let open Result.O in
      Result.concat_map backends
        ~f:(fun (backend : Backend.t) -> backend.runner_libraries)
      >>= fun libs ->
      Lib.DB.find_many (Scope.libs scope) [lib.name]
      >>= fun lib ->
      Result.all
        (List.map info.libraries
           ~f:(Lib.DB.resolve (Scope.libs scope)))
      >>= fun more_libs ->
      Lib.closure (lib @ libs @ more_libs)
    in

    (* Generate the runner file *)
    SC.add_rule sctx (
      let target = Path.relative inline_test_dir main_module_filename in
      let source_modules = Module.Name.Map.values source_modules in
      let files ml_kind =
        Action.Var_expansion.Paths (
          List.filter_map source_modules ~f:(fun m ->
            Module.file m ~dir ml_kind),
          Split)
      in
      let extra_vars =
        List.fold_left
          [ "impl-files", files Impl
          ; "intf-files", files Intf
          ]
          ~init:extra_vars
          ~f:(fun acc (k, v) -> String_map.add acc k v)
      in
      Build.return []
      >>>
      Build.all
        (List.filter_map backends ~f:(fun (backend : Backend.t) ->
           Option.map backend.info.generate_runner ~f:(fun action ->
             SC.Action.run sctx action
               ~extra_vars ~dir ~dep_kind:Required ~targets:Alias ~scope)))
      >>^ (fun actions ->
        Action.with_stdout_to target
          (Action.progn actions))
      >>>
      Build.action_dyn ~targets:[target] ());

    Exe.build_and_link sctx
      ~dir:inline_test_dir
      ~obj_dir:inline_test_dir
      ~program:{ name; main_module_name }
      ~modules
      ~scope
      ~linkages:[Exe.Linkage.native_or_custom (SC.context sctx)]
      ~requires:runner_libs
      ~link_flags:(Build.return ["-linkall"])
      ~flags:(Ocaml_flags.of_list ["-w"; "-24"]);

    let flags =
      let flags =
        List.map backends ~f:(fun backend ->
          backend.Backend.info.flags) @ [info.flags]
      in
      Build.all (
        List.map flags ~f:(fun flags ->
          Super_context.expand_and_eval_set sctx flags
            ~scope
            ~dir
            ~extra_vars
            ~standard:[]))
      >>^ List.concat
    in

    SC.add_alias_action sctx
      (Build_system.Alias.runtest ~dir)
      ~stamp:(List [ Sexp.unsafe_atom_of_string "ppx-runner"
                   ; Quoted_string name
                   ])
      (let module A = Action in
       let exe = Path.relative inline_test_dir (name ^ ".exe") in
       Build.path exe >>>
       Build.fanout
         (Super_context.Deps.interpret sctx info.deps ~dir ~scope)
         flags
       >>^ fun (_deps, flags) ->
       A.chdir dir
         (A.progn
            (A.run (Ok exe) flags ::
             (Module.Name.Map.values source_modules
              |> List.concat_map ~f:(fun m ->
                [ Module.file m ~dir Impl
                ; Module.file m ~dir Intf
                ])
              |> List.filter_map ~f:(fun x -> x)
              |> List.map ~f:(fun fn ->
                A.diff ~optional:true
                  fn (Path.extend_basename fn ~suffix:".corrected"))))))
end)

let linkme = ()
# 26469 "boot.ml"
end
module Watermarks : sig
# 1 "src/watermarks.mli"
(** Watermarking *)

(** Expand watermarks in source files, similarly to what topkg does.

    This is only used when a package is pinned. *)

val subst : ?name:string -> unit -> unit Fiber.t
# 26480 "boot.ml"
end = struct
# 1 "src/watermarks.ml"
open Import
open Jbuilder_opam_file_format

open Fiber.O

let is_a_source_file fn =
  match Filename.extension fn with
  | ".flv"
  | ".gif"
  | ".ico"
  | ".jpeg"
  | ".jpg"
  | ".mov"
  | ".mp3"
  | ".mp4"
  | ".otf"
  | ".pdf"
  | ".png"
  | ".ttf"
  | ".woff" -> false
  | _ -> true

let make_watermark_map ~name ~version ~commit =
  let opam_file = Opam_file.load (name ^ ".opam") in
  let version_num =
    if String.is_prefix version ~prefix:"v" then
      String.sub version ~pos:1 ~len:(String.length version - 1)
    else
      version
  in
  let opam_var name sep =
    match Opam_file.get_field opam_file name with
    | None -> Error (sprintf "variable %S not found in opam file" name)
    | Some value ->
      let err = Error (sprintf "invalid value for variable %S in opam file" name) in
      match value with
      | String (_, s) -> Ok s
      | List (_, l) -> begin
          match
            List.fold_left l ~init:(Ok []) ~f:(fun acc v ->
              match acc with
              | Error _ -> acc
              | Ok l ->
                match v with
                | OpamParserTypes.String (_, s) -> Ok (s :: l)
                | _ -> err)
          with
          | Error _ as e -> e
          | Ok l -> Ok (String.concat ~sep (List.rev l))
        end
      | _ -> err
  in
  String_map.of_list_exn
    [ "NAME"           , Ok name
    ; "VERSION"        , Ok version
    ; "VERSION_NUM"    , Ok version_num
    ; "VCS_COMMIT_ID"  , Ok commit
    ; "PKG_MAINTAINER" , opam_var "maintainer"  ", "
    ; "PKG_AUTHORS"    , opam_var "authors"     ", "
    ; "PKG_HOMEPAGE"   , opam_var "homepage"    " "
    ; "PKG_ISSUES"     , opam_var "bug-reports" " "
    ; "PKG_DOC"        , opam_var "doc"         " "
    ; "PKG_LICENSE"    , opam_var "license"     ", "
    ; "PKG_REPO"       , opam_var "dev-repo"    " "
    ]

let subst_string s ~fname ~map =
  let len = String.length s in
  let longest_var = String.longest (String_map.keys map) in
  let loc_of_offset ~ofs ~len =
    let rec loop lnum bol i =
      if i = ofs then
        let pos =
          { Lexing.
            pos_fname = fname
          ; pos_cnum  = i
          ; pos_lnum  = lnum
          ; pos_bol   = bol
          }
        in
        { Loc.start = pos; stop  = { pos with pos_cnum = pos.pos_cnum + len } }
      else
        match s.[i] with
        | '\n' -> loop (lnum + 1) (i + 1) (i + 1)
        | _    -> loop lnum bol (i + 1)
    in
    loop 1 0 0
  in
  let rec loop i acc =
    if i = len then
      acc
    else
      match s.[i] with
      | '%' -> after_percent (i + 1) acc
      | _ -> loop (i + 1) acc
  and after_percent i acc =
    if i = len then
      acc
    else
      match s.[i] with
      | '%' -> after_double_percent ~start:(i - 1) (i + 1) acc
      | _ -> loop (i + 1) acc
  and after_double_percent ~start i acc =
    if i = len then
      acc
    else
      match s.[i] with
      | '%' -> after_double_percent ~start:(i - 1) (i + 1) acc
      | 'A'..'Z' | '_' -> in_var ~start (i + 1) acc
      | _ -> loop (i + 1) acc
  and in_var ~start i acc =
    if i - start > longest_var then
      loop i acc
    else if i = len then
      acc
    else
      match s.[i] with
      | '%' -> end_of_var ~start (i + 1) acc
      | 'A'..'Z' | '_' -> in_var ~start (i + 1) acc
      | _ -> loop (i + 1) acc
  and end_of_var ~start i acc =
    if i = len then
      acc
    else
      match s.[i] with
      | '%' -> begin
          let var = String.sub s ~pos:(start + 2) ~len:(i - start - 3) in
          match String_map.find map var with
          | None -> in_var ~start:(i - 1) (i + 1) acc
          | Some (Ok repl) ->
            let acc = (start, i + 1, repl) :: acc in
            loop (i + 1) acc
          | Some (Error msg) ->
            let loc = loc_of_offset ~ofs:start ~len:(i + 1 - start) in
            Loc.fail loc "%s" msg
        end
      | _ -> loop (i + 1) acc
  in
  match List.rev (loop 0 []) with
  | [] -> None
  | repls ->
    let result_len =
      List.fold_left repls ~init:(String.length s) ~f:(fun acc (a, b, repl) ->
        acc - (b - a) + String.length repl)
    in
    let buf = Buffer.create result_len in
    let pos =
      List.fold_left repls ~init:0 ~f:(fun pos (a, b, repl) ->
        Buffer.add_substring buf s pos (a - pos);
        Buffer.add_string buf repl;
        b)
    in
    Buffer.add_substring buf s pos (len - pos);
    Some (Buffer.contents buf)

let subst_file fn ~map =
  let s = Io.read_file fn in
  let s =
    if Filename.dirname fn = "." && String.is_suffix fn ~suffix:".opam" then
      "version: \"%%" ^ "VERSION_NUM" ^ "%%\"\n" ^ s
    else
      s
  in
  match subst_string s ~map ~fname:fn with
  | None -> ()
  | Some s -> Io.write_file fn s

let get_name ~files ?name () =
  let package_names =
    List.filter_map files ~f:(fun fn ->
      if Filename.dirname fn = "." then
        match Filename.split_extension fn with
        | s, ".opam" -> Some s
        | _ -> None
      else
        None)
  in
  if package_names = [] then die "@{<error>Error@}: no <package>.opam files found.";
  match name with
  | Some name ->
    if not (List.mem name ~set:package_names) then
      die "@{<error>Error@}: file %s.opam doesn't exist." name;
    name
  | None ->
    let shortest =
      match package_names with
      | [] -> assert false
      | first :: rest ->
        List.fold_left rest ~init:first ~f:(fun acc s ->
          if String.length s < String.length acc then
            s
          else
            acc)
    in
    if List.for_all package_names ~f:(String.is_prefix ~prefix:shortest) then
      shortest
    else
      die "@{<error>Error@}: cannot determine name automatically.\n\
           You must pass a [--name] command line argument."

let subst_git ?name () =
  let rev = "HEAD" in
  let git =
    match Bin.which "git" with
    | Some x -> Path.to_string x
    | None -> Utils.program_not_found "git"
  in
  let env = Env.initial in
  Fiber.fork_and_join
    (fun () ->
       Fiber.fork_and_join
         (fun () ->
            Process.run_capture Strict git ["describe"; "--always"; "--dirty"]
              ~env)
         (fun () ->
            Process.run_capture Strict git ["rev-parse"; rev]
              ~env))
    (fun () ->
       Process.run_capture_lines Strict git ["ls-tree"; "-r"; "--name-only"; rev]
         ~env)
  >>= fun ((version, commit), files) ->
  let version = String.trim version in
  let commit  = String.trim commit  in
  let name = get_name ~files ?name () in
  let watermarks = make_watermark_map ~name ~version ~commit in
  List.iter files ~f:(fun fn ->
    if is_a_source_file fn then
      subst_file fn ~map:watermarks);
  Fiber.return ()

let subst ?name () =
  if Sys.file_exists ".git" then
    subst_git ?name ()
  else
    Fiber.return ()
# 26718 "boot.ml"
end
module Main : sig
# 1 "src/main.mli"
open! Import
open Jbuild

type setup =
  { build_system : Build_system.t
  ; (* Evaluated jbuilds per context names *)
    stanzas      : (Path.t * Scope_info.t * Stanzas.t) list String_map.t
  ; contexts     : Context.t list
  ; packages     : Package.t Package.Name.Map.t
  ; file_tree    : File_tree.t
  ; env          : Env.t
  }

(* Returns [Error ()] if [pkg] is unknown *)
val package_install_file : setup -> Package.Name.t -> (Path.t, unit) result

(** Scan the source tree and discover everything that's needed in order to build
    it. *)
val setup
  :  ?log:Log.t
  -> ?external_lib_deps_mode:bool
  -> ?workspace:Workspace.t
  -> ?workspace_file:string
  -> ?only_packages:Package.Name.Set.t
  -> ?x:string
  -> ?ignore_promoted_rules:bool
  -> ?capture_outputs:bool
  -> unit
  -> setup Fiber.t
val external_lib_deps
  : ?log:Log.t
  -> packages:Package.Name.t list
  -> unit
  -> Build.lib_deps Path.Map.t

val find_context_exn : setup -> name:string -> Context.t

(** Setup the environment *)
val setup_env : capture_outputs:bool -> Env.t

(**/**)

(* This is used to bootstrap jbuilder itself. It is not part of the
   public API. *)
val bootstrap : unit -> unit
# 26767 "boot.ml"
end = struct
# 1 "src/main.ml"
open Import
open Fiber.O

let () = Inline_tests.linkme

type setup =
  { build_system : Build_system.t
  ; stanzas      : (Path.t * Jbuild.Scope_info.t * Jbuild.Stanzas.t) list String_map.t
  ; contexts     : Context.t list
  ; packages     : Package.t Package.Name.Map.t
  ; file_tree    : File_tree.t
  ; env          : Env.t
  }

let package_install_file { packages; _ } pkg =
  match Package.Name.Map.find packages pkg with
  | None -> Error ()
  | Some p ->
    Ok (Path.relative p.path
          (Utils.install_file ~package:p.name ~findlib_toolchain:None))

let setup_env ~capture_outputs =
  let env =
    if capture_outputs || not (Lazy.force Colors.stderr_supports_colors) then
      Env.initial
    else
      Colors.setup_env_for_colors Env.initial
  in
  Env.add env ~var:"INSIDE_DUNE" ~value:"1"

let setup ?(log=Log.no_log)
      ?external_lib_deps_mode
      ?workspace ?(workspace_file="jbuild-workspace")
      ?only_packages
      ?extra_ignored_subtrees
      ?x
      ?ignore_promoted_rules
      ?(capture_outputs=true)
      () =
  let env = setup_env ~capture_outputs in
  let conf =
    Jbuild_load.load ?extra_ignored_subtrees ?ignore_promoted_rules ()
  in
  Option.iter only_packages ~f:(fun set ->
    Package.Name.Set.iter set ~f:(fun pkg ->
      if not (Package.Name.Map.mem conf.packages pkg) then
        let pkg_name = Package.Name.to_string pkg in
        die "@{<error>Error@}: I don't know about package %s \
             (passed through --only-packages/--release)%s"
          pkg_name
          (hint pkg_name
             (Package.Name.Map.keys conf.packages
             |> List.map ~f:Package.Name.to_string))));
  let workspace =
    match workspace with
    | Some w -> w
    | None ->
      if Sys.file_exists workspace_file then
        Workspace.load ?x workspace_file
      else
        { merlin_context = Some "default"
        ; contexts = [Default [
            match x with
            | None -> Native
            | Some x -> Named x
          ]]
        }
  in

  Fiber.parallel_map workspace.contexts ~f:(fun ctx_def ->
    let name = Workspace.Context.name ctx_def in
    Context.create ctx_def ~env ~merlin:(workspace.merlin_context = Some name))
  >>= fun contexts ->
  let contexts = List.concat contexts in
  List.iter contexts ~f:(fun (ctx : Context.t) ->
    Log.infof log "@[<1>Jbuilder context:@,%a@]@." Sexp.pp
      (Context.sexp_of_t ctx));
  let rule_done  = ref 0 in
  let rule_total = ref 0 in
  let gen_status_line () =
    Some (sprintf "Done: %u/%u" !rule_done !rule_total)
  in
  let hook (hook : Build_system.hook) =
    match hook with
    | Rule_started   -> incr rule_total
    | Rule_completed -> incr rule_done
  in
  let build_system =
    Build_system.create ~contexts ~file_tree:conf.file_tree ~hook
  in
  Gen_rules.gen conf
    ~build_system
    ~contexts
    ?only_packages
    ?external_lib_deps_mode
  >>= fun stanzas ->
  Scheduler.set_status_line_generator gen_status_line
  >>>
  Fiber.return
    { build_system
    ; stanzas
    ; contexts
    ; packages = conf.packages
    ; file_tree = conf.file_tree
    ; env
    }

let find_context_exn t ~name =
  match List.find t.contexts ~f:(fun c -> c.name = name) with
  | Some ctx -> ctx
  | None ->
    die "@{<Error>Error@}: Context %S not found!@." name

let external_lib_deps ?log ~packages () =
  Scheduler.go ?log
    (setup () ~external_lib_deps_mode:true
     >>| fun setup ->
     let context = find_context_exn setup ~name:"default" in
     let install_files =
       List.map packages ~f:(fun pkg ->
         match package_install_file setup pkg with
         | Ok path -> Path.append context.build_dir path
         | Error () -> die "Unknown package %S" (Package.Name.to_string pkg))
     in
     let stanzas = Option.value_exn (String_map.find setup.stanzas "default") in
     let internals = Jbuild.Stanzas.lib_names stanzas in
     Path.Map.map
       (Build_system.all_lib_deps setup.build_system
          ~request:(Build.paths install_files))
       ~f:(String_map.filteri ~f:(fun name _ ->
         not (String_set.mem internals name))))

let ignored_during_bootstrap =
  Path.Set.of_list
    (List.map ~f:Path.of_string
       [ "test"
       ; "example"
       ])

(* Called by the script generated by ../build.ml *)
let bootstrap () =
  Colors.setup_err_formatter_colors ();
  let main () =
    let anon s = raise (Arg.Bad (Printf.sprintf "don't know what to do with %s\n" s)) in
    let subst () =
      Scheduler.go (Watermarks.subst () ~name:"jbuilder");
      exit 0
    in
    let display = ref None in
    let display_mode =
      Arg.Symbol
        (List.map Config.Display.all ~f:fst,
         fun s ->
           display := Some (List.assoc s Config.Display.all))
    in
    let concurrency = ref None in
    let set r x = r := Some x in
    Arg.parse
      [ "-j"           , Int (set concurrency), "JOBS concurrency"
      ; "--dev"        , Set Clflags.dev_mode , " set development mode"
      ; "--display"    , display_mode         , " set the display mode"
      ; "--subst"      , Unit subst           ,
        " substitute watermarks in source files"
      ; "--debug-backtraces",
        Set Clflags.debug_backtraces,
        " always print exception backtraces"
      ]
      anon "Usage: boot.exe [-j JOBS] [--dev]\nOptions are:";
    Clflags.debug_dep_path := true;
    let config =
      (* Only load the configuration with --dev *)
      if !Clflags.dev_mode then
        Config.load_user_config_file ()
      else
        Config.default
    in
    let config =
      Config.merge config
        { display     = !display
        ; concurrency = !concurrency
        }
    in
    let config =
      Config.adapt_display config
        ~output_is_a_tty:(Lazy.force Colors.stderr_supports_colors)
    in
    let log = Log.create ~display:config.display () in
    Scheduler.go ~log ~config
      (setup ~log ~workspace:{ merlin_context = Some "default"
                             ; contexts = [Default [Native]] }
         ~extra_ignored_subtrees:ignored_during_bootstrap
         ()
       >>= fun { build_system = bs; _ } ->
       Build_system.do_build bs
         ~request:(Build.path (Path.of_string "_build/default/jbuilder.install")))
  in
  try
    main ()
  with
  | Fiber.Never -> exit 1
  | exn ->
    Report_error.report exn;
    exit 1

let setup = setup ~extra_ignored_subtrees:Path.Set.empty

let find_context_exn t ~name =
  match List.find t.contexts ~f:(fun c -> c.name = name) with
  | Some ctx -> ctx
  | None ->
    die "@{<Error>Error@}: Context %S not found!@." name
# 26981 "boot.ml"
end
let () = Main.bootstrap ()
